"""implement the 'chemfp' entry point, and support 'python -m chemfp'

"""
# Copyright (c) 2020-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse
import pkg_resources
import shutil
import sys
import io

from . import registry
from .cli import cmdsupport


parser = argparse.ArgumentParser(
    prog = "chemfp",
    description = "chemfp tools for cheminformatics fingerprints and structure I/O",
    allow_abbrev = False,
    add_help = False,
    )

def add_global_options(p):
    p.add_argument("--license-file", metavar="FILENAME")
    p.add_argument("--traceback", action="store_true",
                       help = "print the traceback on KeyboardInterrupt")
    cmdsupport.add_version(p)
    p.add_argument("command", nargs="?")
    p.add_argument("rest", nargs=argparse.REMAINDER)


### Deal with entry points

def _resolve_ep(ep, group, expected_types, default_attr):
    try:
        obj = ep.resolve()
    except Exception as err:
        sys.stderr.write("Cannot load %s entry point %s: %s\n" % (
            group, ep, err))
        return None

    if not ep.attrs and default_attr is not None:
        try:
            obj = getattr(obj, default_attr)
        except AttributeError:
            sys.stderr.write("Missing %s entry point %s attribute %s\n" % (
                group, ep, default_attr))
            return None

    if not isinstance(obj, expected_types):
        sys.stderr.write("Unexpected %s entry point %s object: %r\n" % (
                group, ep, obj))
        return None

    try:
        obj.register(ep.name)
    except Exception as err:
        sys.stderr.write("Failed to register %s entry point %s object: %r error: %s\n" % (
                group, ep, obj, err))
        return None

    return obj

def resolve_command(parser, name):
    for ep in pkg_resources.iter_entry_points("chemfp.commands", name):
        obj = _resolve_ep(
            ep,
            "chemfp.commands",
            (registry.CommandConfig,),
            "command_config",
            )
        if obj is None:
            continue
        break
    else:
        parser.error("Unsupported command: %r" % (name,))

    return obj


def _get_from_eps(group, expected_types, default_attr=None):
    d = {}
    for ep in pkg_resources.iter_entry_points(group):
        if ep.name in d:
            continue
        obj = _resolve_ep(ep, group, expected_types, default_attr)
        if obj is None:
            continue
        
        d[ep.name] = (ep, obj)
    return d
            
    
class CommandGroup:
    def __init__(self, name, summary, rank):
        self.name = name
        self.summary = summary
        self.rank = rank
        self.commands = []
        
        
class Command:
    def __init__(self, name, config):
        self.name = name
        self.group_name = config.group_name
        self.summary = config.summary
        self.hidden = config.hidden
        self.aliases = []

        
def print_help():
    import textwrap
    
    group_eps = _get_from_eps(
        "chemfp.command_groups",
        (registry.CommandGroupConfig,),
        )

    groups = dict((name, CommandGroup(name, config.summary, config.rank))
                      for (name, (ep, config)) in group_eps.items())

    other_group = CommandGroup(None, "Other", 99999)
    
    command_eps = _get_from_eps(
        "chemfp.commands",
        (registry.CommandConfig,),
        "command_config",
        )

    commands = {}
    ep_key_to_cmd = {}
    for (name, (ep, config)) in command_eps.items():
        cmd = Command(name, config)
        commands[name] = cmd
        ep_key_to_cmd[(ep.name, ep.attrs)] = cmd

        if not cmd.hidden:
            groups.get(config.group_name, other_group).commands.append(cmd)
        
    
    for ep in pkg_resources.iter_entry_points("chemfp.command_aliases"):
        ep_key = (ep.name, ep.attrs)
        cmd = ep_key_to_cmd.get(ep_key, None)
        if cmd is not None:
            cmd.aliases(ep.name)

    group_list = [group for group in groups.values() if group.commands]
    if other_group.commands:
        group_list.append(other_group)

    group_list.sort(key = lambda group: getattr(group, "rank", 10000))

    ### Output
    
    parser.print_help()
    
    print("\nSummary of the available commands.\n")

    width = shutil.get_terminal_size().columns
    
    for group in group_list:
        print(group.summary)
        print()
        commands = group.commands
        commands.sort(key = lambda cmd: cmd.name)

        for cmd in commands:
            cmd_str = "  " + ", ".join([cmd.name] + cmd.aliases)

            N = 24
            if len(cmd_str) < N-2:
                cmd_str += " " * (N - len(cmd_str))
            else:
                cmd_str += "  "

            print(
                *textwrap.wrap(
                    cmd.summary,
                    initial_indent = cmd_str,
                    subsequent_indent = " " * (N+1),
                    width = width,
                    ),
                sep = "\n",
                )

        print()
            
    

class ShowHelpAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        nargs = 0
        super().__init__(option_strings, dest, nargs=nargs, **kwargs)
        
    def __call__(self, parser, namespace, values, option_string):
        print_help()
        raise SystemExit(0)

parser.add_argument("--help", "-h", action=ShowHelpAction)
add_global_options(parser)

## def get_subcommands():
##     for ep in pkg_resources.iter_entry_points('chemfp.commands'):
##         yield ep.name

def main(args=None):
    traceback = False
    try:
        args = parser.parse_args(args)
        traceback = args.traceback
        #args, remaining = parser.parse_args()
        #print(args, remaining)
        command = args.command
        if command is None:
            print_help()
            raise SystemExit()
    except KeyboardInterrupt:
        raise SystemExit(2)
        
    try:
        command_config = resolve_command(parser, command)
        if traceback:
            command_config.run_command(args.rest, None)
        else:
            try:
                command_config.run_command(args.rest, None)
            except KeyboardInterrupt:
                raise SystemExit(2)
    except (SystemExit, BrokenPipeError) as err:
        # BrokenPipeError and SIGINT could have occurred at the same time
        try:
            sys.stdout.flush()
        except BrokenPipeError:
            cmdsupport.handle_broken_pipe()
        except ValueError:
            # Have something to write to
            sys.stdout = io.StringIO()
        
        if isinstance(err, SystemExit):
            raise
    else:
        try:
            sys.stdout.flush()
        except BrokenPipeError:
            cmdsupport.handle_broken_pipe()


    
if __name__ == "__main__":
    main()
