"Create Open Babel fingerprints"

# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.

#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import sys

import openbabel.openbabel as ob

from . import io
from . import base_toolkit
from . import _text_toolkit
from ._compat import filename_types, myrepr


EXTERNAL_NAME = "Open Babel"

# OBConversion() must be called before trying to find any plugin.
_obconv = ob.OBConversion()

_ob_version = ob.OBReleaseVersion()

SOFTWARE = "OpenBabel/" + _ob_version

_strict_error_handler = io.get_parse_error_handler("strict")

#########

def _convert_for_readstring(content, error_handler, location):
    if isinstance(content, str):
        return content
    try:
        return content.decode("utf8")
    except UnicodeDecodeError as err:
        error_handler.error("Cannot decode the content for Open Babel's ReadString(): %s"
                            % (err,), location)
    return None

def _to_openbabel_string(s):
    if s is None:
        return None
    if isinstance(s, str):
        return s
    return s.decode("utf8")

        

def _convert_tag_for_getdata(tag):
    if tag is None:
        return tag
    if isinstance(tag, str):
        return tag
    return tag.decode("utf8")
        
#########

def _or_string(location):
    filename = location.filename
    if filename is None:
        return "<string>"
    return filename

def _or_bytes(location):
    filename = location.filename
    if filename is None:
        return "<bytes>"
    return filename



def _open_input(filename, obconversion, obmol, register_name):
    if filename is None:
        # Open Babel can't read from stdin, but I can ask it to
        # read from /dev/stdin, or "CON" for MS Windows.
        # (Though "CON" doesn't support reading from a pipe.)

        filename = io.DEV_STDIN
        if filename is None:
            raise ValueError(
                "chemfp cannot convince Open Babel to read from stdin on this operating system")
        success = obconversion.ReadFile(obmol, filename)

    elif isinstance(filename, filename_types):
        # Reset the error log so I might be able to get the error
        # message.  This is not thread-safe.

        errmsg = None
        ob.obErrorLog.ClearLog()
        lvl = ob.obErrorLog.GetOutputLevel()
        ob.obErrorLog.SetOutputLevel(-1) # Suppress messages to stderr

        success = obconversion.ReadFile(obmol, filename)

        # Get any possible error message.
        ob.obErrorLog.SetOutputLevel(lvl) # Restore message level
        if ob.obErrorLog.GetErrorMessageCount():
            errmsg = _get_ob_error(ob.obErrorLog)


        # success == False if this is an empty file or if there was an error.
        # If there's an error, then GetInStream() is None
        if not success and obconversion.GetInStream() is None:
            # I prefer Python's error message, if there is one.
            open(filename).close() # Make sure the file can be opened for reading

            # Report the error from Open Babel, or use my own error message
            # if no such message exists.
            if not errmsg:
                errmsg = "Open Babel is unable to open the %r file" % (register_name,)

            raise IOError(5, errmsg, filename)

    else:
        raise AssertionError("Unsupported filename: %r" % (filename,))

    return success



# Neither the Open Babel format documentation nor API list the
# available options. I've reviewed the source and picked out the ones
# I want to support, for the formats I want to support. You can still
# get low-level access to the options using the "options" option.

class OpenBabelFormatConfig(object):
    is_available = True
    supports_io = True
    alternate_extensions = ()

    class input(object):
        arg_info = (
            ("options", str),
            )
        defaults = {
            "options": None,
            }
        
    output = input


    def _set_options(self, obconversion, in_or_out, args):
        options_str = args.get("options", None)
        if options_str:
            obconversion.SetOptions(options_str, in_or_out)
        return {"options": options_str}

    def _make_obconversion_for_input(self, args):
        obconversion = ob.OBConversion()
        if not obconversion.SetInFormat(self._obformat):
            raise ValueError("Open Babel does not support the %r input format" % (self.record_format,))

        new_args = self._set_options(obconversion, obconversion.INOPTIONS, args)

        return new_args, obconversion

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        if not (source is None or isinstance(source, filename_types)):
            raise ValueError("Open Babel does not support input from a Python file object")
        
        id_tag = _to_openbabel_string(id_tag)

        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)

        # OpenBabel auto-detects gzip compression.
        # Don't double-check to see if the input compression was correct,
        # but make sure it is acceptable.
        if compression not in (None, "", "gz"):
            raise ValueError("Open Babel does not support compression type %r" % (compression,))

        obmol = ob.OBMol()

        success = _open_input(source, obconversion, obmol, self.register_name)

        # We've opened the file. Switch to the iterator.
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(
            metadata,
            iter_generic_records(obconversion, obmol, success, id_tag, error_handler, location),
            location)
        

    def read_ids_and_molecules_from_string(self, content, compression,
                                           id_tag, reader_args, error_handler, location):
        if isinstance(content, str):
            new_content = content
        else:
            assert not isinstance(content, str), "should have already checked"
            new_content = io._decompress_string(content, compression, "read_ids_and_molecules_from_string")
            

        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        id_tag = _to_openbabel_string(id_tag)
        
        obmol = ob.OBMol()
        # Use the strict error handler because the error handler is for records, not for the entire string.
        ob_content = _convert_for_readstring(new_content, _strict_error_handler, location)

        success = obconversion.ReadString(obmol, ob_content)
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(
            metadata,
            iter_generic_records(obconversion, obmol, success, id_tag, error_handler, location),
            location)
        

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        
        obmol = ob.OBMol()
        if id_tag is None:
            def openbabel_parser(content):
                obmol.Clear()
                content = _convert_for_readstring(content, error_handler, None)
                if content is None:
                    return None, None
                success = obconversion.ReadString(obmol, content)
                if not success:
                    error_handler.error("Open Babel cannot parse the %s record"
                                        % (self.register_name,))
                    return None, None
                return obmol.GetTitle(), obmol
        else:
            def openbabel_parser(content):
                obmol.Clear()
                content = _convert_for_readstring(content, error_handler, None)
                if content is None:
                    return None, None
                success = obconversion.ReadString(obmol, content)
                if not success:
                    error_handler.error("Open Babel cannot parse the %s record"
                                        % (self.register_name,))
                    return None, None
                obj = obmol.GetData(id_tag)
                if obj is None:
                    id = None
                else:
                    id = obj.GetValue()
                
                return id, obmol

        openbabel_parser.__name__ = "openbabel_" + self.register_name + "_parse_id_and_molecule"
        return openbabel_parser
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        
        obmol = ob.OBMol()

        content = _convert_for_readstring(content, error_handler, None)
        if content is None:
            return None, None
        success = obconversion.ReadString(obmol, content)
        if not success:
            error_handler.error("Open Babel cannot parse the %s record"
                                % (self.register_name,))
            return None, None

        if id_tag is None:
            id = obmol.GetTitle()
        else:
            id_tag = _to_openbabel_string(id_tag)
            obj = obmol.GetData(id_tag)
            if obj is None:
                id = None
            else:
                id = obj.GetValue()
        
        return id, obmol
        

    def _make_obconversion_for_output(self, args):
        obconversion = ob.OBConversion()
        if not obconversion.SetOutFormat(self._obformat):
            raise ValueError("Open Babel does not support the %r output format" % (self.record_format,))

        new_args = self._set_options(obconversion, obconversion.OUTOPTIONS, args)
        return new_args, obconversion


    # Open Babel has a file writer, but as far as I can tell there's
    # no way from Python to ask it to enable compression. Nor can I
    # pass in stdout or a Python file object. Instead, I implement my
    # own writer by assuming that I can write each record at a time.

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format, new_writer_args),
            gen_write_ob_records(outfile, close, obconversion, error_handler, location, self.register_name),
            location
            )
        

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)

        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_ob_records(outfile, close, obconversion, error_handler, location, self.register_name),
            getvalue,
            location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)

        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_ob_records(outfile, close, obconversion, error_handler, location, self.register_name),
            getvalue,
            location)


    def create_string(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        if id is not None:
            old_id = mol.GetTitle()
            if id == old_id:
                id = None
            else:
                mol.SetTitle(_to_openbabel_string(id))

        result = obconversion.WriteString(mol, False)
        if id is not None:
            mol.SetTitle(old_id)
            
        if not result:
            error_handler.error("Open Babel cannot create the %s string"
                                % (self.register_name,))
            return None
        return result

    def create_bytes(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        if id is not None:
            old_id = mol.GetTitle()
            if id == old_id:
                id = None
            else:
                mol.SetTitle(_to_openbabel_string(id))

        result = obconversion.WriteString(mol, False).encode("utf8")
        if id is not None:
            mol.SetTitle(old_id)
            
        if not result:
            error_handler.error("Open Babel cannot create the %s byte string"
                                % (self.register_name,))
            return None
        return result


def gen_write_ob_records(outfile, close, obconversion, error_handler, location, format_name):
    x = _gen_write_ob_records(outfile, close, obconversion, error_handler, location, format_name)
    next(x)
    return x

# As far as I can tell, there's no way from Python to ask Open Babel
# to write multiple structures to a string. Instead, take care of it
# myself by assuming 

def _gen_write_ob_records(outfile, close, obconversion, error_handler, location, format_name):
    recno = output_recno = 0
    mol = None
    WriteString = obconversion.WriteString

    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        rec = WriteString(mol)
                        if not rec:
                            error_handler.error("Open Babel cannot create the %s string"
                                                % (format_name,), location)
                            continue
                        if not outfile.write_string(rec, error_handler, location):
                            continue
                            
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        if id is not None:
                            mol.SetTitle(id)
                        rec = WriteString(mol)
                        if not rec:
                            error_handler.error("Open Babel cannot create the %s string"
                                                % (format_name,), location)
                            continue

                        if not outfile.write_string(rec, error_handler, location):
                            continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    if id is not None:
                        mol.SetTitle(id)
                    rec = WriteString(mol)
                    if not rec:
                        error_handler.error("Open Babel cannot create the %s string"
                                            % (format_name,), location)
                        continue
                    if not outfile.write_string(rec, error_handler, location):
                        continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                

    finally:
        if close is not None:
            close()
        # Break the cyclical references
        location.save(recno=get_recno(),
                      output_recno=get_output_recno())



def _get_ob_error(log):
    msgs = log.GetMessagesOfLevel(ob.obError)
    return "".join(msgs)



#########

class GenericFormat(OpenBabelFormatConfig):
    def __init__(self, record_format, is_input_format, is_output_format):
        self.register_name = record_format
        self.record_format = record_format
        self._obformat = record_format
        if not is_input_format:
            self.input = None
        if not is_output_format:
            self.output = None


def parse_canonicalization(value):
    if value in ("1", "True", "true", "default"):
        return "default"
    if value in ("0", "False", "false", "none"):
        return None
    if value in ("universal", "anticanonical", "inchified"):
        return value
    raise ValueError("Unsupported canonicalization option %r" % (value,))

def parse_implementation(value):
    if value in ("openbabel", "chemfp"):
        return value
    raise ValueError("Unsupported implementation option %r" % (value,))
    

class BaseSmiFormat(OpenBabelFormatConfig):
    _obformat = "smi"
        
    class output(object):
        arg_info = OpenBabelFormatConfig.output.arg_info + (
            ("isomeric", base_toolkit.parse_boolean),
            ("canonicalization", parse_canonicalization),
            ("explicit_hydrogens", base_toolkit.parse_boolean),
            )
        defaults = OpenBabelFormatConfig.output.defaults.copy()
        defaults.update({
            "isomeric": True,
            "canonicalization": "default",
            "explicit_hydrogens": False,
            })

    def _make_obconversion_for_output(self, writer_args):
        obconversion = ob.OBConversion()
        if not obconversion.SetOutFormat(self._obformat):
            raise ValueError("Open Babel does not support the %r output format" % (self.record_format,))

        OUTOPTIONS = obconversion.OUTOPTIONS
        new_writer_args = {}
        
        options_str = writer_args["options"]
        if options_str:
            obconversion.SetOptions(options_str, OUTOPTIONS)
            new_writer_args["options"] = options_str
        else:
            new_writer_args["options"] = None

        # The "i" flag means "don't include isomeric"
        if writer_args["isomeric"]:
            # User does want isomeric, so leave things alone.
            new_writer_args["isomeric"] = True
        else:
            # Disable isomeric output
            obconversion.AddOption("i", OUTOPTIONS)
            new_writer_args["isomeric"] = False

        # Various types of canonical representations
        canonicalization = writer_args["canonicalization"]
        if canonicalization in (True, "default", None):
            # Generate an Open Babel canonical string
            obconversion.AddOption("c", OUTOPTIONS)
            new_writer_args["canonicalization"] = "default"

        elif canonicalization in (False, "none"):
            # Don't do anything
            new_writer_args["canonicalization"] = "none"

        elif canonicalization == "universal":
            # Generate a universal SMILES
            obconversion.AddOption("U", OUTOPTIONS)
            new_writer_args["canonicalization"] = "universal"

        elif canonicalization == "anticanonical":
            # "anti-canonical" uses randomly assigned atom classes
            obconversion.AddOption("C", OUTOPTIONS)
            new_writer_args["canonicalization"] = "anticanonical"

        elif canonicalization == "inchified":
            # InChI-fied SMILES
            obconversion.AddOption("I", OUTOPTIONS)
            new_writer_args["canonicalization"] = "inchified"

        else:
            raise ValueError("Unsupported Open Babel canonicalization %r" % (canonicalization,))

        # Always write explicit hydrogens
        if writer_args["explicit_hydrogens"]:
            # I don't have a good test case for this
            obconversion.AddOption("h", OUTOPTIONS)
            new_writer_args["explicit_hydrogens"] = True
        else:
            new_writer_args["explicit_hydrogens"] = False


        # disable id in the output; I handle that myself
        obconversion.AddOption("smilesonly", OUTOPTIONS, "smilesonly")

        return new_writer_args, obconversion


class BaseSmiFileFormat(BaseSmiFormat):
    class input(object):
        arg_info = BaseSmiFormat.input.arg_info + (
            ("delimiter", base_toolkit.parse_delimiter),
            ("has_header", base_toolkit.parse_boolean),
            )
        defaults = BaseSmiFormat.input.defaults.copy()
        defaults.update({
            "delimiter": None,
            "has_header": False,
            })

    class output(object):
        arg_info = BaseSmiFormat.output.arg_info + (
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = BaseSmiFormat.output.defaults.copy()
        defaults.update({
            "delimiter": None,
            })
        
        
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        new_reader_args["delimiter"] = delimiter.name
        new_reader_args["has_header"] = has_header = bool(reader_args["has_header"])
        
        infile, close = io.open_compressed_input(source, compression, location, EXTERNAL_NAME)
        
        iter_reader = _text_toolkit.iter_first_two_delimited_fields_as_strings(
            infile, close, delimiter, has_header,
            error_handler, location, "SMILES",
            encoding, encoding_errors)
        
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(
            metadata,
            iter_column_records("SMILES", iter_reader, obconversion, error_handler, location),
            location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                           id_tag, reader_args, error_handler, location):
        is_unicode, infile, close = io._open_string_as_bytes(content, compression, location, EXTERNAL_NAME)
        
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        new_reader_args["delimiter"] = delimiter.name
        new_reader_args["has_header"] = has_header = bool(reader_args["has_header"])

        iter_reader = _text_toolkit.iter_first_two_delimited_fields_as_strings(
            infile, close, delimiter, has_header,
            error_handler, location, "SMILES",
            "utf8", "strict")
        
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args),
            iter_column_records("SMILES", iter_reader, obconversion, error_handler, location),
            location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        obmol = ob.OBMol()

        def openbabel_parser(content):
            obmol.Clear()
            if isinstance(content, str):
                first_line, _, _ = content.partition("\n") # get the first record from the input
                terms = delimiter.split_two_terms_string(first_line.rstrip(u"\r"))
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                smiles, id = terms
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None
                smiles, id_bytes = terms
                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode("utf8")
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s"
                                            % (id_bytes, err))
                        return None, None
            
            smiles = _convert_for_readstring(smiles, error_handler, None)
            if smiles is None:
                return None, None

            success = obconversion.ReadString(obmol, smiles)
            if not success:
                error_handler.error("Open Babel cannot parse the SMILES %s" % (myrepr(smiles),))
                return id, None
            if id is not None:
                obmol.SetTitle(_to_openbabel_string(id))
            return id, obmol

        openbabel_parser.__name__ = "openbabel_" + self.register_name + "_parse_id_and_molecule"
        return openbabel_parser

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        new_reader_args["delimiter"] = delimiter.name

        if isinstance(content, str):
            first_line, _, _ = content.partition(u"\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip(u"\r"))
            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            smiles, id = terms
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None
            smiles, id_bytes = terms
            if id_bytes is None:
                id = None
            else:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id_bytes, err))
                    return None, None
                
            
        smiles = _convert_for_readstring(smiles, error_handler, None)
        if smiles is None:
            return None, None

        obmol = ob.OBMol()
        success = obconversion.ReadString(obmol, smiles)
        if not success:
            error_handler.error("Open Babel cannot parse the SMILES %s" % (myrepr(smiles),))
            return id, None
        if id is not None:
            obmol.SetTitle(id)
        return id, obmol

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        
        return base_toolkit.MoleculeWriter(
            base_toolkit.FormatMetadata(location.filename, self.record_format, new_writer_args),
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           True, obconversion, self.register_name,
                                           encoding, encoding_errors),
            location
            )

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        
        outfile, close, getvalue = io._open_output_string()
        
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           True, obconversion, self.register_name, None, None),
            getvalue,
            location
            )

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           True, obconversion, self.register_name, None, None),
            getvalue,
            location
            )

    def create_string(self, mol, id, writer_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        
        term = obconversion.WriteString(mol, False)
        if not term:
            error_handler.error("Open Babel cannot create the SMILES string")
            return None
            
        if id is None:
            id = mol.GetTitle()
        return delimiter.format_two_terms(term, id)

    def create_bytes(self, mol, id, writer_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        
        term = obconversion.WriteString(mol, False).encode("utf8")
        if not term:
            error_handler.error("Open Babel cannot create the SMILES string")
            return None
            
        if id is None:
            id = mol.GetTitle()
            
        if isinstance(id, str):
            id = id.encode("utf8")
        return delimiter.format_two_terms_bytes(term, id)
        
        
def iter_column_records(format_name, iter_reader, obconversion, error_handler, location):
    x = _iter_column_records(format_name, iter_reader, obconversion, error_handler, location)
    next(x)
    return x

def _iter_column_records(format_name, iter_reader, obconversion, error_handler, location):
    obmol = ob.OBMol()
    def get_mol():
        return obmol
    location.register(get_mol=get_mol)
    
    try:
        yield "Ready for processing!"
        for smi_string, id in iter_reader:
            if not obconversion.ReadString(obmol, smi_string):
                error_handler.error("Open Babel cannot parse the %s %s"
                                    % (format_name, myrepr(smi_string)), location)
            else:
                if id is not None:
                    obmol.SetTitle(id)
                else:
                    # Open Babel will only set the title if it finds one, which
                    # means that a previous title may still be present.
                    # Clear it, just in case.
                    obmol.SetTitle("")
                yield id, obmol
                obmol.Clear()
    finally:
        location.save(mol=None)
        obmol = None
        iter_reader.close()

# The three categories follow OEChem 2.0's lead:
#   "smi" = isomeric, canonical (alternate is "ism")
#   "can" = non-isomeric, canonical
#   "usm" = non-isomeric, non-canonical


class SmiFormat(BaseSmiFileFormat):
    register_name = "smi"
    record_format = "smi"
    alternate_extensions = ("ism", "isosmi")

    class input(object):
        arg_info = BaseSmiFileFormat.input.arg_info
        defaults = BaseSmiFileFormat.input.defaults.copy()

    class output(object):
        arg_info = BaseSmiFileFormat.output.arg_info
        defaults = BaseSmiFileFormat.output.defaults.copy()
        defaults.update({
            "isomeric": True,
            "canonicalization": "default",
            })


class CanFormat(BaseSmiFileFormat):
    register_name = record_format = "can"

    class input(object):
        arg_info = BaseSmiFileFormat.input.arg_info
        defaults = BaseSmiFileFormat.input.defaults.copy()

    class output(object):
        arg_info = BaseSmiFileFormat.output.arg_info
        defaults = BaseSmiFileFormat.output.defaults.copy()
        defaults.update({
            "isomeric": False,
            "canonicalization": "default",
            })

class UsmFormat(BaseSmiFileFormat):
    register_name = record_format = "usm"

    class input(object):
        arg_info = BaseSmiFileFormat.input.arg_info
        defaults = BaseSmiFileFormat.input.defaults.copy()

    class output(object):
        arg_info = BaseSmiFileFormat.output.arg_info
        defaults = BaseSmiFileFormat.output.defaults.copy()
        defaults.update({
            "isomeric": False,
            "canonicalization": False,
            })


class BaseSmiStringFormat(base_toolkit.NotIoFormatConfigMixin, BaseSmiFormat):
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        obmol = ob.OBMol()

        def openbabel_parser(content):
            obmol.Clear()
            
            # Follow the Daylight spec, which says that the SMILES ends with any whitespace
            terms = content.split(None, 1)
            if not terms:
                # In chemfp, the empty SMILES string returns an empty molecule.
                return None, obmol

            success = obconversion.ReadString(obmol, _to_openbabel_string(terms[0]))
            if not success:
                error_handler.error("Open Babel cannot parse the SMILES %s"
                                    % (myrepr(terms[0]),))
                return None, None

            # This is only supposed to have a SMILES. Ignore any possible id.
            return None, obmol

        openbabel_parser.__name__ = "openbabel_" + self.register_name + "_parse_id_and_molecule"
        return openbabel_parser
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)

        obmol = ob.OBMol()

        # Follow the Daylight spec, which says that the SMILES ends with any whitespace
        terms = content.split(None, 1)
        if not terms:
            # In chemfp, the empty SMILES string returns an empty molecule.
            return None, obmol

        success = obconversion.ReadString(obmol, _to_openbabel_string(terms[0]))
        if not success:
            error_handler.error("Open Babel cannot parse the SMILES %s" % (myrepr(terms[0]),))
            return None, None

        # This is only supposed to have a SMILES. Ignore any possible id.
        return None, obmol

    def create_string(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)

        content = obconversion.WriteString(mol, True)
        if not content:
            error_handler.error("Open Babel cannot create the SMILES string")
            return None
        return content




class SmiStringFormat(BaseSmiStringFormat):
    register_name = record_format = "smistring"
    
    input = BaseSmiFormat.input
    
    class output(object):
        arg_info = BaseSmiFormat.output.arg_info
        defaults = BaseSmiFormat.output.defaults.copy()
        defaults.update({
            "isomeric": True,
            "canonicalization": True,
            })

class CanStringFormat(BaseSmiStringFormat):
    register_name = record_format = "canstring"
    
    input = BaseSmiFormat.input
    
    class output(object):
        arg_info = BaseSmiFormat.output.arg_info
        defaults = BaseSmiFormat.output.defaults.copy()
        defaults.update({
            "isomeric": False,
            "canonicalization": True,
            })

class UsmStringFormat(BaseSmiStringFormat):
    register_name = record_format = "usmstring"
    
    input = BaseSmiFormat.input
    
    class output(object):
        arg_info = BaseSmiFormat.output.arg_info
        defaults = BaseSmiFormat.output.defaults.copy()
        defaults.update({
            "isomeric": False,
            "canonicalization": False,
            })
            

######

def iter_read_sdf_structures(infile, close, obconversion, id_tag_bytes, error_handler, location,
                             encoding, encoding_errors):
    assert not isinstance(id_tag_bytes, str), "must be bytes"
    it = _iter_read_sdf_structures(infile, close, obconversion, id_tag_bytes, error_handler, location,
                                   encoding, encoding_errors)
    next(it)
    return it

def _iter_read_sdf_structures(infile, close, obconversion, id_tag_bytes, error_handler, location,
                              encoding, encoding_errors):
    # Register a way to get the current molecule
    mol = ob.OBMol()
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    
    record_reader = _text_toolkit.iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                                                           encoding, encoding_errors)
    try:
        yield "Ready for processing!"
        for (id, record) in record_reader:
            try:
                tmp_record = record.decode(encoding, encoding_errors)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the content for Open Babel's ReadString(): %s"
                                    % (err,), location)
                continue
            success = obconversion.ReadString(mol, tmp_record)
            
            if success:
                yield id, mol
                mol.Clear()
            else:
                # This was not a valid record
                error_handler.error("Could not parse sdf record", location)
                mol.Clear()
            
    finally:
        location.save(mol=None)
        mol = None


class SDFFormat(OpenBabelFormatConfig):
    register_name = record_format = "sdf"
    is_available = True
    _obformat = "sdf"
    
    class input(object):
        arg_info = (
            ("implementation", parse_implementation),
            ("perceive_stereo", base_toolkit.parse_boolean), # XXX not implemented
            ("perceive_0d_stereo", base_toolkit.parse_boolean),  # XXX not implemented
            ("options", str),
            )
        defaults = {
            "implementation": "openbabel",
            "perceive_stereo": False,
            "perceive_0d_stereo": False,
            "options": None,
            }

    class output(object):
        arg_info = (
            ("always_v3000", base_toolkit.parse_boolean),
            ("include_atom_class", base_toolkit.parse_boolean),
            ("include_hcount", base_toolkit.parse_boolean),
            ("options", str),
            )
        defaults = {
            "always_v3000": False,
            "include_atom_class": False,
            "include_hcount": False,
            "options": None,
            }

    
    def _make_obconversion_for_output(self, writer_args):
        obconversion = ob.OBConversion()
        new_writer_args = {}
        if not obconversion.SetOutFormat(self.record_format):
            raise ValueError("Open Babel does not support the %r output format" % (self.record_format,))

        OUTOPTIONS = obconversion.OUTOPTIONS

        options_str = writer_args["options"]
        if options_str:
            obconversion.SetOptions(options_str, OUTOPTIONS)
            new_writer_args["options"] = options_str
        else:
            new_writer_args["options"] = None

        # Force output to be in V3000 format
        if writer_args["always_v3000"]:
            obconversion.AddOption("3", OUTOPTIONS)
            new_writer_args["always_v3000"] = True
        else:
            new_writer_args["always_v3000"] = False

        # Include atom class (will this exist in chemfp?)
        if writer_args["include_atom_class"]:
            obconversion.AddOption("a", OUTOPTIONS)
            new_writer_args["include_atom_class"] = True
        else:
            new_writer_args["include_atom_class"] = False
            

        # Always include zero bond order valence / hydrogen count
        if writer_args["include_hcount"]:
            obconversion.AddOption("H", OUTOPTIONS)
            new_writer_args["include_hcount"] = True
        else:
            new_writer_args["include_hcount"] = False

        return new_writer_args, obconversion

    # I don't like how Open Babel handles errors. I can't force it to
    # recover from errors the way that I can in my own code. On the
    # other hand, there's about a 10% overhead to use chemfp's record
    # reader. I decided to let it be user selectable, but I'm not
    # going to document it and I might change my views in the future.
    
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        implementation = reader_args["implementation"]
        if implementation in (None, "openbabel"):
            return super(SDFFormat, self).read_ids_and_molecules(
                source, compression, id_tag, reader_args, error_handler, location, encoding, encoding_errors)
            
        elif implementation == "chemfp":
            if isinstance(id_tag, str):
                id_tag_bytes = _text_toolkit.get_id_tag_bytes(id_tag, encoding, encoding_errors)
            else:
                id_tag_bytes = id_tag
            new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
            
            infile, close = io.open_compressed_input(source, compression, location, EXTERNAL_NAME)
            
            iter_reader = iter_read_sdf_structures(infile, close, obconversion, id_tag_bytes,
                                                   error_handler, location, encoding, encoding_errors)

            metadata = base_toolkit.FormatMetadata(location.filename,
                                                   self.record_format, new_reader_args)
            return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

        else:
            raise ValueError("Unsupported 'implementation' value %r" % (implementation,))
        

    def read_ids_and_molecules_from_string(self, content, compression,
                                           id_tag, reader_args, error_handler, location):
        encoding = "utf8"
        encoding_errors = "strict"
        implementation = reader_args["implementation"]
        if implementation in (None, "openbabel"):
            return super(SDFFormat, self).read_ids_and_molecules_from_string(
                content, compression, id_tag, reader_args, error_handler, location)
            
        elif implementation in "chemfp":
            if isinstance(id_tag, str):
                id_tag_bytes = _text_toolkit.get_id_tag_bytes(id_tag, encoding, encoding_errors)
            else:
                id_tag_bytes = id_tag
            new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
            is_unicode, infile, close = io._open_string_as_bytes(content, compression, location, EXTERNAL_NAME)
            iter_reader = iter_read_sdf_structures(infile, close, obconversion, id_tag_bytes,
                                                   error_handler, location, encoding, encoding_errors)

            metadata = base_toolkit.FormatMetadata(location.filename,
                                                   self.record_format, new_reader_args)
            return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

        else:
            raise ValueError("Unsupported 'implementation' value %r" % (implementation,))
        
    
class SDF3KFormat(SDFFormat):
    register_name = "sdf3k"
    class output:
        arg_info = SDFFormat.output.arg_info
        defaults = SDFFormat.output.defaults.copy()
        defaults["always_v3000"] = True
        
def gen_write_delimited_structures(outfile, close, error_handler, location, delimiter, include_id,
                                   obconversion, format_name, encoding, encoding_errors):
    x = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter, include_id,
                                        obconversion, format_name, encoding, encoding_errors)
    next(x)
    return x

def _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter, include_id,
                                    obconversion, format_name, encoding, encoding_errors):
    recno = output_recno = 0
    mol = None
    WriteString = obconversion.WriteString

    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    fmt = delimiter.get_format_string()
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        rec = WriteString(mol, True)
                        if not rec:
                            error_handler.error("Open Babel cannot create the %s string"
                                                % (format_name,), location)
                            continue

                        if include_id:
                            id = mol.GetTitle()
                            if not outfile.write_string(fmt % (rec, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(rec + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        if id is not None:
                            mol.SetTitle(id)
                        rec = WriteString(mol).strip()
                        if not rec:
                            error_handler.error("Open Babel cannot create the %s string"
                                                % (format_name,), location)
                            continue

                        if include_id:
                            if id is None:
                                id = mol.GetTitle()
                            if not outfile.write_string(fmt % (rec, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(rec + "\n", error_handler, location):
                                continue
                        output_recno += 1


                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    rec = WriteString(mol).strip()
                    if not rec:
                        error_handler.error("Open Babel cannot create the %s string"
                                            % (format_name,), location)
                        continue

                    if include_id:
                        if id is None:
                            id = mol.GetTitle()
                        if not outfile.write_string(fmt % (rec, id), error_handler, location):
                            continue
                    else:
                        if not outfile.write_string(rec + "\n", error_handler, location):
                            continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]

    finally:
        if close is not None:
            close()
        # Break the cyclical references
        location.save(recno=get_recno(), output_recno=get_output_recno())


# InChI helper class
class _OptionalIdReader(object):
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        new_reader_args["delimiter"] = delimiter.name
        
        infile, close = io.open_compressed_input(source, compression, location, EXTERNAL_NAME)
        
        iter_reader = _text_toolkit.iter_first_two_delimited_fields_as_strings(
            infile, close, delimiter, False,
            error_handler, location, "InChI",
            encoding, encoding_errors)
        
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args),
            iter_column_records("InChI", iter_reader, obconversion, error_handler, location),
            location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                           id_tag, reader_args, error_handler, location):
        is_unicode, infile, close = io._open_string_as_bytes(content, compression, location, EXTERNAL_NAME)
        
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args) # Process "options"
        new_reader_args["delimiter"] = delimiter.name
        iter_reader = _text_toolkit.iter_first_two_delimited_fields_as_strings(
            infile, close, delimiter, False,
            error_handler, location, "InChI",
            "utf8", "strict")
        
        return base_toolkit.IdAndMoleculeReader(
            base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args),
            iter_column_records("InChI", iter_reader, obconversion, error_handler, location),
            location)


    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        obmol = ob.OBMol()

        def openbabel_parser(content):
            obmol.Clear()

            if isinstance(content, str):
                first_line, _, _ = content.partition(u"\n") # get the first record from the input
                terms = delimiter.split_two_terms_string(first_line.rstrip(u"\r"))
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input input")
                    return None, None
                inchi, id = terms
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input input")
                    return None, None
                inchi, id_bytes = terms
                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode("utf8")
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s"
                                            % (id_bytes, err))
                        return None, None
                    
            success = obconversion.ReadString(obmol, _convert_for_readstring(inchi, error_handler, None))
            if not success:
                error_handler.error("Open Babel cannot parse the InChI %s" % (myrepr(inchi),))
                return id, None
            if id is None:
                # Reset the title in case the previous instance had a title and this one doesn't
                obmol.SetTitle("")
            else:
                obmol.SetTitle(id)
            return id, obmol

        openbabel_parser.__name__ = "openbabel_" + self.register_name + "_parse_id_and_molecule"
        return openbabel_parser
            

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)

        if isinstance(content, str):
            first_line, _, _ = content.partition(u"\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip(u"\r"))
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi, id = terms
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi, id_bytes = terms
            if id_bytes is None:
                id = None
            else:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier: %s" % (err,))
                    return None, None
            
        obmol = ob.OBMol()
        
        success = obconversion.ReadString(obmol, _convert_for_readstring(inchi, error_handler, None))
        if not success:
            error_handler.error("Open Babel cannot parse the InChI %s" % (myrepr(inchi),))
            return id, None
        if id is not None:
            obmol.SetTitle(id)
        return id, obmol
    

class _OptionalIdWriter(object):
    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])
        
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           include_id, obconversion, self.inchi_name,
                                           encoding, encoding_errors),
            location
            )

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           include_id, obconversion, self.inchi_name, None, None),
            getvalue,
            location
            )

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        new_writer_args["delimiter"] = delimiter
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)

        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                           include_id, obconversion, self.inchi_name, None, None),
            getvalue,
            location
            )
    
    def create_string(self, mol, id, writer_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        term = obconversion.WriteString(mol, True)
        if not term:
            error_handler.error("Open Babel cannot create the %s string" % (self.inchi_name,))
            return None

        if writer_args["include_id"]:
            if id is None:
                id = mol.GetTitle()
            return delimiter.format_two_terms(term, id)
        else:
            return term + u"\n"

    def create_bytes(self, mol, id, writer_args, error_handler):
        s = self.create_string(mol, id, writer_args, error_handler)
        if s is None:
            return None
        return s.encode("utf8")



_inchi_input_arg_info = (
    ("options", str),
    )
_inchi_input_defaults = {
    "options": None,
    }

_inchi_output_arg_info = (
    ("FixedHLayer", base_toolkit.parse_boolean),
    ("ReconnectedMetals", base_toolkit.parse_boolean),
    ("truncspec", str),
    ("options", str),
    )
_inchi_output_defaults = {
    "FixedHLayer": False,
    "ReconnectedMetals": False,
    "truncspec": None,
    "options": None,
    }

def _make_obconversion_for_inchi_output(writer_args):
    obconversion = ob.OBConversion()
    new_writer_args = {}
    if not obconversion.SetOutFormat("inchi"):
        raise ValueError("Open Babel does not support the 'inchi' output format")

    OUTOPTIONS = obconversion.OUTOPTIONS

    options_str = writer_args["options"]
    if options_str:
        obconversion.SetOptions(options_str, OUTOPTIONS)
        new_writer_args["options"] = options_str
    else:
        new_writer_args["options"] = None

    if writer_args["FixedHLayer"]:
        obconversion.AddOption("F", OUTOPTIONS)
        new_writer_args["FixedHLayer"] = True
    else:
        new_writer_args["FixedHLayer"] = False

    if writer_args["ReconnectedMetals"]:
        obconversion.AddOption("M", OUTOPTIONS)
        new_writer_args["ReconnectedMetals"] = True
    else:
        new_writer_args["ReconnectedMetals"] = False
        

    truncspec = writer_args["truncspec"]
    if truncspec:
        obconversion.AddOption("T", OUTOPTIONS, truncspec)
        new_writer_args["truncspec"] = True
    else:
        new_writer_args["truncspec"] = False

    return new_writer_args, obconversion

def _make_obconversion_for_inchikey_output(writer_args):
    new_writer_args, obconversion = _make_obconversion_for_inchi_output(writer_args)

    # Compute InChI key
    obconversion.AddOption("K", obconversion.OUTOPTIONS)

    return new_writer_args, obconversion

class InChIFormat(_OptionalIdReader, _OptionalIdWriter, OpenBabelFormatConfig):
    register_name = record_format = "inchi"
    _obformat = "inchi"
    inchi_name = "InChI"

    class input(object):
        arg_info = _inchi_input_arg_info + (
            ("delimiter", base_toolkit.parse_delimiter),
            )
        defaults = _inchi_input_defaults.copy()
        defaults.update({
            "delimiter": None,
            "options": None,
            })

    class output(object):
        arg_info = _inchi_output_arg_info + (
            ("delimiter", base_toolkit.parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = _inchi_output_defaults.copy()
        defaults.update({
            "delimiter": None,
            "include_id": True,
            })

    def _make_obconversion_for_output(self, writer_args):
        return _make_obconversion_for_inchi_output(writer_args)


class InChIStringFormat(OpenBabelFormatConfig):
    register_name = record_format = "inchistring"
    _obformat = "inchi"
    inchi_name = "InChI"
    
    supports_io = False

    class input(object):
        arg_info = _inchi_input_arg_info
        defaults = _inchi_input_defaults.copy()

    class output(object):
        arg_info = _inchi_output_arg_info
        defaults = _inchi_output_defaults.copy()

    read_ids_and_molecules = None
    read_ids_and_molecules_from_string = None

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        obmol = ob.OBMol()

        def openbabel_inchistring_parse_id_and_molecule(content):
            obmol.Clear()
            terms = content.split(None, 1)
            if not terms or not terms[0]:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi = terms[0]
            success = obconversion.ReadString(obmol, _to_openbabel_string(inchi))
            if not success:
                error_handler.error("Open Babel cannot parse the InChI %s" % (myrepr(inchi),))
                return None, None
            return None, obmol
        return openbabel_inchistring_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        new_reader_args, obconversion = self._make_obconversion_for_input(reader_args)
        obmol = ob.OBMol()
        terms = content.split(None, 1)
        if not terms or not terms[0]:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
        inchi = terms[0]
        success = obconversion.ReadString(obmol, _to_openbabel_string(inchi))
        if not success:
            error_handler.error("Open Babel cannot parse the InChI %s" % (myrepr(inchi),))
            return None, None
        return None, obmol

    def _make_obconversion_for_output(self, writer_args):
        return _make_obconversion_for_inchi_output(writer_args)

    def create_string(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        result = obconversion.WriteString(mol, True)
        if not result:
            error_handler.error("Open Babel cannot create the InChI string")
            return None
        if isinstance(result, str):
            return result
        return result.decode("utf8")

    def create_bytes(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        result = obconversion.WriteString(mol, True)
        if not result:
            error_handler.error("Open Babel cannot create the InChI string")
            return None
        if isinstance(result, str):
            return result.encode("utf8")
        return result
    
        

class InChIKeyFormat(_OptionalIdWriter, OpenBabelFormatConfig):
    register_name = record_format = "inchikey"
    _obformat = "inchikey"
    inchi_name = "InChIKey"
    
    input = None

    class output(object):
        arg_info = _inchi_output_arg_info + (
            ("delimiter", base_toolkit.parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = _inchi_output_defaults.copy()
        defaults.update({
            "delimiter": None,
            "include_id": True,
            })

    def _make_obconversion_for_output(self, writer_args):
        return _make_obconversion_for_inchikey_output(writer_args)

    
class InChIKeyStringFormat(OpenBabelFormatConfig):
    register_name = record_format = "inchikeystring"
    _obformat = "inchikey"
    inchi_name = "InChIKey"
    supports_io = False

    input = None

    class output(object):
        arg_info = _inchi_output_arg_info
        defaults = _inchi_output_defaults.copy()

    def _make_obconversion_for_output(self, writer_args):
        return _make_obconversion_for_inchikey_output(writer_args)

    def create_string(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        term = obconversion.WriteString(mol, True)
        if not term:
            error_handler.error("Open Babel cannot create the InChIKey string")
            return None
        if isinstance(term, str):
            return term
        return term.decode("utf8")
    
    def create_bytes(self, mol, id, writer_args, error_handler):
        new_writer_args, obconversion = self._make_obconversion_for_output(writer_args)
        term = obconversion.WriteString(mol, True)
        if not term:
            error_handler.error("Open Babel cannot create the InChIKey string")
            return None
        if isinstance(term, str):
            return term.encode("utf8")
        return term


#########

def iter_generic_records(obconversion, obmol, success, id_tag, error_handler, location):
    x = _iter_generic_records(obconversion, obmol, success, id_tag, error_handler, location)
    next(x)
    return x

def _iter_generic_records(obconversion, obmol, success, id_tag, error_handler, location):
    recno = 0
    def get_recno():
        return recno
    def get_mol():
        return obmol

    location.register(get_mol=get_mol, get_recno=get_recno)

    yield # Ready for processing

    try:
        # How do I detect if the input contains a failure?
        if id_tag is None:
            while success:
                recno += 1
                try:
                    id = obmol.GetTitle()
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the record: %s" % (err,), location)
                else:
                    yield id, obmol
                obmol.Clear()
                success = obconversion.Read(obmol)

        else:
            while success:
                recno += 1
                obj = obmol.GetData(id_tag)
                if obj is None:
                    id = None
                else:
                    id = obj.GetValue()
                yield id, obmol
                obmol.Clear()
                success = obconversion.Read(obmol)
    finally:
        location.save(recno=recno)
        obmol = None


#################### register the formats

format_registry = base_toolkit.FormatRegistry(EXTERNAL_NAME, "openbabel")

format_registry.register_formats(
    SmiFormat(),
    CanFormat(),
    UsmFormat(),

    SmiStringFormat(),
    CanStringFormat(),
    UsmStringFormat(),

    SDFFormat(),
    SDF3KFormat(),
    
    InChIFormat(),
    InChIKeyFormat(),
    InChIStringFormat(),
    InChIKeyStringFormat(),
    )

# Add in the other formats

_input_formats = set(format_def.split()[0] for format_def in _obconv.GetSupportedInputFormat())
_output_formats = set(format_def.split()[0] for format_def in _obconv.GetSupportedOutputFormat())

for format_name in _input_formats | _output_formats:
    if format_name in format_registry:
        continue
    format_registry.register_formats(GenericFormat(format_name,
                                                   format_name in _input_formats,
                                                   format_name in _output_formats))

class OpenBabelToolkit(base_toolkit.BaseToolkit):
    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        # Open the file and pass to the format
        exc = None
        try:
            outfile, close = io.open_compressed_output_wrapped(destination, compression, EXTERNAL_NAME,
                                                               encoding, encoding_errors, level)
        except ValueError as err:
            exc = ValueError("Open Babel does not support compression type %r: %s" % (
                compression, err))
        if exc is not None:
            raise exc

        return format_config.open_molecule_writer(destination, outfile, close, writer_args,
                                                  error_handler, location, encoding, encoding_errors)


    def _read_ids_and_molecules_from_string_impl(
            self, format_config, content, compression, id_tag, reader_args, error_handler, location):

        if isinstance(content, str):
            if compression:
                raise ValueError("Compression does not work when the content is a Unicode string. "
                                 "Use a byte string instead.")
            if id_tag is not None and not isinstance(id_tag, str):
                raise ValueError("id_tag must be a string, not bytes, when the content is a string")

        return format_config.read_ids_and_molecules_from_string(content, compression, id_tag, reader_args,
                                                                error_handler, location)
        
        
    
toolkit = OpenBabelToolkit(format_registry)

