# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""Low-level fingerprint functions and global configuration.

The ``bitops`` module contains functions that work on byte-encoded
and/or hex-encoded fingerprints, such as to compute the Tanimoto
between two byte fingerprints, or to count the number of bits in a
fingerprint.

It also contains functions to change the internal configuration of how
chemfp does it's bit-wise operations, and to report that
configuration. Currently only two of those are part of the public API.
"""

import os
import sys

from . import libchemfp_core as _core
from . import _bitops

from ._bitops import (
    hex_isvalid, hex_popcount, hex_intersect_popcount, hex_union_popcount, hex_xor_popcount,
    hex_tanimoto, hex_tversky, hex_contains, hex_contains_bit,
    hex_encode, hex_encode_as_bytes, hex_decode,
    hex_from_bitlist, hex_to_bitlist,
    hex_intersect, hex_union, hex_xor,
    byte_popcount, byte_intersect_popcount, byte_union_popcount, byte_xor_popcount,
    byte_tanimoto, byte_tversky, byte_contains, byte_contains_bit,
    byte_hex_tanimoto, byte_hex_tversky,
    byte_from_bitlist, byte_to_bitlist,
    byte_intersect, byte_union, byte_xor)

    
__all__ = ["byte_popcount", "byte_intersect_popcount",
           "byte_union_popcount", "byte_xor_popcount",
           "byte_tanimoto", "byte_tversky", "byte_contains", "byte_contains_bit",
           "byte_to_bitlist", "byte_from_bitlist",
           "byte_intersect", "byte_union", "byte_xor",
           "hex_isvalid", "hex_popcount", "hex_intersect_popcount", 
           "hex_tanimoto", "hex_tversky", "hex_contains", "hex_contains_bit",
           "byte_hex_tanimoto", "byte_hex_tversky",
           "hex_to_bitlist", "hex_from_bitlist",
           "hex_intersect", "hex_union", "hex_xor",
           "hex_encode", "hex_encode_as_bytes", "hex_decode",
           "get_tanimoto_precision",
           # Don't export the popcount and alignment-related functions
               ]

def byte_difference(fp1, fp2):
    "Deprecated equivalent to byte_xor"
    import warnings
    warnings.warn(
        "byte_difference() has been renamed to byte_xor(). "
        "This compatibility function will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2)
    return byte_xor(fp1, fp2)

def hex_difference(fp1, fp2):
    "Deprecated equivalent to hex_xor"
    import warnings
    warnings.warn(
        "hex_difference() has been renamed to hex_xor(). "
        "This compatibility function will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2)
    return hex_xor(fp1, fp2)
    
    
#hex_decode = binascii.unhexlify

# What's the minimum number of decimal digits needed to distinguish
# between any two Tanimoto scores with N bits?
# 
# I determined these values by enumeration.
_precision_table = [0, 5, 13, 35, 108, 320, 1023, 3201, 10071, 31743, 100224]
# The ratio trends towards 3.157 so the next number is over 300672.
# It doesn't make sense to support more than ~16K bits, so I'll
# stop at 2**18 = 262144 bits.

def get_tanimoto_precision(num_bits: float) -> int:
    """Return the minimum precision needed to distinguish Tanimoto all values with `num_bit` bits

    Given two Tanimoto values from fingerprints of length `num_bits`,
    stored as a Python 64-bit float, how many decimal digits are needed
    to ensure they are distinct?

    For example, for 2048-bit fingerprints you need at least 7 digits::

      >>> "'%.6f' and '%.6f'" % (1/1023, 1/1022)
      "'0.000978' and '0.000978'"
      >>> "'%.7f' and '%.7f'" % (1/1023, 1/1022)
      "'0.0009775' and '0.0009785'"

    This function returns the minumum number of required decimial
    digits, given 1 <= `num_bits` <= 2**18. For example::

      >>> bitops.get_tanimoto_precision(2048)
      7

    This might be used as ('%.7f' % score) or f'{score:.7f}'.

    :param num_bits: The number of bits in the fingerprint
    :type num_bits: integer between 1 and 2**18
    :returns: the precision, as an integer
    """
    if not (1 <= num_bits <= 2**18):
       raise ValueError("num_bits must be between 1 and 2**18, inclusive")
    import bisect
    return bisect.bisect(_precision_table, num_bits)

_sizes = None
def get_sizes():
    global _sizes
    if _sizes is None:
        _sizes = tuple(_bitops.get_size_name(i) for i in range(_bitops.get_num_size_implementations()))
    return _sizes

_methods = None
def get_methods():
    global _methods
    if _methods is None:
        _methods = tuple(_bitops.get_method_name(i) for i in range(_bitops.get_num_methods()))
    return _methods

def get_method_families():
    return [_bitops.get_method_family_name(i) for i in range(_bitops.get_num_method_families())]

def get_size_methods():
    from collections import OrderedDict
    size_methods = OrderedDict()
    for i in range(_bitops.get_num_size_implementations()):
        size_methods[_bitops.get_size_name(i)] = _bitops.get_size_method_name(i)
    return size_methods

def _get_size_index(size_name):
    try:
        return get_sizes().index(size_name)
    except ValueError:
        pass
    raise ValueError("Unknown size_name: %r" % (size_name,))
    
def _get_method_name_index(method_name):
    try:
        return get_methods().index(method_name)
    except ValueError:
        pass
    raise ValueError("Unknown method_name: %r" % (method_name,))
    
def set_size_method(size_name, method_name):
    size_name_index = _get_size_index(size_name)
    method_name_index = _get_method_name_index(method_name)
    _bitops.set_size_method(size_name_index, method_name_index)

def get_options():
    return [_bitops.get_option_name(i) for i in range(_bitops.get_num_options())]

def get_option(option):
    return _bitops.get_option(option)

def set_option(option, value):
    _bitops.set_option(option, value)

def print_report(out=sys.stdout):
    """Print the configuration report to the given file (default: stout)"""
    import platform
    from . import SOFTWARE
    print("== Configuration report for", SOFTWARE, "==", file=out)
    print("Python version:", sys.version.replace("\n", " / "), file=out)
    print("Platform:", platform.platform(), file=out)
    print("Chemfp compiled with:", _core.get_compiler(), file=out)
    print("Number of available threads:", _core.get_num_threads(), file=out)
    print("Available method families:", " ".join(get_method_families()), file=out)
    print("Available methods:", " ".join(get_methods()), file=out)
    print("Size methods:", file=out)
    for size_name, method_name in get_size_methods().items():
        print("  %s: %s" % (size_name, method_name), file=out)
    print("Option settings:", file=out)
    for option in get_options():
        print("  %s: %s" % (option, _bitops.get_option(option)), file=out)

def use_environment_variables(environ=None, outfile=sys.stderr):
    r"""Set the chemfp configuration using environment variables or a dictionary

    By default, process os.environ to find chemfp environment variables
    (which all start with "CHEMFP\_") and use them to configure chemfp
    internals.

    This is meant to be used by any program which wants the same
    configuration system as the core chemfp components.
    """
    method_name_to_index = dict((name, i) for i, name in enumerate(get_methods()))
    size_name_to_index = dict((name, i) for i, name in enumerate(get_sizes()))
    
    if environ is None:
        environ = os.environ

    known = set()
    
    known.add("CHEMFP_LICENSE")
    known.add("CHEMFP_LICENSE_MANAGER")
    known.add("CHEMFP_LICENSE_FILE")
    known.add("CHEMFP_REPORT")
    known.add("CHEMFP_GZIP")
    known.add("CHEMFP_GZCAT")
    known.add("CHEMFP_USE_SYSTEM_GZIP")
    known.add("CHEMFP_PROGRESSBAR")

    # used to configure compilation
    known.add("CHEMFP_OPENMP")
    known.add("CHEMFP_SSSE3")
    known.add("CHEMFP_AVX2")
    
    report = environ.get("CHEMFP_REPORT", "0") == "1"

    # CHEMFP_FAMILIES is a comma-separated list of popcount family names.
    # They are set in the given order, so "popcnt,avx2" sets all of the
    # popcnt implementations then sets all of the avx2 implementations.
    
    name = "CHEMFP_FAMILIES"
    value = environ.get(name, "")
    if value:
        known.add(name)
        if report:
            print("INFO: Setting method families using CHEMFP_FAMILIES=%r" % (value,),
                  file=outfile)
        family_names = value.split(",")

        expanded_family_names = []
        for family_name in family_names:
            if family_name == "all":
                expanded_family_names.extend(get_method_families())
            elif not family_name:
                pass
            else:
                expanded_family_names.append(family_name)

        family_names = expanded_family_names
        for family_name in family_names:
            try:
                _bitops.use_method_family(family_name)
            except ValueError as err:
                print("WARNING: Cannot use method family %r: %s"
                      % (family_name, str(err)), file=outfile)
            else:
                if report:
                    print("INFO: Applied methods from family %s" % (family_name,),
                           file=outfile)

    # Low-level control over each possible size.
    # CHEMFP_SIZE128_18 changes the method for "size128-18"
    for size_name in get_sizes():
        name = "CHEMFP_" + size_name.upper().replace("-", "_")
        known.add(name)
        try:
            method_name = environ[name]
        except KeyError:
            pass
        else:
            size_index = size_name_to_index[size_name]
            try:
                method_index = method_name_to_index[method_name]
            except KeyError:
                print("WARNING: Unable to use $%s = %r: %s" % (
                    (name, method_name, "unsupported method")),
                    file=outfile)
            else:
                try:
                    _bitops.set_size_method(size_index, method_index)
                except ValueError as err:
                    print("WARNING: Unable to use $%s = %r: %s" % (
                        (name, method_name, err)),
                        file=outfile)

    # using CHEMFP_REPORT enables CHEMFP_REPORT_POPCOUNT,
    # CHEMFP_REPORT_INTERSECT, and CHEMFP_REPORT_ALGORITHM
    if report:
        set_option("report-popcount", 1)
        set_option("report-intersect", 1)
        set_option("report-algorithm", 1)

    # Options can also be configured individually
    for option in get_options():
        name = "CHEMFP_" + option.upper().replace("-", "_")
        known.add(name)
        try:
            value = environ[name]
        except KeyError:
            pass
        else:
            try:
                int_value = int(value)
            except ValueError:
                print("WARNING: Unable to use $%s = %r: not an integer value" % (
                      (name, value)), file=outfile)
                continue
            
            try:
                set_option(option, int_value)
            except ValueError as err:
                print("WARNING: Unable to use $%s = %r: %s" % (
                      (name, value, err)), file=outfile)

    known.add("CHEMFP_PRINT_CONFIG")
    if (environ.get("CHEMFP_PRINT_CONFIG", "0") == "1" or report):
        print_report(outfile)


    for k in environ:
        if not k.startswith("CHEMFP_"):
            continue
        if k not in known:
            print("WARNING: Unknown chemfp environment variable %r" % (k,),
                  file=outfile)

if __name__ == "__main__":
    use_environment_variables()
    print_report()
