# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from . import SOFTWARE as CHEMFP_SOFTWARE
from . import pattern_fingerprinter
from . import types
from . import cdk_types
from . import bitops

from . import cdk_toolkit

from ._cdk_toolkit import cdk, java # import through _cdk_toolkit

SOFTWARE = cdk_toolkit.software + " " + CHEMFP_SOFTWARE

ISAROMATIC = cdk.CDKConstants.ISAROMATIC

_no_param_call = types.FingerprintArguments().get_call_method()

class SmartsMatcher(object):
    def __init__(self, pat):
        self.pat = pat
        self.matches = pat.matches # incorporate the underlying 'matches()'

    def matches(self, mol):
        return self.pat.matches(mol)

    def countMatches(self, mol, largest_count, rings):
        # I need unique matches.
        # I don't know how to do that with CDK so do it myself.
        unique = set()
        for match in self.pat.matchAll(mol):
            unique.add(tuple(sorted(match)))
        return len(unique)
                                    

class HydrogenMatcher(object):
    def matches(self, mol):
        for atom in mol.atoms():
            if atom.getAtomicNumber() == 1:
                return 1
            if atom.getImplicitHydrogenCount():
                return 1
        return 0

    def countMatches(self, mol, largest_count, rings):
        num_hydrogens = 0
        for atom in mol.atoms():
            if atom.getAtomicNumber() == 1:
                num_hydrogens += 1
            num_hydrogens += atom.getImplicitHydrogenCount()
            if num_hydrogens >= largest_count:
                return num_hydrogens
        return num_hydrogens
        

def _find_sssr_rings(mol):
    return list(cdk.graph.Cycles.sssr(mol).toRingSet().atomContainers())

class AromaticRings(object):
    def __init__(self):
        # The single ring case is easy; if there's an aromatic atom in a ring
        # then there's a ring
        self._single_matcher = cdk.smiles.smarts.SmartsPattern.create("[aR]")

    def matches(self, mol):
        return self._single_matcher.matches(mol)

    def countMatches(self, mol, largest_count, rings):
        nArom = 0
        for ring in rings:
            if all(bond.getFlag(ISAROMATIC) for bond in ring.bonds()):
                nArom += 1
                if nArom == largest_count:
                    return nArom
        return nArom

def _is_hetereo_aromatic_atom(atom):
    return atom.GetIsAromatic() and atom.GetAtomicNum() not in (1, 6)

class HeteroAromaticRings(object):
    def __init__(self):
        # In the single match case, if there's an aromatic non-carbon atom
        # then it's a hetereo ring
        self._single_matcher = cdk.smiles.smarts.SmartsPattern.create("[aR;!#6]")

    def has_match(self, mol):
        return self._single_matcher.matches(mol)

    def countMatches(self, mol, largest_count, rings):
        nArom = 0
        for ring in rings:
            if (any(atom.getAtomicNumber() != 6 for atom in ring.atoms()) and
                all(bond.getFlag(ISAROMATIC) for bond in ring.bonds())):
                nArom += 1
                if nArom == largest_count:
                    return nArom
        return nArom


ConnectedComponents = cdk.graph.ConnectedComponents
GraphUtil = cdk.graph.GraphUtil
class NumFragments(object):
    def has_match(self, mol):
        return mol.GetNumAtoms() > 0
    def countMatches(self, mol, largest_count, rings):
        return ConnectedComponents(GraphUtil.toAdjList(mol)).nComponents()


_pattern_classes = {
    "<H>": HydrogenMatcher,
    "<aromatic-rings>": AromaticRings,
    "<hetero-aromatic-rings>": HeteroAromaticRings,
    "<fragments>": NumFragments,
    }

def cdk_compile_pattern(pattern, max_count):
    if pattern in _pattern_classes:
        return _pattern_classes[pattern]()
    
    elif pattern.startswith("<"):
        raise NotImplementedError(pattern)
        #return NotImplemented

    # Everything else must be a SMARTS pattern
    return SmartsMatcher(cdk.smiles.smarts.SmartsPattern.create(pattern))
    
class CDKPatternFingerprinter(pattern_fingerprinter.PatternFingerprinter):
    def __init__(self, patterns):
        assert patterns is not None
        super(CDKPatternFingerprinter, self).__init__(patterns, cdk_compile_pattern)
        
    def fingerprint(self, mol):
        bitnos = []
        # Compute only once
        rings = _find_sssr_rings(mol)
        for matcher, largest_count, count_info_tuple in self.matcher_definitions:
            #print matcher, largest_count, count_info_tuple
            if largest_count == 1:
                if matcher.matches(mol):
                    count_info = count_info_tuple[0]
                    bitnos.append(count_info.bit)
            else:
                # There's no good way to get the number of unique
                # matches without iterating over all of them.
                num_matches = matcher.countMatches(mol, largest_count, rings)
                for count_info in count_info_tuple:
                    if num_matches >= count_info.count:
                        bitnos.append(count_info.bit)
                    else:
                        break
        return bitops.byte_from_bitlist(bitnos, self.num_bytes*8)

class _CachedFingerprinters(dict):
    def __missing__(self, name):
        patterns = pattern_fingerprinter._load_named_patterns(name)
        fingerprinter = CDKPatternFingerprinter(patterns)
        self[name] = fingerprinter
        return fingerprinter
_cached_fingerprinters = _CachedFingerprinters()

class SubstructCDKFingerprinter_v1(types.NoFingerprintParametersMixin, cdk_types.CDKBaseFingerprintType):
    """chemfp's Substruct fingerprint implementation for CDK, version 1

    WARNING: these fingerprints have not been validated.

    The Substruct fingerprints are CACTVS/PubChem-like fingerprints designed
    for use across multiple toolkits.

    The ChemFP-Substruct-CDK/1 :class:`.FingerprintType` has no parameters.
    """
    name = "ChemFP-Substruct-CDK/1"
    num_bits = 881
    software = SOFTWARE
    def _make_fingerprinter(self):
        return _cached_fingerprinters["substruct"].fingerprint
    __call__ = _no_param_call
    
class RDMACCSCDKFingerprinter_v1(types.NoFingerprintParametersMixin, cdk_types.CDKBaseFingerprintType):
    """chemfp's RDMACCS fingerprint implementation for CDK , version 1

    The RDMACSS keys are MACCS-166-like fingerprints based on RDKit's
    MACCS116 definition, but designed to be (slightly) more portable
    across multiple chemistry toolkits.

    This version does not define key 44.

    The RDMACSS-CDK/1 :class:`.FingerprintType` has no parameters.
    """
    name = "RDMACCS-CDK/1"
    num_bits = 166
    def _make_fingerprinter(self):
        return _cached_fingerprinters["rdmaccs"].fingerprint
    __call__ = _no_param_call

class RDMACCSCDKFingerprinter_v2(types.NoFingerprintParametersMixin, cdk_types.CDKBaseFingerprintType):
    """chemfp's RDMACCS fingerprint implementation for CDK, version 2

    The RDMACSS keys are MACCS-166-like fingerprints based on RDKit's
    MACCS116 definition, but designed to be (slightly) more portable
    across multiple chemistry toolkits.

    This version defines key 44.

    The RDMACSS-CDK/2 :class:`.FingerprintType` has no parameters.
    """
    name = "RDMACCS-CDK/2"
    num_bits = 166
    def _make_fingerprinter(self):
        return _cached_fingerprinters["rdmaccs2"].fingerprint
    __call__ = _no_param_call

def register_rdmaccs(registry):
    registry.register(RDMACCSCDKFingerprinter_v1)
    registry.register(RDMACCSCDKFingerprinter_v2)

def register_substruct(registry):
    registry.register(SubstructCDKFingerprinter_v1)
