# Copyright (c) 2020-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

"""The chemfp toolkit API wrapper for the CDK toolkit.

This module is also available as ``chemfp.cdk``.
"""

import jpype # Must install JPype so chemfp can interface to the CDK jar

if hasattr(jpype, "sphinx_mock"):
    # This should only happen when we're building the documentation using
    # Sphinx's autodoc with a mock for jpype.  Sphinx's magic mock isn't good
    # enough so provide a concrete mock which is just barely good enough to let
    # autodoc import the module to extract the docstrings.
    class _tk:
        SOFTWARE = "CDK (using sphinx)"
        toolkit = None
        class format_registry:
            @staticmethod
            def get_format(format):
                return None
else:
    from . import _cdk_toolkit as _tk
            
del jpype # Don't need jpype in this module


__all__ = []
def _public(f):
    __all__.append(f.__name__)
    return f

name = "cdk"
software = _tk.SOFTWARE
is_available = True

@_public
def is_licensed():
    """Return True - CDK is always licensed

    :returns: True
    """
    return True


_format_registry = _tk.format_registry
_toolkit = _tk.toolkit

# format API

@_public
def get_formats(include_unavailable=False):
    """Get the list of structure formats that CDK supports

    If *include_unavailable* is True then also include CDK formats
    which aren't available to this specific version of CDK.

    :param include_unavailable: include unavailable formats?
    :type include_unavailable: True or False
    :returns: a list of Format objects
    """
    return _format_registry.get_formats(include_unavailable)

@_public
def get_input_formats():
    """Get the list of supported CDK input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_formats()

@_public
def get_output_formats():
    """Get the list of supported CDK output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_formats()

@_public
def get_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not available.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_format(format)

@_public
def get_input_format(format):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not an input format.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_format(format)

@_public
def get_output_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not an output format.
    
    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_format(format)

@_public
def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.
    
    :param source: the structure data source.
    :type source: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format_from_source(source, format)

@_public
def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: The structure data source.
    :type destination: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format_from_destination(destination, format)

# read/write API

@_public
def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads CDK molecules from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Note: the reader returns a new CDK molecule each time.

    The *reader_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * has_header - True or False
      * sanitize - True or default sanitizes; False for unsanitized processing

    * InChI
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing
      
    * SDF
    
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * strictParsing - True or default for strict parsing; False for lenient parsing

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If None
    then a default Location will be created.

    See :func:`chemfp.cdk_toolkit.read_ids_and_molecules` if you want (id, molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader parameters passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location,
                                   encoding, encoding_errors)

@_public
def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads CDK molecules from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.cdk_toolkit.read_molecules` for details about the other
    parameters.  See :func:`chemfp.cdk_toolkit.read_ids_and_molecules_from_string`
    if you  want to read (id, CDK molecule) pairs instead of just molecules.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

@_public
def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None, encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, CDK molecule) pairs from a structure file

    See :func:`chemfp.cdk_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, CDK molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, CDK molecule) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location,
                                           encoding, encoding_errors)

@_public
def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, CDK molecule) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.cdk_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.cdk_toolkit.read_molecules_from_string`
    if you just want to read the CDK molecules instead of (id, molecule) pairs.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, CDK molecule) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

@_public
def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, CDK molecule) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. However, I haven't
    really noticed much of a performance difference between this and
    :func:`chemfp.cdk_toolkit.parse_id_and_molecule` so you can probably
    so I suggest you use that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.cdk_toolkit.read_molecules` for details about the
    other parameters.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, CDK molecule)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

@_public
def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return a CDK molecule.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.cdk_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.cdk_toolkit.parse_id_and_molecule`
    if you want the (id, CDK molecule) pair instead of just the molecule.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a CDK molecule
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

@_public
def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, CDK molecule) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.cdk_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.cdk_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.cdk_toolkit.parse_molecule`
    if just want the CDK molecule and not the the (id, CDK molecule) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an (id, CDK molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)


@_public
def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert a CDK molecule into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: a CDK molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a Unicode string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

@_public
def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert a CDK molecule into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: a CDK molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a byte string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)


@_public
def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write CDK molecules to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    a CDK molecule, a CDK molecule iterator, or an (id, CDK molecule) pair
    iterator to a file.
    
    Molecules are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.
    
    The *writer_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * isomericSmiles - True to generate isomeric SMILES
      * kekuleSmiles - True to generate SMILES in Kekule form
      * canonical - True to generate a canonical SMILES
      * allBondsExplicit - True to write explict '-' and ':' bonds, even if they can be inferred; default is False
      * allHsExplicit - True to write explicit hydrogen counts; default is False
      * cxsmiles - True to include CXSMILES annotations; default is False

    InChI and InChIKey
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * include_id - True or default to include the id as the second column; False has no id column
      * options - an options string passed to the underlying InChI library
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing

    SDF
    
      * includeStereo - True include stereo information; False or default does not
      * kekulize - True or default creates the connection table with bonds in Kekeule form
      * v3k - True to always export in V3000 format

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer parameters passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

@_public
def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a string.

    See :func:`chemfp.cdk_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)

@_public
def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a text string.

    See :func:`chemfp.cdk_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location, level)


# molecule manipulation API

@_public
def copy_molecule(mol):
    """Return a new CDK molecule which is a copy of the given molecule

    :param mol: the molecule to copy
    :type mol: a CDK molecule
    :returns: a new CDK Mol instance
    """
    return mol.clone()

@_public
def add_tag(mol, tag, value):
    """Add an SD tag value to the CDK molecule

    :param mol: the molecule
    :type mol: a CDK molecule
    :param tag: the SD tag name
    :type tag: string
    :param value: the text for the tag
    :type value: string
    :returns: None
    """
    mol.setProperty(tag, value)

@_public
def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    :param mol: the molecule
    :type mol: a CDK molecule
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    return mol.getProperty(tag)
    
@_public
def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the molecule

    :param mol: the molecule
    :type mol: a CDK molecule
    :returns: a list of (string name, string value) pairs
    """
    it = ((str(x.getKey()), str(x.getValue())) for x in mol.getProperties().items())
    return [(name, value) for (name, value) in it if name[:4] != "cdk:"]

@_public
def get_id(mol):
    """Get the molecule's id from CDK's "cdk:Title" property

    :param mol: the molecule
    :type mol: a CDK molecule
    :returns: a string
    """
    id = mol.getProperty(_tk.TITLE)
    if id is not None:
        id = str(id)
    return id

@_public
def set_id(mol, id):
    """Set the molecule's id as CDK's "cdk:Title" property

    :param mol: the molecule
    :type mol: a CDK molecule
    :param id: the new id
    :type id: string
    :returns: None
    """
    mol.setProperty(_tk.TITLE, id)

# accessor aliases for the default fingerprint type object
    
_aliases = {
    "atom_pairs2d": "CDK-AtomPairs2D",
    "daylight": "CDK-Daylight",
    "ecfp0": "CDK-ECFP0",
    "ecfp2": "CDK-ECFP2",
    "ecfp4": "CDK-ECFP4",
    "ecfp6": "CDK-ECFP6",
    "estate": "CDK-EState",
    "extended": "CDK-Extended",
    "fcfp0": "CDK-FCFP0",
    "fcfp2": "CDK-FCFP2",
    "fcfp4": "CDK-FCFP4",
    "fcfp6": "CDK-FCFP6",
    "graph_only": "CDK-GraphOnly",
    "hybridization": "CDK-Hybridization",
    "klekota_roth": "CDK-KlekotaRoth",
    "maccs": "CDK-MACCS",
    "pubchem": "CDK-Pubchem",
    "shortest_path": "CDK-ShortestPath",
    "substructure": "CDK-Substructure",
    }
    
def __getattr__(name):
    full_name = _aliases.get(name, None)
    if full_name is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")

    from . import get_fingerprint_type
    return get_fingerprint_type(full_name)

def __dir__():
    return sorted(__all__ + list(_aliases))


#### The following was auto-generated by: utils/mol_io_codegen.py cdk

from . import io as _io
from ._compat import (
    Any as _Any,
    BinaryIO as _BinaryIO,
    Literal as _Literal,
    Optional as _Optional,
    Union as _Union,
)


_smistring_fmt = get_format("smistring")


@_public
def from_smistring(
    content: str,
    *,
    kekulise: bool = True,
    errors: str = "strict",
):
    """Parse a SMILES string using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "smistring", reader_args={...}, errors=errors)

    :param kekulise: if true, ensure a valid Kekule intepretation exists
    :type kekulise: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "kekulise": kekulise,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "smistring", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canonical' = 1  (in default bit flags)
      'InChILabelling' = 3
      'AtomAtomMap' = 4
      'AtomicMass' = 8  (in default bit flags)
      'UseAromaticSymbols' = 16
      'StereoTetrahedral' = 256  (in default bit flags)
      'StereoCisTrans' = 512  (in default bit flags)
      'StereoExTetrahedral' = 1024  (in default bit flags)
      'StereoExCisTrans' = 1280  (in default bit flags)
      'AtomicMassStrict' = 2048
      'Stereo' = 1792  (in default bit flags)
      'Cx2dCoordinates' = 4096
      'Cx3dCoordinates' = 8192
      'CxCoordinates' = 12288
      'CxAtomLabel' = 32768
      'CxAtomValue' = 65536
      'CxRadical' = 131072
      'CxMulticenter' = 262144
      'CxPolymer' = 524288
      'CxFragmentGroup' = 1048576
      'AtomAtomMapRenumber' = 33554437
      'CxSmiles' = 12550400
      'CxSmilesWithCoords' = 12562688
      'Unique' = 1  (in default bit flags)
      'Isomeric' = 1800  (in default bit flags)
      'Absolute' = 1801  (in default bit flags)
      'UniversalSmiles' = 1803
      'Default' = 1801  (in default bit flags)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


_smi_fmt = get_format("smi")


@_public
def from_smi(
    content: str,
    *,
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = "to-eol",
    implementation: _Optional[_Literal["cdk", "chemfp"]] = "cdk",
    kekulise: bool = True,
    errors: str = "strict",
):
    """Parse a SMILES string and id using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "smi", reader_args={...}, errors=errors)

    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
    :param implementation: use CDK or chemfp to identify records
    :type implementation: either 'cdk' or 'chemfp'
    :param kekulise: if true, ensure a valid Kekule intepretation exists
    :type kekulise: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "has_header": has_header,
            "delimiter": delimiter,
            "implementation": implementation,
            "kekulise": kekulise,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string and id from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "smi", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canonical' = 1  (in default bit flags)
      'InChILabelling' = 3
      'AtomAtomMap' = 4
      'AtomicMass' = 8  (in default bit flags)
      'UseAromaticSymbols' = 16
      'StereoTetrahedral' = 256  (in default bit flags)
      'StereoCisTrans' = 512  (in default bit flags)
      'StereoExTetrahedral' = 1024  (in default bit flags)
      'StereoExCisTrans' = 1280  (in default bit flags)
      'AtomicMassStrict' = 2048
      'Stereo' = 1792  (in default bit flags)
      'Cx2dCoordinates' = 4096
      'Cx3dCoordinates' = 8192
      'CxCoordinates' = 12288
      'CxAtomLabel' = 32768
      'CxAtomValue' = 65536
      'CxRadical' = 131072
      'CxMulticenter' = 262144
      'CxPolymer' = 524288
      'CxFragmentGroup' = 1048576
      'AtomAtomMapRenumber' = 33554437
      'CxSmiles' = 12550400
      'CxSmilesWithCoords' = 12562688
      'Unique' = 1  (in default bit flags)
      'Isomeric' = 1800  (in default bit flags)
      'Absolute' = 1801  (in default bit flags)
      'UniversalSmiles' = 1803
      'Default' = 1801  (in default bit flags)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "delimiter": delimiter,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_smi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = "to-eol",
    implementation: _Optional[_Literal["cdk", "chemfp"]] = "cdk",
    kekulise: bool = True,
    errors: str = "strict",
):
    """Parse a SMILES string and id file using the CDK toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "smi", reader_args={...}, errors=errors)

    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
    :param implementation: use CDK or chemfp to identify records
    :type implementation: either 'cdk' or 'chemfp'
    :param kekulise: if true, ensure a valid Kekule intepretation exists
    :type kekulise: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "smi"),
        reader_args={
            "has_header": has_header,
            "delimiter": delimiter,
            "implementation": implementation,
            "kekulise": kekulise,
        },
        errors=errors,
    )


@_public
def to_smi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string and id from a CDK molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "smi", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canonical' = 1  (in default bit flags)
      'InChILabelling' = 3
      'AtomAtomMap' = 4
      'AtomicMass' = 8  (in default bit flags)
      'UseAromaticSymbols' = 16
      'StereoTetrahedral' = 256  (in default bit flags)
      'StereoCisTrans' = 512  (in default bit flags)
      'StereoExTetrahedral' = 1024  (in default bit flags)
      'StereoExCisTrans' = 1280  (in default bit flags)
      'AtomicMassStrict' = 2048
      'Stereo' = 1792  (in default bit flags)
      'Cx2dCoordinates' = 4096
      'Cx3dCoordinates' = 8192
      'CxCoordinates' = 12288
      'CxAtomLabel' = 32768
      'CxAtomValue' = 65536
      'CxRadical' = 131072
      'CxMulticenter' = 262144
      'CxPolymer' = 524288
      'CxFragmentGroup' = 1048576
      'AtomAtomMapRenumber' = 33554437
      'CxSmiles' = 12550400
      'CxSmilesWithCoords' = 12562688
      'Unique' = 1  (in default bit flags)
      'Isomeric' = 1800  (in default bit flags)
      'Absolute' = 1801  (in default bit flags)
      'UniversalSmiles' = 1803
      'Default' = 1801  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "smi"),
        writer_args={
            "delimiter": delimiter,
            "flavor": flavor,
        },
        errors=errors,
    )


_sdf_fmt = get_format("sdf")


@_public
def from_sdf(
    content: str,
    *,
    ForceReadAs3DCoordinates: bool = False,
    mode: _Literal["RELAXED", "STRICT"] = "RELAXED",
    AddStereoElements: bool = True,
    InterpretHydrogenIsotopes: bool = True,
    implementation: _Optional[_Literal["cdk", "chemfp"]] = "cdk",
    errors: str = "strict",
):
    """Parse an SDF record using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "sdf", reader_args={...}, errors=errors)

    :param ForceReadAs3DCoordinates: if true, always interpret coordinates as 3D
    :type ForceReadAs3DCoordinates: Boolean (default: False)
    :param mode: strictness mode when parsing a record
    :type mode: 'RELAXED' will attempt to recover, 'STRICT' will not
    :param AddStereoElements: if true, detect and create IStereoElements
    :type AddStereoElements: Boolean (default: True)
    :param InterpretHydrogenIsotopes: if true, interpret D and T as hydrogen isotopes
    :type InterpretHydrogenIsotopes: Boolean (default: True)
    :param implementation: use CDK or chemfp to identify records
    :type implementation: either 'cdk' or 'chemfp'
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sdf_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "ForceReadAs3DCoordinates": ForceReadAs3DCoordinates,
            "mode": mode,
            "AddStereoElements": AddStereoElements,
            "InterpretHydrogenIsotopes": InterpretHydrogenIsotopes,
            "implementation": implementation,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sdf(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    WriteAromaticBondTypes: bool = False,
    WriteMajorIsotopes: bool = True,
    writeProperties: bool = True,
    WriteQueryFormatValencies: bool = False,
    TruncateLongData: bool = False,
    ProgramName: str = "CDK",
    ForceWriteAs2DCoordinates: bool = False,
    WriteDefaultProperties: bool = True,
    writeV3000: bool = False,
    errors: str = "strict",
):
    """Generate an SDF record from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "sdf", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param WriteAromaticBondTypes: if true, write aromatic bonds as bond type 4
    :type WriteAromaticBondTypes: Boolean (default: False)
    :param WriteMajorIsotopes: if true, include isotopic mass for atoms with a specified mass
    :type WriteMajorIsotopes: Boolean (default: True)
    :param writeProperties: if true, write non-molecule data to the data tags
    :type writeProperties: Boolean (default: True)
    :param WriteQueryFormatValencies: if true, write valences in the MDL query format (deprecated)
    :type WriteQueryFormatValencies: Boolean (default: False)
    :param TruncateLongData: if true, truncate data items longer than 200 characters
    :type TruncateLongData: Boolean (default: False)
    :param ProgramName: text to use in the 'program name' section of the second line
    :type ProgramName: a string up to 8 characters long
    :param ForceWriteAs2DCoordinates: if true, write coordinates as 2D
    :type ForceWriteAs2DCoordinates: Boolean (default: False)
    :param WriteDefaultProperties: if true, always include zeros in the empty atom and bond block fields
    :type WriteDefaultProperties: Boolean (default: True)
    :param writeV3000: if true, always write the record in V3000 format
    :type writeV3000: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "WriteAromaticBondTypes": WriteAromaticBondTypes,
            "WriteMajorIsotopes": WriteMajorIsotopes,
            "writeProperties": writeProperties,
            "WriteQueryFormatValencies": WriteQueryFormatValencies,
            "TruncateLongData": TruncateLongData,
            "ProgramName": ProgramName,
            "ForceWriteAs2DCoordinates": ForceWriteAs2DCoordinates,
            "WriteDefaultProperties": WriteDefaultProperties,
            "writeV3000": writeV3000,
        },
        error_handler=error_handler,
    )


@_public
def from_sdf_file(
    source: _Union[None, str, _BinaryIO],
    *,
    ForceReadAs3DCoordinates: bool = False,
    mode: _Literal["RELAXED", "STRICT"] = "RELAXED",
    AddStereoElements: bool = True,
    InterpretHydrogenIsotopes: bool = True,
    implementation: _Optional[_Literal["cdk", "chemfp"]] = "cdk",
    errors: str = "strict",
):
    """Parse an SDF record file using the CDK toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "sdf", reader_args={...}, errors=errors)

    :param ForceReadAs3DCoordinates: if true, always interpret coordinates as 3D
    :type ForceReadAs3DCoordinates: Boolean (default: False)
    :param mode: strictness mode when parsing a record
    :type mode: 'RELAXED' will attempt to recover, 'STRICT' will not
    :param AddStereoElements: if true, detect and create IStereoElements
    :type AddStereoElements: Boolean (default: True)
    :param InterpretHydrogenIsotopes: if true, interpret D and T as hydrogen isotopes
    :type InterpretHydrogenIsotopes: Boolean (default: True)
    :param implementation: use CDK or chemfp to identify records
    :type implementation: either 'cdk' or 'chemfp'
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "sdf"),
        reader_args={
            "ForceReadAs3DCoordinates": ForceReadAs3DCoordinates,
            "mode": mode,
            "AddStereoElements": AddStereoElements,
            "InterpretHydrogenIsotopes": InterpretHydrogenIsotopes,
            "implementation": implementation,
        },
        errors=errors,
    )


@_public
def to_sdf_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    WriteAromaticBondTypes: bool = False,
    WriteMajorIsotopes: bool = True,
    writeProperties: bool = True,
    WriteQueryFormatValencies: bool = False,
    TruncateLongData: bool = False,
    ProgramName: str = "CDK",
    ForceWriteAs2DCoordinates: bool = False,
    WriteDefaultProperties: bool = True,
    writeV3000: bool = False,
    errors: str = "strict",
):
    """Generate an SDF record from a CDK molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param WriteAromaticBondTypes: if true, write aromatic bonds as bond type 4
    :type WriteAromaticBondTypes: Boolean (default: False)
    :param WriteMajorIsotopes: if true, include isotopic mass for atoms with a specified mass
    :type WriteMajorIsotopes: Boolean (default: True)
    :param writeProperties: if true, write non-molecule data to the data tags
    :type writeProperties: Boolean (default: True)
    :param WriteQueryFormatValencies: if true, write valences in the MDL query format (deprecated)
    :type WriteQueryFormatValencies: Boolean (default: False)
    :param TruncateLongData: if true, truncate data items longer than 200 characters
    :type TruncateLongData: Boolean (default: False)
    :param ProgramName: text to use in the 'program name' section of the second line
    :type ProgramName: a string up to 8 characters long
    :param ForceWriteAs2DCoordinates: if true, write coordinates as 2D
    :type ForceWriteAs2DCoordinates: Boolean (default: False)
    :param WriteDefaultProperties: if true, always include zeros in the empty atom and bond block fields
    :type WriteDefaultProperties: Boolean (default: True)
    :param writeV3000: if true, always write the record in V3000 format
    :type writeV3000: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf"),
        writer_args={
            "WriteAromaticBondTypes": WriteAromaticBondTypes,
            "WriteMajorIsotopes": WriteMajorIsotopes,
            "writeProperties": writeProperties,
            "WriteQueryFormatValencies": WriteQueryFormatValencies,
            "TruncateLongData": TruncateLongData,
            "ProgramName": ProgramName,
            "ForceWriteAs2DCoordinates": ForceWriteAs2DCoordinates,
            "WriteDefaultProperties": WriteDefaultProperties,
            "writeV3000": writeV3000,
        },
        errors=errors,
    )


_sdf3k_fmt = get_format("sdf3k")


@_public
def to_sdf3k(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    WriteAromaticBondTypes: bool = False,
    WriteMajorIsotopes: bool = True,
    writeProperties: bool = True,
    WriteQueryFormatValencies: bool = False,
    TruncateLongData: bool = False,
    ProgramName: str = "CDK",
    ForceWriteAs2DCoordinates: bool = False,
    WriteDefaultProperties: bool = True,
    writeV3000: bool = True,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "sdf3k", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param WriteAromaticBondTypes: if true, write aromatic bonds as bond type 4
    :type WriteAromaticBondTypes: Boolean (default: False)
    :param WriteMajorIsotopes: if true, include isotopic mass for atoms with a specified mass
    :type WriteMajorIsotopes: Boolean (default: True)
    :param writeProperties: if true, write non-molecule data to the data tags
    :type writeProperties: Boolean (default: True)
    :param WriteQueryFormatValencies: if true, write valences in the MDL query format (deprecated)
    :type WriteQueryFormatValencies: Boolean (default: False)
    :param TruncateLongData: if true, truncate data items longer than 200 characters
    :type TruncateLongData: Boolean (default: False)
    :param ProgramName: text to use in the 'program name' section of the second line
    :type ProgramName: a string up to 8 characters long
    :param ForceWriteAs2DCoordinates: if true, write coordinates as 2D
    :type ForceWriteAs2DCoordinates: Boolean (default: False)
    :param WriteDefaultProperties: if true, always include zeros in the empty atom and bond block fields
    :type WriteDefaultProperties: Boolean (default: True)
    :param writeV3000: if true, always write the record in V3000 format
    :type writeV3000: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf3k_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "WriteAromaticBondTypes": WriteAromaticBondTypes,
            "WriteMajorIsotopes": WriteMajorIsotopes,
            "writeProperties": writeProperties,
            "WriteQueryFormatValencies": WriteQueryFormatValencies,
            "TruncateLongData": TruncateLongData,
            "ProgramName": ProgramName,
            "ForceWriteAs2DCoordinates": ForceWriteAs2DCoordinates,
            "WriteDefaultProperties": WriteDefaultProperties,
            "writeV3000": writeV3000,
        },
        error_handler=error_handler,
    )


@_public
def to_sdf3k_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    WriteAromaticBondTypes: bool = False,
    WriteMajorIsotopes: bool = True,
    writeProperties: bool = True,
    WriteQueryFormatValencies: bool = False,
    TruncateLongData: bool = False,
    ProgramName: str = "CDK",
    ForceWriteAs2DCoordinates: bool = False,
    WriteDefaultProperties: bool = True,
    writeV3000: bool = True,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from a CDK molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf3k", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param WriteAromaticBondTypes: if true, write aromatic bonds as bond type 4
    :type WriteAromaticBondTypes: Boolean (default: False)
    :param WriteMajorIsotopes: if true, include isotopic mass for atoms with a specified mass
    :type WriteMajorIsotopes: Boolean (default: True)
    :param writeProperties: if true, write non-molecule data to the data tags
    :type writeProperties: Boolean (default: True)
    :param WriteQueryFormatValencies: if true, write valences in the MDL query format (deprecated)
    :type WriteQueryFormatValencies: Boolean (default: False)
    :param TruncateLongData: if true, truncate data items longer than 200 characters
    :type TruncateLongData: Boolean (default: False)
    :param ProgramName: text to use in the 'program name' section of the second line
    :type ProgramName: a string up to 8 characters long
    :param ForceWriteAs2DCoordinates: if true, write coordinates as 2D
    :type ForceWriteAs2DCoordinates: Boolean (default: False)
    :param WriteDefaultProperties: if true, always include zeros in the empty atom and bond block fields
    :type WriteDefaultProperties: Boolean (default: True)
    :param writeV3000: if true, always write the record in V3000 format
    :type writeV3000: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf3k"),
        writer_args={
            "WriteAromaticBondTypes": WriteAromaticBondTypes,
            "WriteMajorIsotopes": WriteMajorIsotopes,
            "writeProperties": writeProperties,
            "WriteQueryFormatValencies": WriteQueryFormatValencies,
            "TruncateLongData": TruncateLongData,
            "ProgramName": ProgramName,
            "ForceWriteAs2DCoordinates": ForceWriteAs2DCoordinates,
            "WriteDefaultProperties": WriteDefaultProperties,
            "writeV3000": writeV3000,
        },
        errors=errors,
    )


_molfile_fmt = get_format("molfile")


@_public
def from_molfile(
    content: str,
    *,
    ForceReadAs3DCoordinates: bool = False,
    mode: _Literal["RELAXED", "STRICT"] = "RELAXED",
    AddStereoElements: bool = True,
    InterpretHydrogenIsotopes: bool = True,
    implementation: _Optional[_Literal["cdk", "chemfp"]] = "cdk",
    errors: str = "strict",
):
    """Parse a molfile using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "molfile", reader_args={...}, errors=errors)

    :param ForceReadAs3DCoordinates: if true, always interpret coordinates as 3D
    :type ForceReadAs3DCoordinates: Boolean (default: False)
    :param mode: strictness mode when parsing a record
    :type mode: 'RELAXED' will attempt to recover, 'STRICT' will not
    :param AddStereoElements: if true, detect and create IStereoElements
    :type AddStereoElements: Boolean (default: True)
    :param InterpretHydrogenIsotopes: if true, interpret D and T as hydrogen isotopes
    :type InterpretHydrogenIsotopes: Boolean (default: True)
    :param implementation: use CDK or chemfp to identify records
    :type implementation: either 'cdk' or 'chemfp'
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _molfile_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "ForceReadAs3DCoordinates": ForceReadAs3DCoordinates,
            "mode": mode,
            "AddStereoElements": AddStereoElements,
            "InterpretHydrogenIsotopes": InterpretHydrogenIsotopes,
            "implementation": implementation,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_molfile(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    WriteAromaticBondTypes: bool = False,
    WriteMajorIsotopes: bool = True,
    WriteQueryFormatValencies: bool = False,
    TruncateLongData: bool = False,
    ProgramName: str = "CDK",
    ForceWriteAs2DCoordinates: bool = False,
    WriteDefaultProperties: bool = True,
    writeV3000: bool = False,
    errors: str = "strict",
):
    """Generate a molfile from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "molfile", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param WriteAromaticBondTypes: if true, write aromatic bonds as bond type 4
    :type WriteAromaticBondTypes: Boolean (default: False)
    :param WriteMajorIsotopes: if true, include isotopic mass for atoms with a specified mass
    :type WriteMajorIsotopes: Boolean (default: True)
    :param WriteQueryFormatValencies: if true, write valences in the MDL query format (deprecated)
    :type WriteQueryFormatValencies: Boolean (default: False)
    :param TruncateLongData: if true, truncate data items longer than 200 characters
    :type TruncateLongData: Boolean (default: False)
    :param ProgramName: text to use in the 'program name' section of the second line
    :type ProgramName: a string up to 8 characters long
    :param ForceWriteAs2DCoordinates: if true, write coordinates as 2D
    :type ForceWriteAs2DCoordinates: Boolean (default: False)
    :param WriteDefaultProperties: if true, always include zeros in the empty atom and bond block fields
    :type WriteDefaultProperties: Boolean (default: True)
    :param writeV3000: if true, always write the record in V3000 format
    :type writeV3000: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _molfile_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "WriteAromaticBondTypes": WriteAromaticBondTypes,
            "WriteMajorIsotopes": WriteMajorIsotopes,
            "WriteQueryFormatValencies": WriteQueryFormatValencies,
            "TruncateLongData": TruncateLongData,
            "ProgramName": ProgramName,
            "ForceWriteAs2DCoordinates": ForceWriteAs2DCoordinates,
            "WriteDefaultProperties": WriteDefaultProperties,
            "writeV3000": writeV3000,
        },
        error_handler=error_handler,
    )


_inchi_fmt = get_format("inchi")


@_public
def from_inchi(
    content: str,
    *,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = "to-eol",
    errors: str = "strict",
):
    """Parse an InChI string and id using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchi", reader_args={...}, errors=errors)

    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "inchi", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "SUCF": None,
            "ChiralFlagON": None,
            "ChiralFlagOFF": None,
            "Snon": None,
            "SRel": None,
            "SRac": None,
            "SUU": None,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": None,
            "DoNotAddH": DoNotAddH,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def from_inchi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = "to-eol",
    errors: str = "strict",
):
    """Parse an InChI string and id file using the CDK toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "inchi", reader_args={...}, errors=errors)

    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "inchi"),
        reader_args={
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_inchi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from a CDK molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchi", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchi"),
        writer_args={
            "SUCF": SUCF,
            "ChiralFlagON": ChiralFlagON,
            "ChiralFlagOFF": ChiralFlagOFF,
            "Snon": Snon,
            "SRel": SRel,
            "SRac": SRac,
            "SUU": SUU,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": AuxNone,
            "DoNotAddH": DoNotAddH,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchistring_fmt = get_format("inchistring")


@_public
def from_inchistring(
    content: str,
    *,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = "to-eol",
    errors: str = "strict",
):
    """Parse an InChI string using the CDK toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchistring", reader_args={...}, errors=errors)

    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    errors: str = "strict",
):
    """Generate an InChI string from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "inchistring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "SUCF": None,
            "ChiralFlagON": None,
            "ChiralFlagOFF": None,
            "Snon": None,
            "SRel": None,
            "SRac": None,
            "SUU": None,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": None,
            "DoNotAddH": DoNotAddH,
            "options": options,
        },
        error_handler=error_handler,
    )


_inchikey_fmt = get_format("inchikey")


@_public
def to_inchikey(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "inchikey", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikey_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "SUCF": None,
            "ChiralFlagON": None,
            "ChiralFlagOFF": None,
            "Snon": None,
            "SRel": None,
            "SRac": None,
            "SUU": None,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": None,
            "DoNotAddH": DoNotAddH,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def to_inchikey_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from a CDK molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchikey", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchikey"),
        writer_args={
            "SUCF": SUCF,
            "ChiralFlagON": ChiralFlagON,
            "ChiralFlagOFF": ChiralFlagOFF,
            "Snon": Snon,
            "SRel": SRel,
            "SRac": SRac,
            "SUU": SUU,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": AuxNone,
            "DoNotAddH": DoNotAddH,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchikeystring_fmt = get_format("inchikeystring")


@_public
def to_inchikeystring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    RecMet: bool = None,
    FixedH: bool = None,
    DoNotAddH: bool = None,
    options: str = None,
    errors: str = "strict",
):
    """Generate an InChIKey string from a CDK molecule

    This is equivalent to calling:
       create_string(mol, "inchikeystring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: a CDK molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param RecMet: Reconnect metals
    :type RecMet: Boolean (default: None)
    :param FixedH: Use fixed hydrogens
    :type FixedH: Boolean (default: None)
    :param DoNotAddH: Do not add hydrogens
    :type DoNotAddH: Boolean (default: None)
    :param options: Configuration string to pass to the InChI API
    :type options: space separated strings
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a CDK molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikeystring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "SUCF": None,
            "ChiralFlagON": None,
            "ChiralFlagOFF": None,
            "Snon": None,
            "SRel": None,
            "SRac": None,
            "SUU": None,
            "RecMet": RecMet,
            "FixedH": FixedH,
            "AuxNone": None,
            "DoNotAddH": DoNotAddH,
            "options": options,
        },
        error_handler=error_handler,
    )


# Aliases

from_smiles = from_smistring

to_smiles = to_smistring

#### End of auto-generated code.
