# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

from . import cdk_toolkit as _tk
from ._cdk_toolkit import cdk, java, _default_builder

# If this attribute exists then we're in a Sphinx build environment
_using_sphinx_mock = hasattr(java, "using_sphinx_mock")

from . import types

from .bitops import byte_from_bitlist

from ._compat import Any, Optional, Union, Literal

software = _tk.software
from . import SOFTWARE as CHEMFP_SOFTWARE

__all__ = []
def _public(obj):
    __all__.append(obj.__name__)
    return obj

_no_param_call = types.FingerprintArguments().get_call_method()


########

# Some constants shared by the fingerprinter and the command-line code.
DEFAULT_SIZE = 1024 # DaylightFingerprinter.DEFAULT_SIZE
DEFAULT_SEARCH_DEPTH = 7 # DaylightFingerprinter.DEFAULT_SEARCH_DEPTH
DEFAULT_PATH_LIMIT = 42000 # DaylightFingerprinter.PATH_LIMIT
DEFAULT_HASH_PSEUDO_ATOMS = False

DEFAULT_PERCEIVE_STEREOCHEMISTRY = False

all_arguments = types.FingerprintArguments()
_add = all_arguments.add_argument

_add("size", decoder=types.positive_int, metavar="INT", default=DEFAULT_SIZE,
         help = "fingerprint size")
_add("searchDepth", decoder=types.nonnegative_int, metavar="INT", default=DEFAULT_SEARCH_DEPTH,
         help = "search depth")
_add("pathLimit", decoder=types.nonnegative_int, metavar="INT", default=DEFAULT_PATH_LIMIT,
         help = "path limit")
_add("hashPseudoAtoms", decoder=types.zero_or_one, metavar="0|1", default=int(DEFAULT_HASH_PSEUDO_ATOMS),
         encoder = types.encode_bool_as_zero_or_one,
         help = "include pseudo-atoms in path enumeration")
_add("perceiveStereochemistry", decoder=types.zero_or_one, metavar="0|1", default=int(DEFAULT_PERCEIVE_STEREOCHEMISTRY),
         encoder = types.encode_bool_as_zero_or_one,
         help = "re-perceive stereochemistry from 2D/3D coordinates")

@_public
class CDKBaseFingerprintType(types.ThreadsafeFingerprinterMixin,
                             types.FingerprintType):
    #: a description of the CDK and chemfp versions
    software = types._sphinx_repr_if(
        _using_sphinx_mock, "...", _tk.software + " " + CHEMFP_SOFTWARE,
        )
    #: a reference to the :mod:`.cdk_toolkit` module
    toolkit = types._sphinx_repr_if(
        _using_sphinx_mock, "<module 'chemfp.cdk_toolkit>'", _tk)
    
    #: a CDK fingerprinter can raise an exception if the molecule isn't correctly prepared
    fingerprinter_can_fail = True

    #### The following was auto-generated by: utils/mol_io_codegen.py --fp cdk

    def from_smistring(
        self,
        content: str,
        *,
        kekulise: bool = True,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string

        This is equivalent to calling::

           mol = fptype.toolkit.from_smistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param kekulise: if true, ensure a valid Kekule intepretation exists
        :type kekulise: Boolean (default: True)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smistring(
            content,
            kekulise=kekulise,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_smi(
        self,
        content: str,
        *,
        has_header: bool = False,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = "to-eol",
        implementation: Optional[Literal["cdk", "chemfp"]] = "cdk",
        kekulise: bool = True,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_smi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param has_header: If true, treat the first line of the SMILES file as a header
        :type has_header: Boolean (default: False)
        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
        :param implementation: use CDK or chemfp to identify records
        :type implementation: either 'cdk' or 'chemfp'
        :param kekulise: if true, ensure a valid Kekule intepretation exists
        :type kekulise: Boolean (default: True)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smi(
            content,
            has_header=has_header,
            delimiter=delimiter,
            implementation=implementation,
            kekulise=kekulise,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_sdf(
        self,
        content: str,
        *,
        ForceReadAs3DCoordinates: bool = False,
        mode: Literal["RELAXED", "STRICT"] = "RELAXED",
        AddStereoElements: bool = True,
        InterpretHydrogenIsotopes: bool = True,
        implementation: Optional[Literal["cdk", "chemfp"]] = "cdk",
        errors: str = "strict",
    ):
        """Generate a fingerprint from an SDF record

        This is equivalent to calling::

           mol = fptype.toolkit.from_sdf(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param ForceReadAs3DCoordinates: if true, always interpret coordinates as 3D
        :type ForceReadAs3DCoordinates: Boolean (default: False)
        :param mode: strictness mode when parsing a record
        :type mode: 'RELAXED' will attempt to recover, 'STRICT' will not
        :param AddStereoElements: if true, detect and create IStereoElements
        :type AddStereoElements: Boolean (default: True)
        :param InterpretHydrogenIsotopes: if true, interpret D and T as hydrogen isotopes
        :type InterpretHydrogenIsotopes: Boolean (default: True)
        :param implementation: use CDK or chemfp to identify records
        :type implementation: either 'cdk' or 'chemfp'
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_sdf(
            content,
            ForceReadAs3DCoordinates=ForceReadAs3DCoordinates,
            mode=mode,
            AddStereoElements=AddStereoElements,
            InterpretHydrogenIsotopes=InterpretHydrogenIsotopes,
            implementation=implementation,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_molfile(
        self,
        content: str,
        *,
        ForceReadAs3DCoordinates: bool = False,
        mode: Literal["RELAXED", "STRICT"] = "RELAXED",
        AddStereoElements: bool = True,
        InterpretHydrogenIsotopes: bool = True,
        implementation: Optional[Literal["cdk", "chemfp"]] = "cdk",
        errors: str = "strict",
    ):
        """Generate a fingerprint from a molfile

        This is equivalent to calling::

           mol = fptype.toolkit.from_molfile(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param ForceReadAs3DCoordinates: if true, always interpret coordinates as 3D
        :type ForceReadAs3DCoordinates: Boolean (default: False)
        :param mode: strictness mode when parsing a record
        :type mode: 'RELAXED' will attempt to recover, 'STRICT' will not
        :param AddStereoElements: if true, detect and create IStereoElements
        :type AddStereoElements: Boolean (default: True)
        :param InterpretHydrogenIsotopes: if true, interpret D and T as hydrogen isotopes
        :type InterpretHydrogenIsotopes: Boolean (default: True)
        :param implementation: use CDK or chemfp to identify records
        :type implementation: either 'cdk' or 'chemfp'
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_molfile(
            content,
            ForceReadAs3DCoordinates=ForceReadAs3DCoordinates,
            mode=mode,
            AddStereoElements=AddStereoElements,
            InterpretHydrogenIsotopes=InterpretHydrogenIsotopes,
            implementation=implementation,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchi(
        self,
        content: str,
        *,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = "to-eol",
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchi(
            content,
            delimiter=delimiter,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchistring(
        self,
        content: str,
        *,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = "to-eol",
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: "to-eol")
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchistring(
            content,
            delimiter=delimiter,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    # Aliases

    from_smiles = from_smistring

    #### End of auto-generated code.


class VariableSizeFingerprint(CDKBaseFingerprintType):
    def __init__(self, fingerprint_kwargs):
        super(VariableSizeFingerprint, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["size"]

class FixedSizeFingerprint(CDKBaseFingerprintType):
    pass

#    public final String getVersionDescription() {
#        StringBuilder sb = new StringBuilder();
#        sb.append("CDK-")
#          .append(getClass().getSimpleName())
#          .append("/")
#          .append(CDK.getVersion()); // could version fingerprints separetely
#        for (Map.Entry<String,String> param : getParameters()) {
#            sb.append(' ').append(param.getKey()).append('=').append(param.getValue());
#        }
#        return sb.toString();
#    }

#### CDK-Daylight -- The Daylight-like fingerprinter

# cdk.fingerprint.Fingerprinter(int size, int searchDepth)
# #type=CDK-Fingerprinter/2.0 searchDepth=7 pathLimit=2000 hashPseudoAtoms=true
#    new SimpleImmutableEntry<>("searchDepth", Integer.toString(searchDepth)),
#    new SimpleImmutableEntry<>("pathLimit", Integer.toString(pathLimit)),
#    new SimpleImmutableEntry<>("hashPseudoAtoms", Boolean.toString(hashPseudoAtoms))

# getSize()
# getSearchDepth()
# setHashPseudoAtoms()
# setPathLimit()


# CDK says "The FingerPrinter assumes that hydrogens are explicitly given!"
# 
DaylightFingerprinter = cdk.fingerprint.Fingerprinter

def make_daylight_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = DaylightFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_daylight_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_daylight_fingerprinter.__qualname__ = "cdk_daylight_fingerprinter"
    cdk_daylight_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_daylight_fingerprinter


@_public
@types.adjust_call_doc
class CDKDaylightFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Daylight-like fingerprint, version 2.0
    
    The CDK-Daylight/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Daylight/2.0"
    _make_fingerprinter = staticmethod(make_daylight_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")
    __call__ = _arguments.get_call_method()
    
def register_daylight(registry):
    registry.register(CDKDaylightFingerprintType_v20)

#### CDK-GraphOnly

# Modified version of DaylightFingerprinter but bond symbols are all ""
GraphOnlyFingerprinter = cdk.fingerprint.GraphOnlyFingerprinter

def make_graphonly_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = GraphOnlyFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_graphonly_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_graphonly_fingerprinter.__qualname__ = "cdk_graphonly_fingerprinter"
    cdk_graphonly_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_graphonly_fingerprinter

@_public
@types.adjust_call_doc
class CDKGraphOnlyFingerprintType_v20(VariableSizeFingerprint):
    """CDK's GraphOnly fingerprint, version 2.0
    
    The CDK-GraphOnly/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-GraphOnly/2.0"
    _make_fingerprinter = staticmethod(make_graphonly_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")
    __call__ = _arguments.get_call_method()

def register_graph_only(registry):
    registry.register(CDKGraphOnlyFingerprintType_v20)

#### CDK-MACCS

# "This class assumes that aromaticity perception, atom typing and adding of
#  implicit hydrogens have been performed prior to generating the fingerprint."
 
MACCSFingerprinter = cdk.fingerprint.MACCSFingerprinter

def make_maccs_fingerprinter():
    fingerprinter = MACCSFingerprinter()
    n = fingerprinter.getSize()
    if n != 166:
        raise AssertionError(("Unexpected MACCSFingerprinter size", n, 166))
    def cdk_maccs_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 166)
    cdk_maccs_fingerprinter.__qualname__ = "cdk_maccs_fingerprinter"
    cdk_maccs_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_maccs_fingerprinter

@_public
@types.adjust_call_doc
class CDKMACCSFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the 166 MACCS keys, version 2.0

    The CDK-MACCS/2.0 fingerprints have no parameters.
    """
    num_bits = 166
    name = "CDK-MACCS/2.0"
    _make_fingerprinter = staticmethod(make_maccs_fingerprinter)
    __call__ = _no_param_call

def register_maccs(registry):
    registry.register(CDKMACCSFingerprintType_v20)

    
#### CDK-EState

# "This fingerprinter generates 79 bit fingerprints using the E-State fragments."

EStateFingerprinter = cdk.fingerprint.EStateFingerprinter

def make_estate_fingerprinter():
    fingerprinter = EStateFingerprinter()
    n = fingerprinter.getSize()
    if n != 79:
        raise AssertionError(("Unexpected EStateFingerprinter size", n, 79))
    def cdk_estate_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 79)
    cdk_estate_fingerprinter.__qualname__ = "cdk_estate_fingerprinter"
    cdk_estate_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_estate_fingerprinter

@_public
@types.adjust_call_doc
class CDKEStateFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the EState fingerprint, version 2.0

    The CDK-EState/2.0 fingerprints have no parameters.
    """
    num_bits = 79
    name = "CDK-EState/2.0"
    _make_fingerprinter = staticmethod(make_estate_fingerprinter)
    __call__ = _no_param_call

def register_estate(registry):
    registry.register(CDKEStateFingerprintType_v20)
    
#### CDK-Extended

ExtendedFingerprinter = cdk.fingerprint.ExtendedFingerprinter

def make_extended_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size >= 26):
        raise ValueError("size must be 26 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)
    fingerprinter = ExtendedFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_extended_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_extended_fingerprinter.__qualname__ = "cdk_extended_fingerprinter"
    cdk_extended_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_extended_fingerprinter


@_public
@types.adjust_call_doc
class CDKExtendedFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Extended fingerprint, version 2.0
    
    The CDK-Extended/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Extended/2.0"
    _make_fingerprinter = staticmethod(make_extended_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")
    __call__ = _arguments.get_call_method()

def register_extended(registry):
    registry.register(CDKExtendedFingerprintType_v20)


#### CDK-Hybridization

# "The FingerPrinter assumes that hydrogens are explicitly given!"
# "Furthermore, if pseudo atoms or atoms with malformed symbols are present, ..."
# "Unlike the {@link Fingerprinter}, this fingerprinter does not take into
#     account aromaticity. Instead, it takes into account SP2 hybridization."

HybridizationFingerprinter = cdk.fingerprint.HybridizationFingerprinter

def make_hybridization_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = HybridizationFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_hybridization_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_hybridization_fingerprinter.__qualname__ = "cdk_hybridization_fingerprinter"
    cdk_hybridization_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_hybridization_fingerprinter

@_public
@types.adjust_call_doc
class CDKHybridizationFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Hybridization fingerprint, version 2.0
    
    The CDK-Hybridization/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Hybridization/2.0"
    _make_fingerprinter = staticmethod(make_hybridization_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")
    __call__ = _arguments.get_call_method()

def register_hybridization(registry):
    registry.register(CDKHybridizationFingerprintType_v20)

#### CDK-KlekotaRoth

# "SMARTS based substructure fingerprint based on <b>Chemical substructures
#  that enrich for biological activity</b> {@cdk.cite Klekota01112008}.
#
#   Presence of 4860 substructures"

# For some reason I am unable to get the KlekotaRothFingerprinter through jpype
# on macOS.
KlekotaRothFingerprinter = getattr(cdk.fingerprint, "KlekotaRothFingerprinter", None)

_klekota_roth_fingerprinter = None
def make_klekota_roth_fingerprinter():
    global _klekota_roth_fingerprinter
    if _klekota_roth_fingerprinter is not None:
        return _klekota_roth_fingerprinter

    fingerprinter = KlekotaRothFingerprinter()
    n = fingerprinter.getSize()
    if n != 4860:
        raise AssertionError(("Unexpected KlekotaRothFingerprinter size", n, 4860))
    def cdk_klekota_roth_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 4860)
    cdk_klekota_roth_fingerprinter.__qualname__ = "cdk_klekota_roth_fingerprinter"
    cdk_klekota_roth_fingerprinter.cdk_fingerprinter = fingerprinter
    _klekota_roth_fingerprinter = cdk_klekota_roth_fingerprinter # cache for future reuse    
    return cdk_klekota_roth_fingerprinter

@_public
@types.adjust_call_doc
class CDKKlekotaRothFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the Klekota-Roth fingerprints

    The CDK-KlekotaRoth/2.0 fingerprints have no parameters.
    """
    num_bits = 4860
    name = "CDK-KlekotaRoth/2.0"
    _make_fingerprinter = staticmethod(make_klekota_roth_fingerprinter)
    __call__ = _no_param_call


def register_klekota_roth(registry):
    registry.register_if(
        KlekotaRothFingerprinter is not None,
        CDKKlekotaRothFingerprintType_v20,
        "jpype is unable to load CDK's KlekotaRothFingerprinter class",
        )
    
#### CDK-Pubchem

# "this class is not thread-safe and uses stores intermediate steps internally."

PubchemFingerprinter = cdk.fingerprint.PubchemFingerprinter

def make_pubchem_fingerprinter():
    fingerprinter = PubchemFingerprinter(_default_builder)
    n = fingerprinter.getSize()
    if n != 881:
        raise AssertionError(("Unexpected PubchemFingerprinter size", n, 881))
    def cdk_pubchem_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 881)
    cdk_pubchem_fingerprinter.__qualname__ = "cdk_pubchem_fingerprinter"
    cdk_pubchem_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_pubchem_fingerprinter


@_public
@types.adjust_call_doc
class CDKPubchemFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the Klekota-Roth fingerprints

    The CDK-Pubchem/2.0 fingerprints have no parameters.
    """
    num_bits = 881
    name = "CDK-Pubchem/2.0"
    _make_fingerprinter = staticmethod(make_pubchem_fingerprinter)
    __call__ = _no_param_call

def register_pubchem(registry):
    registry.register(CDKPubchemFingerprintType_v20)

#### CDK-Substructure

SubstructureFingerprinter = cdk.fingerprint.SubstructureFingerprinter

def make_substructure_fingerprinter():
    fingerprinter = SubstructureFingerprinter()
    n = fingerprinter.getSize()
    if n != 307:
        raise AssertionError(("Unexpected SubstructureFingerprinter size", n, 307))
    def cdk_substructure_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 307)
    cdk_substructure_fingerprinter.__qualname__ = "cdk_substructure_fingerprinter"
    cdk_substructure_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_substructure_fingerprinter

@_public
@types.adjust_call_doc
class CDKSubstructureFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's Substructure fingerprints

    The CDK-Substructure/2.0 fingerprints have no parameters.
    """
    num_bits = 307
    name = "CDK-Substructure/2.0"
    _make_fingerprinter = staticmethod(make_substructure_fingerprinter)
    __call__ = _no_param_call

def register_substructure(registry):
    registry.register(CDKSubstructureFingerprintType_v20)

#### CDK-ShortestPath

# "It is recommended to use atomtyped container before generating the fingerprints."

# "The FingerPrinter calculates fingerprint based on the Shortest
# Paths between two atoms. It also takes into account ring system,
# charges etc while generating a fingerprint."

# "The FingerPrinter assumes that hydrogens are explicitly given!
# Furthermore, if pseudo atoms or atoms with malformed symbols are
# present, their atomic number is taken as one more than the last
# element currently supported in {@link PeriodicTable}.

ShortestPathFingerprinter = cdk.fingerprint.ShortestPathFingerprinter

def make_shortest_path_fingerprinter(size = DEFAULT_SIZE):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = ShortestPathFingerprinter(size)
    def cdk_shortest_path_fingerprinter(mol):
        try:
            fp = fingerprinter.getBitFingerprint(mol)
        except java.lang.Exception as err:
            if "atom has unset atom type" in str(err.getMessage()):
                # Set the atom types in a copy
                mol2 = mol.clone()
                cdk.tools.manipulator.AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(mol2)
                fp = fingerprinter.getBitFingerprint(mol2)
            else:
                raise
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_shortest_path_fingerprinter.__qualname__ = "cdk_shortest_path_fingerprinter"
    cdk_shortest_path_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_shortest_path_fingerprinter

@_public
@types.adjust_call_doc
class CDKShortestPathFingerprintType_v20(VariableSizeFingerprint):
    """CDK's ShortestPath fingerprint, version 2.0
    
    The CDK-ShortestPath/2.0 :class:`.FingerprintType` parameter is:
    
    * size - the number of bits in the fingerprint (default: 1024)

    This is generated by CDK versions older than 2.7.
    """
    name = "CDK-ShortestPath/2.0"
    fingerprinter_can_fail = True
    _make_fingerprinter = staticmethod(make_shortest_path_fingerprinter)
    _arguments = all_arguments.select_names("size")
    _format_string = "size={size}"
    __call__ = _arguments.get_call_method()

@_public
@types.adjust_call_doc
class CDKShortestPathFingerprintType_v27(VariableSizeFingerprint):
    """CDK's ShortestPath fingerprint, version 2.7
    
    The CDK-ShortestPath/2.7 :class:`.FingerprintType` parameter is:
    
    * size - the number of bits in the fingerprint (default: 1024)

    This version is new in CDK 2.7, where the internal PRNG was changed
    from a Mersenne Twister to XorShift, resulting in a different
    bit pattern.
    """
    name = "CDK-ShortestPath/2.7"
    fingerprinter_can_fail = True
    _make_fingerprinter = staticmethod(make_shortest_path_fingerprinter)
    _arguments = all_arguments.select_names("size")
    _format_string = "size={size}"
    __call__ = _arguments.get_call_method()

def register_shortest_path(registry):
    mol = _tk.from_smiles("c1ccccc1O")
    cdk.tools.manipulator.AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(mol)
    fingerprinter = ShortestPathFingerprinter(128)
    fp = fingerprinter.getBitFingerprint(mol)
    bits = list(fp.getSetbits())
    IS_V20 = (bits == [14, 30, 31, 38, 44, 61, 62, 64, 77, 96, 103, 116, 125, 127])
    IS_V27 = (bits == [16, 17, 21, 24, 34, 38, 45, 48, 63, 66, 86, 98, 107, 115])
    registry.register_if(IS_V20, CDKShortestPathFingerprintType_v20)
    registry.register_if(IS_V27, CDKShortestPathFingerprintType_v27)
    
#### CDK-Circular

#  "Circular fingerprints: for generating fingerprints that are
#  functionally equivalent to ECFP-2/4/6 and FCFP-2/4/6 fingerprints,
#  which are partially described by Rogers et al. {@cdk.cite
#  Rogers2010}."

CircularFingerprinter = cdk.fingerprint.CircularFingerprinter

CLASS_ECFP0 = CircularFingerprinter.CLASS_ECFP0
CLASS_ECFP2 = CircularFingerprinter.CLASS_ECFP2
CLASS_ECFP4 = CircularFingerprinter.CLASS_ECFP4
CLASS_ECFP6 = CircularFingerprinter.CLASS_ECFP6

CLASS_FCFP0 = CircularFingerprinter.CLASS_FCFP0
CLASS_FCFP2 = CircularFingerprinter.CLASS_FCFP2
CLASS_FCFP4 = CircularFingerprinter.CLASS_FCFP4
CLASS_FCFP6 = CircularFingerprinter.CLASS_FCFP6

# ECFP

def make_ecfp0_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP0, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp0_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp0_fingerprinter.__qualname__ = "cdk_ecfp0_fingerprinter"
    cdk_ecfp0_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp0_fingerprinter

@_public
@types.adjust_call_doc
class CDKECFP0FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP0 fingerprint, version 2.0
    
    The CDK-ECFP0/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP0/2.0"
    _make_fingerprinter = staticmethod(make_ecfp0_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_ecfp0(registry):
    registry.register(CDKECFP0FingerprintType_v20)

def make_ecfp2_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP2, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp2_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp2_fingerprinter.__qualname__ = "cdk_ecfp2_fingerprinter"
    cdk_ecfp2_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp2_fingerprinter

@_public
@types.adjust_call_doc
class CDKECFP2FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP2 fingerprint, version 2.0
    
    The CDK-ECFP2/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP2/2.0"
    _make_fingerprinter = staticmethod(make_ecfp2_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()


def register_ecfp2(registry):
    registry.register(CDKECFP2FingerprintType_v20)
    
def make_ecfp4_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP4, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp4_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp4_fingerprinter.__qualname__ = "cdk_ecfp4_fingerprinter"
    cdk_ecfp4_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp4_fingerprinter

@_public
@types.adjust_call_doc
class CDKECFP4FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP4 fingerprint, version 2.0
    
    The CDK-ECFP4/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP4/2.0"
    _make_fingerprinter = staticmethod(make_ecfp4_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_ecfp4(registry):
    registry.register(CDKECFP4FingerprintType_v20)

def make_ecfp6_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP6, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp6_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp6_fingerprinter.__qualname__ = "cdk_ecfp6_fingerprinter"
    cdk_ecfp6_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp6_fingerprinter

@_public
@types.adjust_call_doc
class CDKECFP6FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP6 fingerprint, version 2.0
    
    The CDK-ECFP6/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP6/2.0"
    _make_fingerprinter = staticmethod(make_ecfp6_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_ecfp6(registry):
    registry.register(CDKECFP6FingerprintType_v20)
    
# FCFP

def make_fcfp0_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP0, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp0_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp0_fingerprinter.__qualname__ = "cdk_fcfp0_fingerprinter"
    cdk_fcfp0_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp0_fingerprinter

@_public
@types.adjust_call_doc
class CDKFCFP0FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP0 fingerprint, version 2.0
    
    The CDK-FCFP0/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP0/2.0"
    _make_fingerprinter = staticmethod(make_fcfp0_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_fcfp0(registry):
    registry.register(CDKFCFP0FingerprintType_v20)    
    
def make_fcfp2_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP2, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp2_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp2_fingerprinter.__qualname__ = "cdk_fcfp2_fingerprinter"
    cdk_fcfp2_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp2_fingerprinter

@_public
@types.adjust_call_doc
class CDKFCFP2FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP2 fingerprint, version 2.0
    
    The CDK-FCFP2/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP2/2.0"
    _make_fingerprinter = staticmethod(make_fcfp2_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_fcfp2(registry):
    registry.register(CDKFCFP2FingerprintType_v20)

def make_fcfp4_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP4, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp4_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp4_fingerprinter.__qualname__ = "cdk_fcfp4_fingerprinter"
    cdk_fcfp4_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp4_fingerprinter

@_public
@types.adjust_call_doc
class CDKFCFP4FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP4 fingerprint, version 2.0
    
    The CDK-FCFP4/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP4/2.0"
    _make_fingerprinter = staticmethod(make_fcfp4_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_fcfp4(registry):
    registry.register(CDKFCFP4FingerprintType_v20)
    
def make_fcfp6_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP6, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp6_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp6_fingerprinter.__qualname__ = "cdk_fcfp6_fingerprinter"
    cdk_fcfp6_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp6_fingerprinter

@_public
@types.adjust_call_doc
class CDKFCFP6FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP6 fingerprint, version 2.0
    
    The CDK-FCFP6/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP6/2.0"
    _make_fingerprinter = staticmethod(make_fcfp6_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"
    __call__ = _arguments.get_call_method()

def register_fcfp6(registry):
    registry.register(CDKFCFP6FingerprintType_v20)

#### CDK-AtomPairs2D

# "Generates an atom pair 2D fingerprint as implemented in PaDEL given
# an {@link IAtomContainer}, that extends the {@link Fingerprinter}."

# fixed length

AtomPairs2DFingerprinter = cdk.fingerprint.AtomPairs2DFingerprinter
_atom_pairs_2d_fingerprinter = None

def make_atom_pairs_2d_fingerprinter():
    global _atom_pairs_2d_fingerprinter
    x = _atom_pairs_2d_fingerprinter
    if x is not None:
        return x
    
    fingerprinter = AtomPairs2DFingerprinter()
    n = fingerprinter.getSize()
    if n != 780:
        raise AssertionError(("Unexpected AtomPairs2DFingerprinter size", n, 780))
    def cdk_atom_pairs_2d_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 780)
    cdk_atom_pairs_2d_fingerprinter.__qualname__ = "cdk_atom_pairs_2d_fingerprinter"
    cdk_atom_pairs_2d_fingerprinter.cdk_fingerprinter = fingerprinter
    _atom_pairs_2d_fingerprinter = cdk_atom_pairs_2d_fingerprinter # cache for future reuse
    return cdk_atom_pairs_2d_fingerprinter

@_public
@types.adjust_call_doc
class CDKAtomPairs2DFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the atom pairs fingerprint from Yap Chun Wei's PaDEL.

    The CDK-AtomPairs2D/2.0 :class:`.FingerprintType` takes no parameters.
    """
    num_bits = 780
    name = "CDK-AtomPairs2D/2.0"
    _make_fingerprinter = staticmethod(make_atom_pairs_2d_fingerprinter)
    __call__ = _no_param_call


def register_atom_pairs_2d(registry):
    registry.register(CDKAtomPairs2DFingerprintType_v20)

#### CDK-Signature

# Can't support -- variable length
# cdk/fingerprint/SignatureFingerprinter.java

