# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

import argparse

import sys
from .. import cdk_toolkit
from .. import cdk_types

from .. import get_fingerprint_type
from .. import open_fingerprint_writer
from .. import ChemFPError
from . import cmdsupport


############ Command-line parser definition

epilog = """\

By default the CDK structure reader determines the file format
and compression type based on the filename extension. Unknown
filename extensions are treated as a uncompressed SMILES files.

If the data comes from stdin, or the guess based on extension name is
wrong, then use "--in FORMAT" option to change the default input format.
For examples:

   --in smi
   --in sdf.gz

Use `-R` to specify format-specific reader arguments.

Use `--help-formats` for a list of available formats and reader arguments.
"""


def help_formats():
    print("""\
These are the structure file formats that chemfp and read when using
the CDK toolkit.

By default, chemfp uses the filename extension to determine the format
type. If the filename ends with ".gz" or ".zst" then it is intepreted
as a gzip or Zstandard compressed file, and the second-to-last
extension is used to determine the format type. Unknown or unsupported
extensions are interpreted as a SMILES file.

Note: Zstandard support may depend the "zstandard" Python package
and/or the "zstd-jni" Java package. To install the Python package see
https://pypi.org/project/zstandard/ . To get the Java jar file, see
https://github.com/luben/zstd-jni and place it in your CLASSPATH. 

You may instead specify the file format by name (see below), which is
especially important when reading from stdin, which has no associated
filename extension.

The supported filename extensions are:
""")
    print("File Type".center(15) + " Extension(s)")
    print("==========".center(15) + " =============")
    format_lines = []
    format_data = (
            ("SMILES", ["smi", "can", "usm"]),
            ("SDF", ["sdf"]),
            ("InChI", ["inchi"]),
            )
    for format_type, format_names in format_data:
        extensions = set()
        for name in format_names:
            extensions.update(cdk_toolkit.get_format(name).extensions)
        msg = format_type.center(15) + " " + ", ".join(sorted(extensions))
        format_lines.append(format_type.center(15) + " " + ", ".join(format_names))
        print(msg)

    print("""\

The format can also be specified by name using the '--in' option:

   File Type    Format name (append .gz or .zst if compressed)
   ==========   ==============================================""")
    for format_line in format_lines:
        print(format_line)
                      
    print("""\

The input format parsers can be configured with the "-R" option. For
example, the following reader arguments tell the SMILES readers that
the fields are whitespace delimited and the first line is a header.

   -R delimiter=whitespace -R has_header=true

The SMILES format parsers use two additional reader arguments:
   * 'delimiter' specifies the delimiter type. The default is 'to-eol'.
     The other values are 'tab', 'whitespace', 'space' and 'native'.
     Use "-R delimiter=native" to match RDKit's native delimiter
     style, which is 'whitespace'.
   * 'has_header', if false will skip the first line of the SMILES
     file (because it is a header line).
   * 'kekulise': The default of '1' will Kekulize the SMILES. Use '0'
     to skip this step.
   * 'implementation': The default 'cdk' uses CDK's IteratingSMILESReader()
     to parse the SMILES file. The 'chemfp' implementation uses chemfp's
     Python-based SMILES file parser and CDK's SmilesParser() to parse
     parse each SMILES string. The chemfp implementation is slower
     but may have better error-handling and/or reporting.

The SDF format parser supports five reader arguments:
   * 'mode' can be one of 'RELAXED' or 'STRICT'. The default relaxed
     mode supports some records with recoverable errors. The strict
     mode fails to parse those records.
   * 'ForceReadAs3DCoordinates', with the default of '0' interprets
     V2000 records where all z-coordinates == 0.0 as 2D records. The
     value '1' tells CDK to interpret all records as 3D.
   * 'AddStereoElements' with the default of '1' adds 0D stereochemistry
     to V2000 records. The value of '0' skips that step.
   * 'InterpretHydrogenIsotopes with the default of '1' interprets the
     atom symbols 'D' and 'T' as [2H] and [3H], respectively. Use
     '0' to disable this interpretation.
   * 'implementation': The default 'cdk' uses CDK's SDFReaderFactory()
     to parse the SD file. The 'chemfp' implementation uses chemfp's
     SD file parser to parse records, and CDK's MDLReader(),
     MDLV2000Reader(), or MDLV3000Reader() to parse each record. The
     chemfp implementation is about 50% slower than the cdk parser but
     may have better error-handling and/or reporting.

The InChI format parser supports one reader argument:
   * 'delimiter' works the same as it does for the SMILES formats

""")

    

parser = argparse.ArgumentParser(
    description="Generate FPS or FPB fingerprints from a structure file using CDK via JPype",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )
group = parser.add_mutually_exclusive_group()

cmdsupport.add_type_and_using(group)
group.add_argument("--Daylight", action="store_true",
                       help="Make Daylight-like fingerprints using cdk.fingerprinter.Fingerprinter (default)")
group.add_argument("--GraphOnly", action="store_true",
                       help="Make Daylight-like fingerprints (ignoring bond types) using GraphOnlyFingerprinter")
group.add_argument("--MACCS", action="store_true",
                       help="Make 166-bit MACCS keys using MACCSFingerprinter")
group.add_argument("--EState", action="store_true",
                       help="Make 79-bit EState fingerprints using EStateFingerprinter")
group.add_argument("--Extended", action="store_true",
                       help="Make Daylight-like fingerprints extended with ring feature bits, using ExtendedFingerprinter")
group.add_argument("--Hybridization", action="store_true",
                       help="Make Daylight-like fingerprints based on SP2 hybridization instead of aromaticity, using HybridizationFingerprinter")
group.add_argument("--KlekotaRoth", action="store_true",
                       help="Make 4860-bit Klekota-Roth fingerprint, using KlekotaRothFingerprinter")
group.add_argument("--Pubchem", action="store_true",
                       help="Make 881-bit PubChem fingerprint, using PubchemFingerprinter")
group.add_argument("--Substructure", action="store_true",
                       help="Make 307-bit substructure fingerprint, using SubstructureFingerprinter")
group.add_argument("--ShortestPath", action="store_true",
                       help="Make fingerprints based on the shortest path between atoms, ring systems, and more, using ShortestPathFingerprinter")
group.add_argument("--ECFP0", action="store_true",
                       help="Make ECFP0-like circular fingerprints, using CircularFingerprinter(CLASS_ECFP0)")
group.add_argument("--ECFP2", action="store_true",
                       help="Make ECFP2-like circular fingerprints, using CircularFingerprinter(CLASS_ECFP2)")
group.add_argument("--ECFP4", action="store_true",
                       help="Make ECFP4-like circular fingerprints, using CircularFingerprinter(CLASS_ECFP4)")
group.add_argument("--ECFP6", action="store_true",
                       help="Make ECFP6-like circular fingerprints, using CircularFingerprinter(CLASS_ECFP6)")
group.add_argument("--FCFP0", action="store_true",
                       help="Make FCFP0-like circular feature fingerprints, using CircularFingerprinter(CLASS_FCFP0)")
group.add_argument("--FCFP2", action="store_true",
                       help="Make FCFP2-like circular feature fingerprints, using CircularFingerprinter(CLASS_FCFP2)")
group.add_argument("--FCFP4", action="store_true",
                       help="Make FCFP4-like circular feature fingerprints, using CircularFingerprinter(CLASS_FCFP4)")
group.add_argument("--FCFP6", action="store_true",
                       help="Make FCFP6-like circular feature fingerprints, using CircularFingerprinter(CLASS_FCFP6)")
group.add_argument("--AtomPairs2D", action="store_true",
                       help="Make 780-bit atom-pair fingerprints adapted from Yap Chun Wei's PaDEL, using AtomPairs2DFingerprinter")

# Python's argparse doesn't display an argument group in an argument group, but it does accept the options.
# I don't think anyone should use this for CDK, so I'm keeping it hidden-but-accessible for now.

substruct_group = group.add_argument_group(
    "chemfp's version of the 881 bit PubChem substructure keys")
substruct_group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints (you like want to use --Pubchem instead)")

rdmaccs_group = group.add_argument_group(
    "chemfp's version of the 166 bit RDKit/MACCS keys")
rdmaccs_group.add_argument(
    "--rdmaccs", "--rdmaccs/2", dest="rdmaccs/2", action="store_true", help="generate 166 bit RDKit/MACCS fingerprints (version 2) (")
rdmaccs_group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs (there should be no reason to use this instead of --MACCS or --rdmaccs) ")

cdk_arguments = cdk_types.all_arguments
cmdsupport.add_argument_to_argparse(cdk_arguments["size"], parser)
cmdsupport.add_argument_to_argparse(cdk_arguments["searchDepth"], parser)
cmdsupport.add_argument_to_argparse(cdk_arguments["pathLimit"], parser)
cmdsupport.add_argument_to_argparse(cdk_arguments["hashPseudoAtoms"], parser)
cmdsupport.add_argument_to_argparse(cdk_arguments["perceiveStereochemistry"], parser)
                       
parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id  (SD files only)")

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default autodetects from the filename extension)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")
parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_progress_bar(parser)

parser.add_argument(
    "--help-formats", action=cmdsupport.CallFunctionAction,
    default=help_formats,
    help="list the available formats and reader arguments")

parser.add_argument(
    "-R", metavar="NAME=VALUE", dest="reader_settings", action="append",
    help="specify a reader argument")
cmdsupport.add_reader_aliases(parser)

cmdsupport.add_version(parser)
cmdsupport.add_license_check(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")


#########

def run():
    cmdsupport.run(main)

def get_daylight_kwargs(args):
    return {
        "size": args.size,
        "searchDepth": args.searchDepth,
        "pathLimit": args.pathLimit,
        "hashPseudoAtoms": args.hashPseudoAtoms,
        }
def get_circular_kwargs(args):
    return {
        "size": args.size,
        "perceiveStereochemistry": args.perceiveStereochemistry,
        }
    

def main(args=None):
    args = parser.parse_args(args)

    if args.type is not None:
        fptype = cmdsupport.get_fptype_from_type_arg(parser, args.type, "cdk2fps", "cdk")
    elif args.using is not None:
        fptype = cmdsupport.get_fptype_from_using_arg(parser, args.using, "cdk2fps", "cdk")
    elif args.Daylight:
        fptype = get_fingerprint_type("CDK-Daylight", get_daylight_kwargs(args))
    elif args.GraphOnly:
        fptype = get_fingerprint_type("CDK-GraphOnly", get_daylight_kwargs(args))
    elif args.MACCS:
        fptype = get_fingerprint_type("CDK-MACCS")
    elif args.EState:
        fptype = get_fingerprint_type("CDK-EState")
    elif args.Extended:
        fptype = get_fingerprint_type("CDK-Extended", get_daylight_kwargs(args))
    elif args.Hybridization:
        fptype = get_fingerprint_type("CDK-Hybridization", get_daylight_kwargs(args))
    elif args.KlekotaRoth:
        if cdk_types.KlekotaRothFingerprinter is None:
            parser.error("JPype is unable to find CDK's KlekotaRothFingerprinter class")
        fptype = get_fingerprint_type("CDK-KlekotaRoth")
    elif args.Pubchem:
        fptype = get_fingerprint_type("CDK-Pubchem")
    elif args.Substructure:
        fptype = get_fingerprint_type("CDK-Substructure")
    elif args.ShortestPath:
        fptype = get_fingerprint_type("CDK-ShortestPath", {"size": args.size})
        
    elif args.ECFP0:
        fptype = get_fingerprint_type("CDK-ECFP0", get_circular_kwargs(args))
    elif args.ECFP2:
        fptype = get_fingerprint_type("CDK-ECFP2", get_circular_kwargs(args))
    elif args.ECFP4:
        fptype = get_fingerprint_type("CDK-ECFP4", get_circular_kwargs(args))
    elif args.ECFP6:
        fptype = get_fingerprint_type("CDK-ECFP6", get_circular_kwargs(args))
        
    elif args.FCFP0:
        fptype = get_fingerprint_type("CDK-FCFP0", get_circular_kwargs(args))
    elif args.FCFP2:
        fptype = get_fingerprint_type("CDK-FCFP2", get_circular_kwargs(args))
    elif args.FCFP4:
        fptype = get_fingerprint_type("CDK-FCFP4", get_circular_kwargs(args))
    elif args.FCFP6:
        fptype = get_fingerprint_type("CDK-FCFP6", get_circular_kwargs(args))
        
    elif args.AtomPairs2D:
        fptype = get_fingerprint_type("CDK-AtomPairs2D")

    elif args.substruct:
        fptype = get_fingerprint_type("ChemFP-Substruct-CDK")

    elif getattr(args, "rdmaccs/1", False):
        fptype = get_fingerprint_type("RDMACCS-CDK/1")

    elif getattr(args, "rdmaccs/2", False):
        fptype = get_fingerprint_type("RDMACCS-CDK/2")

    else:
        # Default is Daylight fingerprints
        fptype = get_fingerprint_type("CDK-Daylight", get_daylight_kwargs(args))

    try:
        cdk_toolkit.get_input_format(args.format)
    except ValueError as err:
        parser.error("Unsupported format specifier %r: %s" % (args.format, err))
    
    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    reader_settings = cmdsupport.get_reader_settings(parser, args)

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
    
    cmdsupport.check_filenames(parser, cdk_toolkit, args.format, args.filenames)


    # Ready the input reader/iterator
    metadata, num_readers, first_reader, readers = cmdsupport.read_multifile_structure_fingerprints(
        fptype, args.filenames, format = args.format, id_tag = args.id_tag,
        reader_settings = reader_settings, errors = args.errors)

    with first_reader:
        try:
            writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
        except NotImplementedError as err:
            cmdsupport.writer_failure("cdk2fps", str(err))

        except (IOError, ValueError) as err:
            msg = "Cannot open output fingerprint file: %s" % (err,)
            if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
                msg += ".\nUse --out to specify 'fps', 'fps.gz', 'fps.zst', 'fpb', or 'flush'."
            parser.error(msg)

        try:
            cmdsupport.read_write_fingerprints(args, num_readers, readers, writer)
        except ChemFPError as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
    
if __name__ == "__main__":
    main() #pragma: no cover
