# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import os
import sys
import contextlib

import argparse
from .. import __version__ as chemfp_version
from .. import ChemFPError, ParseError, FingerprintIterator
from .. import io

try:
    import chemfp_converters
except ImportError:
    has_chemfp_converters = False
else:
    has_chemfp_converters = True

######
    
def add_version(parser):
    parser.add_argument("--version", action="version",
                        version="%(prog)s " + chemfp_version)
    

###### argparse types

def add_precision(parser):
    parser.add_argument(
        "--precision",
        metavar = "N",
        default = None,
        choices = ("1", "2", "3", "4", "5", "6", "7", "8", "9", "10"),
        help = "Number of digits in Tanimoto score (default: based on the fingerprint size)",
        )
    
    
######

class CallFunctionAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(CallFunctionAction, self).__init__(option_strings, dest, nargs=0, **kwargs)
    def __call__(self, parser, namespace, values, option_string=None):
        self.default()
        raise SystemExit(0)

######

def license_check():
    from .. import is_licensed, get_license_date
    if is_licensed():
        year, month, day = get_license_date()
        if year > 3000:
            # I'll let the future worry about this warning.
            sys.stdout.write("No license key required for this version of chemfp.\n")
        else:
            # Give a warning about possible problems
            sys.stdout.write("The license is valid until %4d-%02d-%02d.\n" % (
                year, month, day))
            check_license_date(sys.stdout)
        raise SystemExit(0)
            
    sys.stdout.write("Unable to use all chemfp functionality.\n")
    check_license_failure(None, sys.stdout)
    raise SystemExit(1)
    

def add_license_check(parser):
    parser.add_argument("--license-check", action=CallFunctionAction,
                            default=license_check,
                            help="Check the license and report results to stdout.")

######

class MyExitStack(contextlib.ExitStack):
    def call_if(self, func):
        if func is None:
            return
        return super().callback(func)
    def enter_if(self, obj):
        if obj is None:
            return
        return super().enter_context(obj)

######

def mutual_exclusion(parser, args, default, groups):
    true_groups = []
    for g in groups:
        x = getattr(args, g)
        if x == "":
            # --type and --using take string arguments but the
            # empty string shouldn't be accepted as False
            true_groups.append(g)
        elif getattr(args, g):
            true_groups.append(g)

    if not true_groups:
        setattr(args, default, True)
    elif len(true_groups) == 1:
        pass
    else:
        parser.error("Cannot specify both --%s and --%s" % (true_groups[0], true_groups[1]))


class AbortHandler(object):
    def error(self, msg, location):
        raise ParseError("Each of the first 100 records contained errors. Final error: %s" % (msg,), location)

def get_cleanup_reader(reader, id_tag, error_handler):
    location = reader.location
    return FingerprintIterator(
        reader.metadata,
        cleanup_ids_iter(reader, id_tag, error_handler, location),
        location = reader.location,
        )

def cleanup_ids_iter(reader, id_tag, error_handler, location):
    try:
        err_count = 0
        for recno, (id, fp) in enumerate(reader):
            if not id:
                if recno == err_count == 99:
                    error_handler = AbortHandler()
                err_count += 1

                if location.record_format == "sdf":
                    if id_tag is None:
                        error_handler.error("Missing title in SD record", location)
                    else:
                        error_handler.error("Missing id tag %r in SD record" % (id_tag,), location)
                elif location.record_format == "smi":
                    error_handler.error("Missing SMILES identifier (second column)", location)
                else:
                    error_handler.error("Missing identifier", location)
                continue

            else:
                # This assumes that these characters are rare
                if "\n" in id:
                    id = id.partition("\n")[0]  # get up to the newline
                if "\r" in id:
                    id = id.replace("\r", "")
                if "\t" in id:
                    id = id.replace("\t", "")
                if "\0" in id:
                    id = id.replace("\0", "")
                if " " in id:
                    id = id.strip()

                if not id:
                    if recno == err_count == 99:
                        error_handler = AbortHandler()
                    err_count += 1
                    if location.record_format == "sdf":
                        if id_tag is None:
                            error_handler.error("Empty title in SD record after cleanup", location)
                        else:
                            error_handler.error("Empty id tag %r in SD record after cleanup"
                                                % (id_tag,), location)
                    elif location.record_format == "smi":
                        error_handler.error("Empty SMILES identifier (second column) after cleanup", location)
                    else:
                        error_handler.error("Empty identifier after cleanup", location)
                    continue
                else:
                    yield id, fp
    finally:
        reader.close()
    

def sys_exit_opener(fptype, source, format, id_tag, reader_settings, error_handler):
    # Normalize to the format type associated with this source/format combination.
    format = fptype.toolkit.get_input_format_from_source(source, format)
    try:
        reader_args = format.get_reader_args_from_text_settings(reader_settings)
    except ValueError as err:
        sys.stderr.write("Could not process a -R option: %s. Exiting.\n" % err)
        raise SystemExit(1)
        
    try:
        return fptype.read_molecule_fingerprints(
            source, format, id_tag, reader_args=reader_args,
            errors=error_handler)
            
    except (IOError, ChemFPError, ValueError) as err:
        sys.stderr.write("Problem reading structure fingerprints: %s. Exiting.\n" % err)
        raise SystemExit(1)

## def iter_all_sources(fptype, filenames, format, id_tag, errors):
##     for filename in filenames:
##         reader = sys_exit_opener(fptype, filename, format, id_tag, errors)
##         for id, fp in reader:
##             yield id, fp

# readers (metadata, num_readers, iter_of_readers)
def read_multifile_structure_fingerprints(fptype, filenames, format, id_tag, reader_settings, errors):
    error_handler = io.get_parse_error_handler(errors)
    if errors == "strict":
        id_error_handler = error_handler
    else:
        id_error_handler = io.get_parse_error_handler("report")
    
    if not filenames:
        reader = sys_exit_opener(fptype, None, format, id_tag, reader_settings, error_handler)
        cleanup_reader = get_cleanup_reader(reader, id_tag, id_error_handler)
        return reader.metadata, 1, cleanup_reader, [cleanup_reader]

    if len(filenames) == 1:
        reader = sys_exit_opener(fptype, filenames[0], format, id_tag, reader_settings, error_handler)
        cleanup_reader = get_cleanup_reader(reader, id_tag, id_error_handler)
        return reader.metadata, 1, cleanup_reader, [cleanup_reader]

    # Get the first reader
    first_reader = sys_exit_opener(fptype, filenames[0], format, id_tag, reader_settings, error_handler)
    first_reader.metadata.sources = filenames

    def multifile_iter(first_reader, filenames):
        yield first_reader
        first_reader.close()
        del first_reader

        for filename in filenames:
            reader = sys_exit_opener(fptype, filename, format, id_tag,
                                         reader_settings, error_handler)
            cleanup_reader = get_cleanup_reader(reader, id_tag, id_error_handler)
            yield cleanup_reader
            reader.close()

    return first_reader.metadata, len(filenames), first_reader, multifile_iter(first_reader, filenames[1:])

def read_write_fingerprints(args, num_readers, readers, writer):
    with writer:
        for reader_i, reader in enumerate(readers):
            with get_progress_bar(args, reader.location, file_count=(reader_i, num_readers)) as progress_bar:
                writer.write_fingerprints(progress_bar(reader))

def is_valid_tag(tag):
    if tag is None:
        return True
    for c in "<>\r\n":
        if c in tag:
            return False
    return True


def get_reader_settings(parser, args):
    if not args.reader_settings:
        return {}
    
    reader_settings = {}
    for r_option in args.reader_settings:
        left, mid, right = r_option.partition("=")
        if mid != "=":
            parser.error("-R setting %r must be in the format NAME=VALUE" % (r_option,))
        reader_settings[left] = right
    return reader_settings

def get_writer_settings(parser, args):
    if not args.writer_settings:
        return {}
    
    writer_settings = {}
    for r_option in args.writer_settings:
        left, mid, right = r_option.partition("=")
        if mid != "=":
            parser.error("-R setting %r must be in the format NAME=VALUE" % (r_option,))
        writer_settings[left] = right
    return writer_settings

def check_filenames(parser, toolkit, format, filenames):
    if not filenames:
        # Read from stdin.
        return None
    for filename in filenames:
        if not os.path.exists(filename):
            parser.error("Structure file %r does not exist" % (filename,))
        try:
            toolkit.get_input_format_from_source(source=filename, format=format)
        except ValueError as err:
            parser.error(str(err))

def add_argument_to_argparse(argument, parser):
    """Add information about the given parameter to an argparse parser

    This calls parser.add_argument() using "--$NAME" as the argument name
    and along with values for `type`, `default`, `action`, `metavar`,
    and `help`.
    """
    # Wrapper function to convert the decoder() API into
    # the argparse API.
    def parse_argument(s):
        try:
            return argument.decoder(s)
        except ValueError as err:
            raise argparse.ArgumentError(None, "--%s %s" % (argument.name, err))

    argparse_kwargs = dict(type=parse_argument,
                           default=argument.default,
                           action=argument.action,
                           metavar=argument.metavar,
                           help=argument.help)

    parser.add_argument("--" + argument.name, **argparse_kwargs)


class DelimiterAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.reader_settings is None:
            namespace.reader_settings = []
        namespace.reader_settings.append("delimiter=%s" % (values,))
        
class HasHeaderAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.reader_settings is None:
            namespace.reader_settings = []
        namespace.reader_settings.append("has_header=1")


def add_delimiter_alias(parser):
    parser.add_argument("--delimiter", choices=("tab", "whitespace", "to-eol", "space"),
                        action=DelimiterAction,
                        help="delimiter style for SMILES and InChI files. Alias for '-R delimiter=VALUE'.")
def add_has_header_alias(parser):
    parser.add_argument("--has-header", action=HasHeaderAction, nargs=0,
                        help="Skip the first line of a SMILES or InChI file Alias for '-R has_header=1'")
    
def add_reader_aliases(parser):
    add_delimiter_alias(parser)
    add_has_header_alias(parser)


def add_type_and_using(parser):
    parser.add_argument("--type", metavar="TYPE_STRING",
                            help="Specify a chemfp type string")
    parser.add_argument("--using", metavar="FILENAME",
                            help="Get the fingerprint type from the metadata of a fingerprint file")

if hasattr(argparse, "BooleanOptionalAction"):
    # This wasn't added until Python 3.9
    BooleanOptionalAction = argparse.BooleanOptionalAction
else:
    class BooleanOptionalAction(argparse.Action):
        def __init__(self,
                     option_strings,
                     dest,
                     default=None,
                     type=None,
                     choices=None,
                     required=False,
                     help=None,
                     metavar=None):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)

                if option_string.startswith('--'):
                    option_string = '--no-' + option_string[2:]
                    _option_strings.append(option_string)

            if help is not None and default is not None:
                help += f" (default: {default})"

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar)

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith('--no-'))

        def format_usage(self):
            return ' | '.join(self.option_strings)

    
def add_progress_bar(parser):
    parser.add_argument(
        "--progress",
        action = BooleanOptionalAction,
        default = None,
        help = "Show a progress bar (default: show unless the output is a terminal)")

def progress_bar_is_enabled(progress, filenames, output):
    if progress is not None:
        # User specified --progress or --no-progress
        return bool(progress)

    stderr_isatty = getattr(sys.stderr, "isatty", None)
    if stderr_isatty is None:
        # File-like object that doesn't implement isatty.
        # Disable because we're in the test suite.
        return False

    if not stderr_isatty():
        # Not a terminal, no progress bar
        return False

    if output is None:
        # Sending to stdout. Don't mix stdout and stderr on the terminal.
        # Can't use isatty() since the output could be pipe'ed to,
        # eg, head, which also writes to stdout.
        # For this case, disable and require the user to enabled.
        return False

    # Output is a file. Check the input.
    if filenames is None or filenames == []:
        # Don't use a progress bar if stdin is a terminal.
        # (Someone typing or copy/pasting input.)
        stdin_isatty = getattr(sys.stdin, "isatty", None)
        if stdin_isatty is None:
            # stdin is a file-like object.
            # Disable because we're in the test suite.
            return False

        return not stdin_isatty()
            
    # Looks good!
    return True
        

def get_progress_bar(args, location, file_count=None):
    is_enabled = progress_bar_is_enabled(args.progress, args.filenames, args.output)
    return location.progress_bar(disable = not is_enabled, file_count=None)

def get_fptype_from_type_arg(parser, typestr, progname, toolkit_name):
    from .. import get_fingerprint_type
    try:
        fptype = get_fingerprint_type(typestr)
    except ValueError as err:
        sys.stderr.write("Unable to use the specified --type: %s\n" % (err,))
        raise SystemExit(1)
    
    if fptype.toolkit is None:
        sys.stderr.write("Unable to use the specified --type: %r is not associated with a toolkit\n" % (
            typestr,))
        raise SystemExit(1)
    
    if fptype.toolkit.name != toolkit_name:
        sys.stderr.write(
            "WARNING: %s uses the %s toolkit but --type %r expects fingerprints using the %s toolkit.\n" % (
                progname, toolkit_name, typestr, fptype.toolkit.name))
    return fptype

def get_fptype_from_using_arg(parser, using_str, progname, toolkit_name):
    from .. import open as chemfp_open, get_fingerprint_type, ChemFPError
    try:
        reader = chemfp_open(using_str)
    except ChemFPError as err:
        sys.stderr.write("Cannot open --using file %r: %s\n" % (using_str, err))
        raise SystemExit(1)
    except (OSError, IOError, ValueError) as err:
        sys.stderr.write("Cannot open --using file %r: %s\n" % (using_str, err))
        raise SystemExit(1)

    with reader:
        typestr = reader.metadata.type
        if typestr is None:
            sys.stderr.write("No fingerprint type found in the --using file %r\n" % (
                using_str,))
            raise SystemExit(1)
        try:
            fptype = get_fingerprint_type(typestr)
        except ValueError as err:
            sys.stderr.write("Unable to use the fingerprint type from the --using file %r: %s\n" % (
                using_str, err))
            raise SystemExit(1)
            
    if fptype.toolkit is None:
        sys.stderr.write("Unable to use the fingerprint type in --using file %r: %r is not associated with a toolkit\n" % (
            using_str, typestr))
        raise SystemExit(1)
    
    if fptype.toolkit.name != toolkit_name:
        sys.stderr.write(
            "WARNING: Fingerprint type %r from the --using file %r is based on the %s toolkit.\n" % (
                typestr, using_str, fptype.toolkit.name))
        sys.stderr.write(
            "WARNING: %s expects fingerprints based on the %s toolkit.\n" % (
                progname, toolkit_name))
    return fptype

# This is used by 'simsearch', 'oe2fps', etc. with a top-level driver.
def run(main_func, *args, **kwargs):
    try:
        main_func(*args, **kwargs)
    except (SystemExit, BrokenPipeError) as err:
        # BrokenPipeError and SIGINT could have occurred at the same time
        try:
            sys.stdout.flush()
        except BrokenPipeError:
            handle_broken_pipe()
        except ValueError:
            # Have something to write to
            sys.stdout = io.StringIO()
        
        if isinstance(err, SystemExit):
            raise
    else:
        try:
            sys.stdout.flush()
        except BrokenPipeError:
            handle_broken_pipe()


def die(msg):
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    sys.stderr.flush()
    raise SystemExit(1)


def writer_failure(tool_name, msg):
    check_license_failure(msg)
    sys.stderr.write("Exiting %s.\n" % (tool_name,))
    raise SystemExit(1)
    

def check_license_failure(msg, outfile=None):
    from .. import get_license, get_license_date
    from .. import _core

    if outfile is None:
        outfile=sys.stderr
    
    def get_reason():
        license = get_license()
        if license is None:
            if "CHEMFP_LICENSE" in os.environ:
                license = os.environ["CHEMFP_LICENSE"]
                import re
                license_pat = re.compile("[0-9]{8}-[ -~]*@([A-Z]{4})?[A-Z]{32}$")
                if not license_pat.match(license) or license.count("@") != 1:
                    return "$CHEMFP_LICENSE does not contain a chemfp license."
                return "$CHEMFP_LICENSE does not appear to be a valid chemfp license."
            
            if "CHEMFP_LICENSE_FILE" in os.environ:
                return "Could not get chemfp license key from $CHEMFP_LICENSE_FILE."
            
            return ("Neither CHEMFP_LICENSE nor CHEMFP_LICENSE_FILE "
                        "environment variables are defined.")

        year, month, day = get_license_date()
        where = {
            1: "The license in the file $CHEMFP_LICENSE_FILE",
            2: "The license in $CHEMFP_LICENSE",
            3: "The license in the file --license-file",
            4: "The embedded license",
            }.get(
                _core.get_license_origin(),
                "The license")
        import time
        now = time.gmtime()
        if (year, month, day) < (now.tm_year, now.tm_mon, now.tm_mday):
            return "%s expired on %04d-%02d-%02d or is not valid." % (
                where, year, month, day)
        else:
            return "%s expires on %04d-%02d-%02d but does not appear to be valid." % (
                where, year, month, day)

    errmsg = get_reason()
    if msg is not None:
        outfile.write(msg)
        if not msg.endswith("\n"):
            outfile.write("\n")
        
    outfile.write("The license key check failed. %s\n" % (errmsg,))
    outfile.write("Email sales@dalkescientific.com to purchase a license key or request a demo license key.\n")


def check_license_date(outfile=None):
    from .. import get_license_date
    if outfile is None:
        outfile = sys.stderr        
    year, month, day = get_license_date()
    if year > 3000:
        # I'll let the future worry about this warning.
        return
    import datetime
    delta = datetime.datetime(year, month, day, 23, 59, 59) - datetime.datetime.utcnow()
    if delta.days < 7:
        if delta.days == 0:
            outfile.write("WARNING: Your chemfp license expires TODAY.\n")
        elif delta.days == 1:
            outfile.write("WARNING: Your chemfp license expires TOMORROW.\n")
        else:
            outfile.write("WARNING: Your chemfp license expires in %d days.\n" % (delta.days,))

def die_no_diversity_license(subcommand, n, outfile=None):
    if outfile is None:
        outfile = sys.stderr
    write_no_diversity_license(subcommand, n, outfile)
    raise SystemExit(1)

def write_no_diversity_license(subcommand, n, outfile):
    if n == 1:
        extra = "the candidates must be a licensed FPB file"
    elif n == 2:
        extra = "the candidates and references must be licensed FPB files"
    elif n == 3:
        extra = "the neighbors must be a licensed FPB file"
    else:
        raise AssertionError("bad count", n)
    outfile.write(f"Cannot run {subcommand}. No license found for this feature.\n")
    outfile.write(f"It requires a valid chemfp diversity license, or {extra}.\n")
    outfile.write("Email sales@dalkescientific.com to purchase a license key or request a demo license key.\n")
    outfile.write("Exiting.\n")
    
    
            
# Following the recommendation at https://docs.python.org/3/library/signal.html#note-on-sigpipe
def handle_broken_pipe():
    # Python flushes standard streams on exit; redirect remaining output
    # to devnull to avoid another BrokenPipeError at shutdown
    devnull = os.open(os.devnull, os.O_WRONLY)
    os.dup2(devnull, sys.stdout.fileno())
    sys.exit(1)  # Python exits with error code 1 on EPIPE

