import argparse
import datetime
import sys
import random

from .. import bitops

from .. import (
    check_metadata_problems,
    load_fingerprints,
    open_fingerprint_writer,
    ChemFPError,
    )

from .. import io as chemfp_io

from . import cmdsupport

TQDM_KWARGS = dict(
    desc = "picks",
    unit = "",
    dynamic_ncols = True,
    )

def get_progress_bar(progress, total):
    from .. import _tqdm
    return _tqdm.get_tqdm_constructor(progress)(
        total = total,
        **TQDM_KWARGS,
        )

def utcnow():
    return datetime.datetime.utcnow().replace(microsecond=0)

### argparse types support functions

def parse_threshold(value):
    try:
        value = float(value)
    except ValueError:
        raise argparse.ArgumentTypeError("must be a float")

    if not (0.0 <= value <= 1.0):
        raise argparse.ArgumentTypeError("must be between 0.0 and 1.0, inclusive")
    
    return value

def parse_num_picks(value):
    if value is None or value == "all":
        return None
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("must be an integer or 'all'")
    if value < 0:
        raise argparse.ArgumentTypeError("must be a non-negative integer or 'all'")
    return value

def parse_batch_size(value):
    if value == "auto":
        return None
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("must be a positive integer or 'auto'")
    if value < 1:
        raise argparse.ArgumentTypeError("must be a positive integer or 'auto'")
    return value

def parse_seed(value):
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("must be an integer")
    
    if value == -1:
        # match the internal API
        return random.randrange(2**64)
    
    if not (0 <= value < 2**64):
        raise argparse.ArgumentTypeError("must an integer between -1 and 18446744073709551615 (2**64-1), inclusive")
    
    return value

### add to parser (shared with spherex)

def add_threshold(parser):
    parser.add_argument(
        "-t",
        "--threshold",
        metavar = "FLOAT",
        type = parse_threshold,
        default = 1.0,
        help = "Maximum similarity (default: 1.0)",
        )

def add_num_picks(parser):
    parser.add_argument(
        "--num-picks", "-n",
        metavar = "N",
        type = parse_num_picks,
        default = None,
        help = "Number of picks (default: 'all')",)

def add_all_equal(parser):
    parser.add_argument(
        "--all-equal",
        action = "store_true",
        help = "Continue picking past --num-picks if the pick score is unchanged",
        )

def add_candidates_format(parser):
    parser.add_argument(
        "--in",
        "--candidates-format",
        dest = "candidates_format",
        help = "Format of the candidates file (default uses filename extension, or 'fps')",
        )

def add_references_and_format(parser):
    parser.add_argument(
        "--references", metavar = "FILENAME",
        help = "Fingerprint file containing reference fingerprints to avoid (the fingerprints you have)",
        )

    parser.add_argument(
        "--references-format",
        metavar = "FORMAT",
        help = "Format of the references file (default uses filename extension, or 'fps')",
        )

def add_seed(parser):
    parser.add_argument(
        "--seed",
        metavar = "N",
        default = -1,
        type = parse_seed,
        help = (
            "Specify the random number generator seed between 0 and 2**64-1, inclusive, or "
            "use -1 to have one picked at random (default: -1)"),
        )


def add_neighbors(parser):
    parser.add_argument(
        "--neighbors",
        metavar="FILENAME",
        help = "For each pick, includes the nearest neighbor and score from FILENAME",
        )
    parser.add_argument(
        "--neighbors-format",
        metavar="FILENAME",
        help = "Format of the neighbors file (default uses filename extension, or 'fps')",
        )

def add_mmap(parser):
    parser.add_argument(
        "--mmap",
        default = True,
        action = cmdsupport.BooleanOptionalAction,
        help = "Don't use mmap to read uncompressed FPB files. "
        "May give better performance on networked file systems, "
        "at the expense of higher memory use.")
    
def add_picker_output_args(parser, precision):
    parser.add_argument(
        "--output",
        "-o",
        metavar = "FILENAME",
        help = "Write output to the named file instead of stdout.",)
    
    parser.add_argument(
        "--out",
        dest = "output_format",
        help = (
            "Output format. Must be one of 'chemfp' (the default), 'csv', 'tsv', or 'excel-tab', "
            "with optional compression"
            ),
        )

    if precision:
        cmdsupport.add_precision(parser)

    parser.add_argument(
        "--save-picks",
        metavar = "FILENAME",
        help = "Write picked fingerprints to the named file.",
        )

    parser.add_argument(
        "--save-picks-format",
        metavar = "FILENAME",
        help = "Specify the format for the picked fingerprints.",
        )

    parser.add_argument(
        "--save-candidates",
        metavar = "FILENAME",
        help = "Write remaining candidate fingerprints to the named file.",
        )

    parser.add_argument(
        "--save-candidates-format",
        metavar = "FILENAME",
        help = "Specify the format for the remaining candidate fingerprints.",
        )

def add_pick_time(parser):
    parser.add_argument(
        "--pick-time",
        default = False,
        action = cmdsupport.BooleanOptionalAction,
        help = "include the elapsed time for each pick",
        )
    
    
def add_date(parser):
    parser.add_argument(
        "--no-date",
        action="store_true",
        help = "Do not include the 'date' metadata in the output header",
        )
    parser.add_argument(
        "--date",
        metavar = "STR",
        help = "An ISO 8601 date (like '2022-02-07T11:10:15') to use for the 'date' metadata in the output header",
        )
    
def add_times(parser):
    parser.add_argument(
        "--times",
        action="store_true",
        help = "Write timing information to stderr",
        )

def add_candidates(parser):
    parser.add_argument(
        "candidates",
        nargs = 1,
        help = "fingerprint file containing candidates (fingerprints to pick from)",
        )

## Get values used parsed args

def get_candidates_arena(parser, args, subcommand, progress):
    candidates_filename = args.candidates[0]
    candidates_format = args.candidates_format
    allow_mmap = args.mmap

    try:
        candidates_arena = load_fingerprints(candidates_filename, format=candidates_format,
                                                 progress=progress, allow_mmap=allow_mmap)
    except NotImplementedError as err:
        cmdsupport.check_license_failure(str(err))
        cmdsupport.die(f"Cannot run {subcommand}. Exiting.")
        
    except ChemFPError as err:
        parser.error(f"Cannot parse candidates file: {err}")
    except (IOError, ValueError) as err:
        parser.error(f"Cannot open candidates file: {err}")

    if len(candidates_arena) < 1:
        parser.error("Candidates file must contain at least one fingerprint")

    if not candidates_arena.popcount_indices:
        parser.error("Candidate fingerprints must have a popcount index")

    return candidates_filename, candidates_arena

def check_metadata_compatability(name1, metadata1, name2, metadata2):
    problems = check_metadata_problems(metadata1, metadata2, name1, name2)
    for problem in problems:
        if problem.severity == "warning":
            print("WARNING:", problem.description, file=sys.stderr)
        elif problem.severity == "error":
            cmdsupport.die(f"ERROR: {problem.description}")


def get_references_arena(parser, args, candidates_arena, subcommand, progress):
    references_filename = args.references
    references_format = args.references_format
    allow_mmap = args.mmap

    if references_filename is None:
        return None, None

    try:
        references_arena = load_fingerprints(
            references_filename,
            format = references_format,
            allow_mmap = allow_mmap,
            )
    except NotImplementedError as err:
        cmdsupport.check_license_failure(str(err))
        cmdsupport.die(f"Cannot run chemfp {subcommand}. Exiting.")
    except ChemFPError as err:
        parser.error(f"Cannot parse references file: {err}")
    except (IOError, ValueError) as err:
        parser.error(f"Cannot open references file: {err}")

    check_metadata_compatability(
        "references", references_arena.metadata,
        "candidates file", candidates_arena.metadata,
        )

    return references_filename, references_arena

def get_neighbors_arena(parser, args, candidates_arena, subcommand, progress):
    neighbors_filename = args.neighbors
    neighbors_format = args.neighbors_format
    allow_mmap = args.mmap

    if neighbors_filename is None:
        return None, None

    try:
        neighbors_arena = load_fingerprints(
            neighbors_filename,
            format = neighbors_format,
            allow_mmap = allow_mmap)
    except NotImplementedError as err:
        cmdsupport.check_license_failure(str(err))
        cmdsupport.die(f"Cannot run chemfp {subcommand}. Exiting.")
    except ChemFPError as err:
        parser.error(f"Cannot parse neighbors file: {err}")
    except (IOError, ValueError) as err:
        parser.error(f"Cannot open neighbors file: {err}")

    check_metadata_compatability(
        "neighbors", neighbors_arena.metadata,
        "candidates file", candidates_arena.metadata,
        )

    return neighbors_filename, neighbors_arena


def get_tanimoto_precision(args, candidates_arena):
    if args.precision is None:
        return bitops.get_tanimoto_precision(candidates_arena.num_bits)
    else:
        return int(args.precision)

def get_output(parser, args, blame_name):
    destination = args.output
    format, compression = chemfp_io.normalize_output_format(
        destination, args.output_format, ("chemfp", ""),
        )
    try:
        f, close = chemfp_io.open_compressed_output(destination, compression, blame_name)
    except IOError as err:
        parser.error(f"Cannot open --output file: {err}")
    return format, f, close

def get_save_picks_writer(parser, args, candidates_arena, subcommand):
    if not args.save_picks:
        return None

    metadata = candidates_arena.metadata.copy(
        sources = [], date = utcnow(),
        )
    try:
        return open_fingerprint_writer(
            args.save_picks,
            metadata = metadata,
            format = args.save_picks_format,
            reorder = False,
            )
    except NotImplementedError as err:
        cmdsupport.check_license_failure(str(err))
        cmdsupport.die(f"Cannot run chemfp {subcommand}. Exiting.")
        
    except (ValueError, IOError, OSError) as err:
        parser.error(f"Cannot open --save-picks file: {err}")

def get_save_candidates_writer(parser, args, candidates_arena, subcommand):
    if not args.save_candidates:
        return None

    metadata = candidates_arena.metadata.copy(
        sources = [], date = utcnow(),
        )
    try:
        return open_fingerprint_writer(
            args.save_candidates,
            metadata = metadata,
            format = args.save_candidates_format,
            )
    except NotImplementedError as err:
        cmdsupport.check_license_failure(str(err))
        cmdsupport.die(f"Cannot run chemfp {subcommand}. Exiting.")
        
    except (ValueError, IOError, OSError) as err:
        parser.error(f"Cannot open --save-candidates file: {err}")

def get_date_str(parser, args):
    if args.no_date:
        if args.date:
            parser.error("Cannot specify both --date and --no-date")
        return None
    import datetime
    if args.date:
        try:
            when = datetime.datetime.fromisoformat(args.date)
        except ValueError as err:
            # if date = "blah" the error ValueError includes the repr()
            # if date = "2021-12-25T25:00:00" it does not contain the repr()
            # I want the repr, so add it unless it already exists
            msg = f"Cannot parse --date: {err}"
            date_repr = repr(args.date)
            if not msg.endswith(date_repr):
                msg = f"{msg}: {date_repr}"
            parser.error(msg)
    else:
        when = utcnow()

    return when.isoformat()

####

def urlquote(s):
    from urllib.parse import quote
    return quote(s)


def report_times(
        show_times,
        init_start_time, init_end_time,
        pick_start_time, pick_end_time,
        start_time, end_time,
        num_picked,
        ):
    if not show_times:
        return
    init_dt = init_end_time - init_start_time
    pick_dt = max(0.0, pick_end_time - pick_start_time)
    total_dt = end_time - start_time

    if pick_dt == 0.0:
        picks_per_second = 0.0
    else:
        picks_per_second = num_picked / pick_dt

    try:
        sys.stderr.write(
            f"T_init: {init_dt:.2f} "
            f"T_pick: {pick_dt:.2f} "
            f"#picks: {num_picked} "
            f"picks/s: {picks_per_second:.2f} "
            f"T_total: {total_dt:.2f}"
            "\n"
            )
    except BrokenPipeError:
        pass

def save_remaining_candidates(
        candidates_arena,
        candidates_writer,
        picker):
    if candidates_writer is None:
        return
    if picker is None:
        return
    
    candidates_writer.write_fingerprints(
        candidates_arena[candidate.candidate_idx] for candidate in picker.candidates)
    
def save_remaining_candidates_spherex(
        candidates_arena,
        candidates_writer,
        picker):
    if candidates_writer is None:
        return
    if picker is None:
        return
    
    candidates_writer.write_fingerprints(
        candidates_arena[candidate] for candidate in picker.candidates)
    
