import argparse
import sys

from .. import io, fpb_io
from .. import registry

def parse_label(s):
    try:
        label = s.encode("ascii")
    except UnicodeEncodeError:
        raise argparse.ArgumentTypeError("label must be ASCII")

    if b"\n" in label:
        raise argparse.ArgumentTypeError("label may not contain a newline")
    return label

parser = argparse.ArgumentParser(
    prog = "fpb_text",
    description = "Show the TEXT sections of an FPB file")
parser.add_argument("--list-labels", action="store_true",
                        help="only show the labels of the TEXT blocks")
parser.add_argument("--label", action="append", type=parse_label,
                        help="only show the named label(s). Can be used more than once.")
parser.add_argument("--in", metavar="FORMAT", dest="source_format",
                        choices=("fpb", "fpb.gz", "fpb.zst"),
                        help="input FPB format")
parser.add_argument("filename")

def fpb_text_command(parser, args):
    if args.label is None:
        labels = None
    else:
        labels = set(args.label)
        
    # In principle I could work directly on the FPB file contents.
    # It's easier to load the file into memory and extract the data from there.
    source = args.filename
    format_name, compression = io.normalize_input_format(source, args.source_format)
    try:
        arena = fpb_io.open_fpb(source, format_name, compression, allow_mmap=True)
    except (ValueError, IOError) as err:
        sys.stderr.write("Cannot open FPB file: %s\n" % (err,))
        raise SystemExit(9)
    
    # If mmap'ed, the arena starts with FPB_MAGIC.
    # If read(), it does not.
    check_magic = arena.arena[:8].startswith(fpb_io.FPB_MAGIC)

    # Don't need to check for errors as open_fpb() did that.
    for chunk in fpb_io.iter_chunks(arena.arena, check_magic=check_magic):
        if chunk.tag == b"TEXT":
            label, nl, content = chunk.content[chunk.start:chunk.end].partition(b"\n")
            if args.list_labels:
                print(label.decode("ascii"))
            elif (labels is None) or (label in labels):
                label = label.decode("ascii")
                print(label)
                print("="*len(label))
                print(content.decode("utf8"))
                print("\n\n")

command_config = registry.CommandConfig.from_argparse(parser, "fp_tools", fpb_text_command)

if __name__ == "__main__":
    command_config.main()
