# Copyright (c) 2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse
from io import TextIOWrapper
import time

from .. import (
    diversity,
    registry,
    )

from ..search import knearest_tanimoto_search_fp

from . import (
    cmdsupport,
    diversity_utils,
    maxmin, # use the same output writers as maxmin
    )

def get_software():
    # This can change if the license changes
    from .. import SOFTWARE
    return SOFTWARE

def get_heapsweep_type(
        *,
        threshold,
        num_picks,
        all_equal,
        randomize,
        seed,
        ):
    terms = ["heapsweep"]
    if threshold is not None:
        terms.append(f"threshold={threshold}")
    if num_picks:
        terms.append(f"num-picks={num_picks}")
        if all_equal:
            terms.append("all-equal=1")
        else:
            terms.append("all-equal=0")
    if randomize:
        terms.append("randomize=1")
    else:
        terms.append("randomize=0")
    # the RNG is used to help find the first hit, even if randomize is False
    terms.append(f"seed={seed}")
    return " ".join(terms)

###
parser = argparse.ArgumentParser(
    description = "Select diverse fingerprints using the heapsweep algorithm",
    allow_abbrev=False,
    )

diversity_utils.add_threshold(parser)
diversity_utils.add_num_picks(parser)
diversity_utils.add_all_equal(parser)

diversity_utils.add_candidates_format(parser)

parser.add_argument(
    "--randomize",
    default = True,
    action = cmdsupport.BooleanOptionalAction,
    help = "Use --randomize (the default) to shuffle the candidates before starting heapsweep",
    )
diversity_utils.add_seed(parser)
diversity_utils.add_mmap(parser)
diversity_utils.add_neighbors(parser)
diversity_utils.add_picker_output_args(parser, precision=True)
diversity_utils.add_pick_time(parser)
diversity_utils.add_date(parser)
diversity_utils.add_times(parser)
cmdsupport.add_progress_bar(parser)

diversity_utils.add_candidates(parser)

####

def heapsweep_command(parser, args):
    with cmdsupport.MyExitStack() as exit_stack:
        return heapsweep_command_with_cleanup(parser, args, exit_stack)

def heapsweep_command_with_cleanup(parser, args, exit_stack):
    start_time = init_start_time = time.time()
    progress = cmdsupport.progress_bar_is_enabled(args.progress, args.candidates, args.output)
    num_picks = args.num_picks
    seed = diversity_utils.parse_seed(args.seed)
    include_pick_time = args.pick_time
    date_str = diversity_utils.get_date_str(parser, args)

    # Figure out the input
    candidates_filename, candidates_arena = diversity_utils.get_candidates_arena(
        parser, args, "heapsweep", progress)
    exit_stack.enter_context(candidates_arena)

    output_format, binary_outfile, binary_outfile_close = diversity_utils.get_output(
        parser, args, "heapsweep")
    exit_stack.call_if(binary_outfile_close)

    # Don't let the wrapper handle the close
    outfile = TextIOWrapper(binary_outfile, write_through=True)
    exit_stack.callback(outfile.detach)
    
    picks_writer = diversity_utils.get_save_picks_writer(parser, args, candidates_arena, "heapsweep")
    exit_stack.enter_if(picks_writer)
    
    candidates_writer = diversity_utils.get_save_candidates_writer(parser, args, candidates_arena, "heapsweep")
    exit_stack.enter_if(candidates_writer)

    neighbors_filename, neighbors_arena = diversity_utils.get_neighbors_arena(
        parser, args, candidates_arena, "heapsweep", progress)
        
    if neighbors_arena is not None:
        if len(neighbors_arena) == 0:
            neighbors_arena.close()
            neighbors_filename = None
            neighbors_arena = None
        else:
            exit_stack.enter_context(neighbors_arena)

    tanimoto_precision = diversity_utils.get_tanimoto_precision(args, candidates_arena)

    # Create the picker
    try:
        picker = diversity.get_heapsweep_picker(
            candidate_arena = candidates_arena,
            randomize = args.randomize,
            seed = seed,
            )
    except NotImplementedError as err:
        if "chemfp diversity license" in str(err):
            cmdsupport.die_no_diversity_license("heapsweep", 1)
        raise
    
    # Check if we can search the neighbors
    if neighbors_arena is not None:
        try:
            test_fp = neighbors_arena.fingerprints[len(neighbors_arena)//2]
            knearest_tanimoto_search_fp(test_fp, neighbors_arena, k=1)
        except NotImplementedError as err:
            if "chemfp license key" in str(err):
                cmdsupport.die_no_diversity_license("heapsweep", 3)
            raise
        del test_fp
    
    # We've got a picker! Let's get to work.
    output_writer = maxmin.get_output_writer(parser, args, output_format, outfile, tanimoto_precision)

    output_writer.write_header(outfile, {
        "num_bits": candidates_arena.num_bits,
        "type": get_heapsweep_type(
            threshold = args.threshold,
            num_picks = num_picks,
            all_equal = args.all_equal,
            randomize = args.randomize,
            seed = seed,
            ),
        "software": get_software(),
        "candidates": candidates_filename,
        "neighbors": neighbors_filename,
        "date": date_str,
        })
    

    elapsed_time = None
    neighbor_id = neighbor_score = None

    output_writer.write_titles(
        outfile,
        include_neighbor = (neighbors_arena is not None),
        include_pick_time = include_pick_time,
        )
    
    pick_start_time = init_end_time = time.time()
    
    try:
        it = picker.iter_indices_and_scores(max_similarity = args.threshold)

        for pick_i, pick_idx, pick_score in maxmin.enumerate_picks(
                progress, it, num_picks, len(candidates_arena), args.all_equal):
            pick_id = candidates_arena.ids[pick_idx]
            
            if picks_writer is not None:
                pick_fp = candidates_arena.fingerprints[pick_idx]
                picks_writer.write_fingerprint(pick_id, pick_fp)

            if neighbors_arena is not None:
                if picks_writer is None:
                    pick_fp = candidates_arena.fingerprints[pick_idx]
                neighbor_hits = knearest_tanimoto_search_fp(
                    pick_fp, neighbors_arena, k=1, threshold=0.0)
                assert neighbor_hits, pick_id
                neighbor_id, neighbor_score = neighbor_hits.get_ids_and_scores()[0]

            if include_pick_time:
                elapsed_time = time.time() - pick_start_time

            output_writer.write_result(
                outfile,
                pick_i, pick_id, pick_score,
                neighbor_id, neighbor_score,
                elapsed_time,
                )

            # Improve interactive use. Otherwise the lag is too long.
            # We expect picks to be slow, so the flush() overhead shouldn't be an issue.
            outfile.flush()
            
    finally:
        # Always print times, even with ^C and pipe errors,
        pick_end_time = end_time = time.time()

        # flush to prevent mixup between stdout and stderr
        try:
            outfile.flush()
        except Exception:
            pass
            
        diversity_utils.report_times(
            args.times,
            init_start_time, init_end_time,
            pick_start_time, pick_end_time,
            start_time, end_time,
            len(picker.picks))
            
        diversity_utils.save_remaining_candidates(candidates_arena, candidates_writer, picker)
    
def main(args=None):
    args = parser.parse_args(args)
    return heapsweep_command(parser, args)

command_config = registry.CommandConfig.from_argparse(parser, "fp_diversity", heapsweep_command)

if __name__ == "__main__":
    main()
