# Copyright (c) 2021-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse
import time

from io import TextIOWrapper

from . import (
    cmdsupport,
    diversity_utils,
    )
from .. import (
    diversity,
    registry,
    )
from ..search import knearest_tanimoto_search_fp

def get_software():
    # This can change if the license changes
    from .. import SOFTWARE
    return SOFTWARE

def get_maxmin_type(
        *,
        num_initial_picks,
        threshold,
        num_picks,
        all_equal,
        randomize,
        seed,
        ):
    terms = ["maxmin"]
    if num_initial_picks:
        terms.append(f"--num-initial-picks={num_initial_picks}")
    if threshold is not None:
        terms.append(f"threshold={threshold}")
    if num_picks is not None:
        terms.append(f"num-picks={num_picks}")
        if all_equal:
            terms.append("all-equal=1")
        else:
            terms.append("all-equal=0")
    if randomize:
        terms.append(f"randomize=1 seed={seed}")
    elif not num_initial_picks:
        # the RNG is used in the heapsweep for the intial pick,
        # which *may* affect the result.
        terms.append(f"randomize=0 seed={seed}")
    else:
        # the RNG is not used
        terms.append("randomize=0")
    return " ".join(terms)

class BaseWriter:
    def __init__(self, precision):
        self.precision = precision
        self.score_formatter = f"%.{precision}f"
        self.time_formatter = "%.2f"

    def _get_titles_and_format_terms(self, include_neighbor, include_pick_time):
        # (i, pick_id, diversity_score, (neighbor_id, neighbor_score)?, pick_time?)
        titles = ["i", "pick_id", "score"]
        format_terms = ["%d", "%s", self.score_formatter]
        
        if include_neighbor:
            titles.extend(["neighbor_id", "neighbor_score"])
            format_terms.extend(["%s", self.score_formatter])

        if include_pick_time:
            titles.append("pick_time")
            format_terms.append(self.time_formatter)

        return titles, format_terms
    
    def write_header(self, outfile, d):
        raise NotImplementedError #pragma: no cover

    def write_result(self, outfile, pick_i, pick_id, pick_score,
                         neighbor_id, neighbor_score, elapsed_time):
        raise NotImplementedError #pragma: no cover

_header_encoders = {
    "candidates": diversity_utils.urlquote,
    "references": diversity_utils.urlquote,
    "neighbors": diversity_utils.urlquote,
    }

class ChemfpWriter(BaseWriter):
    def write_header(self, outfile, d):
        outfile.write("#Diversity/1\n")
        for k, v in d.items():
            if v is not None:
                if k in _header_encoders:
                    v = _header_encoders[k](v)
                outfile.write(f"#{k}={v}\n")
    
    def write_titles(self, outfile, include_neighbor, include_pick_time):
        # (i, pick_id, diversity_score, (neighbor_id, neighbor_score)?, pick_time?)
        titles, format_terms = self._get_titles_and_format_terms(include_neighbor, include_pick_time)
        
        self.line_formatter = "\t".join(format_terms) + "\n"
        outfile.write("\t".join(titles) + "\n")


    def write_result(self, outfile, pick_i, pick_id, pick_score,
                         neighbor_id, neighbor_score, elapsed_time):
        if neighbor_id is None:
            if elapsed_time is None:
                line = self.line_formatter % (pick_i, pick_id, pick_score)
            else:
                line = self.line_formatter % (pick_i, pick_id, pick_score, elapsed_time)
        else:
            if elapsed_time is None:
                line = self.line_formatter % (pick_i, pick_id, pick_score, neighbor_id, neighbor_score)
            else:
                line = self.line_formatter % (pick_i, pick_id, pick_score, neighbor_id, neighbor_score, elapsed_time)
        
        outfile.write(line)

class ExcelWriter(BaseWriter):
    def __init__(self, outfile, precision, dialect):
        super().__init__(precision)
        import csv
        self.csv_writer = csv.writer(outfile, dialect=dialect)
        self.writerow = self.csv_writer.writerow

    def write_header(self, outfile, d):
        pass
                         
    def write_titles(self, outfile, include_neighbor, include_pick_time):
        # (pick_id, diversity_score, (neighbor_id, neighbor_score)?, pick_time?)
        titles, _ = self._get_titles_and_format_terms(include_neighbor, include_pick_time)
        self.writerow(titles[1:])

    def write_result(self, outfile, pick_i, pick_id, pick_score,
                         neighbor_id, neighbor_score, elapsed_time):
        score_formatter = self.score_formatter
        if neighbor_id is None:
            if elapsed_time is None:
                columns = [pick_id, score_formatter % pick_score]
            else:
                columns = [pick_id, score_formatter % pick_score, self.time_formatter % elapsed_time]
        else:
            if elapsed_time is None:
                columns = [pick_id, score_formatter % pick_score,
                               neighbor_id, score_formatter % neighbor_score]
            else:
                columns = [pick_id, score_formatter % pick_score,
                               neighbor_id, score_formatter % neighbor_score, self.time_formatter % elapsed_time]
        self.writerow(columns)

def get_output_writer(
    parser, args, output_format, outfile, precision):

    if output_format in (None, "chemfp"):
        output_format = "chemfp"

    if output_format == "csv":
        dialect = "excel"
    elif output_format == "tsv":
        dialect = "excel-tab"
    else:
        dialect = None
        
    if dialect is not None:
        return ExcelWriter(outfile, precision, dialect = dialect)

    elif (output_format == "chemfp") or (args.output_format is None):
        return ChemfpWriter(precision)

    parser.error(f"Unsupported --out format {args.output_format!r}")

def enumerate_picks(progress, it, num_picks, num_candidates, all_equal):
    if num_picks == 0:
        return

    if num_picks is None:
        with diversity_utils.get_progress_bar(progress, num_candidates) as progress_bar:
            for i, (pick_idx, pick_score) in enumerate(it, 1):
                yield i, pick_idx, pick_score
                progress_bar.update(1)
        return

    i = 0
    with diversity_utils.get_progress_bar(progress, num_picks) as progress_bar:
        while i < num_picks:
            try:
                pick_idx, pick_score = next(it)
            except StopIteration:
                break
            i += 1
            yield i, pick_idx, pick_score
            progress_bar.update(1)

        if all_equal:
            equal_score = pick_score
            while 1:
                try:
                    pick_idx, pick_score = next(it)
                except StopIteration:
                    break
                i += 1
                if pick_score != equal_score:
                    break
                yield i, pick_idx, pick_score
                progress_bar.update(1)
            
###
parser = argparse.ArgumentParser(
    description="Select diverse fingerprints using the MaxMin algorithm",
    allow_abbrev=False,
    )

diversity_utils.add_num_picks(parser)
diversity_utils.add_threshold(parser)
diversity_utils.add_all_equal(parser)

parser.add_argument(
    "--pick-id",
    type = str,
    default = None,
    help = "Candidate id to use for the initial pick (default: use heapsweep)"
    )
parser.add_argument(
    "--pick-index",
    type = int,
    default = None,
    help = "Candidate index to use for the initial pick (default: use heapsweep)"
    )

diversity_utils.add_candidates_format(parser)
diversity_utils.add_references_and_format(parser)

parser.add_argument(
    "--randomize",
    default = True,
    action = cmdsupport.BooleanOptionalAction,
    help = "Use --randomize (the default) to shuffle the candidates before starting MaxMin",
    )
diversity_utils.add_seed(parser)
diversity_utils.add_neighbors(parser)
diversity_utils.add_mmap(parser)
diversity_utils.add_picker_output_args(parser, precision=True)
diversity_utils.add_pick_time(parser)
diversity_utils.add_date(parser)
diversity_utils.add_times(parser)
cmdsupport.add_progress_bar(parser)

diversity_utils.add_candidates(parser)

####

def maxmin_command(parser, args):
    with cmdsupport.MyExitStack() as exit_stack:
        return maxmin_command_with_exit_stack(parser, args, exit_stack)
                
def maxmin_command_with_exit_stack(parser, args, exit_stack):
    start_time = init_start_time = time.time()
    
    progress = cmdsupport.progress_bar_is_enabled(args.progress, args.candidates, args.output)
    seed = diversity_utils.parse_seed(args.seed)
    num_picks = args.num_picks
    include_pick_time = args.pick_time
    date_str = diversity_utils.get_date_str(parser, args)

    if args.pick_id is not None:
        if args.pick_index is not None:
            parser.error("Cannot specify both --pick-id and --pick-index")
        if not args.pick_id:
            parser.error("--pick-id must not be the empty string")
        pick_flag = "--pick-id"
    elif args.pick_index is not None:
        pick_flag = "--pick-index"
    else:
        pick_flag = None
        
    if args.references is not None:
        if pick_flag is not None:
            parser.error(f"Cannot specify both --references and {pick_flag}")
    
    # Figure out the input
    candidates_filename, candidates_arena = diversity_utils.get_candidates_arena(
        parser, args, "maxmin", progress)
    exit_stack.enter_if(candidates_arena)
    
    initial_pick_idx = initial_pick_id = None
    if pick_flag is not None:
        if args.pick_id is not None:
            initial_pick_idx = candidates_arena.get_index_by_id(args.pick_id)
            if initial_pick_idx is None:
                parser.error(f"--pick-id {args.pick_id!r} not found in --candidates file")
        else:
            try:
                initial_pick_idx = range(len(candidates_arena))[args.pick_index]
            except IndexError:
                parser.error(f"--pick-index {args.pick_index} is out of range for --candidates file")
        # Used for the 'type' metadata
        initial_pick_id = candidates_arena.ids[initial_pick_idx]
        
    
    references_filename, references_arena = diversity_utils.get_references_arena(
        parser, args, candidates_arena, "maxmin", progress)
                                                                
    exit_stack.enter_if(references_arena)

    neighbors_filename, neighbors_arena = diversity_utils.get_neighbors_arena(
        parser, args, candidates_arena, "maxmin", progress)
    
    if neighbors_arena is not None:
        if len(neighbors_arena) == 0:
            neighbors_arena.close()
            neighbors_arena = None
            neighbors_filename = None
        else:
            exit_stack.enter_if(neighbors_arena)
    
    # Figure out the output

    output_format, binary_outfile, binary_outfile_close = diversity_utils.get_output(
        parser, args, "maxmin")
    exit_stack.call_if(binary_outfile_close)

    # Don't let the wrapper handle the close
    outfile = TextIOWrapper(binary_outfile, write_through=True)
    exit_stack.callback(outfile.detach)
                   
    picks_writer = diversity_utils.get_save_picks_writer(parser, args, candidates_arena, "maxmin")
    exit_stack.enter_if(picks_writer)
        
    candidates_writer = diversity_utils.get_save_candidates_writer(parser, args, candidates_arena, "maxmin")
    exit_stack.enter_if(candidates_writer)

    tanimoto_precision = diversity_utils.get_tanimoto_precision(args, candidates_arena)

    # Create the picker
    try:
        picker = diversity.get_maxmin_picker(
            candidate_arena = candidates_arena,
            reference_arena = references_arena,
            initial_pick = initial_pick_idx,
            randomize = args.randomize,
            seed = seed,
            )
    except NotImplementedError as err:
        if "chemfp diversity license" in str(err):
            if references_arena is None:
                cmdsupport.die_no_diversity_license("maxmin", 1)
            else:
                cmdsupport.die_no_diversity_license("maxmin", 2)
        raise

    # Check if we can search the neighbors
    if neighbors_arena is not None:
        try:
            test_fp = neighbors_arena.fingerprints[len(neighbors_arena)//2]
            neighbors_arena.knearest_tanimoto_search_fp(test_fp, k=1)
        except NotImplementedError as err:
            if "chemfp license key" in str(err):
                cmdsupport.die_no_diversity_license("maxmin", 3)
            raise
        del test_fp

    # We've got a picker! Let's get to work.
    output_writer = get_output_writer(parser, args, output_format, outfile, tanimoto_precision)
    
    output_writer.write_header(
        outfile, {
            "num_bits": candidates_arena.num_bits,
            "type": get_maxmin_type(
                num_initial_picks = 1 if (initial_pick_id is not None) else 0,
                threshold = args.threshold,
                num_picks = num_picks,
                all_equal = args.all_equal,
                randomize = args.randomize,
                seed = seed,
                ),
            "software": get_software(),
            "candidates": candidates_filename,
            "references": references_filename,
            "neighbors": neighbors_filename,
            "date": date_str,
            })

    elapsed_time = None
    neighbor_id = neighbor_score = None

    output_writer.write_titles(
        outfile,
        include_neighbor = (neighbors_arena is not None),
        include_pick_time = include_pick_time,
        )
    
    pick_start_time = init_end_time = time.time()

    try:
        it = picker.iter_indices_and_scores(max_similarity = args.threshold)
        
        for pick_i, pick_idx, pick_score in enumerate_picks(
                progress, it, num_picks, len(candidates_arena), args.all_equal):
            pick_id = candidates_arena.ids[pick_idx]
            
            if picks_writer is not None:
                pick_fp = candidates_arena.fingerprints[pick_idx]
                picks_writer.write_fingerprint(pick_id, pick_fp)

            if neighbors_arena is not None:
                if picks_writer is None:
                    pick_fp = candidates_arena.fingerprints[pick_idx]
                neighbor_hits = knearest_tanimoto_search_fp(
                    pick_fp, neighbors_arena, k=1, threshold=0.0)
                assert neighbor_hits, pick_id
                neighbor_id, neighbor_score = neighbor_hits.get_ids_and_scores()[0]

            if include_pick_time:
                elapsed_time = time.time() - pick_start_time

            output_writer.write_result(
                outfile,
                pick_i, pick_id, pick_score,
                neighbor_id, neighbor_score,
                elapsed_time,
                )

            # Improve interactive use. Otherwise the lag is too long.
            # We expect picks to be slow, so the flush() overhead shouldn't be an issue.
            outfile.flush()
            
    finally:
        # Always print times, even with ^C and pipe errors,
        pick_end_time = end_time = time.time()

        # flush to prevent mixup between stdout and stderr
        try:
            outfile.flush()
        except Exception:
            pass
            
        diversity_utils.report_times(
            args.times,
            init_start_time, init_end_time,
            pick_start_time, pick_end_time,
            start_time, end_time,
            len(picker.picks))
            
        diversity_utils.save_remaining_candidates(candidates_arena, candidates_writer, picker)
        
def main(args=None):
    args = parser.parse_args(args)
    return maxmin_command(parser, args)


command_config = registry.CommandConfig.from_argparse(parser, "fp_diversity", maxmin_command)


if __name__ == "__main__":
    main()
