# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse

import sys
from chemfp import openbabel_toolkit
from chemfp import _openbabel_toolkit as _tk
from chemfp import openbabel_types

from .. import ChemFPError
from .. import open_fingerprint_writer, get_fingerprint_type
from . import cmdsupport


############ Command-line parser definition

epilog = """\

By default the Open Babel structure reader determines the file format
and compression type based on the filename extension. Unknown
filename extensions are treated as a uncompressed SMILES files.

If the data comes from stdin, or the guess based on extension name is
wrong, then use "--in FORMAT" option to change the default input format.
For examples:

   --in smi
   --in sdf.gz

Use `-R` to specify format-specific reader arguments.

Use `--help-formats` for a list of available formats and reader arguments.
"""

def parse_read_options(description, w, secondary_indent=2):
    pos = description.find("\nRead Options")
    if pos == -1:
        return ""
    lines = description[pos+1:].splitlines(True)[1:]

    leading_space = None
    for i, line in enumerate(lines):
        if line == "\n":
            break
        n = len(line) - len(line.lstrip())
        if leading_space is None:
            leading_space = n
        else:
            leading_space = min(n, leading_space)

    if leading_space is None:
        return ""

    lines = lines[:i]

    secondary_indent_str = " " * secondary_indent
    indent_str = " "*w + " " + secondary_indent_str
    output_lines = []
    for line in lines:
        # Chop
        line = line[leading_space:]
        # Fix the secondary indent
        line = line[:1] + secondary_indent_str + line[1:].strip() + "\n"
        output_lines.append(indent_str + line)
        
    return "".join(output_lines)
    

def help_formats():
    print("""\
chemfp has special support for the SMILES, InChI, and SDF formats when
using the Open Babel toolkit.

For these formats, by default, chemfp uses the filename extension to
determine the format type. If the filename ends with ".gz" or ".zst"
then it is intepreted as a gzip or Zstandard compressed file, and the
second-to-last extension is used to determine the format type. Unknown
or unsupported extensions are then tested against Open Babel format
names (see below), and if still unknown, interpreted as a SMILES file.
""")
    
    try:
        import zstandard
    except ImportError:
        print("""\
Note: To enable Zstandard compression, please install the "zstandard"
Python package from https://pypi.org/project/zstandard/ .
""")

    print("""\
You will need to use "-R implementation=chemfp" to enable zst support for 
the SDF format.

You may instead specify the file format by name (see below), which is
especially important when reading from stdin, which has no associated
filename extension.

These specially supported filename extensions are:
""")
    print("File Type".center(15) + " Extension(s)")
    print("==========".center(15) + " =============")
    format_lines = []
    format_data = (
            ("SMILES", ["smi", "can", "usm"]),
            ("SDF", ["sdf"]),
            ("InChI", ["inchi"]),
            )
    for format_type, format_names in format_data:
        extensions = set()
        for name in format_names:
            extensions.update(openbabel_toolkit.get_format(name).extensions)
        msg = format_type.center(15) + " " + ", ".join(sorted(extensions))
        format_lines.append(format_type.center(15) + " " + ", ".join(format_names))
        print(msg)

    print("""\

The format can also be specified by name using the '--in' option:

   File Type    Format name (append .gz or .zst if compressed)
   ==========   ===========""")
    for format_line in format_lines:
        print(format_line)

    print("""\

The input format parsers can be configured with the "-R" option. For
examples, the following reader arguments tell the SMILES readers that
the fields are whitespace delimited and the first line is a header.

   -R delimiter=whitespace -R has_header=true

All of the readers support the 'options' reader argument, which is a
string passed directly to OBConversion(). This is a compact way to
encode all of the Open Babel parameters used in the conversion. For
example, 'ab"text"', would set option 'a' to True, and option 'b' to
the string "text".

The SMILES format parsers use two additional reader arguments:
   * 'delimiter' specifies the delimiter type. The default is 'to-eol'.
     The other values are 'tab', 'whitespace', 'space' and 'native'.
     Use "-R delimiter=native" to match Open Babel's native delimiter
     style, which is 'to-eol'.
   * 'has_header', if false will skip the first line
     of the SMILES file (because it is a header line).

The SDF format parser supports one additional reader argument:
   * 'implementation': if "openbabel" or "native", use Open Babel's
     native SDF parser. If "chemfp" use chemfp's own implementation
     to find SDF records, which are then passed to Open Babel for
     parsing. This gives more fine-grained error reporting, and
     supports zst compression, and with similar performance.
  (Note: Open Babel supports additional options.)

The InChI format parser supports one additional reader argument:
   * 'delimiter' works the same as it does for the SMILES formats

In addition, you may specify an Open Babel formats, either by one of
the following format names, or by reading a filename ending with one
of the format names, optionally with a .gz suffix. Zstandard
compression is not supported by the native Open Babel reader.

""")
    _ignore = set(["smi", "can", "usm", "sdf", "inchi"])
    raw_format_names = []
    for fmt in openbabel_toolkit.get_input_formats():
        if fmt.supports_io and fmt.name not in _ignore:
            raw_format_names.append(fmt.name)
    raw_format_names.sort()

    ob = openbabel_toolkit._ob
    lengths = [len(name) for name in raw_format_names]
    if not lengths:
        w = 0
    else:
        w = max(lengths)
    w = max(w, len("Format"))
    w += 2
    print("Format".center(w), "Description and options")
    print("="*w, "==========================")
    for name in raw_format_names:
        obformat = ob.OBFormat_FindType(name)
        if obformat is None:
            continue
        description = obformat.Description()
        title = description.partition("\n")[0].strip()
        print(name.center(w), title)
        s = parse_read_options(description, w, 2)
        if s:
            print(s, end="")

    print("""
You will need to consult the Open Babel documentation
(see https://openbabel.org/wiki/List_of_extensions ) and
implementation for full details about how these options work.
""")

parser = argparse.ArgumentParser(
    description="Generate FPS or FPB fingerprints from a structure file using Open Babel",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )
group = parser.add_mutually_exclusive_group()
group.add_argument("--FP2", action="store_true",
                   help="linear fragments up to 7 atoms")
group.add_argument("--FP3", action="store_true",
                   help="SMARTS patterns specified in the file patterns.txt")
group.add_argument("--FP4", action="store_true",
                   help="SMARTS patterns specified in the file SMARTS_InteLigand.txt")

if openbabel_types.HAS_MACCS:
    # Added in Open Babel 2.3
    group.add_argument("--MACCS", action="store_true",
                       help="Open Babel's implementation of the MACCS 166 keys")
else:
    group.add_argument("--MACCS", action="store_true", #pragma: no cover
                       help="(Not available using your version of Open Babel)")

if openbabel_types.HAS_ECFP:
    group.add_argument("--ECFP0", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 0")
    group.add_argument("--ECFP2", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 2")
    group.add_argument("--ECFP4", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 4")
    group.add_argument("--ECFP6", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 6")
    group.add_argument("--ECFP8", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 8")
    group.add_argument("--ECFP10", action="store_true",
                       help="ECFP (circular) fingerprints with diameter 10")
    # Fix a problem where argparse allows --ECFP1 as an alias for --ECFP10
    group.add_argument('--ECFP1', help=argparse.SUPPRESS)
else:
    parser.set_defaults(ECFP0=False, ECFP2=False, ECFP4=False, ECFP6=False,
                        ECFP8=False, ECFP10=False, ECFP1=False)


           
group.add_argument(
    "--substruct", action="store_true", help="ChemFP substructure fingerprints")

group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="166 bit RDKit/MACCS fingerprints (version 2)")
group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")

if openbabel_types.HAS_ECFP:
    ecfp_group = parser.add_argument_group("ECFP argument")
    for name in ("nBits",):
        cmdsupport.add_argument_to_argparse(openbabel_types.ecfp_arguments[name], ecfp_group)

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id  (SD files only)")

cmdsupport.add_type_and_using(parser)

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default autodetects from the filename extension)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")
parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_progress_bar(parser)

parser.add_argument(
    "--help-formats", action=cmdsupport.CallFunctionAction,
    default=help_formats,
    help="list the available formats and reader arguments")
parser.add_argument(
    "-R", metavar="NAME=VALUE", dest="reader_settings", action="append",
    help="specify a reader argument")
cmdsupport.add_reader_aliases(parser)

cmdsupport.add_version(parser)
cmdsupport.add_license_check(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")


#########

def _check_ecfp():
    if not openbabel_types.ECFP_AVAILABLE:
        sys.stderr.write("This version of Open Babel does not support ECFP fingerprints.\n")
        raise SystemExit(1)
    if not openbabel_types.HAS_ECFP:
        sys.stderr.write("Chemfp does not support pre-3.0 Open Babel ECFP fingerprints.\n")
        raise SystemExit(1)

    
def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    # argparse implements abbreviations based on unique prefix. There
    # is no way to disable that in Python 2.7. I found myself doing
    # "--ECFP1", because I was thinking of radius, not diameter. I
    # ended up with ECFP10, which is not at all what I expected.
    if args.ECFP1:
        parser.error("unrecognized arguments: --ECFP1")
        
    cmdsupport.mutual_exclusion(parser, args, "FP2",
                                ("FP2", "FP3", "FP4", "MACCS",
                                 "ECFP0", "ECFP2", "ECFP4", "ECFP6", "ECFP8", "ECFP10",
                                 "substruct", "rdmaccs", "rdmaccs/1",
                                 "type", "using"))

    if args.FP2:
        fptype = get_fingerprint_type("OpenBabel-FP2")
    elif args.FP3:
        fptype = get_fingerprint_type("OpenBabel-FP3")
    elif args.FP4:
        fptype = get_fingerprint_type("OpenBabel-FP4")
    elif args.MACCS:
        if not openbabel_types.HAS_MACCS:
            parser.error(  #pragma: no cover
                "--MACCS is not supported in your Open Babel installation (%s)" % (
                    _tk.OBReleaseVersion(),))
        fptype = get_fingerprint_type("OpenBabel-MACCS")
    elif args.ECFP0:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP0", {"nBits": args.nBits})
    elif args.ECFP2:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP2", {"nBits": args.nBits})
    elif args.ECFP4:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP4", {"nBits": args.nBits})
    elif args.ECFP6:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP6", {"nBits": args.nBits})
    elif args.ECFP8:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP8", {"nBits": args.nBits})
    elif args.ECFP10:
        _check_ecfp()
        fptype = get_fingerprint_type("OpenBabel-ECFP10", {"nBits": args.nBits})
            
                                       
    elif args.substruct:
        fptype = get_fingerprint_type("ChemFP-Substruct-OpenBabel")
    elif args.rdmaccs:
        fptype = get_fingerprint_type("RDMACCS-OpenBabel")
    elif getattr(args, "rdmaccs/1"):
        fptype = get_fingerprint_type("RDMACCS-OpenBabel/1")
    elif args.type is not None:
        fptype = cmdsupport.get_fptype_from_type_arg(parser, args.type, "ob2fps", "openbabel")
    elif args.using is not None:
        fptype = cmdsupport.get_fptype_from_using_arg(parser, args.using, "ob2fps", "openbabel")
    else:
        parser.error("should not get here") #pragma: no cover

    try:
        openbabel_toolkit.get_input_format(args.format)
    except ValueError as err:
        parser.error("Unsupported format specifier %r: %s" % (args.format, err))
    
    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    reader_settings = cmdsupport.get_reader_settings(parser, args)

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
    
    cmdsupport.check_filenames(parser, openbabel_toolkit, args.format, args.filenames)


    # Ready the input reader/iterator
    metadata, num_readers, first_reader, readers = cmdsupport.read_multifile_structure_fingerprints(
        fptype, args.filenames, format = args.format, id_tag = args.id_tag,
        reader_settings = reader_settings, errors = args.errors)

    with first_reader:
        try:
            writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
        except NotImplementedError as err:
            cmdsupport.writer_failure("ob2fps", str(err))

        except (IOError, ValueError) as err:
            msg = "Cannot open output fingerprint file: %s" % (err,)
            if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
                msg += ".\nUse --out to specify 'fps', 'fps.gz', 'fps.zst', 'fpb', or 'flush'."
            parser.error(msg)

        try:
            cmdsupport.read_write_fingerprints(args, num_readers, readers, writer)
        except ChemFPError as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
    
if __name__ == "__main__":
    main() #pragma: no cover
