# Copyright (c) 2010-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import sys
import textwrap
import argparse

from .. import ChemFPError, get_fingerprint_type
from .. import openeye_toolkit
from .. import openeye_types as oetypes
from . import cmdsupport
from .. import open_fingerprint_writer

##### Handle command-line argument parsing

# Build up some help text based on the atype and btype fields
atype_options = "\n  ".join(textwrap.wrap(" ".join(sorted(dict(oetypes._atype_flags)))))
btype_options = "\n  ".join(textwrap.wrap(" ".join(sorted(dict(oetypes._btype_flags)))))
assert oetypes.OEGRAPHSIM_API_VERSION != "1", "chemfp no longer supports OEGraphSim version 1"

from openeye.oegraphsim import (
    OEGetFPAtomType, OEFPAtomType_DefaultPathAtom,
    OEFPAtomType_DefaultCircularAtom, OEFPAtomType_DefaultTreeAtom,
    OEGetFPBondType, OEFPBondType_DefaultPathBond,
    OEFPBondType_DefaultCircularBond, OEFPBondType_DefaultTreeBond,
    )
type_help = """\
ATYPE is one or more of the following, separated by the '|' character

  %(atype_options)s

The following shorthand terms and expansions are also available:
 DefaultPathAtom = %(defaultpathatom)s
 DefaultCircularAtom = %(defaultcircularatom)s
 DefaultTreeAtom = %(defaulttreeatom)s
and 'Default' selects the correct value for the specified fingerprint.

Examples:
  --atype Default
  --atype "Arom|AtmNum|FCharge|HCount"
  --atype Arom,AtmNum,FCharge,HCount

BTYPE is one or more of the following, separated by the '|' character

  %(btype_options)s
  
The following shorthand terms and expansions are also available:
 DefaultPathBond = %(defaultpathbond)s
 DefaultCircularBond = %(defaultcircularbond)s
 DefaultTreeBond = %(defaulttreebond)s
and 'Default' selects the correct value for the specified fingerprint.

Examples:
   --btype Default
   --btype Order|InRing

To simplify command-line use, a comma may be used instead of a '|' to
separate different fields. Example:
  --atype AtmNum,HvyDegree

""" % dict(atype_options=atype_options,
           btype_options=btype_options,
           defaultpathatom=OEGetFPAtomType(OEFPAtomType_DefaultPathAtom),
           defaultcircularatom=OEGetFPAtomType(OEFPAtomType_DefaultCircularAtom),
           defaulttreeatom=OEGetFPAtomType(OEFPAtomType_DefaultTreeAtom),
           defaultpathbond=OEGetFPBondType(OEFPBondType_DefaultPathBond),
           defaultcircularbond=OEGetFPBondType(OEFPBondType_DefaultCircularBond),
           defaulttreebond=OEGetFPBondType(OEFPBondType_DefaultTreeBond),

)


# Extra help text after the parameter descriptions
epilog = type_help + """\

By default, chemfp will use the filename extension to determine the
structure file format type and possible compression. Most of the file
readers support configuration parameters. Use the '-R' option to
specify those parameters.

Use '--help-formats' to list available formats and reader parameters.

"""

def help_formats():
    print("""\
These are the structure file formats that chemfp can read when using
the OEChem toolkit.

By default, chemfp uses the filename extension to determine the format
type. If the filename ends with ".gz" then it is intepreted as a gzip
compressed file, and the second-to-last extension is used to determine
the format type. Unknown or unsupported extensions are interpreted as
a SMILES file.

(The OEChem structure file readers do not support Zstandard
compression.)

You may instead specify the file format by name (see below), which is
especially important when reading from stdin, which has no associated
filename extension.

The supported filename extensions are:
""")
    print("File Type".center(15) + " Extension(s)")
    print("==========".center(15) + " =============")
    format_lines = []
    format_data = (
            ("SMILES", ["smi", "can", "usm"]),
            ("SDF", ["sdf"]),
            ("InChI", ["inchi"]),
            ("Tripos Mol2", ["mol2", "mol2h"]),
            ("PDB", ["pdb"]),
            ("XYZ", ["xyz"]),
            ("SKC", ["skc"]),
            ("Macromodel", ["mmod"]),
            ("ChemDraw CDX", ["cdx"]),
            ("OE binary", ["oeb"]),
            ("OEB compressed", ["oez"]),
            ("CIF", ["cif"]),
            ("mmCIF", ["mmcif"]),
            ("FASTA", ["fasta"]),
            ("CSV", ["csv"]),
            )
    for format_type, format_names in format_data:
        try:
            openeye_toolkit.get_input_format(format_names[0])
        except ValueError:
            try:
                openeye_toolkit.get_format(format_names[0])
            except ValueError:
                errmsg = "(Unsupported in this OEChem installation)"
            else:
                errmsg = "(Unsupported input format for this OEChem installation)"
            
            msg = (format_type.center(15) + " " + errmsg)
            format_lines.append(
                format_type.center(15) + " " +
                ", ".join(format_names) + " " + errmsg)
        else:
            extensions = set()
            for name in format_names:
                extensions.update(openeye_toolkit.get_format(name).extensions)
            msg = format_type.center(15) + " " + ", ".join(sorted(extensions))
            format_lines.append(format_type.center(15) + " " + ", ".join(format_names))
        print(msg)

    print("""
Append a '.gz' to the filename to indicate that the contents are
gzip-compressed.

The format can also be specified by name using the '--in' option:

   File Type    Format name
   ==========   =============""")
    
    for format_line in format_lines:
        print(format_line)

    print("""
Append a '.gz' to the format name to indicate that the contents are
gzip-compressed.

The input format parsers can be configured with the "-R" option. For
example, the following reader arguments tell the SMILES readers that
the fields are whitespace delimited and the first line is a header.

   -R delimiter=whitespace -R has_header=true

All formats handle the following two reader arguments:

  aromaticity - one of 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
      (this can also be set via the older '--aromaticity' command-line option)

  flavor - a '|' or ',' separated list of flavor names, or a numeric value.
       A leading '-' means to remove the given flavor. Examples include:
       
       o  Canon,Strict  -- the bitwise merger of the format's Canon and Strict values
       o  Default,-Kekule -- the format's Default flavor but without the Kekule bits
                      (every flavor has a Default)
       o  42  -- the specific OEChem flavor value 42

The SMILES and InChI formats also handle reader arguments for the
delimiter style and the presence of an initial header line using the
following:

   delimiter - one of 'to-eol' (Daylight/OEChem style), 'tab',
        'whitespace', 'space', or 'native' (for the native toolkit style)

   has_header - '1' if the first line contains a header, else '0'.

The supported format, default reader arguments, and input flavors are:
""")

    from openeye.oechem import OEIFlavor_Generic_Rings
    def print_wrapped(prefix, args):
        line = False
        indent = len(prefix) - len(prefix.lstrip()) + 4
        for arg in args:
            if line is False:
                line = prefix + ": " + arg
            elif len(line) + len(arg) > 75:
                print(line + ",")
                line = " " * indent + arg
            else:
                line += ", " + arg
        if line is not False:
            print(line)
            
    for fmt in sorted(
            (fmt for fmt in openeye_toolkit.get_input_formats() if fmt.supports_io and fmt.name != "json"),
            key=lambda fmt: fmt.name):
        # There's no good way to get the actual default values from the chemfp API
        reader_args = openeye_toolkit.read_molecules_from_string(b"", fmt).metadata.args
        print("Format:", fmt.name)
        for k, v in sorted(reader_args.items()):
            if v is None:
                v_str = "<N/A>"
            elif k == "flavor":
                v_str = "Default"
                if v & OEIFlavor_Generic_Rings == OEIFlavor_Generic_Rings:
                    v = v ^ OEIFlavor_Generic_Rings
                #v_str += " (bitflags: 0x%x = %d)" % (v, v)
                
            elif k == "has_header":
                if v:
                    v_str = "1"
                else:
                    v_str = "0"
            elif k == "delimiter":
                if v == " ":
                    v_str = "space"
                else:
                    v_str = v
            else:
                v_str = str(v)
                
            print("    %s: %s" % (k, v_str))
            
            if k == "flavor":
                input_flavor = fmt._format_config.input.flavor
                if not input_flavor.canonical_options:
                    print("      no flavor flags available")
                else:
                    flavor_strings = []
                    bits = 0
                    for opt in input_flavor.canonical_options:
                        opt_flavor = input_flavor.flavors[opt]
                        if v & opt_flavor == opt_flavor:
                            flavor_strings.append(opt)
                            bits |= opt_flavor
                    assert bits == v, (v, flavor_strings, bits)
                    if not flavor_strings:
                        print("        default flags: <none>")
                    else:
                        print_wrapped("        default flags", flavor_strings)
                
                    print_wrapped("        available flags", input_flavor.canonical_options)
                    
        print()
    
    print("""
See https://docs.eyesopen.com/toolkits/cpp/oechemtk/molreadwrite.html#flavored-input-and-output
for documentation about the flavors for each format.
""")
    

parser = argparse.ArgumentParser(
    description="Generate FPS or FPB fingerprints from a structure file using OEChem",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,    
    )

path_group = parser.add_argument_group("path, circular, and tree fingerprints")
path_group.add_argument(
    "--path", action="store_true", help="generate path fingerprints (default)")
path_group.add_argument(
    "--circular", action="store_true", help="generate circular fingerprints")
path_group.add_argument(
    "--tree", action="store_true", help="generate tree fingerprints")

path_group.add_argument(
    "--numbits", action="store", type=int, metavar="INT", default=4096,
    help="number of bits in the fingerprint (default=4096)")
path_group.add_argument(
    "--minbonds", action="store", type=int, metavar="INT", default=0,
    help="minimum number of bonds in the path or tree fingerprint (default=0)")
path_group.add_argument(
    "--maxbonds", action="store", type=int, metavar="INT", default=None,
    help="maximum number of bonds in the path or tree fingerprint (path default=5, tree default=4)")
cmdsupport.add_argument_to_argparse(oetypes.circular_arguments["minradius"], path_group)
cmdsupport.add_argument_to_argparse(oetypes.circular_arguments["maxradius"], path_group)


# The expansion of 'Default' differs based on the fingerprint type
path_group.add_argument(
    "--atype", metavar="ATYPE", default="Default",
    help="atom type flags, described below (default=Default)")
path_group.add_argument(
    "--btype", metavar="BTYPE", default="Default",
    help="bond type flags, described below (default=Default)")

maccs_group = parser.add_argument_group("166 bit MACCS substructure keys")
maccs_group.add_argument(
    "--maccs166", action="store_true", help="generate MACCS fingerprints")

substruct_group = parser.add_argument_group("881 bit ChemFP substructure keys")
substruct_group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints")

rdmaccs_group = parser.add_argument_group("ChemFP version of the 166 bit RDKit/MACCS keys")
rdmaccs_group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="generate 166 bit RDKit/MACCS fingerprints (version 2)")
rdmaccs_group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")

parser.add_argument(
    "--aromaticity", metavar="NAME", choices=["openeye", "daylight", "tripos", "mdl", "mmff"],
    default=None,
    help="use the named aromaticity model (same as '-R aromaticity=NAME')")

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id (SD files only)")

cmdsupport.add_type_and_using(parser)

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default guesses from filename)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")
parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_progress_bar(parser)

parser.add_argument(
    "--help-formats", action=cmdsupport.CallFunctionAction,
    default=help_formats,
    help="list the available formats and reader arguments")
parser.add_argument(
    "-R", metavar="NAME=VALUE", dest="reader_settings", action="append", default=None,
    help="specify a reader argument")
cmdsupport.add_reader_aliases(parser)

cmdsupport.add_version(parser)
cmdsupport.add_license_check(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")

def _get_atype_and_btype(args, atom_description_to_value, bond_description_to_value):
    try:
        atype = atom_description_to_value(args.atype)
    except ValueError as err:
        parser.error("--atype must contain '|' separated atom terms: %s" % (err,))
    try:
        btype = bond_description_to_value(args.btype)
    except ValueError as err:
        parser.error("--btype must contain '|' separated atom terms: %s" % (err,))
    return atype, btype

#######

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    supported_fingerprints = ("maccs166", "path", "substruct", "rdmaccs", "rdmaccs/1",
                              "circular", "tree", "type", "using")
    cmdsupport.mutual_exclusion(parser, args, "path", supported_fingerprints)

    # This allows None as an aromaticity
    if not openeye_toolkit.is_valid_aromaticity(args.aromaticity):
        parser.error("Unsupported aromaticity specifier: %r" % (args.aromaticity,))  #pragma: no cover - argparser handles this

    if args.maccs166:
        # Create the MACCS keys fingerprinter
        fptype = get_fingerprint_type("OpenEye-MACCS166")
    elif args.path:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits") #pragma: no cover - argparser handles this

        if not (0 <= args.minbonds):
            parser.error("--minbonds must be 0 or greater")
        if args.maxbonds is None:
            args.maxbonds = 5
        if not (args.minbonds <= args.maxbonds):
            parser.error("--maxbonds must not be smaller than --minbonds")
        atype, btype = _get_atype_and_btype(args, oetypes.path_atom_description_to_value,
                                            oetypes.path_bond_description_to_value)
        fptype = get_fingerprint_type(
            "OpenEye-Path",
            fingerprint_kwargs = dict(
                numbits = args.numbits,
                minbonds = args.minbonds,
                maxbonds = args.maxbonds,
                atype = atype,
                btype = btype),
            )
    elif args.circular:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits")

        if not (0 <= args.minradius):
            parser.error("--minradius must be 0 or greater")
        if not (args.minradius <= args.maxradius):
            parser.error("--maxradius must not be smaller than --minradius")
        atype, btype = _get_atype_and_btype(args, oetypes.circular_atom_description_to_value,
                                            oetypes.circular_bond_description_to_value)

        fptype = get_fingerprint_type(
            "OpenEye-Circular",
            fingerprint_kwargs = dict(
                numbits = args.numbits,
                minradius = args.minradius,
                maxradius = args.maxradius,
                atype = atype,
                btype = btype),
            )
        
    elif args.tree:
        if not (16 <= args.numbits <= 65536):
            parser.error("--numbits must be between 16 and 65536 bits")

        if not (0 <= args.minbonds):
            parser.error("--minbonds must be 0 or greater")
        if args.maxbonds is None:
            args.maxbonds = 4
        if not (args.minbonds <= args.maxbonds):
            parser.error("--maxbonds must not be smaller than --minbonds")
        atype, btype = _get_atype_and_btype(args, oetypes.tree_atom_description_to_value,
                                            oetypes.tree_bond_description_to_value)

        fptype = get_fingerprint_type(
            "OpenEye-Tree",
            fingerprint_kwargs = dict(
                numbits = args.numbits,
                minbonds = args.minbonds,
                maxbonds = args.maxbonds,
                atype = atype,
                btype = btype),
            )
        
    elif args.substruct:
        fptype = get_fingerprint_type(
            "ChemFP-Substruct-OpenEye",
            )

    elif args.rdmaccs:
        fptype = get_fingerprint_type(
            "RDMACCS-OpenEye",
            )
    elif getattr(args, "rdmaccs/1"):
        fptype = get_fingerprint_type(
            "RDMACCS-OpenEye/1",
            )
    elif args.type is not None:
        fptype = cmdsupport.get_fptype_from_type_arg(parser, args.type, "oe2fps", "openeye")
    elif args.using is not None:
        fptype = cmdsupport.get_fptype_from_using_arg(parser, args.using, "oe2fps", "openeye")
    else:
        parser.error("ERROR: fingerprint not specified?") #pragma: no cover - should not get here

    assert fptype is not None

    # Check that this is a supported format.
    try:
        openeye_toolkit.get_input_format(args.format)
    except ValueError:
        parser.error("Unsupported format specifier: %r" % (args.format,))
        
    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    # Override anything set by -R
    # (Note: you can override with something more specific, like "oechem.aromaticity=mdl")
    if args.reader_settings is None:
        args.reader_settings = []
    if args.aromaticity is not None:
        args.reader_settings.append( "aromaticity=" + args.aromaticity )
    reader_settings = cmdsupport.get_reader_settings(parser, args)
    
    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
    
    cmdsupport.check_filenames(parser, openeye_toolkit, args.format, args.filenames)

    # Ready the input reader/iterator
    metadata, num_readers, first_reader, readers = cmdsupport.read_multifile_structure_fingerprints(
        fptype, args.filenames, format=args.format, id_tag=args.id_tag,
        reader_settings=reader_settings, errors=args.errors)

    with first_reader:
        try:
            writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
        except NotImplementedError as err:
            cmdsupport.writer_failure("oe2fps", str(err))

        except (IOError, ValueError) as err:
            msg = "Cannot open output fingerprint file: %s" % (err,)
            if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
                msg += ".\nUse --out to specify 'fps', 'fps.gz', 'fps.zst', 'fpb', or 'flush'."
            parser.error(msg)

        try:
            cmdsupport.read_write_fingerprints(args, num_readers, readers, writer)
        except BrokenPipeError:
            cmdsupport.handle_broken_pipe()
        except (ChemFPError, IOError, EOFError) as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
    
if __name__ == "__main__":
    main()  #pragma: no cover - cannot (easily) be tested by unitttest
