# Copyright (c) 2010-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import sys
import re
import itertools
import argparse

from .. import Metadata, ChemFPError, FingerprintIterator
from .. import encodings
from .. import io
from .. import open_fingerprint_writer
from .. import text_toolkit

from . import cmdsupport
from .. import _compat
from .. import registry

def _check_num_bits(num_bits,  # from the user
                    fp_num_bits, # not None if the fp decoder know it exactly
                    num_bytes, # length of decoded fp in bytes
                    parser):
    """Check that the number of fingerprint bits and bytes match the user input

    Difficulties: some fingerprints have only a byte length, and the user
    doesn't have to specify the input.

    Returns the number of bits, or calls parser.error if there are problems
    """
    if fp_num_bits is not None:
        # The fingerprint knows exactly how many bits it contains
        if num_bits is None:
            # The user hasn't specified, so go with the exact number
            return fp_num_bits

        # If the user gave a value, make sure it matches
        if num_bits != fp_num_bits:
            parser.error(
                ("the first fingerprint has %(fp_num_bits)s bits which "
                 "is not the same as the --num-bits value of %(num_bits)s") % dict(
                    num_bits=num_bits, fp_num_bits=fp_num_bits))
            raise AssertionError("should not get here")
        
        return num_bits

    # If the number of bits isn't specified, assume it's exactly
    # enough to fill up the fingerprint bytes.
    if num_bits is None:
        return num_bytes * 8

    # The user specified the number of bits. The first fingerprint
    # has a number of bytes. This must be enough to hold the bits,
    # but only up to 7 bits larger.
    if (num_bits+7)//8 != num_bytes:
        parser.error(
            ("The byte length of the first fingerprint is %(num_bytes)s so --num-bits "
             "must be %(min)s <= num-bits <= %(max)s, not %(num_bits)s") % dict(
                num_bytes=num_bytes, min=num_bytes*8-7, max=num_bytes*8,
                num_bits=num_bits))
        raise AssertionError("should not get here")

    # Accept what the user suggested
    return num_bits

parser = argparse.ArgumentParser(
    description="Extract a fingerprint tag from an SD file and generate FPS or FPB fingerprints",
    #epilog=epilog,
    #formatter_class=argparse.RawDescriptionHelpFormatter,
    )

parser.add_argument("--id-tag", metavar="TAG", default=None,
            help="get the record id from TAG instead of the first line of the record")
parser.add_argument("--fp-tag", metavar="TAG", 
                    help="get the fingerprint from tag TAG (required)")

parser.add_argument("--in", metavar="FORMAT", dest="format",
                        choices=("sdf", "sdf.gz", "sdf.zst"), default=None,
                        help='Specify the input format (one of "sdf", "sdf.gz", or "sdf.zst")')
                    

parser.add_argument("--num-bits", metavar="INT", type=int,
                    help="use the first INT bits of the input. Use only when the "
                    "last 1-7 bits of the last byte are not part of the fingerprint. "
                    "Unexpected errors will occur if these bits are not all zero.")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="strict",
    help="how should structure parse errors be handled? (default=strict)")

parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")
parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
    help="output format, one of 'fps', 'fps.gz', 'fps.zst', 'fpb', or 'flush' "
         "(default guesses from output filename, or is 'fps')")
parser.add_argument("--software", metavar="TEXT",
                    help="use TEXT as the software description")
parser.add_argument("--type", metavar="TEXT",
                    help="use TEXT as the fingerprint type description")

cmdsupport.add_progress_bar(parser)

cmdsupport.add_version(parser)
cmdsupport.add_license_check(parser)

parser.add_argument(
    "filenames", nargs="*", help="input SD files (default is stdin)", default=None)


# TODO:
# Currently I guess the format type based solely on the extension.
# Do I want to support encoding of the fps output?

#parser.add_argument(
#    "--compress", action="store", metavar="METHOD", default="auto",
#    help="use METHOD to compress the output (default='auto', 'none', 'gzip', 'bzip2')")


# This adds --cactvs, --base64 and other decoders to the command-line arguments
encodings._add_decoding_group(parser)

# Support the "--pubchem" option
shortcuts_group = parser.add_argument_group("shortcuts")

class AddSubsKeys(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.cactvs=True
        # the 1.3 is solely based on the version of the document at
        #  ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
        namespace.software="CACTVS/unknown"
        namespace.type="CACTVS-E_SCREEN/1.0 extended=2"
        namespace.fp_tag="PUBCHEM_CACTVS_SUBSKEYS"

shortcuts_group.add_argument("--pubchem", nargs=0, action=AddSubsKeys,
   help = ("decode CACTVS substructure keys used in PubChem. Same as "
           "--software=CACTVS/unknown --type 'CACTVS-E_SCREEN/1.0 extended=2' "
           "--fp-tag=PUBCHEM_CACTVS_SUBSKEYS --cactvs"))

###############

def run():
    cmdsupport.run(main)

_illegal_value_pat = re.compile(r"[\000-\037]")

def main(args=None):    
    args = parser.parse_args(args)
    return _main(parser, args)
    
def _main(parser, args):
    if not args.fp_tag:
        parser.error("argument --fp-tag is required")
    if args.num_bits is not None and args.num_bits <= 0:
        parser.error("--num-bits must be a positive integer")

    fp_decoder_name, fp_decoder = encodings._extract_decoder(parser, args)

    cmdsupport.check_filenames(parser, text_toolkit, "sdf", args.filenames)

    for attr in ("software", "type"):
        description = getattr(args, attr, None)
        if description is None:
            continue
        m = _illegal_value_pat.search(description)
        if m is None:
            continue
        parser.error("--%(attr)s description may not contain the character %(c)r" % dict(
                attr=attr, c = m.group(0)))

    error_handler = io.get_parse_error_handler(args.errors)

    if args.format == "sdf":
        compression = ""
    elif args.format == "sdf.gz":
        compression = "gz"
    elif args.format == "sdf.zst":
        compression = "zst"
    else:
        compression = "auto"    

    # Set up the error messages for missing id or fingerprints.
    if args.id_tag is None:
        MISSING_FP = "Missing fingerprint tag %(tag)r"
    else:
        MISSING_FP = "Missing fingerprint tag %(tag)r in for id %(id)r"

    #### Handle 0 or more filenames
    def read_multifile_fingerprints(filenames):
        if not filenames:
            reader = get_fingerprint_reader(None)
            return reader.metadata, 1, reader, [reader]

        if len(filenames) == 1:
            reader = get_fingerprint_reader(filenames[0])
            return reader.metadata, 1, reader, [reader]

        # Get the first reader
        first_reader = get_fingerprint_reader(filenames[0])
        first_reader.metadata.sources = filenames

        def multifile_iter(first_reader, filenames):
            yield first_reader
            first_reader.close()
            for filename in filenames:
                reader = get_fingerprint_reader(filename)
                yield reader
                reader.close()

        return first_reader.metadata, len(filenames), first_reader, multifile_iter(first_reader, filenames[1:])
        
    ### Read from a single source
    def get_fingerprint_reader(source):
        # Read the (cleaned-up) ids and values
        try:
            reader = text_toolkit.read_sdf_ids_and_values(
                source, args.id_tag, args.fp_tag, compression=compression, errors=args.errors)
        except OSError as err:
            # Can fail in gzio if the first file is supposed to be in gzip format, but is not.
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
        except ValueError as err:
            if "Please install the zstandard module" in str(err):
                sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
                raise SystemExit(1)
            raise

        encoded_fps = cmdsupport.cleanup_ids_iter(reader, args.id_tag, error_handler, reader.location)

        try:
            metadata, decoded_fps = decode_fingerprints(encoded_fps, reader.location, error_handler)
        except (ChemFPError, IOError) as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)

        return FingerprintIterator(metadata, decoded_fps, reader.location)

    def decode_fingerprints(encoded_fps_reader, location, error_handler):
        decoded_iter = decode_fingerprint_iter(encoded_fps_reader, location, error_handler)
        try:
            id, fp, num_bits = next(decoded_iter)
        except StopIteration:
            # No fingerprints? Make a new empty stream
            metadata = Metadata(date = io.utcnow())
            chained_reader = iter([])
        else:
            # Got the first fingerprint
            expected_num_bytes = len(fp)

            # Verify that they match
            expected_num_bits = _check_num_bits(args.num_bits, num_bits, expected_num_bytes, parser)


            chained_reader = itertools.chain( [(id, fp)], (x[:2] for x in decoded_iter) )
            metadata = Metadata(num_bits = expected_num_bits,
                                software = args.software,
                                type = args.type,
                                sources = args.filenames,
                                date = io.utcnow())
            
        return metadata, chained_reader
            
    # Decoded encoded fingerprints, yielding (id, fp, num_bits)
    def decode_fingerprint_iter(encoded_fp_reader, location, error_handler):
        expected_num_bits = -1
        expected_fp_size = None
        bad_count = 0

        for id, encoded_fp in encoded_fp_reader:
            assert id, "this check was supposed to be done in cmdsupport.cleanup_ids()"
            ## if not id:
            ##     msg = MISSING_ID % dict(id=id, where=location.where(),
            ##                             tag=args.id_tag)
            ##     error_handler.error(msg)
            ##     continue

            if not encoded_fp:
                msg = MISSING_FP % dict(id=id, where=location.where(),
                                        tag=args.fp_tag)
                error_handler.error(msg, location)
                if expected_fp_size is None:
                    bad_count += 1
                    if bad_count == 100:
                        cmdsupport.AbortHandler().error(msg, location)
                        return
                continue

            # Decode the fingerprint, and complain if it isn't decodeable.
            try:
                num_bits, fp = fp_decoder(encoded_fp)
            except (ValueError, TypeError) as err:
                msg = ("Could not %(decoder_name)s decode tag %(tag)r value %(encoded_fp)s: %(err)s" %
                       dict(decoder_name=fp_decoder_name, tag=args.fp_tag,
                            err=err, encoded_fp=_compat.myrepr(encoded_fp)))
                error_handler.error(msg, location)
                if expected_fp_size is None:
                    bad_count += 1
                    if bad_count == 100:
                        cmdsupport.AbortHandler().error(msg, location)
                        return
                continue

            if num_bits != expected_num_bits:
                if expected_num_bits == -1:
                    expected_num_bits = num_bits
                else:
                    msg = ("Tag %(tag)r value %(encoded_fp)s has %(got)d bits but expected %(expected)d" %
                           dict(tag=args.fp_tag, encoded_fp=_compat.myrepr(encoded_fp),
                                got=num_bits, expected=expected_num_bits))
                    error_handler.error(msg, location)
                    continue

            if len(fp) != expected_fp_size:
                if expected_fp_size is None:
                    expected_fp_size = len(fp)
                else:
                    msg = ("Tag %(tag)r value %(encoded_fp)r has %(got)d bytes but expected %(expected)d" %
                           dict(tag=args.fp_tag, encoded_fp=encoded_fp,
                                got=len(fp), expected=expected_fp_size))
                    error_handler.error(msg, location)
                    continue

            yield id, fp, num_bits
            

    metadata, num_readers, first_reader, reader_iter = read_multifile_fingerprints(args.filenames)

    with first_reader:
        output_location = io.Location.from_destination(args.output)
        try:
            writer = open_fingerprint_writer(args.output, metadata, format=args.output_format,
                                             location=output_location)
        except NotImplementedError as err:
            cmdsupport.writer_failure("sdf2fps", str(err))
        except IOError as err:
            parser.error("Cannot open output file: %s" % (err,))
        except ValueError as err:
            parser.error("Cannot open output file: %s: %r"
                          % (err, output_location.filename))

        try:
            cmdsupport.read_write_fingerprints(args, num_readers, reader_iter, writer)
        except ChemFPError as err:
            sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
            raise SystemExit(1)
        except ValueError as err:
            if "Please install the zstandard module" in str(err):
                sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
                raise SystemExit(1)
            raise


command_config = registry.CommandConfig.from_argparse(parser, "fp_generation", _main)
    
            
if __name__ == "__main__":
    main()
