# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse
import itertools
import re
import sys
import time

import chemfp

from .. import registry
from .. import io, SOFTWARE, bitops
from .. import search

from . import cmdsupport

def get_precision(num_bytes, precision):
    if precision is None:
        precision = bitops.get_tanimoto_precision(num_bytes*8)
    return precision

def check_id_field(field, field_name):
    if field is None:
        return
    for c, name in ( ("\t", "tab"),
                     ("\n", "newline"),
                     ("\r", "control-return"),
                     ("\0", "NUL")):
        if c in field:
            parser.error(f"--{field_name} must not contain the {name} character: {field:!r}")
    
def make_output_type(pattern, k=None, threshold=None, alpha=None, beta=None):
    if alpha is None:
        name = "Tanimoto"
    else:
        name = "Tversky"
        pattern += " alpha=%(alpha)s beta=%(beta)s"
    s = pattern % {"name": name, "k": k, "threshold": threshold, "alpha": alpha, "beta": beta}
    assert "None" not in s # safety check
    return s

class BaseWriter:
    precision = None
    
    def set_precision(self, precision):
        self.score_formatter = score_formatter = f"%.{precision}f"
        self.hit_formatter = f"\t%s\t{score_formatter}"
        self.id_id_score_formatter = f"%s\t%s\t{score_formatter}\n"

    def write_magic(self, outfile):
        pass

    def detach(self):
        pass

class ChemfpHeaderWriter:
    def write_header(self, outfile, d):
        lines = []
        for name in ("num_bits", "type", "software", "queries", "targets"):
            value = d.get(name, None)
            if value is not None:
                lines.append(("#%s=%s\n" % (name, value)).encode("utf8"))
        for name in ("query_sources", "target_sources"):
            for value in d.get(name, []):
                lines.append(("#%s=%s\n" % (name[:-1], value)).encode("utf8"))
        outfile.writelines(lines)
    
class ChemfpCountWriter(BaseWriter, ChemfpHeaderWriter):
    def write_magic(self, outfile):
        outfile.write(b"#Count/1\n")
        
    def write_counts(self, outfile, counts):
        for query_id, hit_count in counts:
            outfile.write(f"{hit_count}\t{query_id}\n".encode("utf8"))

    def write_NxN_counts(self, outfile, original_ids, original_index_to_current_index, counts):
        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            count = counts[current_index]
            outfile.write(b"%d\t%s\n" % (count, original_id.encode("utf8")))
                             
        
    
class ChemfpSimsearchWriter(BaseWriter, ChemfpHeaderWriter):
    def write_magic(self, outfile):
        outfile.write(b"#Simsearch/1\n")
        
    def write_result(self, outfile, result):
        hit_formatter = self.hit_formatter
        for query_id, row in result:
            outfile.write(b"%d\t%s" % (len(row), query_id.encode("utf8")))
            for hit in row.get_ids_and_scores():
                outfile.write((hit_formatter % hit).encode("utf8"))
            outfile.write(b"\n") # XXX flush?
        
    def write_NxN_result(self, outfile, original_ids, original_index_to_current_index, precision, results):
        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            result = results[current_index]
            outfile.write(b"%d\t%s" % (len(result), original_id.encode("utf8")))
            if len(result) > 0:
                target_ids = [original_ids[new_index] for new_index in result.get_ids()]
                #print("len", len(result), target_ids, result.get_ids_and_scores(),file=open("/dev/stderr", "w"))
                s = result.format_ids_and_scores_as_bytes(target_ids, precision)
                outfile.write(b"\t")
                outfile.write(s)

            outfile.write(b"\n") # XXX flush?
    
    
class BaseExcelWriter(BaseWriter):
    def __init__(self, dialect):
        self.dialect = dialect
        self.outfile = None

    def write_magic(self, outfile):
        import io
        import csv
        self.text_outfile = text_outfile = io.TextIOWrapper(outfile, write_through=True)
        self.csv_writer = csv.writer(text_outfile, dialect=self.dialect)

    def detach(self):
        # prevent TextIOWrapper's __del__ from forwarding close() to outfile.close()
        self.text_outfile.detach()
                
class ExcelCountWriter(BaseExcelWriter):
    def write_header(self, outfile, d):
        self.csv_writer.writerow(["query_id", "count"])
        
    def write_counts(self, outfile, counts):
        writerow = self.csv_writer.writerow
        for query_id, hit_count in counts:
            writerow([query_id, hit_count])

    def write_NxN_counts(self, outfile, original_ids, original_index_to_current_index, counts):
        writerow = self.csv_writer.writerow
        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            count = counts[current_index]
            writerow([original_id, count])

class ExcelSimsearchWriter(BaseExcelWriter):
    def __init__(self, dialect, include_empty, empty_target_id, empty_score):
        super().__init__(dialect)
        self.include_empty = include_empty
        self.empty_target_id = empty_target_id
        self.empty_score = empty_score
        
    def write_header(self, outfile, d):
        self.csv_writer.writerow(["query_id", "target_id", "score"])

    def write_result(self, outfile, result):
        writerow = self.csv_writer.writerow
        score_formatter = self.score_formatter
        for query_id, row in result:
            if row:
                for target_id, score in row.get_ids_and_scores():
                    writerow([query_id, target_id, score_formatter % score])
            elif self.include_empty:
                writerow([query_id, self.empty_target_id, self.empty_score])
                
    def write_NxN_result(self, outfile, original_ids, original_index_to_current_index, precision, results):
        score_formatter = self.score_formatter
        writerow = self.csv_writer.writerow
        writerows = self.csv_writer.writerows
        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            result = results[current_index]
            if result:
                target_ids = [original_ids[new_index] for new_index in result.get_ids()]
                writerows(
                    [original_id, target_id, score_formatter % score]
                    for target_id, score in zip(target_ids, result.get_scores())
                    )
            elif self.include_empty:
                writerow([original_id, self.empty_target_id, self.empty_score])
                
def check_writer_args(parser, args):
    destination = args.output
    format, compression = io.normalize_output_format(
        destination, args.output_format, ("chemfp", ""),
        )

    if args.output_format is not None:
        if format not in (None, "chemfp", "csv", "tsv"):
            parser.error(f"Unsupported --out format {args.output_format!r}")

    return format, compression

def get_count_writer(output_format, args):
    if output_format in (None, "chemfp"):
        return ChemfpCountWriter()
    elif output_format == "csv":
        return ExcelCountWriter(dialect = "excel")
    elif output_format == "tsv":
        return ExcelCountWriter(dialect = "excel-tab")
    raise AssertionError(output_format)

def get_simsearch_writer(output_format, args):
    if output_format in ("csv", "tsv"):
        dialect = {
            "csv": "excel",
            "tsv": "excel-tab",
            }[output_format]
        return ExcelSimsearchWriter(
            dialect = dialect,
            include_empty = args.include_empty,
            empty_target_id = args.empty_target_id,
            empty_score = args.empty_score,
            )
    return ChemfpSimsearchWriter()

    
#### The NxM cases

# Helper class to keep track of the times
class ElapsedTime(object):
    __slots__ = ("read_time", "search_time", "output_time")
    def __init__(self):
        self.read_time = 0.0
        self.search_time = 0.0
        self.output_time = 0.0
    
def report_threshold(output_writer, outfile, query_arenas, targets, threshold, alpha, beta, precision, elapsed_time):
    precision = get_precision(targets.metadata.num_bytes, precision)
    def search_function(query_arena):
        return chemfp.threshold_tversky_search(query_arena, targets, threshold=threshold,
                                               alpha=alpha, beta=beta)
    _report_search(output_writer, outfile, precision, query_arenas, search_function, elapsed_time)

def report_knearest(output_writer, outfile, query_arenas, targets, k, threshold, alpha, beta, precision, elapsed_time):
    precision = get_precision(targets.metadata.num_bytes, precision)
    def search_function(query_arena):
        return chemfp.knearest_tversky_search(query_arena, targets, k=k, threshold=threshold,
                                              alpha=alpha, beta=beta)
    _report_search(output_writer, outfile, precision, query_arenas, search_function, elapsed_time)

def _report_search(output_writer, outfile, precision, query_arenas, search_function, elapsed_time):
    output_writer.set_precision(precision)
    search_time = output_time = 0.0
    start_time = time.time()
    write_result = output_writer.write_result
    for query_arena in query_arenas:
        t1 = time.time()
        result = search_function(query_arena)
        t2 = time.time()
        search_time += t2-t1
        write_result(outfile, result)
        t3 = time.time()
        output_time += t3-t2
    end_time = time.time()

    elapsed_time.search_time = search_time
    elapsed_time.output_time = output_time
    elapsed_time.read_time = (end_time - start_time) - (output_time + search_time)


def report_counts(output_writer, outfile, query_arenas, targets, threshold, alpha, beta, elapsed_time):
    search_time = output_time = 0.0
    start_time = time.time()
    write_counts = output_writer.write_counts
    for query_arena in query_arenas:
        t1 = time.time()
        counts = chemfp.count_tversky_hits(query_arena, targets, threshold, alpha, beta)
        t2 = time.time()
        search_time += t2-t1
        write_counts(outfile, counts)
        t3 = time.time()
        output_time += t3-t2
    end_time = time.time()
        
    elapsed_time.search_time = search_time
    elapsed_time.output_time = output_time
    elapsed_time.read_time = (end_time - start_time) - (output_time + search_time)


#### The NxN cases

class UnchangedDict(object):
    def __getitem__(self, x):
        return x

unchanged_dict = UnchangedDict()

class OriginalIds(object):
    def __init__(self, ids):
        self.ids = ids
    def __iter__(self):
        return iter(self.ids)
    def __getitem__(self, x):
        return x

def do_NxN_searches(args, k, threshold, target_filename, alpha, beta,
                        precision, exit_stack, output_format, output_compression):
    # Ensure that the file is closed
    start_time = time.time()
    try:
        fps = chemfp.open(target_filename, format=args.target_format, allow_mmap=(not args.no_mmap))
    except (IOError, ValueError, chemfp.ChemFPError) as err:
        sys.stderr.write("Cannot open targets file: %s\n" % (err,))
        raise SystemExit(1)
    
    with fps:
        t2 = time.time()
        _do_NxN_searches(start_time, t2-start_time, fps, args, k, threshold, target_filename, alpha, beta,
                             precision, exit_stack, output_format, output_compression)
    
def _do_NxN_searches(start_time, open_time, fps, args, k, threshold, target_filename, alpha, beta,
                         precision, exit_stack, output_format, output_compression):
    t1 = time.time()
    # load_fingerprints sorts the fingerprints based on popcount
    # I want the output to be in the same order as the input.
    # This means I need to do some reordering. Consider:
    #   0003  ID_A
    #   010a  ID_B
    #   1000  ID_C
    # I use this to generate:
    #   original_ids = ["ID_A", "ID_B", "ID_C"]
    #   targets.ids = [2, 0, 1]
    #   original_index_to_current_index = {2:0, 0:1, 1:2}
    #   current_index_to_original_index = {0:2, 1:0, 2:1}
    
    # If the file is empty then the num_bytes is None.  This will
    # cause downstream code to fail. The easiest solution (perhaps a
    # hack) is to force it to have 1 byte.
    if fps.metadata.num_bytes is None:
        fps.metadata = fps.metadata.copy(num_bytes=1)
        
    if getattr(fps, "popcount_indices", b""):
        # Already ordered; don't need to reorder
        targets = fps.copy()
        exit_stack.enter_context(targets)
        original_ids = OriginalIds(fps.ids)
        original_index_to_current_index = unchanged_dict
        ## current_index_to_original_index = unchanged_dict
    else:
        # preserve alignment, if present
        alignment = getattr(fps, "alignment", None)
        # Either the input is an unordered FPB or an FPS
        original_ids = []
        def get_index_to_id(fps):
            for i, (id, fp) in enumerate(fps):
                original_ids.append(id)
                yield i, fp

        targets = chemfp.load_fingerprints(get_index_to_id(fps), fps.metadata, alignment=alignment)
        exit_stack.enter_context(targets)
        original_index_to_current_index = dict((id, i) for (i, id) in enumerate(targets.ids))
        ## current_index_to_original_id = dict((i, original_ids[original_index])
        ##                                         for i, original_index in enumerate(targets.ids))
    t2 = time.time()
    read_time = t2-t1

    try:
        outfile, close = io.open_compressed_output(args.output, output_compression, "simsearch")
    except (IOError, ValueError) as err:
        sys.stderr.write("Cannot open output file: %s\n" % (err,))
        raise SystemExit(1)
    exit_stack.call_if(close)

    if args.count:
        type = make_output_type("Count threshold=%(threshold)s NxN=full",
                                threshold=threshold, alpha=alpha, beta=beta)
        output_writer = get_count_writer(output_format, args)
        output_writer.write_magic(outfile)
    else:
        type = make_output_type("%(name)s k=%(k)s threshold=%(threshold)s NxN=full",
                                k=k, threshold=threshold, alpha=alpha, beta=beta)
        output_writer = get_simsearch_writer(output_format, args)
        output_writer.write_magic(outfile)

    exit_stack.callback(output_writer.detach)

    if alpha is None:
        assert beta is None
        alpha = beta = 1.0

    output_writer.write_header(outfile, {
        "num_bits": targets.metadata.num_bits,
        "software": SOFTWARE,
        "type": type,
        "targets": target_filename,
        "target_sources": targets.metadata.sources})

    t3 = time.time()
    output_time = t3-t2

    if args.count:
        # The implementation will use Tanimoto if alpha=beta=1.0
        start_search_time = time.time()
        counts = search.count_tversky_hits_symmetric(targets, threshold, alpha, beta,
                                                     batch_size=args.batch_size)
        end_search_time = time.time()

        output_writer.write_NxN_counts(outfile, original_ids, original_index_to_current_index, counts)
        end_write_time = time.time()
    else:
        precision = get_precision(targets.metadata.num_bytes, precision)
        output_writer.set_precision(precision)

        start_search_time = time.time()
        if k == "all":
            results = search.threshold_tversky_search_symmetric(targets, threshold, alpha, beta,
                                                                batch_size=args.batch_size)

        else:
            results = search.knearest_tversky_search_symmetric(targets, k, threshold, alpha, beta,
                                                              batch_size=args.batch_size)
        end_search_time = time.time()

        output_writer.write_NxN_result(outfile, original_ids, original_index_to_current_index, precision, results)

        end_write_time = time.time()

    output_time += end_write_time - end_search_time
    search_time = end_search_time - start_search_time

    total_time = time.time() - start_time
    if args.times:
        sys.stderr.write("open %.2f read %.2f search %.2f output %.2f total %.2f\n" % (
            open_time, read_time, search_time, output_time, total_time))

####

def int_or_all(s):
    if s == "all":
        return s
    return int(s)

_alpha_pat = re.compile(r"[0-9]+(\.[0-9]*)?$")
def alpha_beta(error, name, s):
    m = _alpha_pat.match(s)
    if m is None:
        error("%s value must be a decimal value between 0.0 and 100.0, not %r" % (name, s))
    value = float(s)
    if value > 100.0:
        error("%s value must be a decimal value between 0.0 and 100.0, not %r" % (name, s))
    decimal = m.group(1)
    if decimal and len(decimal) >= 5:
        error("%s precision is too high; no more than four decimal places are supported, not %r"
               % (name, s))
    return value

# the "2fps" options need a way to say "get the options from --reference"
# ob2fps --reference targets.fps | simsearch -k  5 --threshold 0.5 targets.fps

parser = argparse.ArgumentParser(
    description="Search an FPS or FPB file for similar fingerprints")
parser.add_argument("-k" ,"--k-nearest", help="select the k nearest neighbors (use 'all' for all neighbors)",
                    default=None, type=int_or_all)
parser.add_argument("-t" ,"--threshold", help="minimum similarity score threshold",
                    default=None, type=float)
parser.add_argument("--alpha", default=None, 
                    help="Tversky alpha parameter (default: 1.0)")
parser.add_argument("--beta", default=None, 
                    help="Tversky beta parameter (default: the value of --alpha)")
parser.add_argument("--queries", "-q", help="filename containing the query fingerprints")
parser.add_argument("--NxN", action="store_true",
                    help="use the targets as the queries, and exclude the self-similarity term")
parser.add_argument("--query", help="query as a structure record (default format: 'smi')")
parser.add_argument("--hex-query", help="query in hex")
parser.add_argument("--query-id", default=None,
                    help="id for the query or hex-query (default: 'Query1'")
parser.add_argument("--query-format", "--in", metavar="FORMAT", dest="query_format",
                    help="input query format (default uses the file extension, else 'fps')")
parser.add_argument("--target-format", metavar="FORMAT", dest="target_format",
                    help="input target format (default uses the file extension, else 'fps')")
parser.add_argument("--query-type", metavar="STRING",
                    help="fingerprint type string if the queries are structures (default: use the target fingerprint type)")

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag containing the record id if --query-format is an SD file)")
parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")
parser.add_argument(
    "-R", metavar="NAME=VALUE", dest="reader_settings", action="append", default=None,
    help="specify a reader argument")
cmdsupport.add_reader_aliases(parser)

parser.add_argument("-o", "--output", metavar="FILENAME",
                    help="output filename (default is stdout)")

parser.add_argument(
    "--out",
    dest = "output_format",
    metavar = "FORMAT",
    default = None,
    help="Output format. One of 'chemfp', 'csv', or 'tsv' (default: based on filename, or 'chemfp')",
    )
parser.add_argument(
    "--include-empty",
    default = True,
    action = cmdsupport.BooleanOptionalAction,
    help = "In csv or tsv output, include a line for queries with no hits (the default)",
    )
parser.add_argument(
    "--empty-target-id",
    default = "*",
    help = "In csv or tsv output, the target id for a query with no hits (default: '*')",
    )
parser.add_argument(
    "--empty-score",
    default = "NaN",
    help = "In csv or tsv output, the score for a query with no hits (default: 'NaN')",
    )

cmdsupport.add_precision(parser)

parser.add_argument("-c", "--count", help="report counts", action="store_true")

parser.add_argument("-b", "--batch-size", help="batch size",
                    default=100, type=int)

parser.add_argument("--scan", help="scan the file to find matches (low memory overhead)",
                    action="store_true")
parser.add_argument("--memory", help="build and search an in-memory data structure (faster for multiple queries)",
                    action="store_true")
parser.add_argument("--no-mmap", action="store_true",
                    help="don't use mmap to read uncompressed FPB files. "
                         "May give better performance on networked file systems, "
                         "at the expense of higher memory use.")

parser.add_argument("--times", help="report load and execution times to stderr",
                    action="store_true")

cmdsupport.add_version(parser)
cmdsupport.add_license_check(parser)

parser.add_argument("target_filename", nargs=1, help="target filename", default=None)

## Something to enable multi-threading
#parser.add_argument("-j", "--jobs", help="number of jobs ",
#                    default=10, type=int)


def _get_query_id(id1, id2=None):
    if id1 is not None:
        return id1
    if id2 is not None:
        return id2
    return "Query1"

def print_query_type_help(query_type, err):
    from chemfp.types import FingerprintTypeValueError, tokenize_type_string
    
    if not isinstance(err, FingerprintTypeValueError):
        return

    try:
        tokenized_type = tokenize_type_string(query_type)
    except FingerprintTypeValueError:
        # Can't even parse the string.
        # Print the available fingerprint types.
        family = None
    else:
        try:
            family = chemfp.get_fingerprint_family(tokenized_type.name)
        except FingerprintTypeValueError:
            family = None

    if family is None:
        sys.stderr.write("Available fingerprint types:\n")
        for name in sorted(chemfp.get_fingerprint_family_names()):
            if "/" in name:
                sys.stderr.write(f"  {name}\n")
        return
    
    if not query_type:
        return

    defaults = family.get_defaults()
    if defaults:
        sys.stderr.write("Default settings for fingerprint type %s:\n" % (family.name,))
        for k, v in sorted(defaults.items()):
            sys.stderr.write("  %s: %r\n" % (k, v))

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)
    return _main(parser, args)
    
def _main(parser, args):
    with cmdsupport.MyExitStack() as exit_stack:
        try:
            return main_with_cleanup(parser, args, exit_stack)
        except NotImplementedError as err:
            cmdsupport.check_license_failure(str(err))
            cmdsupport.die("Cannot run simsearch. Exiting.")
    
def main_with_cleanup(parser, args, exit_stack):    
    if chemfp.is_licensed():
        # Give a warning about possible problems
        cmdsupport.check_license_date()
    
    target_filename = args.target_filename[0]

    threshold = args.threshold
    k = args.k_nearest

    if args.precision is None:
        precision = None
    else:
        precision = int(args.precision)

    if args.count and k is not None and k != "all":
        parser.error("--count search does not support --k-nearest")

    if args.alpha is None:
        if args.beta is not None:
            parser.error("Must specify --alpha when --beta is specified")
        alpha = beta = None
    else:
        alpha = alpha_beta(parser.error, "--alpha", args.alpha)
            
        if args.beta is None:
            # if not specified, let beta = alpha
            beta = alpha
        else:
            beta = alpha_beta(parser.error, "--beta", args.beta)

    if chemfp.is_licensed():
        pass
    elif (alpha in (1.0, None)) and (beta in (1.0, None)):
        pass
    else:
        raise NotImplementedError("Tversky search requires a valid chemfp license.")

    # People should not use this without setting parameters.  On the
    # other hand, I don't want an error message if there are no
    # parameters. This solution seems to make sense.

    if threshold is None:
        if k is None:
            # If nothing is set, use defaults of --threshold 0.0 -k 3
            if args.count:
                # Rather pointless to use a 0.0 threshold.
                # Keep the chemfp 3.x default of 0.7
                threshold = 0.7
            else:
                threshold = 0.0
            k = 3
        else:
            # only k is set; search over all possible matches
            threshold = 0.0
    else:
        if k is None:
            # only threshold is set; search for all hits above that threshold
            k = "all"

    if k == "all":
        pass
    elif k < 0:
        parser.error("--k-nearest must be non-negative or 'all'")

    if not (0.0 <= threshold <= 1.0):
        parser.error("--threshold must be between 0.0 and 1.0, inclusive")

    if args.batch_size < 1:
        parser.error("--batch-size must be positive")

    bitops.use_environment_variables()

    output_format, output_compression = check_writer_args(parser, args)

    if args.NxN:
        if args.scan:
            parser.error("Cannot specify --scan with an --NxN search")
        if args.query:
            parser.error("Cannot specify --query with an --NxN search")
        if args.hex_query:
            parser.error("Cannot specify --hex-query with an --NxN search")
        if args.queries:
            parser.error("Cannot specify --queries with an --NxN search")
        do_NxN_searches(args, k, threshold, target_filename, alpha, beta,
                            precision=precision,
                            exit_stack=exit_stack,
                            output_format=output_format,
                            output_compression=output_compression,
                            )
        return
            
    if args.scan and args.memory:
        parser.error("Cannot specify both --scan and --memory")

    if bool(args.query) + bool(args.hex_query) + bool(args.queries) > 1:
        parser.error("Can only specify at most one of --query, --hex-query, or --queries")

    # Verify that the query_id, if present, is in the valid format
    query_id = args.query_id
    if query_id is not None:
        for c, name in ( ("\t", "tab"),
                         ("\n", "newline"),
                         ("\r", "control-return"),
                         ("\0", "NUL")):
            if c in query_id:
                parser.error("--query-id must not contain the %s character" %
                             (name,))
    
    # Open the target file. This reads just enough to get the header.
    if target_filename is not None and target_filename.lower().endswith(".flush"):
        cmdsupport.die("Simsearch cannot use flush files as input. Use chemfp_converters to convert it to FPS or FPB format.")

    try:
        targets = chemfp.open(target_filename, format=args.target_format, allow_mmap=(not args.no_mmap))
    except chemfp.ChemFPError as err:
        sys.stderr.write("Cannot parse targets file: %s\n" % (err,))
        raise SystemExit(1)
    except (OSError, IOError, ValueError) as err:
        sys.stderr.write("Cannot open targets file: %s\n" % (err,))
        raise SystemExit(1)

    exit_stack.enter_if(targets)
    
    if args.hex_query is not None:
        query_id = _get_query_id(query_id)
        try:
            query_fp = bitops.hex_decode(args.hex_query)
        except (TypeError, Exception) as err:
            parser.error("--hex-query is not a hex string: %s" % (err,))

        for problem in chemfp.check_fingerprint_problems(query_fp, targets.metadata,
                                                         query_name="query fingerprint",
                                                         target_name=repr(target_filename)):
            if problem.severity == "error":
                sys.stderr.write("ERROR: %s\n" % (problem.description,))
                raise SystemExit(1)

        num_bits = targets.metadata.num_bits
        if num_bits is None:
            num_bits = len(query_fp) * 8
        query_metadata = chemfp.Metadata(num_bits=num_bits, num_bytes=len(query_fp))
        queries = chemfp.Fingerprints(query_metadata,
                                      [(query_id, query_fp)])
        query_filename = None

    elif args.query is not None:
        # Specified a structure on the command-line
        if args.query_type is not None:
            # Specified a fingerprint type on the command-line
            try:
                fptype = chemfp.get_fingerprint_type(args.query_type)
            except ValueError as err:
                sys.stderr.write("Unable to use specified --query-type: %s\n" % (err,))
                print_query_type_help(args.query_type, err)
                raise SystemExit(1)
        else:
            # Use the fingerprint from the target file, if available
            try:
                fptype = targets.get_fingerprint_type()
            except ValueError as err:
                sys.stderr.write("Unable to use the fingerprint type from %r\n"
                                 "ERROR: %s\n" % (target_filename, err.args[0]))

                raise SystemExit(1)

        format = args.query_format
        if format is None:
            format = "smi"

        try:
            format = fptype.toolkit.get_input_format(format)
        except ValueError as err:
            sys.stderr.write("Cannot use the --query-format: %s\n" % (err,))
            raise SystemExit(1)

        try:
            id, query_fp = fptype.parse_id_and_molecule_fingerprint(args.query, format)
        except chemfp.ParseError as err:
            sys.stderr.write("Cannot use the --query: %s\n" % (err,))
            raise SystemExit(1)
        query_id = _get_query_id(query_id, id)

        queries = chemfp.Fingerprints(fptype.get_metadata(),
                                      [(query_id, query_fp)])
        query_filename = None

    else:
        # Need to figure out if the input contains fingerprints or structures
        query_format = args.query_format
        query_filename = args.queries

        try:
             base_format, compression = io.normalize_input_format(query_filename, query_format, default=(None, ""))
        except ValueError as err:
            sys.stderr.write("Cannot use the --query-format: %s\n" % (err,))
            raise SystemExit(1)

        if base_format is None or base_format in ("fps", "fpb"):
            # fingerprint format
            try:
                queries = chemfp.open(query_filename, format=query_format)
            except chemfp.ChemFPError as err:
                sys.stderr.write("Cannot parse queries file: %s\n" % (err,))
                raise SystemExit(1)
            except (OSError, IOError, ValueError) as err:
                sys.stderr.write("Cannot open queries file: %s\n" % (err,))
                raise SystemExit(1)
            exit_stack.enter_if(queries)
        else:
            # structure format
            # Need a fingerprint type to convert them to fingerprints.
            if args.query_type is not None:
                # Specified a fingerprint type on the command-line
                try:
                    fptype = chemfp.get_fingerprint_type(args.query_type)
                except ValueError as err:
                    sys.stderr.write("Unable to use specified --query-type: %s\n" % (err,))
                    print_query_type_help(args.query_type, err)
                    raise SystemExit(1)
            else:
                # Use the fingerprint from the target file, if available
                try:
                    fptype = targets.get_fingerprint_type()
                except ValueError as err:
                    sys.stderr.write("The queries do not appear to be fingerprints so assuming they are structures.\n")
                    sys.stderr.write("Cannot convert the query structures to fingerprints:\n")
                    if targets.metadata is None or not targets.metadata.type:
                        sys.stderr.write("  The target file %r does not specify a fingerprint type\n"
                                             % (target_filename,))
                    else:
                        sys.stderr.write("  Cannot interpret the target fingerprint type: %s\n"
                                             % (err,))
                    sys.stderr.write("Specify a fingerprint --query-format or extension if the queries contain fingerprints.\n")
                    raise SystemExit(1)
        
            try:
                format = fptype.toolkit.get_input_format_from_source(query_filename, query_format)
            except ValueError as err:
                sys.stderr.write("Cannot use the --query-format: Not a fingerprint format and %s\n" % (err,))
                raise SystemExit(1)
        
            id_tag = args.id_tag
            if not cmdsupport.is_valid_tag(id_tag):
                sys.stderr.write("The --id-tag %r contains one or more unsupported characters.\n" % (id_tag,))
                raise SystemExit(1)

            reader_settings = cmdsupport.get_reader_settings(parser, args)

            metadata, num_readers, first_reader, readers = cmdsupport.read_multifile_structure_fingerprints(
                fptype, [query_filename], format=format, id_tag=id_tag,
                reader_settings = reader_settings, errors=args.errors)
            queries = first_reader
            exit_stack.enter_if(queries)

    batch_size = args.batch_size
    # If the targets file is empty then the number of bytes is None.
    # The metadata check between the queries and targets will fail.
    # The solution is easy - replace the targets metadata.
    if targets.metadata.num_bytes is None:
        targets.metadata = targets.metadata.copy(num_bytes=queries.metadata.num_bytes)

    query_arena_iter = queries.iter_arenas(batch_size)
    
    start_time = t1 = time.time()

    first_query_arena = None
    
    try:
        for first_query_arena in query_arena_iter:
            break
    except chemfp.ChemFPError as err:
        sys.stderr.write("Cannot parse queries file: %s\n" % (err,))
        raise SystemExit(1)

    if args.scan:
        # Leave the targets as-is
        pass
    elif args.memory:
        try:
            targets = chemfp.load_fingerprints(targets)
        except NotImplementedError as err:
            raise NotImplementedError(str(err) + "\n" + "(You might try using --scan to avoid an in-memory search.)")
    if not first_query_arena:
        # No input. Leave as-is
        pass
    elif len(first_query_arena) <= min(20, batch_size):
        # Figure out the optimal search. If there is a
        # small number of inputs (< ~20) then a scan
        # of the FPS file is faster than an arena search.
        pass
    else:
        targets = chemfp.load_fingerprints(targets)

    problems = chemfp.check_metadata_problems(queries.metadata, targets.metadata,
                                              query_name="queries", target_name="targets")
    for problem in problems:
        if problem.severity == "error":
            sys.stderr.write("ERROR: " + problem.description + "\n")
            raise SystemExit(1)
        elif problem.severity == "warning":
            sys.stderr.write("WARNING: " + problem.description + "\n")
    t2 = time.time()
    open_time = t2 - t1
    t1 = t2
    try:
        outfile, close = io.open_compressed_output(args.output, output_compression, "simsearch")
    except (IOError, ValueError) as err:
        sys.stderr.write("Cannot open output file: %s\n" % (err,))
        raise SystemExit(1)
    
    exit_stack.call_if(close)

    if args.count:
        type = make_output_type("Count threshold=%(threshold)s",
                                threshold=threshold, alpha=alpha, beta=beta)
        output_writer = get_count_writer(output_format, args)
    else:
        type = make_output_type("%(name)s k=%(k)s threshold=%(threshold)s",
                                k=k, threshold=threshold, alpha=alpha, beta=beta)
        output_writer = get_simsearch_writer(output_format, args)

    exit_stack.callback(output_writer.detach)
    output_writer.write_magic(outfile)
    
    if alpha is None:
        assert beta is None
        alpha = beta = 1.0

    output_writer.write_header(outfile, {
        "num_bits": targets.metadata.num_bits,
        "software": SOFTWARE,
        "type": type,
        "queries": query_filename,
        "targets": target_filename,
        "query_sources": queries.metadata.sources,
        "target_sources": targets.metadata.sources})

    t2 = time.time()
    output_time = t2-t1

    elapsed_time = ElapsedTime()
    if first_query_arena:
        query_arenas = itertools.chain([first_query_arena],
                                       query_arena_iter)

        try:
            if args.count:
                report_counts(output_writer, outfile, query_arenas, targets,
                              threshold = threshold, alpha=alpha, beta=beta,
                              elapsed_time = elapsed_time)
            elif k == "all":
                report_threshold(output_writer, outfile, query_arenas, targets,
                                 threshold = threshold, alpha=alpha, beta=beta,
                                 precision = precision,
                                 elapsed_time = elapsed_time)
                                 
            else:
                report_knearest(output_writer, outfile, query_arenas, targets,
                                k = k, threshold = threshold, alpha=alpha, beta=beta,
                                 precision = precision,
                                elapsed_time = elapsed_time)
                                
        except chemfp.ChemFPError as err:
            sys.stderr.write("Cannot parse targets file: %s\n" % (err,))
            raise SystemExit(1)
            
    output_time += elapsed_time.output_time
    total_time = time.time() - start_time
    
    if args.times:
        if args.output is None:
            # If writing to stdout, flush it so times go after the output.
            try:
                sys.stdout.flush()
            except Exception:
                pass
        sys.stderr.write("open %.2f read %.2f search %.2f output %.2f total %.2f\n" % (
            open_time, elapsed_time.read_time, elapsed_time.search_time, output_time, total_time))

command_config = registry.CommandConfig.from_argparse(parser, "fp_search", _main)

 
if __name__ == "__main__":
    main()
