# Copyright (c) 2021-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import argparse
import array
import itertools
import math
import sys
import time

from io import TextIOWrapper

from .. import ChemFPError, open as chemfp_open
from .. import io
from .. import registry
from .. import diversity

from . import cmdsupport
from . import diversity_utils

def get_software():
    # This can change if the license changes
    from .. import SOFTWARE
    return SOFTWARE

def get_spherex_type(
        *,
        num_initial_picks,
        num_picks,
        threshold,
        randomize,
        seed,
        directed_metadata,
        ):
    if randomize is None:
        if directed_metadata:
            randomize = False
        else:
            randomize = True

    terms = ["spherex"]
    
    if num_initial_picks > 0:
        terms.append(f"num-initial-picks={num_initial_picks}")

    if threshold is not None:
        terms.append(f"threshold={threshold}")
        
    if num_picks is not None:
        terms.append(f"num-picks={num_picks}")

    if randomize:
        terms.append(f"randomize=1 seed={seed}")
    else:
        terms.append("randomize=0")

    if directed_metadata:
        d = directed_metadata.copy()
        dise_type = d.pop("dise-type")
        terms.append(f"directed={dise_type}")
        other_metadata = d
    else:
        other_metadata = {}
    
    return " ".join(terms), other_metadata

def parse_nonempty(value):
    if not value:
        raise argparse.ArgumentTypeError("must not be the empty string")
    return value

parser = argparse.ArgumentParser(
    description="Select diverse fingerprints using the sphere exclusion algorithm",
    allow_abbrev=False,
    )

diversity_utils.add_threshold(parser)
diversity_utils.add_num_picks(parser)

parser.add_argument(
    "--dise",
    action = "store_true",
    help = "Use directed sphere exclusion",
    )
parser.add_argument(
    "--dise-references",
    metavar = "FILENAME",
    help = "DISE reference structures or fingerprints (default uses the Gobbi & Lee structures)",
    )
parser.add_argument(
    "--dise-format",
    metavar = "FORMAT",
    help = "format of the DISE reference file (default uses the file extension, else 'fps')",
    )
parser.add_argument(
    "--ranks",
    metavar = "FILENAME",
    help = "file containing fingerprint rank values",
    )
parser.add_argument(
    "--ranks-default",
    metavar = "N",
    type = int,
    default = 2**32-1,
    help = "default rank value if candidate id not found in the ranks file (default: 2**32-1)",
    )
parser.add_argument(
    "--ranks-format",
    metavar = "FORMAT",
    choices = ("tsv",),
    default = "tsv",
    help = "format for the ranks file (only 'tsv' is supported)",
    )
parser.add_argument(
    "--ranks-has-header",
    action = "store_true",
    help = "skip the first line of the ranks file",
    )

    
diversity_utils.add_candidates_format(parser)

diversity_utils.add_references_and_format(parser)

def add_initial_ids(parser):
    parser.add_argument(
        "--pick-id", action="append", type = parse_nonempty,
        help = "Initial candidate id (if no reference file). Can be used more than once.")

    parser.add_argument(
        "--pick-id-file", metavar="FILENAME",
        help = "File containing initial candidate ids, one per line")

add_initial_ids(parser)

parser.add_argument(
    "--randomize",
    default = None,
    action = cmdsupport.BooleanOptionalAction,
    help = (
        "Use --randomize (default for undirected picking) to randomly pick from the available candidates, "
        "or --no-random (default for directed picking) to pick the candidate with the smallest arena index."
        )
    )
diversity_utils.add_seed(parser)
diversity_utils.add_mmap(parser)

parser.add_argument(
    "--include-hits",
    action = "store_true",
    help = "include ids and scores for fingerprints in each sphere",
    )
diversity_utils.add_picker_output_args(parser, precision=True)

parser.add_argument(
    "--include-empty",
    default = True,
    action = cmdsupport.BooleanOptionalAction,
    help = "In csv and tsv format with --include-hits, include picks with no hits (the default)",
    )
parser.add_argument(
    "--empty-score",
    default = "NaN",
    help = "The score if --include-empty outputs a pick with no hits (default: 'NaN')",
    )
parser.add_argument(
    "--empty-hit-id",
    default = "*",
    help = "The hit id if --include-empty outputs a pick with no hits (default: '*')",
    )

diversity_utils.add_pick_time(parser)
diversity_utils.add_date(parser)
diversity_utils.add_times(parser)
cmdsupport.add_progress_bar(parser)
diversity_utils.add_candidates(parser)

####
def get_initial_candidates_from_file(
        parser, pick_id_file, candidates_arena):
    try:
        f = open(pick_id_file)
    except Exception as err:
        parser.error(f"Cannot open --pick-id-file: {err}")
    
    with f:
        initial_candidates = []
        for lineno, line in enumerate(f, 1):
            candidate_id = line.rstrip("\r\n")
            idx = candidates_arena.get_index_by_id(candidate_id)
            if idx is None:
                parser.exit(
                    3,
                    f"pick id {candidate_id!r} not found in --candidates file "
                    f"{pick_id_file!r}, line {lineno}"
                    )
            initial_candidates.append(idx)
    return initial_candidates
            

def get_references_arena_or_initial_candidates(parser, args, candidates_arena, *,
                                                   ranks=None, progress=True):
    references_filename = args.references

    pick_id_list = args.pick_id
    pick_id_file = args.pick_id_file

    n = bool(references_filename) + bool(pick_id_list) + bool(pick_id_file)
    if n > 1:
        parser.error("May only specify at most one of --references, --pick-id, or --pick-id-file")

    references_arena = initial_candidates = None

    if references_filename is not None:
        references_filename, references_arena = diversity_utils.get_references_arena(
            parser, args, candidates_arena, "spherex", progress = progress)
        

    elif pick_id_list is not None:
        initial_candidates = []
        for id in pick_id_list:
            idx = candidates_arena.get_index_by_id(id)
            if idx is None:
                parser.error(f"--pick-id {id!r} not found in --candidates file")
            initial_candidates.append(idx)

    elif pick_id_file is not None:
        initial_candidates = get_initial_candidates_from_file(
            parser, pick_id_file, candidates_arena)
        
    else:
        initial_candidates = []

    return references_filename, references_arena, initial_candidates


def get_ranks(parser, args, arena):
    use_dise = args.dise or (args.dise_references is not None)
    use_ranks = args.ranks is not None

    if not (use_dise or use_ranks):
        return None, {}

    if use_dise and use_ranks:
        if args.dise_references:
            parser.error("May only use one of --dise-references or --ranks")
        else:
            parser.error("May only use one of --dise or --ranks")
        
    if use_dise:
        return get_dise_ranks(
            arena,
            ref_filename = args.dise_references,
            ref_format = args.dise_format,
            )
    else:
        return get_file_ranks(
            arena,
            ranks_filename = args.ranks,
            ranks_format = args.ranks_format,
            ranks_has_header = args.ranks_has_header,
            ranks_default = args.ranks_default,
            )

def die9(msg):
    sys.stderr.write(f"{msg}\n")
    raise SystemExit(9)
    
def get_dise_ranks(arena, ref_filename, ref_format):
    with cmdsupport.MyExitStack() as exit_stack:
        return _get_dise_ranks(arena, ref_filename, ref_format, exit_stack)

def _get_dise_ranks(arena, ref_filename, ref_format, exit_stack):
    # TODO: What if the fingerprint type doesn't exist or can't be created?
    if ref_filename is not None:
        # can be fingerprints or structures
        try:
             base_format, compression = io.normalize_input_format(ref_filename, ref_format, default=(None, ""))
        except ValueError as err:
            die9(f"Cannot use the --dise-format: {err}")

        if base_format is None or base_format in ("fps", "fpb"):
            # fingerprint format
            try:
                refs = chemfp_open(ref_filename, format=ref_format)
            except ChemFPError as err:
                die9(f"Cannot parse --dise-references file: {err}")
            except (OSError, IOError, ValueError) as err:
                die9(f"Cannot open --dise-references file: {err}")
            exit_stack.enter_context(refs)

            diversity_utils.check_metadata_compatability(
                "candidates", arena.metadata, "--dise-references fingerprints", refs.metadata)
            ranker = diversity.DISERanker.from_fingerprints(refs)
            directed_metadata = {
                "dise-type": "fingerprints",
                "dise-fingerprints": ref_filename,
                }
            if refs.metadata.type is not None:
                directed_metadata["dise-fingerprint-type"] = refs.metadata.type
                
        else:
            # structure format
            if arena.metadata.type is None:
                die9(
                    "Cannot generate fingerprints for the --dise-references structures: "
                    "No candidate fingerprint type available"
                    )
            try:
                fptype = arena.get_fingerprint_type()
            except (ChemFPError, ValueError) as err:
                msg = f"Unable to use candidate fingerprint type {arena.metadata.type!r}: {err}"
                die9(f"Cannot generate fingerprints for the --dise-references structures: {msg}")

            # Check now for validity so it's easier to report the cause of an error
            try:
                fptype.toolkit.get_input_format_from_source(ref_filename, ref_format)
            except ValueError as err:
                die9(f"Cannot use the --dise-format: Not a fingerprint format and {err}")

            try:
                reader = fptype.read_molecule_fingerprints(ref_filename, ref_format)
            except (IOError, OSError, ValueError) as err:
                die9(f"Cannot open --dise-references structure file: {err}")
                
            try:
                with reader:
                    ranker = diversity.DISERanker.from_fingerprints(reader)
            except ChemFPError as err:
                die9(f"Cannot process all --dise-references structures: {err}")
            directed_metadata = {
                "dise-type": "structures",
                "dise-structures": ref_filename,
                }
                
    else:
        if arena.metadata.type is None:
            die9(
                "Cannot generate --dise fingerprints for the candidates: "
                "No candidate fingerprint type available"
                )
        try:
            fptype = arena.get_fingerprint_type()
        except (ChemFPError, ValueError) as err:
            msg = f"Unable to use candidate fingerprint type {arena.metadata.type!r}: {err}"
            die9(f"Cannot generate --dise fingerprints for the candidates: {msg}")
            
        ranker = diversity.DISERanker.from_dise_paper(fptype)
        
        directed_metadata = {
            "dise-type": "DISE-paper",
            }

    ranks = ranker.rank_arena(arena)
    return ranks, directed_metadata

def get_file_ranks(arena, ranks_filename, ranks_format, ranks_has_header, ranks_default):
    assert ranks_format == "tsv", (ranks_format, "unknown")

    # map id -> value (last value wins in case of duplicate)
    values = dict.fromkeys(arena.ids, ranks_default)
    id_to_idx = dict(zip(arena.ids, range(len(arena))))
    skip_first = ranks_has_header

    try:
        infile = open(ranks_filename)
    except (IOError, OSError) as err:
        die9(f"Cannot open --ranks file: {err}")
    with infile:
        for lineno, line in enumerate(infile, 1):
            if skip_first:
                skip_first = False
                continue
            terms = line.split("\t")
            if len(terms) < 2:
                die9(f"Need at least two tab-separated columns, on line {lineno} of {ranks_filename!r}.")
            id, value_str = terms[:2]

            idx = arena.get_index_by_id(id)
            if idx is None:
                die9(f"Unknown candidate id {id!r}, on line {lineno} of {ranks_filename!r}")
            id_to_idx[id] = idx

            try:
                value = int(value_str)
            except ValueError:
                try:
                    value = float(value_str)
                except ValueError:
                    die9(f"Cannot interpret rank value, on line {lineno} of {ranks_filename!r}: {value_str.rstrip()!r}")
                if math.isnan(value):
                    die9(f"NaN value not supported, on line {lineno} of {ranks_filename!r}: {value_str.rstrip()!r}")
                
            values[id] = value
            

    # Order from smallest to largest
    items = [(value, id) for (id, value) in values.items()]
    items.sort()

    # Replace the values with the rank
    prev_value = None
    rank = -1
    for (value, id) in items:
        if value != prev_value:
            rank += 1
            prev_value = value
        values[id] = rank

    # Create the ranks array, initialized with default values
    ranks = array.array("I", [0] * len(arena))
    for id, rank in values.items():
        ranks[id_to_idx[id]] = rank

    return ranks, {
        "dise-type": "ranks",
        "dise-ranks": ranks_filename,
        "dise-default-rank": ranks_default,
        }

#### output

TQDM_KWARGS = dict(
    desc = "picks",
    unit = "",
    dynamic_ncols = True,
    )

class Precision:
    def __init__(self, precision):
        self.score_formatter = f"%.{precision}f"

BASE_TITLES = ["i", "pick_id", "count", "pick_time"]

class BaseWriter:
    def write_titles(self, outfile, include_pick_time):
        raise NotImplementedError # pragma: no cover

    def write_result(self, outfile, pick_i, pick_idx, pick_id, pick):
        raise NotImplementedError # pragma: no cover

_header_encoders = {
    "candidates": diversity_utils.urlquote,
    "references": diversity_utils.urlquote,
    "dise-fingerprints": diversity_utils.urlquote,
    "dise-structures": diversity_utils.urlquote,
    "dise-ranks": diversity_utils.urlquote,
    }

class BaseChemfpWriter(BaseWriter):
    def write_header(self, outfile, d):
        outfile.write("#Spherex/1\n")
        for k, v in d.items():
            if v is not None:
                if k in _header_encoders:
                    v = _header_encoders[k](v)
                outfile.write(f"#{k}={v}\n")

class ChemfpCountWriter(BaseChemfpWriter):
    def write_titles(self, outfile, include_pick_time):
        if include_pick_time:
            titles = BASE_TITLES[:]
        else:
            titles = BASE_TITLES[:-1]
        outfile.write("\t".join(titles) + "\n")
        
    def write_result(self, outfile, pick_i, pick_idx, pick_id, count, elapsed_time):
        if elapsed_time is None:
            outfile.write(f"{pick_i}\t{pick_id}\t{count}\n")
        else:
            outfile.write(f"{pick_i}\t{pick_id}\t{count}\t{elapsed_time:.2f}\n")

class ChemfpHitsWriter(Precision, BaseChemfpWriter):
    def write_titles(self, outfile, include_pick_time):
        if include_pick_time:
            titles = BASE_TITLES[:]
        else:
            titles = BASE_TITLES[:-1]
        titles.append("hits")
        outfile.write("\t".join(titles) + "\n")
        self.hit_formatter = f"\t%s\t{self.score_formatter}"
        
    def write_result(self, outfile, pick_i, pick_idx, pick_id, hits, elapsed_time):
        count = len(hits)
        hit_formatter = self.hit_formatter
        if count:
            extra = "".join((hit_formatter % hit) for hit in hits.get_ids_and_scores())
        else:
            extra = "\t"
        if elapsed_time is None:
            outfile.write(f"{pick_i}\t{pick_id}\t{count}{extra}\n")
        else:
            outfile.write(f"{pick_i}\t{pick_id}\t{count}\t{elapsed_time:.2f}{extra}\n")

class BaseExcelWriter(BaseWriter):
    def write_header(self, outfile, d):
        pass
    
class ExcelCountWriter(BaseExcelWriter):
    def __init__(self, outfile, dialect):
        import csv
        self.csv_writer = csv_writer = csv.writer(outfile, dialect=dialect)
        self.writerow = csv_writer.writerow

    def write_titles(self, outfile, include_pick_time):
        if include_pick_time:
            self.writerow(BASE_TITLES[1:4])
        else:
            self.writerow(BASE_TITLES[1:3])
        
    def write_result(self, outfile, pick_i, pick_idx, pick_id, count, elapsed_time):
        if elapsed_time is None:
            self.writerow([pick_id, count])
        else:
            self.writerow([pick_id, count, f"{elapsed_time:.2f}"])

class ExcelHitsWriter(Precision, BaseExcelWriter):
    def __init__(self, outfile, dialect, precision, include_empty, empty_hit_id, empty_score):
        super().__init__(precision)
        self.include_empty = include_empty
        self.empty_hit_id = empty_hit_id
        self.empty_score = empty_score
        
        import csv
        self.csv_writer = csv_writer = csv.writer(outfile, dialect=dialect)
        self.writerow = csv_writer.writerow

    def write_titles(self, outfile, include_pick_time):
        titles = [BASE_TITLES[1], "hit_id", "hit_score"]
        if include_pick_time:
            titles.append(BASE_TITLES[3])
        self.writerow(titles)

    def write_result(self, outfile, pick_i, pick_idx, pick_id, picks, elapsed_time):
        if picks:
            score_formatter = self.score_formatter
            write_row = self.writerow
            if elapsed_time is None:
                for hit_id, score in picks.get_ids_and_scores():
                    write_row([pick_id, hit_id, score_formatter % score])
            else:
                elapsed_time_str = f"{elapsed_time:.2f}"
                for hit_id, score in picks.get_ids_and_scores():
                    write_row([pick_id, hit_id, score_formatter % score, elapsed_time_str])
        elif self.include_empty:
            if elapsed_time is None:
                self.writerow([pick_id, self.empty_hit_id, self.empty_score])
            else:
                self.writerow([pick_id, self.empty_hit_id, self.empty_score,
                                   f"{elapsed_time:.2f}"])

def get_count_writer(parser, args, outfile, output_format):
    if output_format == "csv":
        return ExcelCountWriter(outfile, dialect="excel")
    
    if output_format == "tsv":
        return ExcelCountWriter(outfile, dialect="excel-tab")

    if output_format in (None, "chemfp") or args.output_format is None:
        return ChemfpCountWriter()
    
    parser.error(f"Unsupported --out format {args.output_format!r}")
    
def get_hits_writer(parser, args, outfile, output_format, precision):
    dialect = None
    
    if output_format == "csv":
        dialect = "excel"
    elif output_format == "tsv":
        dialect = "excel-tab"

    if dialect is not None:
        return ExcelHitsWriter(
            outfile,
            dialect=dialect,
            precision=precision,
            include_empty = args.include_empty,
            empty_hit_id = args.empty_hit_id,
            empty_score = args.empty_score,
            )
    
    if output_format in (None, "chemfp") or args.output_format is None:
        return ChemfpHitsWriter(precision)

    parser.error(f"Unsupported --out format {args.output_format!r}")

def enumerate_picks(progress, it, picker, num_picks):
    if num_picks == 0:
        return

    if num_picks is None:
        # Pick all. Each pick may remove many fingerprints, so
        # use the number of remaining candidates as the guide.
        before = len(picker.candidates)
        with diversity_utils.get_progress_bar(progress, before) as progress_bar:
            for pick_i, (pick_idx, pick) in enumerate(it, 1):
                yield pick_i, pick_idx, pick
                
                after = len(picker.candidates)
                progress_bar.update(before - after)
                before = after
        return
                
    it = itertools.islice(it, 0, num_picks)
    with diversity_utils.get_progress_bar(progress, num_picks) as progress_bar:
        for pick_i, (pick_idx, pick) in enumerate(it, 1):
            yield pick_i, pick_idx, pick
            progress_bar.update(1)

####

def spherex_command(parser, args):
    with cmdsupport.MyExitStack() as exit_stack:
        return spherex_command_with_exit_stack(parser, args, exit_stack)

def spherex_command_with_exit_stack(parser, args, exit_stack):
    start_time = init_start_time = time.time()
    
    progress = cmdsupport.progress_bar_is_enabled(args.progress, args.candidates, args.output)
    seed = diversity_utils.parse_seed(args.seed)
    num_picks = args.num_picks
    include_pick_time = args.pick_time
    date_str = diversity_utils.get_date_str(parser, args)

    # Figure out the input
    candidates_filename, candidates_arena = diversity_utils.get_candidates_arena(
        parser, args, "spherex", progress)
    exit_stack.enter_context(candidates_arena)

    ranks, directed_metadata = get_ranks(parser, args, candidates_arena)
    
    references_filename, references_arena, initial_indices = get_references_arena_or_initial_candidates(
        parser, args, candidates_arena, ranks=ranks, progress=progress)
    exit_stack.enter_if(references_arena)

    if initial_indices is None:
        initial_pick_ids = None
    else:
        initial_pick_ids = [candidates_arena.ids[idx] for idx in initial_indices]

    # Figure out the output

    output_format, binary_outfile, binary_outfile_close = diversity_utils.get_output(
        parser, args, "spherex")
    exit_stack.call_if(binary_outfile_close)

    # Don't let the wrapper handle the close
    outfile = TextIOWrapper(binary_outfile, write_through=True)
    exit_stack.callback(outfile.detach)
    
    picks_writer = diversity_utils.get_save_picks_writer(parser, args, candidates_arena, "spherex")
    exit_stack.enter_if(picks_writer)
    
    candidates_writer = diversity_utils.get_save_candidates_writer(parser, args, candidates_arena, "spherex")
    exit_stack.enter_if(candidates_writer)

    # Create the picker
    try:
        picker = diversity.get_sphere_exclusion_picker(
            candidate_arena = candidates_arena,
            reference_arena = references_arena,
            initial_picks = initial_indices,
            threshold = args.threshold,
            randomize = args.randomize,
            seed = seed,
            ranks = ranks,
            )
    except NotImplementedError as err:
        if "chemfp diversity license" in str(err):
            if references_arena is None:
                cmdsupport.die_no_diversity_license("spherex", 1)
            else:
                cmdsupport.die_no_diversity_license("spherex", 2)
        raise


    include_hits = args.include_hits
    if include_hits:
        tanimoto_precision = diversity_utils.get_tanimoto_precision(
            args, candidates_arena)
        writer = get_hits_writer(parser, args, outfile, output_format, tanimoto_precision)
        it = picker.iter_indices_and_neighbors()
    else:
        writer = get_count_writer(parser, args, outfile, output_format)
        it = picker.iter_indices_and_counts()


    spherex_type_str, other_metadata = get_spherex_type(
                num_initial_picks = len(initial_pick_ids) if initial_pick_ids else 0,
                num_picks = num_picks,
                threshold = args.threshold,
                randomize = args.randomize,
                seed = seed,
                directed_metadata = directed_metadata,
                )
    writer.write_header(
        outfile, {
            "num_bits": candidates_arena.num_bits,
            "type": spherex_type_str,
            **other_metadata,
            "software": get_software(),
            "candidates": candidates_filename,
            "references": references_filename,
            "date": date_str,
        })
    writer.write_titles(outfile, include_pick_time=include_pick_time)

    init_end_time = pick_start_time = time.time()
    elapsed_time = None
                
    try:
        ids = candidates_arena.ids
        for pick_i, pick_idx, pick in enumerate_picks(progress, it, picker, num_picks):
            pick_id = ids[pick_idx]
            
            if picks_writer is not None:
                pick_fp = candidates_arena.fingerprints[pick_idx]
                picks_writer.write_fingerprint(pick_id, pick_fp)

            if include_pick_time:
                elapsed_time = time.time() - pick_start_time
            writer.write_result(outfile, pick_i, pick_idx, pick_id, pick, elapsed_time)
            outfile.flush()
    finally:
        pick_end_time = end_time = time.time()

        diversity_utils.report_times(
            args.times,
            init_start_time, init_end_time,
            pick_start_time, pick_end_time,
            start_time, end_time,
            len(picker.picks))

        diversity_utils.save_remaining_candidates_spherex(
            candidates_arena, candidates_writer, picker)

def main(args=None):
    args = parser.parse_args(args)
    return spherex_command(parser, args)

command_config = registry.CommandConfig.from_argparse(parser, "fp_diversity", spherex_command)

if __name__ == "__main__":
    main()
