# Copyright (c) 2010-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""FPS file similarity search and search result implementations.

Chemfp implements similarity search methods which work directly on FPS
files. This might be useful in a streaming environment (where the FPS
data is generated on-the-fly and not saved), and where you have at
most a handful of queries. In that case, an FPS search is faster than
an arena-based search because the FPS parsing overhead is about the
same, but the FPS search have the arena creation or memory overhead an
in-memory search would have.


"""

import ctypes
import itertools
import array

from . import libchemfp_core as _core
from . import _fps

from . import ParseError

def _chemfp_error(err, location):
    msg = _core.strerror(err)
    if -40 <= err <= -30:
        return ParseError(msg, location)
    elif err == -2:
        return MemoryError(msg)
    else:
        # This shouldn't happen
        return RuntimeError(msg)

def require_matching_sizes(query_arena, target_reader):
    query_num_bits = query_arena.metadata.num_bits
    assert query_num_bits is not None, "arenas must define num_bits"
    target_num_bits = target_reader.metadata.num_bits
    if (target_num_bits is not None):
        if query_num_bits != target_num_bits:
            raise ValueError("query_arena has %d bits while target_reader has %d bits" % (query_num_bits, target_num_bits))

    query_num_bytes = query_arena.metadata.num_bytes
    assert query_num_bytes is not None, "arenas must define num_bytes"
    target_num_bytes = target_reader.metadata.num_bytes
    if target_num_bytes is None:
        raise ValueError("target_reader missing num_bytes metadata")
    if query_num_bytes != target_num_bytes:
        raise ValueError("query_arena uses %d bytes while target_reader uses %d bytes" % (query_num_bytes, target_num_bytes))


def report_errors(problem_report):
    for (severity, error, msg_template) in problem_report:
        if severity == "error":
            raise TypeError(msg_template % dict(metadata1 = "query",
                                                metadata2 = "target"))

######## count Tanimoto search #########

def _fp_to_arena(query_fp, metadata):
    if len(query_fp) != metadata.num_bytes:
        raise ValueError("query fingerprint has %d bytes while target_reader uses %d bytes" % (
            len(query_fp), metadata.num_bytes))
    from . import arena
    return arena.FingerprintArena(metadata, 1, 0, 0, len(query_fp), query_fp, b"", [None])

def count_tanimoto_hits_fp(query_fp, target_reader, threshold=0.7):
    """Count the number of hits in *target_reader* at least *threshold* similar to the *query_fp*

    This uses Tanimoto similarity.
    """
    return count_tanimoto_hits_arena(_fp_to_arena(query_fp, target_reader.metadata), target_reader, threshold)[0]

def count_tanimoto_hits_arena(query_arena, target_reader, threshold=0.7):
    """For each fingerprint in *query_arena*, count the number of hits in *target_reader* at least *threshold* similar to it

    This uses Tanimoto similarity.
    """

    require_matching_sizes(query_arena, target_reader)
    counts = array.array("i", (0 for i in range(len(query_arena))))

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    #location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            err, num_lines = _fps.fps_count_tanimoto_hits(
                query_arena.metadata.num_bits,
                query_arena.start_padding, query_arena.end_padding,
                query_arena.storage_size, query_arena.arena, 0, -1,
                block, 0, -1,
                threshold, counts)
            lineno += num_lines
            if err:
                lineno += 1
                raise _chemfp_error(err, location)

        return list(counts)
    finally:
        location.save(lineno=get_lineno())



######## threshold Tanimoto search #########

class TanimotoCell(ctypes.Structure):
    _fields_ = [("score", ctypes.c_double),
                ("query_index", ctypes.c_int),
                ("id_start", ctypes.c_long),
                ("id_end", ctypes.c_long)]


def threshold_tanimoto_search_fp(query_fp, target_reader, threshold=0.7):
    """Find matches in the target reader which are at least threshold similar to the query fingerprint

    :returns: an :class:`.FPSSearchResult` instance contain the result.
    """
    ids = []
    scores = []
    fp_size = len(query_fp)
    num_bits = fp_size * 8
    if fp_size != target_reader.metadata.num_bytes:
        raise ValueError("query fingerprint has %d bytes while target_reader uses %d bytes" % (
            fp_size, target_reader.metadata.num_bytes))
    NUM_CELLS = 1000
    cells = (TanimotoCell*NUM_CELLS)()

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _fps.fps_threshold_tanimoto_search(
                    num_bits, 0, 0, fp_size, query_fp, 0, -1,
                    block, start, end,
                    threshold, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, location)

                for cell in itertools.islice(cells, 0, num_cells):
                        ids.append(block[cell.id_start:cell.id_end].decode("utf8"))
                        scores.append(cell.score)
                if start == end:
                    break
        return FPSSearchResult(ids, scores)
    finally:
        location.save(lineno=get_lineno())

def threshold_tanimoto_search_arena(query_arena, target_reader, threshold):
    """Find matches in the target reader which are at least threshold similar to the query arena fingerprints

    :returns: an :class:`.FPSSearchResults` instance containing a list of query results.
    """
    
    require_matching_sizes(query_arena, target_reader)

    if not query_arena:
        return FPSSearchResults([], [])
    results = [FPSSearchResult([], [], query_id=query_id) for query_id in query_arena.ids]
    
    # Compute at least 100 tanimotos per query, but at most 10,000 at a time
    # (That's about 200K of memory)
    NUM_CELLS = max(10000, len(query_arena) * 100)
    cells = (TanimotoCell*NUM_CELLS)()

    lineno = target_reader.location.lineno

    def get_lineno():
        return lineno
    target_reader.location.register(get_lineno=get_lineno)
    
    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _fps.fps_threshold_tanimoto_search(
                    query_arena.metadata.num_bits,
                    query_arena.start_padding, query_arena.end_padding,
                    query_arena.storage_size, query_arena.arena, 0, -1,
                    block, start, end,
                    threshold, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, target_reader.location)

                for cell in itertools.islice(cells, 0, num_cells):
                    id = block[cell.id_start:cell.id_end]
                    id = id.decode("utf8", "surrogateescape")
                    result = results[cell.query_index]
                    result.ids.append(id)
                    result.scores.append(cell.score)

                if start == end:
                    break
    finally:
        target_reader.location.save(lineno=get_lineno())

    return FPSSearchResults(query_arena.ids, results)
            
######### k-nearest Tanimoto search, with threshold


# Support for peering into the chemfp_fps_heap data structure

def _make_knearest_search(num_queries, k):
    class TanimotoHeap(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int),
                    ("heap_state", ctypes.c_int),
                    ("indices", ctypes.POINTER(ctypes.c_int*k)),
                    ("ids", ctypes.POINTER(ctypes.c_char_p*k)),
                    ("scores", ctypes.POINTER(ctypes.c_double*k))]

    class KNearestSearch(ctypes.Structure):
        _fields_ = [("queries_start", ctypes.c_char_p),
                    ("num_queries", ctypes.c_int),
                    ("query_fp_size", ctypes.c_int),
                    ("query_storage_size", ctypes.c_int),
                    ("k", ctypes.c_int),
                    ("search_state", ctypes.c_int),
                    ("threshold", ctypes.c_double),
                    ("scaled_alpha", ctypes.c_int),
                    ("scaled_beta", ctypes.c_int),
                    ("heaps", ctypes.POINTER(TanimotoHeap*num_queries)),
                    ("num_targets_processed", ctypes.c_int),
                    ("_all_ids", ctypes.c_void_p),
                    ("_all_scores", ctypes.c_void_p)]

    return KNearestSearch()


def knearest_tanimoto_search_fp(query_fp, target_reader, k=3, threshold=0.0):
    """Find the nearest k matches in the target reader which are at least threshold similar to the query fingerprint

    This uses Tanimoto similarity.

    :returns: an :class:`.FPSSearchResult` instance contain the result.
    """
    query_arena = _fp_to_arena(query_fp, target_reader.metadata)
    return knearest_tanimoto_search_arena(query_arena, target_reader, k, threshold)[0]

def knearest_tanimoto_search_arena(query_arena, target_reader, k=3, threshold=0.0):
    """Find the nearest k matches in the target reader which are at least threshold similar to the query arena fingerprints

    This uses Tanimoto similarity.

    :returns: an :class:`.FPSSearchResults` instance containing a list of query results.
    """
    require_matching_sizes(query_arena, target_reader)
    if k < 0:
        raise ValueError("k must be non-negative")

    num_queries = len(query_arena)
    search = _make_knearest_search(num_queries, k)

    _fps.fps_knearest_search_init(
        search,
        query_arena.metadata.num_bits,
        query_arena.start_padding, query_arena.end_padding,
        query_arena.storage_size, query_arena.arena, 0, -1,
        k, threshold, 1.0, 1.0)

    location = target_reader.location
    first_lineno = location.lineno
    try:
        for block in target_reader.iter_blocks():
            err = _fps.fps_knearest_tanimoto_search_feed(search, block, 0, -1)
            if err:
                first_lineno += 1
                raise _chemfp_error(err, location)

        _fps.fps_knearest_search_finish(search)

        results = []
        for query_index, query_id in enumerate(query_arena.ids):
            heap = search.heaps[0][query_index]
            ids = []
            for i in range(heap.size):
                id = ctypes.string_at(heap.ids[0][i])
                id = id.decode("utf8", "surrogateescape")
                ids.append(id)
            scores = heap.scores[0][:heap.size]
            results.append(FPSSearchResult(ids, scores, query_id))
        return FPSSearchResults(query_arena.ids, results)

    finally:
        location.save(lineno=first_lineno + search.num_targets_processed)
        _fps.fps_knearest_search_free(search)

######## Tversky

######## count Tversky search #########

def count_tversky_hits_fp(query_fp, target_reader, threshold, alpha=1.0, beta=1.0):
    """Count the number of hits in *target_reader* at least *threshold* similar to the *query_fp*

    This uses Tversky similarity with the specified values of alpha and beta.
    """
    
    return count_tversky_hits_arena(_fp_to_arena(query_fp, target_reader.metadata),
                                    target_reader, threshold, alpha, beta)[0]

def count_tversky_hits_arena(query_arena, target_reader, threshold, alpha=1.0, beta=1.0):
    """Count the number of hits in *target_reader* at least *threshold* similar to the *query_fp*

    This uses Tversky similarity with the specified values of alpha and beta.
    """
    require_matching_sizes(query_arena, target_reader)
    counts = array.array("i", (0 for i in range(len(query_arena))))

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    #location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            err, num_lines = _fps.fps_count_tversky_hits(
                query_arena.metadata.num_bits,
                query_arena.start_padding, query_arena.end_padding,
                query_arena.storage_size, query_arena.arena, 0, -1,
                block, 0, -1,
                threshold, alpha, beta, counts)
            lineno += num_lines
            if err:
                lineno += 1
                raise _chemfp_error(err, location)

        return list(counts)
    finally:
        location.save(lineno=get_lineno())



######## threshold Tversky search #########

def threshold_tversky_search_fp(query_fp, target_reader, threshold, alpha=1.0, beta=1.0):
    """Find matches in the target reader which are at least threshold similar to the query fingerprint

    This uses Tversky similarity with the specified values of alpha and beta.

    :returns: an :class:`.FPSSearchResult` instance contain the result.
    """
    ids = []
    scores = []
    fp_size = len(query_fp)
    if fp_size != target_reader.metadata.num_bytes:
        raise ValueError("query fingerprint has %d bytes while target_reader uses %d bytes" % (
            fp_size, target_reader.metadata.num_bytes))
    num_bits = fp_size * 8
        
    NUM_CELLS = 1000
    cells = (TanimotoCell*NUM_CELLS)()

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _fps.fps_threshold_tversky_search(
                    num_bits, 0, 0, fp_size, query_fp, 0, -1,
                    block, start, end,
                    threshold, alpha, beta, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, location)

                for cell in itertools.islice(cells, 0, num_cells):
                        ids.append(block[cell.id_start:cell.id_end].decode("utf8"))
                        scores.append(cell.score)
                if start == end:
                    break
        return FPSSearchResult(ids, scores)
    finally:
        location.save(lineno=get_lineno())

def threshold_tversky_search_arena(query_arena, target_reader, threshold, alpha=1.0, beta=1.0):
    """Find matches in the target reader which are at least threshold similar to the query arena fingerprints

    This uses Tversky similarity with the specified values of alpha and beta.

    :returns: an :class:`.FPSSearchResults` instance containing a list of query results.
    """
    
    require_matching_sizes(query_arena, target_reader)

    if not query_arena:
        return FPSSearchResults([], [])
    results = [FPSSearchResult([], [], query_id=query_id) for query_id in query_arena.ids]
    
    # Compute at least 100 Tverskys per query, but at most 10,000 at a time
    # (That's about 200K of memory)
    NUM_CELLS = max(10000, len(query_arena) * 100)
    cells = (TanimotoCell*NUM_CELLS)()

    lineno = target_reader.location.lineno

    def get_lineno():
        return lineno
    target_reader.location.register(get_lineno=get_lineno)
    
    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _fps.fps_threshold_tversky_search(
                    query_arena.metadata.num_bits,
                    query_arena.start_padding, query_arena.end_padding,
                    query_arena.storage_size, query_arena.arena, 0, -1,
                    block, start, end,
                    threshold, alpha, beta, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, target_reader.location)

                for cell in itertools.islice(cells, 0, num_cells):
                    id = block[cell.id_start:cell.id_end]
                    id = id.decode("utf8", "surrogateescape")
                    result = results[cell.query_index]
                    result.ids.append(id)
                    result.scores.append(cell.score)

                if start == end:
                    break
    finally:
        target_reader.location.save(lineno=get_lineno())

    return FPSSearchResults(query_arena.ids, results)
            
######### k-nearest Tversky search, with threshold


# Support for peering into the chemfp_fps_heap data structure

def knearest_tversky_search_fp(query_fp, target_reader, k=3, threshold=0.0, alpha=1.0, beta=1.0):
    """Find the nearest k matches in the target reader which are at least threshold similar to the query fingerprint

    This uses Tversky similarity with the specified values of alpha and beta.

    :returns: an :class:`.FPSSearchResult` instance contain the result.
    """
    query_arena = _fp_to_arena(query_fp, target_reader.metadata)
    return knearest_tversky_search_arena(query_arena, target_reader, k, threshold, alpha, beta)[0]

def knearest_tversky_search_arena(query_arena, target_reader, k=3, threshold=0.0, alpha=1.0, beta=1.0):
    """Find the nearest k matches in the target reader which are at least threshold similar to the query arena fingerprints

    This uses Tversky similarity with the specified values of alpha and beta.

    :returns: an :class:`.FPSSearchResults` instance containing a list of query results.
    """
    require_matching_sizes(query_arena, target_reader)
    if k < 0:
        raise ValueError("k must be non-negative")

    num_queries = len(query_arena)
    search = _make_knearest_search(num_queries, k)

    _fps.fps_knearest_search_init(
        search,
        query_arena.metadata.num_bits,
        query_arena.start_padding, query_arena.end_padding,
        query_arena.storage_size, query_arena.arena, 0, -1,
        k, threshold, alpha, beta)

    location = target_reader.location
    first_lineno = location.lineno
    try:
        for block in target_reader.iter_blocks():
            err = _fps.fps_knearest_tversky_search_feed(search, block, 0, -1)
            if err:
                first_lineno += 1
                raise _chemfp_error(err, location)

        _fps.fps_knearest_search_finish(search)

        results = []
        for query_index, query_id in enumerate(query_arena.ids):
            heap = search.heaps[0][query_index]
            ids = []
            for i in range(heap.size):
                id = ctypes.string_at(heap.ids[0][i])
                id = id.decode("utf8", "surrogateescape")
                ids.append(id)
            scores = heap.scores[0][:heap.size]
            results.append(FPSSearchResult(ids, scores, query_id))
        return FPSSearchResults(query_arena.ids, results)

    finally:
        location.save(lineno=first_lineno + search.num_targets_processed)
        _fps.fps_knearest_search_free(search)



########
def _reorder_row(ids, scores, name):
    indices = list(range(len(ids)))
    if name == "decreasing-score-plus":
        indices.sort(key=lambda i: (-scores[i], ids[i]))
    elif name == "increasing-score-plus":
        indices.sort(key=lambda i: (scores[i], ids[i]))
    elif name == "decreasing-score":
        indices.sort(key=lambda i: -scores[i])
    elif name == "increasing-score":
        indices.sort(key=lambda i: scores[i])
    elif name == "decreasing-id":
        indices.sort(key=lambda i: ids[i], reverse=True)
    elif name == "increasing-id":
        indices.sort(key=lambda i: ids[i])
    elif name == "reverse":
        ids.reverse()
        scores.reverse()
        return
    elif name == "move-closest-first":
        if len(ids) <= 1:
            # Short-circuit when I don't need to do anything
            return
        x = max(scores)
        i = scores.index(x)
        ids[0], ids[i] = ids[i], ids[0]
        scores[0], scores[i] = scores[i], scores[0]
        return
    else:
        raise ValueError("Unknown sort order")

    new_ids = [ids[i] for i in indices]
    new_scores = [scores[i] for i in indices]
    ids[:] = new_ids
    scores[:] = new_scores

class FPSSearchResult:
    """Search results for a query fingerprint against a target FPS reader.

    The results contains a list of hits. Hits contain a target id and
    score. The hits can be reordered based on id or score.
    """
    def __init__(self, ids, scores, query_id=None):
        # The ids for the hits.
        self.ids = ids
        #: The similarity scores for the hits.
        self.scores = scores
        #: The id of the query fingerprint, if available, otherwise None.
        self.query_id = query_id
        
    def __len__(self):
        """Return the number of hits."""
        return len(self.ids)
    
    def __bool__(self):
        return bool(self.ids)
    
    def __iter__(self):
        """Iterate through the pairs of (target id, score) using the current ordering"""
        return zip(self.ids, self.scores)
    
    def __getitem__(self, i):
        """Return the (id, score) pair for the given index."""
        return (self.ids[i], self.scores[i])
    
    def clear(self):
        """Remove all ids and scores from this result
        
        .. deprecated:: 4.0

          This function will likely be removed in a future version of
          chemfp as it doesn't seem useful.
        """
        self.ids = []
        self.scores = []

    def get_ids(self):
        """The list of target identifiers in the current ordering.

        This returns the same list each time.
        """
        return self.ids
    
    def get_scores(self):
        """The list of target scores, in the current ordering.
        
        This returns the same list each time.
        """
        return self.scores
    
    def get_ids_and_scores(self):
        """The list of (target identifier, target score) pairs, in the current ordering"""
        return list(zip(self.ids, self.scores))
    
    def reorder(self, order="decreasing-score"):
        """Reorder the hits based on the requested ordering.
        
        The available orderings are:
          * increasing-score - sort by increasing score
          * decreasing-score - sort by decreasing score
          * increasing-score-plus - sort by increasing score, break ties by increasing index
          * decreasing-score-plus - sort by decreasing score, break ties by increasing index
          * increasing-id - sort by increasing target id
          * decreasing-id - sort by decreasing target id
          * move-closest-first - move the hit with the highest score to the first position
          * reverse - reverse the current ordering
        """
        _reorder_row(self.ids, self.scores, order)
        
    def to_pandas(self, *, columns=["target_id", "score"]):
        """Return a pandas DataFrame with the target ids and scores
        
        The first column contains the ids, the second column contains
        the ids. The default columns headers are "target_id" and
        "score". Use `columns` to specify different headers.

        :param columns: column names for the returned DataFrame
        :type columns: a list of two strings
        :returns: a pandas DataFrame
        """
        from . import pandas_support
        return pandas_support.to_pandas_from_row(self, columns)

class FPSSearchResults:
    """Search results for a query arena against a target FPS reader."""
    def __init__(self, query_ids, results):
        #: A list of query ids, one for each result.
        #: This comes from the query arena's `ids`.
        self.query_ids = query_ids
        self._results = results

    def __len__(self):
        """The number of search results in this collection"""
        return len(self._results)

    def __getitem__(self, i):
        """Return a :class:`.SearchResult` by index"""
        return self._results[i]

    def __iter__(self):
        """Iterate through the search results"""
        return iter(self._results)

    def iter_ids(self):
        """For each search result, yield the list of target identifiers"""
        for result in self._results:
            yield result.ids

    def iter_scores(self):
        """For each search result, yield the list of target scores"""
        for result in self._results:
            yield result.scores
            
    def iter_ids_and_scores(self):
        """For each search result, yield the list of target (id, score) tuples"""
        for result in self._results:
            yield list(zip(result.ids, result.scores))

    def reorder_all(self, order="decreasing-score"):
        """Reorder the hits for all of the rows based on the requested *order*.
        
        The available orderings are:

        * increasing-score - sort by increasing score
        * decreasing-score - sort by decreasing score
        * increasing-id - sort by increasing target id
        * decreasing-id - sort by decreasing target id
        * move-closest-first - move the hit with the highest score to the first position
        * reverse - reverse the current ordering
        """
        for result in self._results:
            _reorder_row(result.ids, result.scores, order)

    def clear_all(self):
        """Remove all hits from all of the search results

        .. deprecated:: 4.0

          This function will likely be removed in a future version of
          chemfp as it doesn't seem useful.
        """
        import warnings
        warnings.warn(
            "FPSSearchResults.clear_all() will be removed in a future chemfp release",
            DeprecationWarning,
            stacklevel=2)
        
        for result in self._results:
            result.clear()

    def to_pandas(self, *,
                      columns = ["query_id", "target_id", "score"],
                      empty = ("*", None),
                      ):
        """Return a pandas DataFrame with query_id, target_id and score columns.

        Each query has zero or more hits. Each hit becomes a row in the
        output table, with the query id in the first column, the hit
        target id in the second, and the hit score in the third.

        If a query has no hits then by default a row is added with the
        query id, '*' as the target id, and None as the score (which
        pandas will treat as a NA value).

        Use `empty` to specify different behavior for queries with no
        hits.  If `empty` is None then no row is added to the table. If
        `empty` is a 2-element tuple the first element is used as the
        target id and the second is used as the score.

        :param columns: column names for the returned DataFrame
        :type columns: a list of three strings
        :param empty: the target id and score used for queries with
            no hits, or None to not include a row for that case
        :type empty: a list of two strings, or None
        :returns: a pandas DataFrame
        """

        from . import pandas_support
        return pandas_support.to_pandas_from_table(self, columns, empty)
