# high-level API similar to the conversion command-line tools

# Copyright (c) 2021-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

"""This module should not be imported directly.

It contains internal implementation details of the high-level API
available from the top-level chemfp module.

This module is included in the documentation because parts of this
module are returned to the user, and are part of the public API.

"""

import time

from .. import _filename_types, Metadata
from .. import io
from .. import open_fingerprint_writer
from . import highlevel_utils

def get_source_filenames(sources):
    names = []
    seen_none = False
    for source in sources:
        if source is None:
            if seen_none:
                raise ValueError("Cannot specify 'None' source multiple times")
            seen_none = True
        elif isinstance(source, _filename_types):
            names.append(str(source))
        elif hasattr(source, "read"):
            filename = getattr(source, "name", None)
            if filename is not None:
                names.append(filename)
    return names

def get_output_filename(destination):
    if isinstance(destination, _filename_types):
        return str(destination)
    # Decided to not do this until there was a
    # stronger use case. 
    ## elif hasattr(destination, "write"):
    ##     filename = getattr(destination, "name", None)
    ##     if filename is not None:
    ##         return filename
    return None

class ConversionInfo:
    """Store information about the conversion to FPS or FPB file.

    An instance of this object is returned by the
    :func:`.rdkit2fps`, :func:`.oe2fps`, :func:`.ob2fps`, :func:`.cdk2fps`,
    :func:`.convert2fps`, and :func:`.sdf2fps` functions.
    """
    def __init__(
            self,
            description,
            source_locations,
            destination_location,
            output_filename,
            output_format,
            times,
            ):
        assert isinstance(description, str), description
        assert isinstance(times, dict), times
        
        self.description = description
        self.source_locations = source_locations
        self.destination_location = destination_location
        self.output_filename = output_filename
        self.output_format = output_format
        self.times = times
        self._num_input_records = -1

    @property
    def num_input_records(self):
        "the number of input records read"
        if self._num_input_records != -1:
            return self._num_input_records
        num_input_records = 0
        for location in self.source_locations:
            recno = location.recno
            assert recno is not None, location
            num_input_records += recno
        self._num_input_records = num_input_records
        return num_input_records
        
    @property
    def num_output_records(self):
        "the number of output records written"
        return self.destination_location.recno

    def get_description(self):
        "Return a human-readable description of the conversion run"
        if self.destination_location is None:
            # overwrite = False and there was a file
            return self.description

        num_input_records = self.num_input_records
        num_output_records = self.num_output_records
        
        msg = (
            f"{self.description}. "
            f"#input_records={num_input_records}, "
            f"#output_records={num_output_records}"
            )
        total = highlevel_utils.format_time(self.times["total"])
        return f"{msg} (total: {total})"
                
    def __repr__(self):
        return f"ConversionInfo({self.get_description()!r})"
    
    def __str__(self):
        return self.get_description()

    def open_output(self, *, allow_mmap = True):
        """Open the newly created output fingerprint file for reading.

        This is the same as calling `chemfp.open` using the output
        filename and format, along with the optional `allow_mmap`.

        :param allow_mmap: if True, use mmap to open uncompressed FPB files, otherwise read the contents
        :type allow_mmap: boolean
        :returns: a :class:`chemfp.fps_io.FPSReader` or :class:`chemfp.arena.FingerprintArena`
        """

        if self.output_filename is None:
            raise ValueError("Cannot open fingerprints: No output filename available from the conversion")
        from .. import open as open_fingerprints
        return open_fingerprints(
            self.output_filename,
            format = self.output_format,
            allow_mmap = allow_mmap,
            )
    
    def load_output(self, reorder=True, allow_mmap=True, *, progress = True):
        """Load the fingerprints which were written to the created output file.

        This is the same as calling :func:`chemfp.load_fingerprints` using the
        output filename and format. The `reorder`, `allow_mmap` and
        `progress` values are passed down as well.

        :param reorder: Specify if fingerprints should be reordered for better performance
        :type reorder: True or False
        :param allow_mmap: Allow chemfp to use mmap on FPB files, instead of reading the file's contents into memory
        :type allow_mmap: True or False
        :param progress: Enable or disable progress bars, optionally specifying the progress bar constructor
        :type progress: True, False, or a callable
        :returns: :class:`chemfp.arena.FingerprintArena`
        """

        if self.output_filename is None:
            raise ValueError("Cannot load fingerprints: No output filename available from the conversion")
        from .. import load_fingerprints
        return load_fingerprints(
            self.output_filename,
            reorder = reorder,
            format = self.output_format,
            allow_mmap = allow_mmap,
            progress = progress,
            )
    
def convert2fps(
        *,
        source,
        destination,
        fptype,
        input_format,
        output_format,
        reader_args,
        id_tag,
        errors,
        fingerprint_kwargs,
        id_prefix,
        id_template,
        id_cleanup,
        overwrite,
        reorder,
        tmpdir,
        max_spool_size,
        progress,
        ):
    start_time = time.time()
    overwrite_msg = highlevel_utils.get_overwrite_msg(overwrite, destination)
    if overwrite_msg is not None:
        times = {"total": time.time() - start_time}
        return ConversionInfo(
            description = overwrite_msg,
            source_locations = None,
            destination_location = None,
            output_filename = get_output_filename(destination),
            output_format = output_format,
            times = times,
            )
        
    if source is None:
        sources = [source]
        description = "Converted structures from stdin"
    elif isinstance(source, _filename_types):
        sources = [source]
        description = f"Converted structures from {source!r}"
    elif hasattr(source, "read"):
        sources = [source]
        description = "Converted structures from a file object"
    else:
        # Pass in a list
        sources = source
        n = len(sources)
        description = f"Converted structures from a list of sources of length {n}"

    source_filenames = get_source_filenames(sources)
    
    del source
    num_sources = len(sources)

    error_handler = io.get_parse_error_handler(errors)
    
    id_processor = highlevel_utils.get_id_processor(
        id_prefix = id_prefix,
        id_template = id_template,
        id_cleanup = id_cleanup,
        id_tag = id_tag,
        errors = error_handler,
        )
    
    source_locations = []

    # Defer making the writer until the first source is open.
    # This prevents making an empty output file 
    writer = None
    writer_location = None
    try:
        for source_i, source in enumerate(sources):
            with fptype.read_molecule_fingerprints(
                    source,
                    format = input_format,
                    id_tag = id_tag,
                    errors = errors,
                    reader_args = reader_args) as reader:

                location = reader.location
                source_locations.append(location)

                with location.progress_bar(
                        progress = progress,
                        file_count = (source_i, num_sources),
                        ) as progress_bar:
                    if writer is None:
                        writer = open_fingerprint_writer(
                            destination,
                            format = output_format,
                            reorder = reorder,
                            metadata = fptype.get_metadata(
                                sources = source_filenames,
                                ),
                            )
                        writer_location = writer.location

                    writer.write_fingerprints(
                        progress_bar(
                            id_processor(reader, location)
                            )
                        )
        if writer is None:
            # This can only be reached if passes an empty list of sources.
            # Create an empty file.
            writer = open_fingerprint_writer(
                destination,
                format = output_format,
                metadata = fptype.get_metadata(),
                )
            writer_location = writer.location

    finally:
        if writer is not None:
            writer.close()

    end_time = time.time()
    return ConversionInfo(
        description = description,
        source_locations = source_locations,
        destination_location = writer_location,
        output_filename = get_output_filename(destination),
        output_format = output_format,
        times = {"total": end_time - start_time},
        )

class _abort_handler:
    @staticmethod
    def error(errmsg, location=None):
        handler = io.get_parse_error_handler("strict")
        handler.error(
            f"Each of the first 100 records contained errors. Final error: {errmsg}",
            location,
            )

def sdf2fps(
        source,
        destination,
        id_tag,
        fp_tag,
        input_format,
        output_format,
        metadata,
        decoder,
        pubchem,
        errors,
        id_prefix,
        id_template,
        id_cleanup,
        overwrite,
        reorder,
        tmpdir,
        max_spool_size,
        progress,
        ):
    from .. import text_toolkit
    start_time = time.time()
    overwrite_msg = highlevel_utils.get_overwrite_msg(overwrite, destination)
    if overwrite_msg is not None:
        times = {"total": time.time() - start_time}
        return ConversionInfo(
            description = overwrite_msg,
            source_locations = None,
            destination_location = None,
            output_filename = get_output_filename(destination),
            output_format = output_format,
            times = times,
            )


    if input_format is None:
        compression = None
    elif input_format == "sdf":
        compression = ""
    elif input_format == "sdf.gz":
        compression = "gz"
    elif input_format == "sdf.zst":
        compression = "zst"
    else:
        raise ValueError(f"Unsupported input_format: {input_format!r}. "
                             "Must be one of 'sdf', 'sdf.gz', or 'sdf.zst'")

    if source is None:
        sources = [source]
        description = "Converted SDF records from stdin"
    elif isinstance(source, _filename_types):
        sources = [source]
        description = f"Converted SDF records from {source!r}"
    elif hasattr(source, "read"):
        sources = [source]
        description = "Converted SDF records from a file object"
    else:
        # Pass in a list
        sources = source
        n = len(sources)
        description = f"Converted SDF records from a list of {n} sources"

    source_filenames = get_source_filenames(sources)
    del source
    
    if pubchem:
        if fp_tag is None:
            fp_tag = "PUBCHEM_CACTVS_SUBSKEYS"
            
        if decoder is None:
            decoder = "cactvs"
            
        if metadata is None:
            metadata = Metadata(
                software = "CACTVS/unknown",
                type = "CACTVS-E_SCREEN/1.0 extended=2",
                num_bits = 881,
                sources = source_filenames,
                )
        else:
            # It's probably None, but also override any wrong number of bits
            if metadata.num_bits != 881:
                metadata = metadata.copy(num_bits=881, num_bytes=111)

    if metadata is None:
        # Set a default (and mostly empty) metadata appropriately.
        # Don't have a num_bits yet
        metadata = Metadata(
            sources = source_filenames,
            )

    num_bits = metadata.num_bits
    if num_bits is None:
        num_bytes = None
    else:
        num_bytes = (num_bits + 7) // 8
    
    # Figure out how to decode the fingerprint tags
    if decoder is None:
        # Default uses hex-decoding
        from .. import encodings
        decoder_f = encodings.from_hex
        decoder_name = "hex "
        
    elif isinstance(decoder, str):
        # Specify a decoder name
        from .. import encodings
        decoder_f = encodings._decoder_table.get(decoder, None)
        decoder_name = f"{decoder} "
        if decoder_f is None:
            # TODO: need a better API for this
            names = [repr(name) for name in encodings._decoder_table if name != "decoder"]
            names.sort()
            s = ", ".join(names)
            raise ValueError(f"Unknown decoder {decoder!r}. Available values are {s} or a callable")
        
    elif callable(decoder):
        # Can pass in your own function
        decoder_f = decoder
        decoder_name = ""
    else:
        raise TypeError("decoder must be None, a string, or a callable")
        
    decoder = decoder_f
    del decoder_f

    # given the (id, fp) pairs, yield (id, decoded fp) pairs
    def get_id_and_decoded_fp_iter(id_fp_iter, decoder, location, error_handler):
        nonlocal num_bits, num_bytes, metadata

        LEN = len # slightly faster inner loop by switching to a local lookup
        err_count = 0
        def incr_error():
            nonlocal error_handler, err_count
            if recno == err_count == 99:
                error_handler = _abort_handler
            err_count += 1
            
        for recno, (id, encoded_fp) in enumerate(id_fp_iter):
            if not encoded_fp:
                incr_error()
                if fp_tag is None:
                    msg = f"Missing fingerprint for id {id!r}"
                else:
                    msg = f"Missing fingerprint tag {fp_tag!r} for id {id!r}"
                error_handler.error(msg, location)
                continue
            
            try:
                decoded_num_bits, decoded_fp = decoder(encoded_fp)
            except ValueError as err:
                incr_error()
                if fp_tag is None:
                    msg = (
                        f"Cannot {decoder_name}decode the title for id {id!r} as a fingerprint: {err}"
                        )
                else:
                    msg = f"Cannot {decoder_name}decode fingerprint tag {fp_tag!r} for id {id!r}: {err}"
                error_handler.error(msg, location)
                continue

            decoded_num_bytes = LEN(decoded_fp)
            if num_bytes is None:
                # This only happens when we've read the first fingerprint and don't
                # know what the fingerprint size should be.
                # We can use the first fingerprint to figure out the size.
                num_bytes = decoded_num_bytes

                if decoded_num_bits is None:
                    # The number of bits is based on the byte size
                    num_bits = num_bytes * 8
                else:
                    # The decoder returned an exact number of bits
                    num_bits = decoded_num_bits
                
                # Update the metadata if it previously was None
                if num_bits != metadata.num_bits:
                    metadata = metadata.copy(num_bits=num_bits)
            else:
                # We know what the fingerprint size should be.
                if decoded_num_bits is None:
                    # Can only compare the fingerprint byte lengths
                    if decoded_num_bytes != num_bytes:
                        error_handler.error(
                            f"Fingerprint for id {id!r} has {8*decoded_num_bytes} bits, expected {8*num_bytes}",
                            location)
                        continue
                elif decoded_num_bits != num_bits:
                    error_handler.error(
                        f"Fingerprint for id {id!r} has {decoded_num_bits} bits, expected {num_bits}",
                        location)
                    continue
            
            yield id, decoded_fp

    num_sources = len(sources)
    
    source_locations = []

    # Defer making the writer until the first source is open.
    # This prevents making an empty output file if there are no sources
    # and ensures that the num_bits is set correctly from the first source.
    writer = None
    writer_location = None
    
    error_handler = io.get_parse_error_handler(errors)
    
    id_processor = highlevel_utils.get_id_processor(
        id_prefix = id_prefix,
        id_template = id_template,
        id_cleanup = id_cleanup,
        id_tag = id_tag,
        errors = error_handler,
        )
    
    try:
        for source_i, source in enumerate(sources):
            with text_toolkit.read_sdf_ids_and_values(
                    source, id_tag, fp_tag,
                    compression=compression, errors=error_handler) as reader:

                location = reader.location
                source_locations.append(location)

                # The progress bar needs to start here so the reader location is correct
                # even during the initial read of the initial source.
                with location.progress_bar(
                        progress = progress,
                        file_count=(source_i, num_sources),
                        ) as progress_bar:

                    clean_id_fp_iter = id_processor(reader, location)
                    id_fp_iter = get_id_and_decoded_fp_iter(clean_id_fp_iter, decoder, location, error_handler)

                    
                    if writer is None:
                        try:
                            # set 'num_bits' if not yet known
                            id, fp = next(id_fp_iter)
                        except StopIteration:
                            # There's no content.
                            id = fp = None

                        writer = open_fingerprint_writer(
                            destination,
                            format = output_format,
                            metadata = metadata,
                            reorder = reorder,
                            tmpdir = tmpdir,
                            max_spool_size = max_spool_size,
                            )
                        writer_location = writer.location
                        if id is not None:
                            writer.write_fingerprint(id, fp)
                
                    writer.write_fingerprints(progress_bar(id_fp_iter))

        if writer is None:
            # This can only be reached if passes an empty list of sources.
            # Create an empty file.
            writer = open_fingerprint_writer(
                destination,
                format = output_format,
                metadata = metadata,
                )
    finally:
        if writer is not None:
            writer.close()
            
    return ConversionInfo(
        description = description,
        source_locations = source_locations,
        destination_location = writer_location,
        output_filename = get_output_filename(destination),
        output_format = output_format,
        times = {"total": time.time() - start_time},
        )
