# high-level API similar to the diversity command-line tools

# Copyright (c) 2021-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## maxmin

"""This module should not be imported directly.

It contains internal implementation details of the high-level API
available from the top-level chemfp module.

This module is included in the documentation because parts of this
module are returned to the user, and are part of the public API.

"""

import time
import random

from .. import check_metadata_problems
from .. import diversity
from .. import _tqdm

from .highlevel_utils import (
    format_time,
    get_fptype,
    highlevel_load,
    report_problems,
    record_time,
    update_docstrings,
    )

def check_num_picks(num_picks):
    if num_picks is None:
        return
    if isinstance(num_picks, str) and num_picks == "all":
        return
    if isinstance(num_picks, int) and num_picks >= 0:
        return
    raise ValueError("num_picks must be a non-negative integer or 'all'")

def get_num_picks(num_picks, max_num_picks):
    if num_picks is None or num_picks == "all":
        return max_num_picks
    return min(num_picks, max_num_picks)

def get_threshold(threshold):
    if not (0.0 <= threshold <= 1.0):
        raise ValueError("threshold must be between 0.0 and 1.0, inclusive")
    return threshold

def get_seed(seed):
    err = False
    if seed is None:
        pass
    elif not isinstance(seed, int):
        err = True
    elif seed == -1:
        seed = random.Random().randrange(2**64)
    elif not (0 <= seed < 2**64):
        err = True

    if err:
        raise ValueError(
            "seed must be None, -1, or a value between 0 and 2**64-1, inclusive"
            )
    return seed

def get_candidate_indices_by_ids(arena, ids):
    indices = []
    for id in ids:
        idx = arena.get_index_by_id(id)
        if idx is None:
            raise ValueError(f"Cannot find id {id!r} in the candidates")
        indices.append(idx)
    return indices

class _CloseArenas:
    def __init__(self, candidates_close, references_close=None):
        self._candidates_close = candidates_close
        self._references_close = references_close
        
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        close = self._candidates_close
        if close is not None:
            self._candidates_close = None
            try:
                close()
            except Exception:
                pass
        
        close = self._references_close
        if close is not None:
            self._references_close = None
            try:
                close()
            except Exception:
                pass

class BaseMaxMinSearch(_CloseArenas):
    """This is the base class for objects returned by :func:`chemfp.maxmin`

    It contains the query parameters, search results, and timings.

    In addition, it is a context manager for any files which may have
    been opened.
    """
    def __init__(self, candidates, references, seed, num_picked, threshold, times,
                     picker, result, candidates_close, references_close):
        self.candidates = candidates
        self.references = references
        self.seed = seed
        self.num_picked = num_picked
        self.threshold = threshold
        self.times = times
        self.picker = picker
        self.result = result
        _CloseArenas.__init__(self, candidates_close, references_close)

    def get_description(self):
        "Return a human-readable description of the simsearch run"
        msg = f"picked {self.num_picked} fps. similarity <= {self.threshold}, #candidates={len(self.candidates)}"
        if self.references is not None:
            msg += f", #references={len(self.references)}"
        if self.seed is not None:
            msg += f", seed={self.seed}"
            
        pick_str = format_time(self.times["pick"])
        total_str = format_time(self.times["total"])
        msg += f" (pick: {pick_str}, total: {total_str})"
        return msg

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({self.get_description()!r}, picker={self.picker}, result={self.result})"
    
    def __str__(self):
        return self.get_description()

    @property
    def picks(self):
        "Shortcut for `self.picker.picks`. See :meth:`.MaxMinPicker.pick`."
        return self.picker.picks

class MaxMinSearch(BaseMaxMinSearch):
    """maxmin() returns an instance of this type when `include_scores=False`"""
    def __len__(self):
        "Shortcut for `len(maxmin_search.result)`. See :meth:`.Picks.__len__`."
        return len(self.result)
    
    def __getitem__(self, i):
        "Shortcut for `maxmin_search.result[i]`. See :meth:`.Picks.__getitem__`."
        return self.result[i]
    
    def __iter__(self):
        "Shortcut for `iter(maxmin_search.result)`. See :meth:`.Picks.__iter__`."
        return iter(self.result)
    
    def get_indices(self):
        return self.result.get_indices()
    
    def get_ids(self):
        return self.result.get_ids()
    
    def to_pandas(self, *, column="pick_id"):
        return self.result.to_pandas(column=column)
    
    def as_numpy(self):
        return self.result.as_numpy()
    
    def as_ctypes(self):
        return self.result.as_ctypes()

update_docstrings(MaxMinSearch, diversity.Picks, "maxmin_search")

class MaxMinScoreSearch(BaseMaxMinSearch):
    """maxmin() returns an instance of this type when `include_scores=True`"""
    def __getitem__(self, i):
        "Shortcut for `len(maxmin_search.result)`. See :meth:`.PicksAndScores.__len__`."
        return self.result[i]
    
    def __len__(self):
        "Shortcut for `maxmin_search.result[i]`. See :meth:`.PicksAndScores.__getitem__`."
        return len(self.result)
    
    def __iter__(self):
        "Shortcut for `iter(maxmin_search.result)`. See :meth:`.PicksAndScores.__iter__`."
        return iter(self.result)
    
    def get_ids(self):
        return self.result.get_ids()

    def get_scores(self):
        return self.result.get_scores()

    def get_ids_and_scores(self):
        return self.result.get_ids_and_scores()

    def get_indices(self):
        return self.result.get_indices()
    
    def get_indices_and_scores(self):
        return self.result.get_indices_and_scores()
    
    def to_pandas(self, columns=["pick_id", "score"]):
        return self.result.to_pandas(columns=columns)
    
    def as_numpy(self):
        return self.result.as_numpy()
    
    def as_ctypes(self):
        return self.result.as_ctypes()

update_docstrings(MaxMinScoreSearch, diversity.PicksAndScores, "maxmin_search")
    
def progress_pick_n(progress, picker, num_picks, threshold):
    if num_picks == 0:
        return picker.picks
    
    progress_bar = _tqdm.get_tqdm_constructor(progress)(
        desc = "picks",
        unit = "",
        dynamic_ncols = True,
        total = num_picks)

    with progress_bar:
        num_picked = 0
        for i in range(1, num_picks+1):
            picked = picker.pick_n(1)
            if not picked:
                break
            num_picked += 1
            progress_bar.update(1)

        # Can't implement all_equal as there are no scores

    return picker.picks
    
def progress_pick_n_with_scores(progress, picker, num_picks, threshold, all_equal):
    if num_picks == 0:
        # I need a PicksAndScores so pick 0 elements.
        return picker.pick_n_with_scores(0, threshold)

    progress_bar = _tqdm.get_tqdm_constructor(progress)(
        desc = "picks",
        unit = "",
        dynamic_ncols = True,
        total = num_picks)

    with progress_bar:
        result = None
        for expected_len in range(1, num_picks+1):
            result = picker.pick_n_with_scores(1, threshold, result)
            if len(result) != expected_len:
                break
            progress_bar.update(1)
        else:
            # Picked num_picks. Should I pick more?
            if all_equal:
                # Keep picking items with the same threshold.
                threshold = result.as_ctypes()[-1].score
                expected_len = len(result) + 1
                while 1:
                    result = picker.pick_n_with_scores(1, threshold, result)
                    if len(result) != expected_len:
                        break
                    progress_bar.update(1) # will now be > num_picks
                    expected_len += 1

    return result

    
def maxmin(
        candidates,
        references,
        initial_pick,
        candidates_format,
        references_format,
        num_picks,
        threshold,
        all_equal,
        randomize,
        seed,
        include_scores,
        progress,
        ):
    times = {}
    start_time = time.time()

    check_num_picks(num_picks)
    threshold = get_threshold(threshold)
    seed = get_seed(seed)

    candidates_close = references_close = None
    try:
        with record_time("load_candidates", times):
            candidates, candidates_close = highlevel_load(
                candidates, candidates_format, progress, "'candidates'")

        if references is None:
            if initial_pick is None:
                # pick the first point using heapsweep
                pass
            elif isinstance(initial_pick, int):
                # treat as an index
                try:
                    # normalize negatives, and do range checking
                    initial_pick = range(len(candidates))[initial_pick]
                except IndexError:
                    raise ValueError("initial_pick index is out of range")
            else:
                initial_pick = candidates.get_index_by_id(initial_pick)
                if initial_pick is None:
                    raise ValueError("initial_pick id not found in candidates")
        else:
            if initial_pick is not None:
                raise ValueError("Cannot specify both 'references' and 'initial_pick'")
            
        num_picks = get_num_picks(num_picks, len(candidates))

        if references is not None:
            with record_time("load_references", times):
                references_arena, references_close = highlevel_load(
                    references, references_format, progress, "'references'")

            problems = check_metadata_problems(candidates.metadata, references_arena.metadata,
                                                   "candidates", "references")
            report_problems(problems)


            picker = diversity.MaxMinPicker.from_candidates_and_references(
                candidates, references_arena, randomize=randomize, seed=seed)
        elif initial_pick is None:
            picker = diversity.MaxMinPicker.from_candidates(
                candidates, randomize=randomize, seed=seed)
        else:
            picker = diversity.MaxMinPicker.from_candidates_and_initial_pick(
                candidates, initial_pick, randomize=randomize, seed=seed)

        with record_time("pick", times):
            if include_scores:
                result = progress_pick_n_with_scores(
                    progress, picker, num_picks, threshold, all_equal)
                klass = MaxMinScoreSearch
                                                         
            elif all_equal:
                progress_pick_n_with_scores(
                    progress, picker, num_picks, threshold, all_equal)
                                                
                result = picker.picks
                klass = MaxMinSearch
            else:
                result = progress_pick_n(progress, picker, num_picks, threshold)
                klass = MaxMinSearch
    
    except:
        if candidates_close is not None:
            candidates_close()
        if references_close is not None:
            references_close()
        raise
    
    times["total"] = time.time() - start_time
    
    return klass(
        candidates = candidates,
        references = references,
        seed = seed,
        num_picked = len(result),
        threshold = threshold,
        times = times,
        picker = picker,
        result = result,
        candidates_close = candidates_close,
        references_close = references_close,
        )

## heapsweep

class BaseHeapSweepSearch(_CloseArenas):
    """This is the base class for objects returned by :func:`chemfp.heapsweep`

    It contains the query parameters, search results, and timings.

    In addition, it is a context manager for any files which may have
    been opened.
    """
    def __init__(self, candidates, seed, num_picked, threshold, times,
                     picker, result, candidates_close):
        self.candidates = candidates
        self.seed = seed
        self.num_picked = num_picked
        self.threshold = threshold
        self.times = times
        self.picker = picker
        self.result = result
        _CloseArenas.__init__(self, candidates_close, None)
        
    def get_description(self):
        "Return a human-readable description of the simsearch run"
        msg = f"picked {self.num_picked} fps. similarity <= {self.threshold}, #candidates={len(self.candidates)}"
        if self.seed is not None:
            msg += f", seed={self.seed}"
        
        pick_str = format_time(self.times["pick"])
        total_str = format_time(self.times["total"])
        msg += f" (pick: {pick_str}, total: {total_str})"
        return msg

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({self.get_description()!r}, picker={self.picker}, result={self.result})"

    def __str__(self):
        return self.get_description()

    @property
    def picks(self):
        "Shortcut for `self.picker.picks`. See :meth:`.MaxMinPicker.pick`."
        return self.picker.picks


class HeapSweepSearch(BaseHeapSweepSearch):
    """heapsweep() returns an instance of this class when `include_scores=False`"""
    def __len__(self):
        "Shortcut for `len(maxmin_search.result)`. See :meth:`.Picks.__len__`."
        return len(self.result)
    
    def __getitem__(self, i):
        "Shortcut for `maxmin_search.result[i]`. See :meth:`.Picks.__getitem__`."
        return self.result[i]
    
    def __iter__(self):
        "Shortcut for `iter(maxmin_search.result)`. See :meth:`.Picks.__iter__`."
        return iter(self.result)
    
    def get_indices(self):
        return self.result.get_indices()
    
    def get_ids(self):
        return self.result.get_ids()
    
    def to_pandas(self, *, column="pick_id"):
        return self.result.to_pandas(column=column)
    
    def as_numpy(self):
        return self.result.as_numpy()
    
    def as_ctypes(self):
        return self.result.as_ctypes()

update_docstrings(HeapSweepSearch, diversity.Picks, "heapsweep_search")
    
class HeapSweepScoreSearch(BaseHeapSweepSearch):
    """heapsweep() returns an instance of this class when `include_scores=True`"""
    def __getitem__(self, i):
        "Shortcut for `len(heapsweep_search.result)`. See :meth:`.PicksAndScores.__len__`."
        return self.result[i]
    
    def __len__(self):
        "Shortcut for `heapsweep_search.result[i]`. See :meth:`.PicksAndScores.__getitem__`."
        return len(self.result)
    
    def __iter__(self):
        "Shortcut for `iter(heapsweep_search.result)`. See :meth:`.PicksAndScores.__iter__`."
        return iter(self.result)
    
    def get_ids(self):
        return self.result.get_ids()

    def get_scores(self):
        return self.result.get_scores()

    def get_ids_and_scores(self):
        return self.result.get_ids_and_scores()

    def get_indices(self):
        return self.result.get_indices()
    
    def get_indices_and_scores(self):
        return self.result.get_indices_and_scores()
    
    def to_pandas(self, columns=["pick_id", "score"]):
        return self.result.to_pandas(columns=columns)
    
    def as_numpy(self):
        return self.result.as_numpy()
    
    def as_ctypes(self):
        return self.result.as_ctypes()

update_docstrings(HeapSweepScoreSearch, diversity.PicksAndScores, "heapsweep_search")


def heapsweep(
        candidates,
        candidates_format,
        num_picks,
        threshold,
        all_equal,
        randomize,
        seed,
        include_scores,
        progress,
        ):
    times = {}
    start_time = time.time()

    check_num_picks(num_picks)
    
    with record_time("load_candidates", times):
        candidates, candidates_close = highlevel_load(
            candidates, candidates_format, progress, "'candidates'")

    try:
        num_picks = get_num_picks(num_picks, len(candidates))

        picker = diversity.HeapSweepPicker.from_candidates(
            candidates, randomize=randomize, seed=seed)

        with record_time("pick", times):
            if include_scores:
                result = progress_pick_n_with_scores(
                    progress, picker, num_picks, threshold, all_equal)
                klass = HeapSweepScoreSearch
                                                        
            elif all_equal:
                progress_pick_n_with_scores(
                    progress, picker, num_picks, threshold, all_equal)
                result = picker.picks
                klass = HeapSweepSearch
            else:
                result = progress_pick_n(progress, picker, num_picks, threshold)
                klass = HeapSweepSearch
                                             
    except:
        if candidates_close is not None:
            candidates_close()
        raise
    
    times["total"] = time.time() - start_time

    return klass(
        candidates = candidates,
        seed = seed,
        num_picked = len(result),
        threshold = threshold,
        times = times,
        picker = picker,
        result = result,
        candidates_close = candidates_close,
        )

        
## spherex

class BaseSpherexSearch(_CloseArenas):
    """This is the base class for objects returned by :func:`chemfp.spherex`

    It contains the query parameters, search results, and timings.

    In addition, it is a context manager for any files which may have
    been opened.
    """
    def __init__(
            self,
            candidates,
            references,
            num_initial_picks,
            threshold,
            seed,
            direction_msg,
            picker,
            num_picked,
            result,
            times,
            candidates_close,
            references_close,
            ):
        self.candidates = candidates
        self.references = references
        self.num_initial_picks = num_initial_picks
        self.threshold = threshold
        self.seed = seed
        self.direction_msg = direction_msg
        self.picker = picker
        self.num_picked = num_picked
        self.result = result
        self.times = times
        _CloseArenas.__init__(self, candidates_close, references_close)
    
    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({self.get_description()!r}, picker={self.picker}, result={self.result})"
    
    def __str__(self):
        return self.get_description()

    def get_description(self):
        "Return a human-readable description of the simsearch run"
        msg = f"picked {self.num_picked}. "
        msg += f"similarity <= {self.threshold}, #candidates={len(self.candidates)}"
        
        if self.num_initial_picks:
            msg += f"#initial_picks={self.num_initial_picks}"
            
        if self.references is not None:
            msg += f", #references={len(self.references)}"

        if self.seed is None:
            if self.direction_msg is None:
                msg += ", seed=None (directed by popcount)"
            else:
                msg += f", seed=None (directed by {self.direction_msg})"
        else:
            if self.direction_msg is None:
                msg += f", seed={self.seed} (random pick)"
            else:
                msg += f", seed={self.seed} (directed by {self.direction_msg})"
        
        pick_str = format_time(self.times["pick"])
        total_str = format_time(self.times["total"])
        msg += f" (pick: {pick_str}, total: {total_str})"
        return msg

class SpherexSearch(BaseSpherexSearch):
    """spherex() returns an instance of this class when `include_counts` and `include_neighbors` are False"""
    def __len__(self):
        "Shortcut for `len(maxmin_search.result)`. See :meth:`.Picks.__len__`."
        return len(self.result)
    
    def __getitem__(self, i):
        "Shortcut for `maxmin_search.result[i]`. See :meth:`.Picks.__getitem__`."
        return self.result[i]
    
    def __iter__(self):
        "Shortcut for `iter(maxmin_search.result)`. See :meth:`.Picks.__iter__`."
        return iter(self.result)
    
    def get_indices(self):
        return self.result.get_indices()
    
    def get_ids(self):
        return self.result.get_ids()
    
    def to_pandas(self, *, column="pick_id"):
        return self.result.to_pandas(column=column)
    
    def as_numpy(self):
        return self.result.as_numpy()
    
    def as_ctypes(self):
        return self.result.as_ctypes()

update_docstrings(SpherexSearch, diversity.Picks, "spherex_search")

class SpherexCountSearch(BaseSpherexSearch):
    """spherex() returns an instance of this class when `include_counts` is True"""
    def __getitem__(self, i):
        "Shortcut for `len(heapsweep_search.result)`. See :meth:`.PicksAndCounts.__len__`."
        return self.result[i]
    
    def __len__(self):
        "Shortcut for `heapsweep_search.result[i]`. See :meth:`.PicksAndCounts.__getitem__`."
        return len(self.result)
    
    def __iter__(self):
        "Shortcut for `iter(heapsweep_search.result)`. See :meth:`.PicksAndCounts.__iter__`."
        return iter(self.result)
    
    def get_ids(self):
        return self.result.get_ids()

    def get_counts(self):
        return self.result.get_counts()

    def get_ids_and_counts(self):
        return self.result.get_ids_and_counts()

    def get_indices(self):
        return self.result.get_indices()
    
    def get_indices_and_counts(self):
        return self.result.get_indices_and_counts()
    
    def to_pandas(self, columns=["pick_id", "count"]):
        return self.result.to_pandas(columns=columns)
    

update_docstrings(SpherexCountSearch, diversity.PicksAndCounts, "spherex_search")

class SpherexNeighborSearch(BaseSpherexSearch):
    """spherex() returns an instance of this class when `include_neighbors` is True"""
    def __iter__(self):
        "Shortcut for `iter(heapsweep_search.result)`. See :meth:`.PicksAndNeighbors.__iter__`."
        return iter(self.result)

    def __getitem__(self, i):
        "Shortcut for `heapsweep_search.result[i]`. See :meth:`.PicksAndNeighbors.__getitem__`."
        return self.result[i]

    def get_all_neighbors(self):
        return self.result.get_all_neighbors()
        
    def get_indices_and_neighbors(self):
        return self.result.get_indices_and_neighbors()
    
    def get_ids_and_hits(self):
        return self.result.get_ids_and_hits()

    def get_counts(self):
        return self.result.get_counts()
        
    def get_ids_and_counts(self):
        return self.result.get_ids_and_counts()
        
    def get_indices_and_counts(self):
        return self.result.get_indices_and_counts()
        
    def to_pandas(
            self,
            *,
            columns = ["pick_id", "neighbor_id", "score"],
            empty = ("*", None)):
        return self.result.to_pandas(columns=columns, empty=empty)
        

update_docstrings(SpherexNeighborSearch, diversity.PicksAndNeighbors, "spherex_search")

def spherex(
        candidates,
        references,
        initial_picks,
        
        candidates_format,
        references_format,
        
        num_picks,
        ranks,
        dise,
        dise_type,
        dise_references,
        dise_references_format,

        randomize,
        seed,
        
        threshold,
        include_counts,
        include_neighbors,
        progress,
        ):
    
    start_time = time.time()
    times = {}

    if include_counts and include_neighbors:
        raise ValueError("Cannot specify both include_counts and include_neighbors")

    num_directions = (
        bool(ranks is not None) +
        bool(dise) +
        bool(dise_references is not None)
        )
    if num_directions > 1:
        raise ValueError("Cannot only specify at most one of ranks, dise, or dise_references")

    pick_type = None
    if references is None:
        if initial_picks is None:
            pass
        else:
            # must be an id or an index, or a list there-of
            if isinstance(initial_picks, int):
                initial_picks = [initial_picks]
                pick_type = int
            elif isinstance(initial_picks, str):
                initial_picks = [initial_picks]
                pick_type = str
            elif not initial_picks:
                # an empty list is just fine
                initial_picks = None
            else:
                x = initial_picks[0]
                if isinstance(x, int):
                    pick_type = int
                else:
                    pick_type = int
                if not all(isinstance(x, pick_type) for x in initial_picks[1:]):
                    raise ValueError(
                        "initial_picks values must be all integers or all strings"
                        )
    else:
        if initial_picks is None:
            pass
        else:
            raise ValueError("Cannot currently specify both references and initial_picks")

    check_num_picks(num_picks)
    threshold = get_threshold(threshold)
    seed = get_seed(seed)

    candidates_close = references_close = None
    with record_time("load_candidates", times):
        candidates, candidates_close = highlevel_load(
            candidates, candidates_format, progress, "'candidates'")
        
    num_picks = get_num_picks(num_picks, len(candidates))

    try:
        if num_picks is None:
            num_picks = len(candidates)
        else:
            num_picks = min(num_picks, len(candidates))

        direction_msg = None
        if ranks is not None:
            if len(ranks) != len(candidates):
                raise ValueError(
                    f"The number of ranks ({len(ranks)}) must equal "
                    f"the number of candidates ({len(candidates)})"
                    )
            direction_msg = "input ranks"
        elif (dise) or (dise_references is not None):
            dise_arena_close = None
            try:
                if dise:
                    # Use DISE given a fingerprint type, either from
                    # the candidates or specified by the user.
                    with record_time("get_fptype", times):
                        if dise_type is None:
                            dise_type = candidates.get_fingerprint_type()
                        else:
                            dise_type = get_fptype(dise_type)
                    # Use the fingerprint type to rank based on the SMILES
                    # in the DISE paper.
                    with record_time("generate_fps", times):
                        ranker = diversity.DISERanker.from_dise_paper(dise_type)
                    direction_msg = "DISE paper references"
                else:
                    # Rank based on the specified arena
                    with record_time("load_dise_references", times):
                        dise_arena, dise_arena_close = highlevel_load(
                            dise_references, dise_references_format, progress, "'dise_references'")
                    ranker = diversity.DISERanker(dise_arena)
                    direction_msg = f"{len(dise_arena)} reference fps"

                problems = check_metadata_problems(
                    candidates.metadata,
                    ranker.dise_arena.metadata,
                    "candidate fingerprints", "DISE fingerprints")
                report_problems(problems)

                # Rank the candidates
                with record_time("dise_rank", times):
                    ranks = ranker.rank_arena(candidates, rng=seed)
            finally:
                if dise_arena_close is not None:
                    dise_arena_close()

        if initial_picks is None:
            if references is None:
                picker = diversity.SphereExclusionPicker.from_candidates(
                    candidates,
                    threshold=threshold,
                    randomize=randomize,
                    seed=seed,
                    ranks=ranks,
                    )
            else:
                references_arena, references_close = highlevel_load(
                    references, references_format, progress, "'references'")

                problems = check_metadata_problems(
                    candidates.metadata,
                    references_arena.metadata,
                    "candidate fingerprints", "reference fingerprints")
                report_problems(problems)
                picker = diversity.SphereExclusionPicker.from_candidates_and_references(
                    candidates,
                    references_arena,
                    threshold=threshold,
                    randomize=randomize,
                    seed=seed,
                    ranks=ranks,
                    )
        else:
            assert references is None, "should have been checked earlier"
            # Figure out the pick indices
            if pick_type is int:
                # leave it as-is
                pass
            elif pick_type is str:
                initial_picks = get_candidate_indices_by_ids(candidates, initial_picks)
            else:
                raise AssertionError("how is pick_type None?", pick_type)
            picker = diversity.SphereExclusionPicker.from_candidates_and_initial_picks(
                candidates,
                initial_picks,
                threshold = threshold,
                randomize = randomize,
                seed = seed,
                ranks = ranks)


        with record_time("pick", times):
            if include_counts:
                result = picker.pick_n_with_counts(num_picks)
                klass = SpherexCountSearch
            elif include_neighbors:
                result = picker.pick_n_with_neighbors(num_picks)
                klass = SpherexNeighborSearch
            else:
                result = picker.pick_n(num_picks)
                klass = SpherexSearch
    except:
        if references_close is not None:
            references_close()
        if candidates_close is not None:
            candidates_close()
        raise
    
    times["total"] = time.time() - start_time
    return klass(
        candidates = candidates,
        references = references,
        num_initial_picks = 0 if initial_picks is None else len(initial_picks),
        
        threshold = threshold,
        seed = seed,
        direction_msg = direction_msg,
        
        picker = picker,
        num_picked = len(result),
        result = result,
        times = times,
        
        candidates_close = candidates_close,
        references_close = references_close,
        )
