import itertools
import os
import string
import sys
import time

from .. import load_fingerprints, ChemFPError, ParseError
from .. import libchemfp_core

from .. import io
from ..arena import FingerprintArena
from .. import _filename_types

def get_format_desc(format):
    if format is None:
        return ""
    return f" using format {format!r}"

def get_overwrite_msg(overwrite, destination):
    if overwrite:
        return None
    if not isinstance(destination, _filename_types):
        return None
    if not os.path.exists(destination):
        return None
    return f"Output file {destination!r} exists and overwrite is false"
            

def highlevel_load(source, format, progress, label):
    if isinstance(source, FingerprintArena):
        return source, None
    try:
        try:
            arena = load_fingerprints(
                source,
                format=format,
                progress=progress,
                )
            return arena, arena.close
        except (ParseError, ChemFPError):
            # Doing an unusual except-raise-except so I can
            # create errmsg and location in one spot.
            errmsg = f"Unable to read {label} fingerprints{get_format_desc(format)}"
            location = io.Location.from_source(source)
            raise
    except ParseError as err:
        location = err.location or location
        raise ParseError(f"{errmsg}: {err.msg}", location)
    except ChemFPError as err:
        s = str(ParseError(f"{errmsg}: {err}", location))
        raise ChemFPError(s)


def report_problems(problems):
    for problem in problems:
        if problem.severity == "info":
            continue
        if problem.severity != "error":
            sys.stderr.write(str(problem) + "\n")
            continue
        raise problem


def format_time(dt):
    if dt < 0.001:
        return f"{dt*1000*1000:.2f} us"
    if dt < 1:
        return f"{dt*1000:.2f} ms"
    if dt < 60:
        return f"{dt:.2f} s"

    import datetime

    return str(datetime.timedelta(seconds=int(dt)))


def get_fptype(type):
    if not isinstance(type, str):
        return type
    from .. import types, get_fingerprint_type

    try:
        return get_fingerprint_type(type)
    except types.FingerprintTypeUnknownError as err:
        err.include_help()
        raise


class record_time:
    def __init__(self, name, times):
        self.name = name
        self.times = times

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, *args):
        self.times[self.name] = time.time() - self.start_time


class finish_time:
    def __init__(self, times):
        self.times = times

    def __enter__(self):
        return self

    def __exit__(self, *args):
        end_time = time.time()
        start_time = self.times.pop("start")
        self.times["total"] = end_time - start_time

### Clean up reader ids

class AbortHandler(object):
    def error(self, msg, location):
        raise ParseError(
            f"Each of the first 100 records contained errors. Final error: {msg}",
            location)

_abort_handler = AbortHandler()

class IdProcessor:
    def __call__(self, id_fp_iter, location):
        raise NotImplementedError

def as_is_id_processor(id_fp_iter, location):
    return id_fp_iter

    
class CleanupIdProcessor(IdProcessor):
    def __init__(self, id_tag, error_handler):
        self.id_tag = id_tag
        self.error_handler = error_handler
    
    def __call__(self, id_fp_iter, location):
        error_handler = self.error_handler
        cleanup_id = libchemfp_core.cleanup_id
        
        err_count = 0
        for recno, (id, fp) in enumerate(id_fp_iter):
            if not id:
                if recno == err_count == 99:
                    error_handler = _abort_handler
                err_count += 1

                if location.record_format == "sdf":
                    if self.id_tag is None:
                        msg = "Missing title in SD record"
                    else:
                        msg = f"Missing id tag {self.id_tag!r} in SD record"
                elif location.record_format == "smi":
                    msg = "Missing SMILES identifier (second column)"
                else:
                    msg = "Missing identifier"
                error_handler.error(msg, location)
                continue

            id = cleanup_id(id)

            if not id:
                if recno == err_count == 99:
                    error_handler = _abort_handler
                err_count += 1

                if location.record_format == "sdf":
                    if self.id_tag is None:
                        msg = "Empty title in SD record after cleanup"
                    else:
                        msg = f"Empty id tag {self.id_tag!r} in SD record after cleanup"
                elif location.record_format == "smi":
                    msg = "Empty SMILES identifier (second column) after cleanup"
                else:
                    msg = "Empty identifier after cleanup"
                error_handler.error(msg, location)
                continue

            yield id, fp

class PrefixIdProcessor(IdProcessor):
    def __init__(self, prefix):
        self.prefix = prefix
        self._i_counter = itertools.count(1)
        
    def __call__(self, id_fp_iter, location):
        prefix = self.prefix
        for (id, fp), i in zip(id_fp_iter, self._i_counter):
            new_id = f"{prefix}{i}"
            yield new_id, fp

            
class FunctionIdProcessor(IdProcessor):
    def __init__(self, func, id_tag, id_cleanup, error_handler):
        self.func = func
        self.id_tag = id_tag
        self.id_cleanup = id_cleanup
        self.error_handler = error_handler
        self._i_counter = itertools.count(1)
        
    def __call__(self, id_fp_iter, location):
        func = self.func
        id_tag = self.id_tag
        error_handler = self.error_handler

        id_cleanup = self.id_cleanup
        cleanup_id = libchemfp_core.cleanup_id
            
        for (id, fp), i in zip(id_fp_iter, self._i_counter):
            new_id = func(i, id, id_tag, location, error_handler)
            if new_id is None:
                # Allow the callback to skip records
                continue
            if id_cleanup:
                if not new_id:
                    error_handler.error(
                        "Template function result is an empty string",
                        location,
                        )
                    continue

                new_id = cleanup_id(new_id)
                if not new_id:
                    error_handler.error(
                        "Template function result after cleanup is an empty string",
                        location,
                        )
                    continue
                
            yield new_id, fp


    
class TemplateFormatMap(dict):
    known_field_names = {"i", "i0", "id", "clean_id", "recno", "first_word", "first_line"}
    def __init__(self):
        self.location = None
        
    def __missing__(self, name):
        if name == "i0":
            return self["i"] - 1
        if name == "clean_id":
            return libchemfp_core.cleanup_id(self["id"])
        if name == "recno":
            return self.location.recno
        if name == "first_word":
            s = self.location.first_line
            if s is None:
                return ""
            return s.split(None, 1)[0]
        if name == "first_line":
            s = self.location.first_line
            if s is None:
                return ""
            return s
        raise KeyError(name)
            
def get_template_processor(id_template, id_cleanup, error_handler):
    formatter = string.Formatter()
    try:
        parsed_fields = list(formatter.parse(id_template))
    except ValueError as err:
        raise ValueError(f"Cannot parse id_template {id_template!r}: {err}") from None

    for literal_text, field_name, format_spec, conversion in parsed_fields:
        if field_name is None:
            # can happen if there is text after the last field
            continue
        if not field_name:
            raise ValueError(f"Cannot use id_template {id_template!r}: Positional fields are not supported")
        if field_name not in TemplateFormatMap.known_field_names:
            raise ValueError(f"Cannot use id_template {id_template!r}: Unsupported field name {field_name!r}")
    # Should I check that at least one field name is specified?

    # Check that we can format
    format_map = TemplateFormatMap()
    location = io.Location("test")
    location.save(recno=3, record=b"This is a test\n")
    format_map.location = location
    format_map["i"] = 9
    format_map["id"] = "APD"

    # XXX How can this fail?
    try:
        id_template.format_map(format_map)
    except ValueError as err:
        raise ValueError(f"Unable to use id_template {id_template!r}: {err}") from None

    i_counter = itertools.count(1)
    def template_processor(id_fp_iter, location):
        format_map.location = location
        cleanup_id = libchemfp_core.cleanup_id
        for (id, fp), i in zip(id_fp_iter, i_counter):
            format_map["i"] = i
            format_map["id"] = id if (id is not None) else ""
            try:
                new_id = id_template.format_map(format_map)
            except Exception as err:
                error_handler.error(f"Unable to apply id_template {id_template!r}: {err}", location)
                continue
            if id_cleanup:
                if not new_id:
                    error_handler.error(
                        f"Template {id_template!r} result is an empty string",
                        location,
                        )
                    continue
                new_id = cleanup_id(new_id)
                if not new_id:
                    error_handler.error(
                        f"Template {id_template!r} result after cleanup is an empty string",
                        location,
                        )
                    continue
            yield new_id, fp
            
    return template_processor

            
def get_id_processor(*, id_prefix, id_template, id_cleanup, id_tag, errors):
    if id_prefix is not None and id_template is not None:
        raise TypeError("Cannot specify both 'id_prefix' and 'id_template'")
    
    error_handler = io.get_parse_error_handler(errors)
    
    if id_prefix is None and id_template is None:
        if id_cleanup:
            return CleanupIdProcessor(id_tag, error_handler)
        else:
            return as_is_id_processor

    if id_prefix is not None:
        if id_cleanup:
            # The only thing to clean up is the prefix
            id_prefix = libchemfp_core.cleanup_id(id_prefix)
        return PrefixIdProcessor(prefix = id_prefix)

    if id_template is None:
        raise AssertionError("should not get here")
    
    if isinstance(id_template, str):
        return get_template_processor(id_template, id_cleanup, error_handler)

    if callable(id_template):
        return FunctionIdProcessor(
            id_template,
            id_tag = id_tag,
            id_cleanup = id_cleanup,
            error_handler = error_handler,
            )

    raise ValueError("id_template must be a string or a callable")

#### Add docstrings to the high-level shortcuts

# Return the prefix needed to add a new line to a docstring.
# This is the min(line.count_leading_spaces() for line in lines[1:])
def get_docstring_indent(lines):
    indent = None
    for line in lines[1:]:
        s = line.lstrip(" ")
        if not s:
            # ignore indentation of blank lines
            continue
        delta = len(line) - len(s)
        if indent is None:
            indent = delta
        else:
            indent = min(indent, delta)
            
    if indent is None:
        return ""
    else:
        return " "*indent


# The high-level API returns an object containing shortcuts
# that forward to the underlying result object.

# Copy the original (underlying) docstring but insert a
# comment that it's a shortcut.

#  SingleQuerySimsearch forwards to SearchResult
#  MultiQuerySimsearch forwards to SearchResults


def update_docstrings(new_class, search_class, obj_str="obj"):
    search_class_name = search_class.__name__

    # Only document the function in new_class but not its parent
    new_methods = set(dir(new_class)) - set(dir(new_class.__base__))

    # Figure out which functions to update
    for name in dir(search_class):
        if name not in new_methods:
            continue
        # All of the docstrings are supposed to be there.
        docstring = getattr(search_class, name).__doc__
        if docstring is None:
            raise AssertionError(name, search_class)
        doc_lines = docstring.splitlines()
        indent = get_docstring_indent(doc_lines)

        # This is the high-level forwarding function.
        f = getattr(new_class, name)

        # Does it have a doctring to insert?
        old_f_doc = getattr(f, "__doc__", None)
        if old_f_doc is None:
            # Otherwise, create the default
            old_f_doc = f"Shortcut for `{obj_str}.result.{name}()`. See :meth:`.{search_class_name}.{name}`."
        
        doc_lines[1:1] = [f"{indent}", f"{indent}{old_f_doc}"]

        f.__doc__ = "\n".join(doc_lines) + "\n"
