# high-level API similar to the simsearch command-line tool

# Copyright (c) 2021-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""This module should not be imported directly.

It contains internal implementation details of the high-level API
available from the top-level chemfp module.

This module is included in the documentation because parts of this
module are returned to the user, and are part of the public API.

"""

import time

from .. import _core
from .. import search

from .. import (
    check_fingerprint_problems,
    check_metadata_problems,
    get_num_threads,
)

from .highlevel_utils import (
    finish_time,
    format_time,
    get_fptype,
    highlevel_load,
    report_problems,
    record_time,
    update_docstrings,
)

# Wrap the KeyError with something which has a more informative message.
class LookupKeyError(KeyError):
    def __str__(self):
        s = super().__str__()
        return f"No target fingerprint found with the query_id: {s}"


def get_k_and_threshold(k, threshold):
    if k is not None:
        if k < 1:
            raise ValueError("k must be positive")

    if threshold is not None:
        if not (0.0 <= threshold <= 1.0):
            raise ValueError("threshold must be between 0.0 and 1.0, inclusive")

    if threshold is None:
        if k is None:
            threshold = 0.0
            k = 3
        else:
            threshold = 0.0

    return k, threshold


def get_alpha_and_beta(alpha, beta):
    if alpha is None:
        if beta is None:
            return (1.0, 1.0)
        raise TypeError("Must specify 'alpha' if 'beta' is specified")
    elif beta is None:
        beta = alpha

    if not (0.0 <= alpha <= 100.0):
        raise ValueError("alpha must be in the range [0.0, 100.0]")

    if not (0.0 <= beta <= 100.0):
        raise ValueError("beta must be in the range [0.0, 100.0]")

    return alpha, beta


class BaseSimsearch:
    """This is the base class for the objects returned by :func:`.simsearch`

    It contains the query parameters, search results, and timings.

    In addition, it is a context manager for any files which may have
    been opened.
    """
    def __init__(
        self,
        *,
        num_queries,
        num_targets,
        k,
        threshold,
        alpha,
        beta,
        NxN,
        times,
        result,
        query_fp=None,
        queries=None,
        targets=None,
        queries_close=None,
        targets_close=None,
    ):
        self.num_queries = num_queries
        self.num_targets = num_targets
        self.k = k
        self.threshold = threshold
        self.alpha = alpha
        self.beta = beta
        self.NxN = NxN
        self.times = times
        self.result = result

        self.query_fp = query_fp
        self.queries = queries
        self.targets = targets
        self._queries_close = queries_close
        self._targets_close = targets_close

    def get_description(self):
        "Return a human-readable description of the simsearch run"
        if self.alpha == self.beta == 1.0:
            what = "Tanimoto"
        else:
            what = f"Tversky(a={self.alpha}, b={self.beta})"

        if self.k is None:
            msg = f"{what} >= {self.threshold}"
        elif self.threshold == 0:
            msg = f"{self.k}-nearest {what}"
        else:
            msg = f"{self.k}-nearest {what} >= {self.threshold}"

        if self.NxN:
            full_msg = f"{msg} NxN search. #fingerprints={self.num_targets}"
        else:
            full_msg = f"{msg} search. #queries={self.num_queries}, #targets={self.num_targets}"

        search_time = self.times["search"]
        total_time = self.times["total"]
        time_msg = (
            f" (search: {format_time(search_time)} total: {format_time(total_time)})"
        )
        return full_msg + time_msg

    def __repr__(self):
        return f'{self.__class__.__name__}({self.get_description()!r}, result={self.result!r})'

    def __str__(self):
        return self.get_description()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        "Close any associated files"
        close = self._queries_close
        if close is not None:
            self._queries_close = None
            try:
                close()
            except Exception:
                pass
        
        close = self._targets_close
        if close is not None:
            self._targets_close = None
            try:
                close()
            except Exception:
                pass


class SingleQuerySimsearch(BaseSimsearch):
    def __len__(self):
        "Shortcut for `len(simsearch.result)`. See :meth:`.SearchResult.__len__`."
        return len(self.result)

    def __iter__(self):
        "Shortcut for `iter(simsearch.result)`.  See :meth:`.SearchResult.__iter__`."
        return iter(self.result)

    def as_buffer(self):
        return self.result.as_buffer()

    def get_indices(self):
        return self.result.get_indices()

    def get_ids(self):
        return self.result.get_ids()

    def iter_ids(self):
        return self.result.iter_ids()

    def get_scores(self):
        return self.result.get_scores()

    def get_ids_and_scores(self):
        return self.result.get_ids_and_scores()

    @property
    def query_id(self):
        "Shortcut for `simsearch.result.query_id`. See :attr:`.SearchResult.query_id`."
        return self.result.query_id

    def to_pandas(self, *, columns=["target_id", "score"]):
        return self.result.to_pandas(columns=columns)

    def get_indices_and_scores(self):
        return self.result.get_indices_and_scores()

    def as_ctypes(self):
        return self.result.as_ctypes()

    def as_numpy_array(self):
        return self.result.as_numpy_array()

    def get_indices_as_numpy_array(self):
        return self.result.get_indices_as_numpy_array()

    def get_scores_as_numpy_array(self):
        return self.result.get_scores_as_numpy_array()
    
    def reorder(self, ordering="decreasing-score-plus"):
        return self.result.reorder(ordering)

    def min(self):
        return self.result.min()

    def max(self):
        return self.result.max()

    def count(self, min_score=None, max_score=None, interval="[]"):
        return self.result.count(min_score, max_score, interval)

    def cumulative_score(self, min_score=None, max_score=None, interval="[]"):
        return self.result.cumulative_score(min_score, max_score, interval)

    def format_ids_and_scores_as_bytes(self, ids=None, precision=4):
        return self.result.format_ids_and_scores_as_bytes(ids, precision)

update_docstrings(SingleQuerySimsearch, search.SearchResult)
 
class MultiQuerySimsearch(BaseSimsearch):
    def __iter__(self):
        "Shortcut for `iter(simsearch.result)`. See :attr:`.SearchResults.__iter__`."
        return iter(self.result)

    def __len__(self):
        "Shortcut for len(simsearch.result). See :attr:`.SearchResults.__len__`."
        return len(self.result)

    @property
    def shape(self):
        return self.result.shape

    def __getitem__(self, i):
        "Shortcut for simsearch.result[i]. See :attr:`.SearchResults.__getitem__`."
        return self.result[i]

    def iter_indices(self):
        return self.result.iter_indices()

    def iter_ids(self):
        return self.result.iter_ids()

    def iter_scores(self):
        return self.result.iter_scores()

    def iter_indices_and_scores(self):
        return self.result.iter_indices_and_scores()

    def iter_ids_and_scores(self):
        return self.result.iter_ids_and_scores()

    def to_pandas(self, *,
                      columns = ["query_id", "target_id", "score"],
                      empty = ("*", None),
                      ):
        return self.result.to_pandas(columns=columns, empty=empty)

    def count_all(self, min_score=None, max_score=None, interval="[]"):
        return self.result.count_all(min_score, max_score, interval)
    
    def cumulative_score_all(self, min_score=None, max_score=None, interval="[]"):
        return self.result.cumulative_score_all(min_score, max_score, interval)
    
    def reorder_all(self, order="decreasing-score-plus"):
        return self.result.reorder_all(order)
    
    def to_csr(self, dtype=None):
        return self.result.to_csr(dtype)
    
update_docstrings(MultiQuerySimsearch, search.SearchResults)

class NxNSimsearch(BaseSimsearch):
    def __iter__(self):
        "Shortcut for `iter(simsearch.result)`. See :attr:`.SearchResults.__iter__`."
        return iter(self.result)

    def __len__(self):
        "Shortcut for len(simsearch.result). See :attr:`.SearchResults.__len__`."
        return len(self.result)

    @property
    def shape(self):
        return self.result.shape

    def __getitem__(self, i):
        "Shortcut for simsearch.result[i]. See :attr:`.SearchResults.__getitem__`."
        return self.result[i]

    def iter_indices(self):
        return self.result.iter_indices()

    def iter_ids(self):
        return self.result.iter_ids()

    def iter_scores(self):
        return self.result.iter_scores()

    def iter_indices_and_scores(self):
        return self.result.iter_indices_and_scores()

    def iter_ids_and_scores(self):
        return self.result.iter_ids_and_scores()

    def to_pandas(self, *,
                      columns = ["query_id", "target_id", "score"],
                      empty = ("*", None),
                      ):
        return self.result.to_pandas(columns=columns, empty=empty)

    def count_all(self, min_score=None, max_score=None, interval="[]"):
        return self.result.count_all(min_score, max_score, interval)
    
    def cumulative_score_all(self, min_score=None, max_score=None, interval="[]"):
        return self.result.cumulative_score_all(min_score, max_score, interval)
    
    def reorder_all(self, order="decreasing-score-plus"):
        return self.result.reorder_all(order)
    
    def to_csr(self, dtype=None):
        return self.result.to_csr(dtype)
    
update_docstrings(NxNSimsearch, search.SearchResults)


def get_progress_bar(*, progress, desc, unit, total):
    from .. import _tqdm
    return _tqdm.get_tqdm_constructor(progress)(
        desc=desc,
        unit=unit,
        total=total,
        dynamic_ncols = True,
        )
    
    
def simsearch_NxN(targets, targets_close, k, threshold, alpha, beta,
                      include_lower_triangle, ordering, progress, times):
    from .. import search, _tqdm
    progress_bar = get_progress_bar(
        progress = progress,
        desc = "queries",
        unit = " fps",
        total = len(targets),
        )

    def batch_callback(results, start, end):
        progress_bar.update(end - start)

    with progress_bar:
        if k is None:
            if alpha == beta == 1.0:
                # Can't use the more generic Tversky interface since
                # that doesn't support include_lower_triangle.
                hits = search.threshold_tanimoto_search_symmetric(
                    targets, threshold, include_lower_triangle=include_lower_triangle,
                    batch_callback=batch_callback,
                )
            else:
                hits = search.threshold_tversky_search_symmetric(
                    targets, threshold, alpha, beta,
                    batch_callback=batch_callback,
                )
            if ordering is not None:
                hits.reorder_all(ordering)
        else:
            hits = search.knearest_tversky_search_symmetric(
                targets, k, threshold, alpha, beta, batch_callback=batch_callback
            )
            # knearest is already sorted by decreasing score
            if ordering is not None and ordering != DECREASING_SCORE:
                hits.reorder_all(ordering)

    
    return NxNSimsearch(
        num_queries=len(targets),
        num_targets=len(targets),
        k=k,
        threshold=threshold,
        alpha=alpha,
        beta=beta,
        NxN=True,
        times=times,
        result=hits,
        queries=targets,
        targets=targets,
        targets_close=targets_close,
    )


def simsearch_queries(queries, queries_close, targets, targets_close,
                          k, threshold, alpha, beta, ordering, progress, times):
    progress_bar = get_progress_bar(
        progress = progress,
        desc = "queries",
        unit = " fps",
        total = len(queries),
        )

    # Want about 1 update per second
    batch_size = 1000

    num_targets = len(targets)
    if num_targets:
        # Assuming 100M fps/second
        batch_size = int(100_000_000 / num_targets)
        batch_size = max(batch_size, get_num_threads())

    def batch_callback(results, start, end):
        progress_bar.update(end - start)

    if k is None:
        hits = search.threshold_tversky_search_arena(
            queries,
            targets,
            threshold,
            alpha,
            beta,
            batch_size=batch_size,
            batch_callback=batch_callback,
        )
        if ordering is not None:
            hits.reorder_all(ordering)
    else:
        hits = search.knearest_tversky_search_arena(
            queries,
            targets,
            k,
            threshold,
            alpha,
            beta,
            batch_size=batch_size,
            batch_callback=batch_callback,
        )
        if ordering is not None and ordering != DECREASING_SCORE:
            hits.reorder_all(ordering)


    progress_bar.close()
    
    return MultiQuerySimsearch(
        num_queries=len(queries),
        num_targets=len(targets),
        k=k,
        threshold=threshold,
        alpha=alpha,
        beta=beta,
        NxN=False,
        times=times,
        result=hits,
        queries=queries,
        targets=targets,
        queries_close=queries_close,
        targets_close=targets_close,
    )


def simsearch_query(query_fp, targets, targets_close, k, threshold, alpha, beta, ordering, times):
    if k is None:
        result = targets.threshold_tversky_search_fp(query_fp, threshold, alpha, beta)
        if ordering:
            result.reorder(ordering)
    else:
        result = targets.knearest_tversky_search_fp(query_fp, k, threshold, alpha, beta)
        if ordering and ordering != DECREASING_SCORE:
            result.reorder(ordering)

        
    return SingleQuerySimsearch(
        num_queries=1,
        num_targets=len(targets),
        k=k,
        threshold=threshold,
        alpha=alpha,
        beta=beta,
        NxN=False,
        result=result,
        times=times,
        query_fp=query_fp,
        targets=targets,
        targets_close=targets_close,
    )

DECREASING_SCORE = "decreasing-score" # help reduce the chance of typos

_supported_orderings = set()
    
for i in range(1000):
    try:
        _supported_orderings.add(_core.get_ordering(i))
    except IndexError:
        break
del i

assert DECREASING_SCORE in _supported_orderings
    
def _check_ordering(reorder):
    if reorder is None:
        return
    
    if reorder in _supported_orderings:
        return

    reordering_list = ", ".join([repr(s) for s in sorted(_supported_orderings)])
    raise ValueError(
        f"Unsupported ordering: {reorder!r}. "
        f"Available values are: {reordering_list}"
    )
    

def simsearch(
    targets,
    query,
    query_fp,
    query_id,
    queries,
    NxN,
    query_format,
    target_format,
    type,
    k,
    threshold,
    alpha,
    beta,
    include_lower_triangle,
    ordering,
    progress,
):

    queries_close = targets_close = None
    times = {"start": time.time()}

    k, threshold = get_k_and_threshold(k, threshold)
    alpha, beta = get_alpha_and_beta(alpha, beta)

    num_query_args = (
        (query is not None)
        + (query_fp is not None)
        + (query_id is not None)
        + (queries is not None)
        + (NxN is not None)
    )
    if num_query_args == 0:
        raise TypeError("Must specify one of 'query', 'query_fp', 'query_id', 'queries', or 'NxN'")
    if num_query_args != 1:
        raise TypeError("Must specify only one of 'query', 'query_fp', 'query_id', 'queries', or 'NxN'")

    # Do an early check so we don't do a huge search only to fail at the very
    # end with a typo.
    _check_ordering(ordering)
    
    with record_time("load_targets", times):
        targets, targets_close = highlevel_load(targets, target_format, progress, "'targets'")

    try:
        # Load the query
        if NxN:
            # No need to load queries
            with finish_time(times), record_time("search", times):
                return simsearch_NxN(targets, targets_close, k, threshold, alpha, beta,
                                         include_lower_triangle, ordering, progress, times)

        if queries is not None:
            with record_time("load_queries", times):
                queries, queries_close = highlevel_load(queries, query_format, progress, "'queries'")

            problems = check_metadata_problems(
                queries.metadata,
                targets.metadata,
                query_name="'queries'",
                target_name="'targets'",
                )
            report_problems(problems)

            with finish_time(times), record_time("search", times):
                return simsearch_queries(
                    queries, queries_close, targets, targets_close,
                    k, threshold, alpha, beta, ordering, progress, times
                )

        if query is not None:
            with record_time("get_fptype", times):
                if query_format is None:
                    query_format = "smi"
                if type is None:
                    fptype = targets.get_fingerprint_type()
                else:
                    fptype = get_fptype(type)
            with record_time("generate_fp", times):
                query_fp = fptype.parse_molecule_fingerprint(query, format=query_format)

        if query_id is not None:
            query_fp = targets.get_fingerprint_by_id(query_id)
            if query_fp is None:
                raise LookupKeyError(query_id)

        problems = check_fingerprint_problems(query_fp, targets)
        report_problems(problems)

        with finish_time(times), record_time("search", times):
            return simsearch_query(query_fp, targets, targets_close,
                                       k, threshold, alpha, beta, ordering, times)
    except:
        # If I opened a file, then close it.
        if queries_close is not None:
            queries_close()
        if targets_close is not None:
            targets_close()
        raise
