# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""Helper module for working with file I/O.

Only the :class:`.Location` class is part of the public API and meant to
be used directly. The public API also returns :class:`.ProgressBar`
objects, which are part of the public API.

The other parts of this module are not part of the public API. (Let me
know if anything else should be part of the public API.)

"""

__all__ = ["Location"]

import re
import os
import sys
import binascii
from io import BytesIO, StringIO

from datetime import datetime

from . import ParseError
from ._compat import filename_types, myrepr

if sys.platform.startswith("win"):
    DEV_STDIN = "CON"
else:
    if os.path.exists("/dev/stdin"):
        DEV_STDIN = "/dev/stdin"
    else:
        DEV_STDIN = None

def _stdin():
    stdin = sys.stdin
    return getattr(stdin, "buffer", stdin)

def _stdout():
    stdout = sys.stdout
    return getattr(stdout, "buffer", stdout)

        
def utcnow():
    return datetime.utcnow().isoformat().split(".", 1)[0]


############ error handlers
def _get_text(msg, location):
    if location is None:
        return str(msg)
    return "%s, %s" % (msg, location.where())

class ErrorHandler:
    ## def warning(self, msg, location=None, extra=None):
    ##     sys.stderr.write("WARNING: %s\n" % _get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        raise NotImplementedError("error() must be implemented in the subclass")
    

class IgnoreHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     pass
    
    def error(self, msg, location=None, extra=None):
        pass

class ReportHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     sys.stderr.write("WARNING: %s\n" % _get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        sys.stderr.write("ERROR: %s. Skipping.\n" % _get_text(msg, location))

class LogHandler(ErrorHandler):
    def __init__(self, logger="chemfp"):
        if isinstance(logger, filename_types):
            import logging
            logger = logging.getLogger(logger)
        self._logger = logger

    ## def warning(self, msg, location=None, extra=None):
    ##     self._logger.warning(_get_text(msg, location))

    def error(self, msg, location=None, extra=None):
        self._logger.error(_get_text(msg, location))

class StrictHandler(ErrorHandler):
    ## def warning(self, msg, location=None, extra=None):
    ##     pass

    def error(self, msg, location=None, extra=None):
        if location is not None:
            # Need to make a copy because the main location object
            # wil be cleared out as part of the exception unwinding,
            # which will clear record-specific fields.
            location = location.copy()
        raise ParseError(msg, location).with_traceback(None)

_error_handlers = {
    "ignore": IgnoreHandler(),
    "report": ReportHandler(),
    "strict": StrictHandler(),
}

def get_parse_error_handler(errors):
    if isinstance(errors, str):
        try:
            return _error_handlers[errors]
        except KeyError:
            pass
        if "log" not in _error_handlers:
            _error_handlers["log"] = handler = LogHandler("chemfp")
        if errors == "log":
            return handler
            
        raise ValueError("'errors' value must be an ErrorHandler or one of %s, not %r"
                         % (", ".join(map(repr, sorted(_error_handlers))), errors))

    return errors


############

_compression_extensions = {
    ".gz": "gz",
    ".gzip": "gz",
    ".bz2": "bz2",
    ".bzip": "bz2",
    ".bzip2": "bz2",
    ".xz": "xz",
    ".zst": "zst",
    }
# _compression_regex = "|".join(re.escape(ext) for ext in _compression_extensions)

_format_pat = re.compile("^([a-zA-Z0-9]+)(\\.[a-zA-Z0-9]+)?$")

def _normalize_format_name(format):
    # This must be of the form <name> [. <compression> ]
    m = _format_pat.match(format)
    if m is None:
        raise ValueError("format name must be in the form 'name[.compression]':  %r"
                         % (format,))

    format_name = m.group(1)
    ext = m.group(2)
    if ext is None:
        compression = ""
    else:
        compression = ext[1:]  # remove the leading "."
    return format_name, compression

def _determine_format_from_filename(filename, default):
    # The filename could have 0, 1 or 2 extensions
    base, ext = os.path.splitext(filename)
    if ext == "":
        # No extensions, use the default
        return default
    ext = ext.lower()

    # If it's not a compression extension then it's a format indicator
    if ext not in _compression_extensions:
        # the [1:] is to remove the leading "."
        return (ext[1:], "")

    # Found a compression, now look for the format
    compression = _compression_extensions[ext]

    base, ext = os.path.splitext(base)
    if ext == "":
        # compression found but not the actual format type
        return (default[0], compression)

    # The [1:] is to remove the leading "."
    format_name = ext[1:].lower()
    
    return (format_name, compression)


    
def normalize_input_format(source, format, default=("fps", "")):
    if format is not None:
        # Either 0 or 1 dots
        terms = format.split(".")
        if terms:
            if len(terms) == 1:
                return terms[0], ""
            elif len(terms) == 2:
                return terms[0], terms[1]
        raise ValueError("Could not understand format specification %r" % (format,))
        
    if source is None:
        # Read from stdin
        filename = None
    elif isinstance(source, filename_types):
        # This is a filename
        filename = source
    elif hasattr(source, "read"):
        # This is a Python file object
        filename = getattr(source, "name", None)
    else:
        raise ValueError("Unsupported source type %r" % (source,))

    if filename is None:
        # Reading from stdin or an unnamed file-like object with no
        # specified format. Not going to sniff the input. Instead, just
        return default

    return _determine_format_from_filename(filename, default)
        
def get_compressed_input_format(source, base_format):
    compression = normalize_input_format(source, None, (base_format, None))[1]
    if compression:
        return f"{base_format}.{compression}"
    else:
        return base_format
    

def normalize_output_format(destination, format, default=("fps", "")):
    if format is not None:
        # Either 0 or 1 dots
        terms = format.split(".")
        if terms:
            if len(terms) == 1:
                return terms[0], ""
            elif len(terms) == 2:
                return terms[0], terms[1]
        raise ValueError("Could not understand format specification %r" % (format,))
    
    if destination is None:
        # Write to stdout
        filename = None
    elif isinstance(destination, filename_types):
        # This is a filename
        filename = destination
    elif hasattr(destination, "write"):
        # This is a Python file object
        filename = getattr(destination, "name", None)
    else:
        raise ValueError("Unknown destination type %r" % (destination,))

    if filename is None:
        # Writing to stdout or an unnamed file-like object with no
        # specified format.
        return default
        
    return _determine_format_from_filename(filename, default)

def get_compressed_output_format(destination, base_format):
    compression = normalize_output_format(destination, None, (base_format, None))[1]
    if compression:
        return f"{base_format}.{compression}"
    else:
        return base_format

def get_source_name(source):
    if source is None:
        return "<stdin>"
    
    if isinstance(source, filename_types):
        return source

    filename = getattr(source, "name", None)
    if isinstance(filename, filename_types):
        return filename

    return "<file>"

def get_destination_name(destination):
    if destination is None:
        return "<stdout>"

    if isinstance(destination, filename_types):
        return destination
    
    filename = getattr(destination, "name", None)
    if isinstance(filename, filename_types):
        return filename

    return "<file>"

 
####

def _do_nothing():
    pass

def open_binary_output(destination, blame_name=None, level=None):
    if destination is None:
        return _stdout(), _do_nothing
    if not isinstance(destination, filename_types):
        return destination, _do_nothing
    base, ext = os.path.splitext(destination)
    ext = ext.lower()
    if ext not in _compression_extensions:
        f = open(destination, "wb")
        close = f.close
    else:
        # I don't like how I wrap and then unwrap the output.
        f, close = open_compressed_output(destination, _compression_extensions[ext], blame_name, level)
    return f, close

def open_compressed_output_wrapped(destination, compression, blame_name, encoding, encoding_errors, level):
    f, close = open_compressed_output(destination, compression, blame_name, level)
    return OutputToBytes(f, encoding, encoding_errors), close

def _get_gzip_compresslevel(level):
    if level is None:
        return 9

    if isinstance(level, str):
        if level == "min":
            return 1
        if level == "max" or level == "default":
            return 9
        raise ValueError("Unsupported gzip compression level %r" % (level,))

    if level < 1:
        return 1
    if level > 9:
        return 9
    
    return level

def _get_zst_level(level):
    if level is None:
        return 3

    if isinstance(level, str):
        if level == "min":
            return 1
        if level == "default":
            return 3
        if level == "max":
            return 19
        raise ValueError("Unsupported zst compression level %r" % (level,))

    if level < 1:
        return 1
    if level > 19:
        return 19
    
    return level

def get_zlib_level(level):
    if level is None:
        return -1

    if isinstance(level, str):
        if level == "min":
            return 1
        if level == "default":
            return 9
        if level == "max":
            return 9
        raise ValueError("Unsupported zlib compression level %r" % (level,))
    
    if level < 1:
        return 1
    if level > 9:
        return 9
    return level

# Used to prevent zstandard's writer from passing the close to the file object
class _Destination:
    def __init__(self, destination):
        self.write = destination.write
    def close(self):
        pass

def open_compressed_output(destination, compression, blame_name, level=None):
    # Raises a ValueError if the compression isn't supported
    if not compression:
        if destination is None:
            return _stdout(), _do_nothing
        
        if isinstance(destination, filename_types):
            f = open(destination, "wb")
            return f, f.close

        # XX double-check that it's open in binary?
        return destination, _do_nothing

    if compression == "gz":
        import gzip
        compresslevel = _get_gzip_compresslevel(level)
        if destination is None:
            gzip_envvar, gzip_args = get_gzip_args(compresslevel)
            if gzip_args is not None and not sys.stdout.isatty():
                # Need the check for tty as otherwise I get:
                #   "gzip: standard output is a terminal -- ignoring"
                import subprocess
                try:
                    p = subprocess.Popen(gzip_args,
                                             stdin=subprocess.PIPE)
                except OSError as err:
                    sys.stderr.write("Cannot exec $%s arguments %r: %s\n" % (
                        gzip_envvar, gzip_args, err))
                    sys.stderr.write("Using an internal gzip writer.\n")
                else:
                    return _get_subprocess_file_and_close(p, p.stdin)

            f = gzip.GzipFile(mode="wb", fileobj=_stdout(), compresslevel=compresslevel)
        elif isinstance(destination, filename_types):
            gzip_envvar, gzip_args = get_gzip_args(compresslevel)
            if gzip_args is not None:
                import subprocess
                g = open(destination, "wb")
                try:
                    p = subprocess.Popen(gzip_args,
                                             stdin=subprocess.PIPE,
                                             stdout=g)
                except OSError as err:
                    sys.stderr.write("Cannot exec $%s arguments %r: %s\n" % (
                        gzip_envvar, gzip_args, err))
                    sys.stderr.write("Using an internal gzip writer.\n")
                    g.close()
                else:
                    return _get_subprocess_file_and_close(p, p.stdin, g.close)
            
            f = gzip.open(destination, "wb", compresslevel=compresslevel)
        else:
            f = gzip.GzipFile(mode="wb", fileobj=destination, compresslevel=compresslevel)
        return f, f.close

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        level = _get_zst_level(level)
        if destination is None:
            cctx = zstd.ZstdCompressor(level)
            writer = cctx.stream_writer(_stdout())
            return writer, writer.close
        
        elif isinstance(destination, filename_types):
            f = open(destination, "wb")
            try:
                cctx = zstd.ZstdCompressor(level)
                writer = cctx.stream_writer(f)
            except:
                f.close()
                raise

            return writer, make_multicloser(writer.close, f.close)
        
        else:
            cctx = zstd.ZstdCompressor(level)
            # zstandard's compressor forwards close() to
            # the underlying file handle, so you can't
            # pass in a BytesIO.
            # chemfp doesn't take control of close().
            writer = cctx.stream_writer(_Destination(destination))
            return writer, writer.close
    
    if compression == "bz2":
        try:
            import bz2
        except ImportError:
            raise ValueError("Please install the bz2 module for bzip2 compression support")
        
        if destination is None:
            if not os.path.exists("/dev/stdout"):
                raise ValueError("Python's bz2 library cannot write compressed data to stdout "
                               "on this platform")
            f = bz2.BZ2File("/dev/stdout", "wb")
        elif isinstance(destination, filename_types):
            f = bz2.BZ2File(destination, "wb")
        else:
            raise ValueError("Python's bz2 library does not support writing to a file object")
        return f, f.close

    if compression == "xz":
        raise ValueError("chemfp does not yet support xz compression")

    if blame_name is None:
        raise ValueError("Unsupported compression type %r" % (compression,))
    else:
        raise ValueError("%s does not support compression type %r" % (blame_name, compression))
    
def make_multicloser(*close_funcs):
    def close():
        for close_func in close_funcs:
            close_func()
    return close

def get_gzcat_args():
    envvar = "CHEMFP_GZCAT"
    s = os.environ.get(envvar, "")
    if not s:
        envvar = "CHEMFP_GZIP"
        s = os.environ.get(envvar, "")
        if not s:
            return None, None
        extra = ["-dc"]
    else:
        extra = []
    import shlex
    terms = shlex.split(s)
    if not terms:
        return None, None
    #print("Using", terms, file=sys.stderr)
    return envvar, terms + extra

def get_gzip_args(compresslevel):
    envvar = "CHEMFP_GZIP"
    s = os.environ.get(envvar, "")
    if not s:
        return None, None
    import shlex
    terms = shlex.split(s)
    if not terms:
        return None, None
    # interpose the required arguments first, which may allow possible override
    terms = [terms[0]] + ["-c", "-%d"%(compresslevel,)] + terms[1:]
    #print("Using", terms, file=sys.stderr)
    return envvar, terms

def _get_subprocess_file_and_close(p, f, *extra_close):
    def close():
        f.close()
        #returncode = p.wait()
        p.wait()
        for func in extra_close:
            func()
        ### Can't use this because the close() is embedded in too much
        ### code which expects that close() will never fail.
        ## if returncode != 0:
        ##     raise IOError("Subprocess (%r) exited with unexpected error code %d" % (
        ##         p.args, returncode))
    return f, close

class ZstdStreamWrapper:
    def __init__(self, stream):
        self._stream = stream
        self.close = stream.close
        self.flush = stream.flush
        self.isatty = stream.isatty
        self.read = stream.read
        if hasattr(stream, "read1"):
            self.read1 = stream.read1
        self.readable = stream.readable
        self.readall = stream.readall
        if hasattr(stream, "readinto"):
            self.readinto = stream.readinto
        if hasattr(stream, "readinto1"):
            self.readinto1 = stream.readinto1
        self.readline = stream.readline
        self.readlines = stream.readlines
        self.seek = stream.seek
        self.seekable = stream.seekable
        self.tell = stream.tell
        self.writable = stream.writable
        self.write = stream.write
        #self.writelines = stream.writelines

    @property
    def closed(self):
        return self._stream.closed

    ### chemfp expects a few other API features
    
    _it = None
    def __iter__(self):
        it = self._it
        if it is None:
            self._it = it = self._get_iter()
        return it
            
    def _get_iter(self):
        prev = b""
        while 1:
            s = self.read(100000)
            if not s:
                if prev:
                    yield prev
                break
            s = prev + s
            lines = s.splitlines(True)
            if lines[-1][-1:] == "\n":
                prev = b""
                yield from lines
            else:
                prev = lines[-1]
                yield from lines[:-1]

    def writelines(self, lines):
        # zstandard's ZstdCompressionWriter does not support writelines()
        write = self.write
        for line in lines:
            write(line)

# The Zstandard decompressor on close will close the underlying
# object, even if it's a BytesIO where the decompressor isn't in
# charge of the I/O lifetime. In that case, pass it a wrapper which
# ignores the close.

class NoCloseWrapper:
    def __init__(self, obj):
        self.read = obj.read
        
    def close(self):
        pass

# I have a tough time ensuring that I don't get_position()
# on a closed file. Report the current offset or, if that
# fails, report the larger of the previous value or the
# end position.
    
def get_max_position(tell_function, end_position, exceptions = (ValueError,)):
    position = 0
    if end_position is None:
        def get_position():
            nonlocal position
            try:
                position = tell_function()
            except exceptions:
                return position
            return position
    else:
        def get_position():
            nonlocal position
            try:
                position = tell_function()
            except exceptions:
                return max(position, end_position)
            return position
    return get_position

    
def register_fileobj_position(location, f, end_position=None):
    if location.position_units is not None:
        # Already registered
        return
    if end_position is None:
        # Figure it out from the file object
        try:
            fileno = f.fileno()
        except (AttributeError, os.error):
            pass
        else:
            end_position = get_fileno_size(fileno)

    # Either have end_position or it's None
    try:
        tell_function = f.tell
    except AttributeError:
        tell_function = None
        # Can't get f.tell, so can't do anything
        return

    # Is it usable?
    try:
        current_position = tell_function()
    except (ValueError, os.error):
        # Can't f.tell(), so can't do anything
        return

    location.register(get_position = get_max_position(tell_function, end_position))
    location.save(
        start_position = current_position,
        end_position = end_position,
        position_units = "bytes",
        )

def get_filename_size(filename):
    try:
        return os.stat(filename).st_size
    except os.error:
        return None

def get_fileno_size(fileno):
    try:
        return os.fstat(fileno).st_size
    except os.error:
        return None
    
def register_gzio_position(location, gzf, end_position):
    # If this zlib doesn't have a gzoffset function then
    # the GZFile won't implement the method, leaving
    # no way to have a progress bar.
    gzoffset = getattr(gzf, "gzoffset", None)
    if gzoffset is not None:
        location.register(
            get_position = get_max_position(gzoffset, end_position),
            )
        location.save(
            start_position = None if (gzoffset is None) else gzoffset(),
            end_position = end_position,
            position_units = "bytes",
            )

def open_bytes_io(content, location):
    infile = BytesIO(content)
    close = infile.close
    register_fileobj_position(location, infile, len(content))
    return infile, close
                            
def open_compressed_input(source, compression, location, blame_name):
    if location is None:
        location = Location()
        
    # Raises a ValueError if the compression isn't supported
    if not compression:
        if source is None:
            stdin_fileobj = _stdin()
            register_fileobj_position(location, stdin_fileobj)
            return stdin_fileobj, _do_nothing
        elif isinstance(source, filename_types):
            f = open(source, "rb")
            register_fileobj_position(location, f)
            return f, f.close
        else:
            # Are we reading strings or bytes?
            s = source.read(0)
            if s != b"":
                # It reads strings, while chemfp expects bytes. I can either
                # transparently convert it to bytes with an arbitrary encoding,
                # which will introduce invisible performance overhead, or I
                # can be explicit and require bytes. I chose the latter
                raise ValueError("The file-like 'source' object must be opened in binary mode, not text")
            register_fileobj_position(location, source)
            return source, _do_nothing

    if compression == "gz":
        import gzip
        if source is None:
            gzcat_args = None
            if not sys.stdin.isatty():
                try:
                    fileno = sys.stdin.fileno()
                except Exception:
                    fileno = -1
                if fileno == 0:
                    gzcat_envvar, gzcat_args = get_gzcat_args()
            
            if gzcat_args is not None:
                import subprocess
                # Read from stdin
                try:
                    p = subprocess.Popen(gzcat_args, stdout=subprocess.PIPE)
                except OSError as err:
                    sys.stderr.write("Cannot exec $%s arguments %r: %s\n" % (
                        gzcat_envvar, gzcat_args, err))
                    sys.stderr.write("Using an internal gzip reader.\n")
                else:
                    # I can't track the file position in this case
                    return _get_subprocess_file_and_close(p, p.stdout)

            from . import gzio
            fd = gzio.get_fd(sys.stdin)
            if fd is not None:
                f = gzio.gzdopen_rb(fd, name="<stdin>")
                register_gzio_position(location, f, get_fileno_size(fd))
            else:
                fileobj = _stdin()
                f = gzip.GzipFile(fileobj=fileobj)
                register_fileobj_position(location, fileobj)
                
        elif isinstance(source, filename_types):
            gzcat_envvar, gzcat_args = get_gzcat_args()
            if gzcat_args is not None:
                import subprocess
                try:
                    p = subprocess.Popen(gzcat_args + [source], stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE)
                except OSError as err:
                    sys.stderr.write("Cannot exec $%s arguments %r: %s\n" % (
                        gzcat_envvar, gzcat_args, err))
                    sys.stderr.write("Using an internal gzip reader.\n")
                else:
                    try:
                        returncode = p.wait(0.01) # See if it failed
                    except subprocess.TimeoutExpired:
                        returncode = None

                    # TODO: do I need to check for SIGTERM?
                    if returncode is None or returncode == 0:
                        # I can't track the file position in this case
                        return _get_subprocess_file_and_close(p, p.stdout, p.stderr.close)

                    msg = p.stderr.read().strip()
                    
                    sys.stderr.write(
                        "Unexpected exit for $%s arguments %r (exit code %d): %s\n" % (
                            gzcat_envvar, gzcat_args, returncode, msg))
                    sys.stderr.write("Using an internal gzip reader.\n")
                    
                    p.stdout.close()
                    p.stderr.close()
                    # fall through

            from . import gzio
            if gzio.is_available:
                f = gzio.gzopen_rb(source)
                register_gzio_position(location, f, get_filename_size(source))
            else:
                f = gzip.open(source, "rb")
                register_fileobj_position(location, f.fileobj)
        else:
            f = gzip.GzipFile(mode="rb", fileobj=source)
            register_fileobj_position(location, source)

        return f, f.close

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        if source is None:
            dctx = zstd.ZstdDecompressor()
            stdin_fileobj = _stdin()
            reader = dctx.stream_reader(stdin_fileobj)
            register_fileobj_position(location, stdin_fileobj)
            return ZstdStreamWrapper(reader), reader.close
            
        elif isinstance(source, filename_types):
            f = open(source, "rb")
            try:
                dctx = zstd.ZstdDecompressor()
                reader = dctx.stream_reader(f)
            except:
                f.close()
                raise

            register_fileobj_position(location, f)
            return ZstdStreamWrapper(reader), make_multicloser(reader.close, f.close)
        
        else:
            dctx = zstd.ZstdDecompressor()
            reader = dctx.stream_reader(NoCloseWrapper(source))
            register_fileobj_position(location, source)
            return ZstdStreamWrapper(reader), reader.close


    if compression == "bz2":
        try:
            import bz2
        except ImportError:
            raise ValueError("Please install the bz2 module for bzip2 compression support")
        if source is None:
            # bz2 doesn't support Python objects. On some platforms
            # I can work around the problem
            if not os.path.exists("/dev/stdin"):
                raise ValueError("Python's bz2 library cannot read compressed data from stdin "
                                 "on this platform")
            f = bz2.BZ2File("/dev/stdin", "rb")
        elif isinstance(source, filename_types):
            f = bz2.BZ2File(source, "rb")
        else:
            # Well, I could emulate it, but I'm not going to
            raise NotImplementedError("Python's bz2 library does not support reading from file objects")
        # TODO: add position tracker
        return f, f.close

    if compression == "xz":
        raise ValueError("chemfp does not yet support xz decompression")

    if blame_name is None:
        raise ValueError("Unsupported compression type %r" % (compression,))
    else:
        raise ValueError("%s does not support compression type %r" % (blame_name, compression))


def write_fps1_magic(outfile):
    outfile.write(b"#FPS1\n")

def write_fps1_header(outfile, metadata):
    if metadata is None:
        return 0
    lines = []
    if metadata.num_bits is not None:
        if not isinstance(metadata.num_bits, int):
            raise ValueError("metadata num_bits must be an integer")
        lines.append(b"#num_bits=%d\n" % metadata.num_bits)

    if metadata.type is not None:
        assert "\n" not in metadata.type
        lines.append(b"#type=" + metadata.type.encode("ascii")+b"\n") # type cannot contain non-ASCII characters

    if metadata.software is not None:
        assert "\n" not in metadata.software
        lines.append(b"#software=" + metadata.software.encode("utf8")+b"\n")

    if metadata.aromaticity is not None:
        lines.append(b"#aromaticity=" + metadata.aromaticity.encode("ascii") + b"\n")

    for source in metadata.sources:
        # Ignore newlines in the source filename, if present
        source = source.replace("\n", "")
        lines.append(b"#source=" + source.encode("utf8")+b"\n")

    if metadata.date is not None:
        date = metadata.date
        if not isinstance(date, str):
            date = date.isoformat().split(".", 1)[0]
        lines.append(b"#date=" + date.encode("ascii")+b"\n") # date cannot contain non-ASCII characters

    # Don't use outfile.writelines(lines) because some writers (zstandard, I'm looking at you)
    # don't implement it.
    for line in lines:
        outfile.write(line)
    return len(lines)


def write_fps1_fingerprint(outfile, fp, id):
    if "\t" in id:
        raise ValueError("Fingerprint ids must not contain a tab: %r" % (id,))
    if "\n" in id:
        raise ValueError("Fingerprint ids must not contain a newline: %r" % (id,))
    if not id:
        raise ValueError("Fingerprint ids must not be the empty string")
    
    outfile.write(binascii.hexlify(fp) + b"\t" + id.encode("utf8") + b"\n")


## def write_fps1_output(reader, destination, metadata=None):
##     if metadata is None:
##         metadata = reader.metadata
##     hexlify = binascii.hexlify
##     write_fps1_magic(outfile)
##     write_fps1_header(outfile, metadata)
##     
##     for i, (id, fp) in enumerate(reader):
##         if "\t" in id:
##             raise ValueError("Fingerprint ids must not contain a tab: %r in record %d" %
##                              (id, i+1))
##         if "\n" in id:
##             raise ValueError("Fingerprint ids must not contain a newline: %r in record %d" %
##                              (id, i+1))
##         if not id:
##             raise ValueError("Fingerprint ids must not be the empty string in record %d" %
##                              (i+1,))
##         outfile.write("%s\t%s\n" % (hexlify(fp), id))


def _compress_string(content, compression, function_name, level):
    if not compression:
        return content
    if compression == "gz":
        import gzip
        gzip_compress = getattr(gzip, "compress", None)
        if gzip_compress is not None:
            return gzip_compress(content, _get_gzip_compresslevel(level))
        f = BytesIO()
        g = gzip.GzipFile(fileobj=f, mode="wb", compresslevel=_get_gzip_compresslevel(level))
        g.write(content)
        g.close()
        return f.getvalue()        

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        bio = BytesIO()
        cctx = zstd.ZstdCompressor(_get_zst_level(level))
        writer = cctx.stream_writer(_Destination(bio))
        writer.write(content)
        writer.close()
        return bio.getvalue()

    if compression == "zlib":
        import zlib
        return zlib.compress(content, get_zlib_level(level))

    raise ValueError("%s does not support compression type %r" % (function_name, compression))


def _decompress_string(content, compression, function_name):
    if compression in (None, ""):
        return content
    if compression == "gz":
        import gzip

        gzip_decompress = getattr(gzip, "decompress", None)
        if gzip_decompress is not None:
            return gzip_decompress(content)
        bio = BytesIO(content)
        infile = gzip.GzipFile(fileobj=bio)
        return infile.read()        

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        bio = BytesIO(content)
        dctx = zstd.ZstdDecompressor()
        reader = dctx.stream_reader(bio)
        s = reader.read()
        reader.close()
        return s
        
    
    if compression == "zlib":
        import zlib
        return zlib.decompress(content)
    else:
        raise ValueError("%s does not support compression type %r" % (function_name, compression))
    


# Helper function to read from a (possibly compressed) string, as a file
def _open_compressed_input_bytes(content, compression, location, blame_name):
    if compression in (None, ""):
        f = BytesIO(content)
        register_fileobj_position(location, f, len(content))
        return f

    if compression == "gz":
        import gzip
        bio = BytesIO(content)
        infile = gzip.GzipFile(fileobj=bio)
        register_fileobj_position(location, bio, len(content))
        return infile

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        bio = BytesIO(content)
        dctx = zstd.ZstdDecompressor()
        reader = dctx.stream_reader(bio)
        register_fileobj_position(location, bio, len(content))
        return ZstdStreamWrapper(reader)
        

    if blame_name is None:
        raise ValueError("Unsupported compression %r" % (compression,))
    else:
        raise ValueError("%s does not support compression type %r" % (blame_name, compression))

# Pass the entire content as a single block.

class ByteStringFile:
    def __init__(self, content):
        self.content = content
    def read(self, buffersize=-1):
        if buffersize == 0:
            return b""
        content = self.content
        self.content = b""
        return content
    def __iter__(self):
        content = self.read()
        if content:
            return iter(content.splitlines(True))
        return iter([])
    
def _open_string_as_bytes(content, compression, location, blame_name):
    if isinstance(content, str):
        if compression:
            raise ValueError("Compression does not work when the content is a Unicode string. Use a byte string instead.")
        is_unicode = True
        content_bytes = content.encode("utf8")
    else:
        is_unicode = False
        content_bytes = content
        
    if not compression:
        #f = ByteStringFile(content_bytes) # XXX what was this for????
        f, close = open_bytes_io(content_bytes, location)
        return is_unicode, f, close
    
    infile = _open_compressed_input_bytes(content_bytes, compression, location, blame_name)
    return is_unicode, infile, infile.close

# Helper function to create a (possibly compressed) bytes output
def _open_compressed_output_bytes(compression, blame_name, level):
    encoding, encoding_errors = "utf8", "strict"
    if compression not in (None, "", "gz", "zst"):
        if blame_name is None:
            raise ValueError("Unsupported compression %r" % (compression,))
        else:
            raise ValueError("%s does not support compression type %r" % (blame_name, compression))

    if compression in (None, ""):
        bio = BytesIO()
        wrapped_bio = OutputToBytes(bio, encoding, encoding_errors)
        return wrapped_bio, None, bio.getvalue

    if compression == "gz":
        import gzip
        bio = BytesIO()
        gzout = gzip.GzipFile(fileobj=bio, mode="wb",
                              compresslevel=_get_gzip_compresslevel(level))
        wrapped_bio = OutputToBytes(gzout, encoding, encoding_errors)
        return wrapped_bio, gzout.close, bio.getvalue

    if compression == "zst":
        try:
            import zstandard as zstd
        except ImportError as err:
            raise ValueError("Please install the zstandard module for zst compression support: %s"
                                 % (err,))
        bio = BytesIO()
        cctx = zstd.ZstdCompressor(_get_zst_level(level))
        writer = cctx.stream_writer(_Destination(bio))
        wrapped_bio = OutputToBytes(ZstdStreamWrapper(writer), encoding, encoding_errors)
        return wrapped_bio, writer.close, bio.getvalue

        

    raise AssertionError(compression)

class OutputToBytes:
    def __init__(self, bio, encoding, encoding_errors):
        self.bio = bio
        self.encoding = encoding
        self.encoding_errors = encoding_errors

    def write_bytes(self, content, error_handler, location, encoding, encoding_errors):
        if encoding != self.encoding:
            # Need to transcode
            try:
                s = content.decode(encoding, encoding_errors)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the input record bytes to string: %s" % (err,),
                                    location)
                return False
            try:
                content = s.encode(self.encoding, self.encoding_errors)
            except UnicodeEncodeError as err:
                error_handler.error("Cannot transcode the record string to byte output: %s" % (err,),
                                    location)
                return
        self.bio.write(content)
        return True

    def write_string(self, content, error_handler, location):
        try:
            s = content.encode(self.encoding, self.encoding_errors)
        except UnicodeDecodeError as err:
            error_handler.error("Cannot encode the input record as bytes: %s" % (err,),
                                location)
            return False
        
        self.bio.write(s)
        return True

    def write_id_then_remainder(self, id, remainder, error_handler, location,
                                encoding, encoding_errors):
        try:
            id_s = id.encode(self.encoding, self.encoding_errors)
        except UnicodeDecodeError as err:
            error_handler.error("Cannot encode the identifier %r as bytes: %s" % (id, err),
                                location)
            return False

        if encoding != self.encoding:
            # Need to transcode
            try:
                s = remainder.decode(encoding, encoding_errors)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the input record bytes to string: %s" % (err,),
                                    location)
                return False
            try:
                remainder = s.encode(self.encoding, self.encoding_errors)
            except UnicodeEncodeError as err:
                error_handler.error("Cannot transcode the record string to byte output: %s" % (err,),
                                    location)
                return False
        
        self.bio.write(id_s + remainder)
        return True

        
    def writelines(self, lines, error_handler, location):
        encoding = self.encoding
        encoding_errors = self.encoding_errors
        try:
            self.bio.writelines(line.encode(encoding, encoding_errors) for line in lines)
        except UnicodeEncodeError as err:
            error_handler.error(
                "Failure during encoding of output lines to byte output: %s: Record is likely corrupted."
                % (err,), location)
        except NotImplementedError:
            # likely raised by zst.ZstdCompressionWriter
            write = self.bio.write
            try:
                for line in lines:
                    write(line.encode(encoding, encoding_errors))
            except UnicodeEncodeError as err:
                error_handler.error(
                    "Failure during encoding of output lines to byte output: %s: Record is likely corrupted."
                    % (err,), location)
        return True

    
class OutputToString:
    def __init__(self):
        self.sio = StringIO()

    def write_bytes(self, content, error_handler, location, encoding, encoding_errors):
        try:
            s = content.decode(encoding, encoding_errors)
        except UnicodeDecodeError as err:
            error_handler.error("Cannot decode record's bytes: %s" % (err,), location)
            return False
        self.sio.write(s)
        return True

    def write_string(self, content, error_handler, location):
        self.sio.write(content)
        return True

    def write_id_then_remainder(self, id, remainder, error_handler, location,
                                encoding, encoding_errors):
        try:
            s = remainder.decode(encoding, encoding_errors)
        except UnicodeDecodeError as err:
            error_handler.error("Cannot decode record's bytes: %s" % (err,),
                                location)
            return False
        self.sio.write(id)
        self.sio.write(s)
        return True

    def writelines(self, lines, error_handler, location):
        self.sio.writelines(lines)
        return True

    def getvalue(self):
        return self.sio.getvalue()


    
def _open_output_string():
    output = OutputToString()
    return output, None, output.getvalue

####

class ProgressBar:
    """Use a tdqm progress bar to track location progress.

    The progress shown depends on the capabilities of the
    :class:`.Location`.
    
    * If position information is available, use it, along
      with optional units and end position information.

    * Otherwise, display the record or iteration progress.

    See :meth:`.Location.progress_bar` for details about
    how to configure and create a ProgressBar. This
    class should not be called directly.
    """
    def __init__(self, tqdm, start_position, get_position):
        """This constructor is not part of the public API"""
        self.tqdm = tqdm

        if get_position is None:
            # Need to track the iterations myself
            assert start_position is None
            self._manage_own_position = True
            self._position = start_position = 0
            get_position = self._get_position
        else:
            self._manage_own_position = False
            
        self.start_position = start_position
        self._get_position = get_position
        self.current_position = start_position

    def _get_position(self):
        return self._position
    
    def update(self):
        """Update the toolbar to the current progress.

        Use this to update the progress bar manually
        when not being used as an iterator wrapper.
        
        Example of use::

          with location.progress_bar() as progress_bar:
            for i, mol in enumerate(mol_reader):
              if i % 100 == 0:
                # only update after every 100 molecules
                progress_bar.update()
        """
        position = self._get_position()
        delta = position - self.current_position
        self.current_position = position
        self.tqdm.update(delta)

    def __call__(self, it):
        """Wrap an iterator and update the toolbar after yielding each item

        Example of use::
        
            with location.progress_bar() as progress_bar:
                writer.write_fingerprints(progress_bar(reader))

        :meta public:
        """
        if self._manage_own_position:
            for i, obj in enumerate(it):
                yield obj
                self._position += 1
                self.update()
        else:
            for i, obj in enumerate(it):
                yield obj
                self.update()

    def __enter__(self):
        """Context manager to close the progress bar upon completion.

        :meta public:
        """
        return self
    
    def __exit__(self, type, value, traceback):
        """Close the progress bar when the context ends.

        :meta public:
        """
        self.update() # Ensure at least one update even on empty input
        self.close()

    def close(self):
        """Close the progress bar."""
        self.tqdm.close()

####

_where_template = {
    (0, 0, 0): "<unknown position>",
    (0, 0, 1): "record #%(recno)d",
    (0, 1, 0): "line %(lineno)d",
    (0, 1, 1): "line %(lineno)d, record #%(recno)d",
    (1, 0, 0): "file %(filename)r",
    (1, 0, 1): "file %(filename)r, record #%(recno)d",
    (1, 1, 0): "file %(filename)r, line %(lineno)d",
    (1, 1, 1): "file %(filename)r, line %(lineno)d, record #%(recno)d",
    }

class Location:
    """Get location and other internal reader and writer state information

    A Location instance gives a way to access information like
    the current record number, line number, and molecule object.

     >>> import chemfp
     >>> with chemfp.read_molecule_fingerprints("RDKit-MACCS166",
     ...                        "ChEBI_lite.sdf.gz", id_tag="ChEBI ID") as reader:
     ...   for id, fp in reader:
     ...     if id == "CHEBI:3499":
     ...         print("Record starts at line", reader.location.lineno)
     ...         print("Record byte range:", reader.location.offsets)
     ...         print("Number of atoms:", reader.location.mol.GetNumAtoms())
     ...         break
     ... 
     [08:18:12]  S group MUL ignored on line 103
     Record starts at line 3599
     Record byte range: (138171, 141791)
     Number of atoms: 36

    * :attr:`filename` - a string describing the source or destination
    * :attr:`lineno` - the line number for the start of the file
    * :attr:`mol` - the toolkit molecule for the current record
    * :attr:`offsets` - the (start, end) byte positions for the current record
    * :attr:`position` - the approximate current position in the input file.
      (This should only be used for progress information.)
    * :attr:`start_position` - the position when the location was attached
      to the file. Usually 0.
    * :attr:`end_position` - the expected final position.
    * :attr:`position_units` - a short name for how the position is measured.
      For files this usually "bytes".
    * :attr:`recno` - the current record number
    * :attr:`record` - the record as a byte string
    * :attr:`record_format` - the record format, like "sdf" or "can"
    * :attr:`output_recno` - the number of records written successfully

    Most of the readers and writers do not support all of the
    properties.  Unsupported properties return a None. The ``filename``
    is a read/write attribute and the other attributes are read-only.
    
    If you don't pass a location to the readers and writers then they
    will create a new one based on the source or destination,
    respectively.  You can also pass in your own Location, created as
    :class:`.Location` if you have an actual filename, or
    :meth:`Location.from_source` or :meth:`Location.from_destination` if
    you have a more generic source or destination.
    """
    _get_recno = None
    _get_output_recno = None
    _get_lineno = None
    _get_offsets = None
    _get_mol = None
    _get_record = None
    _get_position = None
    
    _start_position = None
    _end_position = None
    _position_units = None
    _record_format = None
    
    def __init__(self, filename=None):
        """Use *filename* as the location's filename"""
        #: A string describing the source or destination, or None (read/write)
        self.filename = filename 

    def __repr__(self):
        """Return a string like 'Location("<stdout>")'"""
        return "Location(%r)" % (self.filename,)

    def progress_bar(self, **kwargs):
        r"""Return a :class:`.ProgressBar` given the location and kwargs values.

        The :class:`.ProgressBar` is a wrapper around tqdm.  if the
        kwarg ``progress`` is True (the default) then the progress bar
        will be displayed. If False then a progress bar will not be
        used. If it is callable, then it will be called instead of
        calling the tqdm constructor. It must understand the tqdm
        constructor parameters.

        The remaining kwargs are passed to the tqdm constructor. If
        unspecified, the :meth:`.progress_bar` method will add some
        kwargs depending on what is available from the
        :class:`.Location`.

        If ``desc`` is not specified, the a default is created based on
        the optional ``file_count`` kwarg followed by the location's
        filename.

        A ProgressBar only shows the progress for a single file. When
        processing multiple files, use ``file_count`` to add information
        to the default ``desc`` about the how many files have been
        processed and how many are going to be processed.

        If ``file_count`` is not None then it must be a 2-element tuple
        containing the current file index (starting at 0) and the number
        of files to process. The number of files to process may be None
        if that number if not known. The ``file_count`` (1, 5) results
        in the description "(2/5)" and the ``file_count`` (1, None) results
        in the description "(2)".

        The progress bar will show one of several types of progress,
        depending on what information is available from the location.

        If it's possible to get a ``position`` (typically the byte
        position in the input file or the compressed input file) then
        that is used. The tqdbm ``units`` is set if the location
        has a ``position_units`` and, if ``position_units`` is
        "bytes" then tqdm's ``unit_scale`` is set to 1. The tqdm
        ``initial`` is set to the current position.
        
        Otherwise, if it's possible to get a `recno` then it is used for
        the progress information, with tqdm ``units`` set to "\ recs"
        and the tqdm ``initial`` is set to the current record number
        (which is likely 0).

        Otherwise, the tqdm ``units`` will be "\ its", which is the tqdm
        default.

        The tqdm ```total`` are set to the location's ``end_position``,
        if available. This must only be available when it's possible
        to get the current ``position``.

        The tqdm ``dynamic_ncols`` is set to True, to allow
        terminal resizing to update the tqdm output.

        Again, if you set the values in the kwargs then they will
        not be overwritten but will be passed directly to the tqdm
        constructor (or to the ``progress`` callable).

        :param progress: True to show a progress bar, False to not show one, 
           or a callable to use instead of the tqdm constructor
        :type progress: boolean or callable
        :param file_count: optional file index and number of files, used as the start of 'desc'
        :type file_count: None, or (i, None), or (i, N)
        :param \**kwargs: kwargs passed to the tqdm constructor or progress callable.
        :returns: a :class:`.ProgressBar`
        """
        from . import _tqdm
        # The 'progress' kwarg may specify an alternate tqdm-like constructor.
        progress = kwargs.pop("progress", True)
        tqdm_constructor = _tqdm.get_tqdm_constructor(progress)
        
        # Set a kwargs, only if it's not present or None
        def set(name, value):
            if kwargs.get(name, None) is None:
                kwargs[name] = value

        # Use the filename as the description
        desc = self.filename
        if desc is None:
            # Unknown filename, probably from a file-like object
            desc = ""
        file_count = kwargs.pop("file_count", None)
        if file_count is not None:
            file_i, file_N = file_count
            if file_N is None:
                prefix = f"({file_i+1}) "
            elif file_N == 1:
                prefix = ""
            else:
                prefix = f"({file_i+1}/{file_N}) "
            desc = prefix + desc
        set("desc", desc)

        # Allow dynamic resizing
        set("dynamic_ncols", True)

        get_position = self._get_position
        if get_position is None:
            # No way to get size. Can we get a record count?
            get_position = self._get_recno
            if get_position is None:
                # Nope. Will update a counter for each update()
                # This is the tqdm default
                set("unit", " its")
            else:
                # Report the number of records
                set("unit", " recs")
        else:
            # Use the appropriate units
            position_units = self.position_units
            if position_units is not None:
                if self.position_units == "bytes":
                    set("unit", "bytes")
                    # Use k/M/G
                    set("unit_scale", 1)
                else:
                    set("unit", " " + self.position_units)

        # If there is an estimated end position, then use it
        set("total", self.end_position)

        # Where are we?
        if get_position is None:
            start_position = None
        else:
            start_position = self.start_position
            if start_position is None:
                start_position = get_position()
                set("initial", start_position)
        
        # Start
        tqdm = tqdm_constructor(**kwargs)
        return ProgressBar(tqdm, start_position, get_position)
        
    def where(self):
        """Return a human readable description about the current reader or writer state.

        The description will contain the filename, line number, record
        number, and up to the first 40 characters of the first line of
        the record, if those properties are available.
        """
        filename = self.filename
        lineno = self.lineno
        recno = self.recno

        template = _where_template[ (filename is not None, lineno is not None, recno is not None) ]
        s = template % {"filename": filename, "lineno": lineno, "recno": recno}

        first_line = self.first_line

        if first_line:  # Don't show None and don't show blank lines
            if len(first_line) > 40:
                t = myrepr(first_line[:40])
                t = t[:-1] + " ..." + t[-1]
                s += ": first line starts %s" % (t,)
            else:
                s += ": first line is %s" % (myrepr(first_line),)
        return s

    @classmethod
    def from_source(cls, source):
        """Create a Location instance based on the source

        If *source* is a string then it's used as the filename.
        If *source* is None then the location filename is "<stdin>".
        If *source* is a file object then its ``name`` attribute
        is used as the filename, or None if there is no attribute.
        """
        if source is None:
            return cls("<stdin>")
        if isinstance(source, filename_types):
            return cls(source)
        return cls(getattr(source, "name", None))

    @classmethod
    def from_destination(cls, destination):
        """Create a Location instance based on the destination
        
        If *destination* is a string then it's used as the filename.
        If *destination* is None then the location filename is "<stdout>".
        If *destination* is a file object then its ``name`` attribute
        is used as the filename, or None if there is no attribute.
        """
        if destination is None:
            return cls("<stdout>")
        if isinstance(destination, filename_types):
            return cls(destination)
        return cls(getattr(destination, "name", None))


    def clear_registry(self):
        """(Do not use) Remove all registered handlers.
        
        Part of the internal API, and subject to change.
        
        :meta private:
        """
        self._get_recno = None
        self._get_output_recno = None
        self._get_lineno = None
        self._get_offsets = None
        self._get_mol = None
        self._get_record = None
        self._get_position = None

    def register(self, **kwargs):
        """(Do not use) Used by the reader or writer to add property callback handlers.

        Part of the internal API, and subject to change.

        :meta private:
        """
        for k, v in kwargs.items():
            if k in ("get_recno", "get_output_recno", "get_lineno", "get_offsets",
                         "get_mol", "get_record", "get_position"):
                setattr(self, "_" + k, v)
            else:
                raise KeyError(k)

    def save(self, **kwargs):
        """(Do not use) Used by the reader or writer to specify final values.

        Part of the internal API, and subject to change.

        :meta private:
        """
        for k, v in kwargs.items():
            if k in ("recno", "output_recno", "lineno", "offsets", "mol", "record", "position"):
                def recall_value(value=v):
                    return value
                setattr(self, "_get_" + k, recall_value)
            elif k  == "record_format":
                self._record_format = v
            elif k  == "start_position":
                self._start_position = v
            elif k  == "end_position":
                self._end_position = v
            elif k  == "position_units":
                self._position_units = v
            else:
                raise KeyError(k)

    def end(self):
        """(Do not use) Save the current values of any handlers and remove the handlers.

        Part of the internal API, and subject to change.

        :meta private:
        """
        # Save anything which wasn't saved
        for k in ("recno", "output_recno", "lineno", "offsets", "mol", "record", "position"):
            f = getattr(self, "_get_" + k)
            if f is None:
                continue
            def recall_value(value=f()):
                return value
            setattr(self, "_get_" + k, recall_value)
            
    def get_registry(self):
        """(Do not use) Get the current registry as a dict.

        Part of the internal API, and subject to change.

        :meta private:
        """
        return {"get_recno": self._get_recno,
                "get_output_recno": self._get_output_recno,
                "get_lineno": self._get_lineno,
                "get_offsets": self._get_offsets,
                "get_mol": self._get_mol,
                "get_record": self._get_record,
                "get_position": self._get_position,
                }

    @property
    def recno(self):
        """The current record number

        For writers this is the number of records sent to
        the writer, and output_recno is the number of records
        sucessfully written to the file or string.
        """
        _get_recno = self._get_recno
        if _get_recno is None:
            return None
        return _get_recno()

    @property
    def output_recno(self):
        """The number of records actually written to the file or string.

        The value ``recno - output_recno`` is the number of records
        sent to the writer but which had an error and could not be
        written to the output.
        """
        _get_output_recno = self._get_output_recno
        if _get_output_recno is None:
            return None
        return _get_output_recno()

    @property
    def lineno(self):
        """The current line number, starting from 1.

        This will be 0 if file processing has not yet started.
        """
        _get_lineno = self._get_lineno
        if _get_lineno is None:
            return None
        return _get_lineno()

    @property
    def offsets(self):
        """The (start, end) byte offsets, starting from 0.

        *start* is the record start byte position and *end* is
        one byte past the last byte of the record.
        """
        _get_offsets = self._get_offsets
        if _get_offsets is None:
            return None
        return _get_offsets()

    @property
    def mol(self):
        """The molecule object for the current record."""
        _get_mol = self._get_mol
        if _get_mol is None:
            return None
        return _get_mol()

    @property
    def record(self):
        """The current record as an uncompressed byte string."""
        _get_record = self._get_record
        if _get_record is None:
            return None
        return _get_record()

    @property
    def record_format(self):
        """The name of the record format.

        This is a string like "smi", "fps", or "sdf", 
        without any compression suffix.
        """
        return self._record_format

    @property
    def position(self):
        """The (approximate) current position in the file.

        The `position`, along with `end_position` should only be
        used for progress information. They do not correspond to
        the start or end position of the current record, nor
        to the sum of the record sizes.
        
        For example, for compressed files this may be be the location
        of the end of the most recently read compressed data.

        Several records may have the same `position`, and `position`
        may equal `end_position` even if more records are available.
        
        The position is measured in `position_units` units, typically
        "bytes" for input readers, and "records" for output writers.
        """
        _get_position = self._get_position
        if _get_position is None:
            return None
        return _get_position()

    @property
    def start_position(self):
        """The position when the location was attached to the file

        The `start_position` and `end_position`, along with
        `position` are meant for progress information and are not
        necessarily the number of bytes in the current file.

        For input files this is typically 0.
        
        The position is measured in `position_units` units, typically
        "bytes" for input readers, and "records" for output writers.
        """
        return self._start_position

    @property
    def end_position(self):
        """The (expected) end position in the file

        The `end_position`, along with `position` is meant for
        progress information and is not necessarily the number of
        bytes in the current file.

        For input files this is typically the file size, if available.
        Note that the actual file size may change while the file is
        being processed, in which case `end_position` will be
        invalid.

        The position is measured in `position_units` units, typically
        "bytes" for input readers, and "records" for output writers.
        """
        return self._end_position

    @property
    def position_units(self):
        """The units used to measure `position` and `end_position`

        This is typically "bytes" for input readers and "records" for
        output readers.
        """
        return self._position_units


    @property
    def first_line(self):
        """The first line of the current record, as a string.

        The newline and any preceeding control return characters
        are not included.

        The first line is decoded as UTF-8, with unknown
        characters replaced with '?'. Use :attr:`.first_line_bytes`
        if this is not appropriate.
        """
        _get_record = self._get_record
        if _get_record is None:
            return None
        record = _get_record()
        if record is None:
            return None
        first_line, _, _ = record.partition(b"\n")
        return first_line.rstrip(b"\r").decode("utf8", errors="replace")

    @property
    def first_line_bytes(self):
        """The first line of the current record, as a byte string.
        
        The newline and any preceeding control return characters
        are not included.

        If you want a text/Unicode string and the input record
        is UTF-8 encoded then use :attr:`.first_line`.
        """
        _get_record = self._get_record
        if _get_record is None:
            return None
        record = _get_record()
        if record is None:
            return None
        first_line, _, _ = record.partition(b"\n")
        return first_line.rstrip(b"\r")
        

    def copy(self):
        """(Do not use) Return a shallow copy of this location

        Part of the internal API, and subject to change.

        :meta private:
        """
        loc = Location(self.filename)
        if self._record_format is not None:
            loc.save(record_format=self._record_format)
        if self._start_position is not None:
            loc.save(start_position=self._start_position)
        if self._end_position is not None:
            loc.save(end_position=self._end_position)
        if self._position_units is not None:
            loc.save(position_units=self._position_units)

        d = {}
        for get_func_name, get_func in self.get_registry().items():
            if get_func is None:
                continue
            assert get_func_name.startswith("get_"), get_func_name
            name = get_func_name[4:]
            value = get_func()
            d[name] = value
        loc.save(**d)
        return loc
