# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""The chemfp toolkit API wrapper for the Open Babel toolkit.

This module is also available as ``chemfp.openbabel``.

"""

import openbabel  # Open Babel must be installed to use chemfp.openbabel_toolkit

if hasattr(openbabel, "sphinx_mock"):
    # This should only happen when we're building the documentation using
    # Sphinx's autodoc with a mock for Open Babel.  Sphinx's magic mock isn't
    # good enough so provide some concrete mocks which are just barely good
    # enough to let autodoc import the module to extract the docstrings.
    class _tk:
        SOFTWARE = "OpenBabel (using sphinx)"
        class format_registry:
            @staticmethod
            def get_format(format_name):
                return None
        toolkit = None
    class _ob:
        PairData = None
        CommentData = None
else:
    # We have a real Open Babel installation.
    from . import _openbabel_toolkit as _tk
    if getattr(openbabel, "__version__", "2.0")[:2] == "2.":
        # TODO: remove support for Open Babel 2.0
        # (3.0.0 was released Oct 10, 2019.)
        import openbabel as _ob
    else:
        import openbabel.openbabel as _ob
del openbabel

__all__ = []
def _public(f):
    __all__.append(f.__name__)
    return f

name = "openbabel"
software = _tk.SOFTWARE
is_available = True

@_public
def is_licensed():
    """Return True - Open Babel is always licensed

    :returns: True
    """
    return True

_format_registry = _tk.format_registry
_toolkit = _tk.toolkit

# format API

@_public
def get_formats(include_unavailable=False):
    """Get the list of structure formats that Open Babel supports

    If *include_unavailable* is True then also include Open Babel formats
    which aren't available to this specific version of Open Babel.

    :param include_unavailable: include unavailable formats?
    :type include_unavailable: True or False
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_formats(include_unavailable)

@_public
def get_input_formats():
    """Get the list of supported Open Babel input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_formats()

@_public
def get_output_formats():
    """Get the list of supported Open Babel output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_formats()

@_public
def get_format(format_name):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if Open Babel does not implement
    the format *format_name* or that format is not available.

    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_format(format_name)

@_public
def get_input_format(format_name):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError if Open Babel does not implement
    the format *format_name* or that format is not an input format.

    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format(format_name)

@_public
def get_output_format(format_name):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if Open Babel does not implement
    the format *format_name* or that format is not an output format.
    
    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format(format_name)

@_public
def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param source: the structure data source.
    :type source: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format_from_source(source, format)

@_public
def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: the structure data source.
    :type destination: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format_from_destination(destination, format)

# read/write API

@_public
def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads OBMol molecules from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Note: the reader will clear and reuse the OBMol instance. Make a copy
    if you want to keep the molecule around.

    The *reader_args* dictionary parameters depend on the format. Every
    Open Babel format supports an "options" entry, which is passed to
    SetOptions(). See that documentation for details. Some formats support
    additional parameters:

    * SMILES and InChI
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * has_header - True or False

    * SDF
    
      * implementation - if "openbabel" or None, use the Open Babel record parser;
        if "chemfp", use chemfp's own record parser, which has better location tracking

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If None
    then a default Location will be created.

    See :func:`chemfp.openbabel_toolkit.read_ids_and_molecules` if you want
    (id, OBMol) pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OBMol molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location,
                                   encoding, encoding_errors)

@_public
def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads OBMol molecules from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.openbabel_toolkit.read_molecules` for details about the other
    parameters.  See :func:`chemfp.openbabel_toolkit.read_ids_and_molecules_from_string`
    if you  want to read (id, OBMol) pairs instead of just molecules.

    Note: the reader will clear and reuse the OBMol instance. Make a
    copy if you want to keep the molecule around.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OBMol molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

@_public
def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None,
                           encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, OBMol molecule) pairs from a structure file

    See :func:`chemfp.openbabel_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, OBMol)
    pairs instead of just the molecules.
    
    Note: the reader will clear and reuse the OBMol instance. Make a
    copy if you want to keep the molecule around.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, OBMol) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location,
                                           encoding, encoding_errors)

@_public
def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, OBMol) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.openbabel_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.openbabel_toolkit.read_molecules_from_string`
    if you just want to read the OBMol molecules instead of (id, OBMol) pairs.
    
    Note: the reader will clear and reuse the OBMol instance. Make a
    copy if you want to keep the molecule around.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, OBMol) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

@_public
def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, OBMol) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. The function
    will reuse the OBMol for successive calls, so make a copy if you want
    to keep it around. However, I haven't really noticed much of a performance
    difference between this and :func:`chemfp.openbabel_toolkit.parse_id_and_molecule`
    so I suggest you use that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.openbabel_toolkit.read_molecules` for details about the
    other parameters.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, OBMol)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

@_public
def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return an OBMol molecule.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.openbabel_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.openbabel_toolkit.parse_id_and_molecule`
    if you want the (id, OBMol) pair instead of just the molecule.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an OBMol molecule
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

@_public
def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, OBMol) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.openbabel_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.openbabel_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.openbabel_toolkit.parse_molecule`
    if just want the OBMol molecule and not the the (id, OBMol) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an (id, OBMol molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)


@_public
def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert an OBMol into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an Open Babel molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a Unicode string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

@_public
def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert an OBMol into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an Open Babel molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a byte string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)


@_public
def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write Open Babel molecules to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    an OBMol molecule, an OBMol molecule iterator, or an (id, OBMol molecule) pair
    iterator to a file.

    Molecules are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.

    The *writer_args* dictionary parameters depend on the format. Every
    format supports an ``options`` entry, which is passed to Open Babel's
    ``SetOptions()``. See the Open Babel documentation for details. Some
    formats supports additional parameters:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * isomeric - True to write isomeric SMILES, False or default is non-isomeric
      * canonicalization - True, "default", or None uses Open Babel's own canonicalization
        algorithm; False or "none" to use no canonicalization; "universal" generates a
        universal SMILES; "anticanonical" generates a SMILES with randomly assigned
        atom classes; "inchified" uses InChI-fied SMILES

    * InChI and InChIKey
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * include_id - True or default to include the id as the second column; False has no id column

    * SDF
    
      * always_v3000 - True to always write V3000 files; False or default to
        write V3000 files only if needed.
      * include_atom_class - True to include atom class; False or default does not
      * include_hcount - True to include hcount; False or default does not

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats (does not affect Open Babel)
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

@_public
def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write Open Babel molecule records to a string.

    See :func:`chemfp.openbabel_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting Open Babel molecules
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)


@_public
def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write Open Babel molecule records to a byte string

    See :func:`chemfp.openbabel_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats (does not affect Open Babel)
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting Open Babel molecules
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location, level)


##### Molecule API

@_public
def copy_molecule(mol):
    """Return a new OBMol molecule which is a copy of the given Open Babel molecule

    :param mol: the molecule to copy
    :type mol: an Open Babel molecule
    :returns: a new OBMol instance
    """
    return _ob.OBMol(mol)


# I don't know when one should use PairData vs. CommentData.
# However, this is what Noel O'Boyle recommends doing in pybel.
_PairData = _ob.PairData
_CommentData = _ob.CommentData

@_public
def add_tag(mol, tag, value):
    """Add an SD tag value to the Open Babel molecule

    Raises a KeyError if the tag is a special internal Open Babel name.

    :param mol: the molecule
    :type mol: an Open Babel molecule
    :param tag: the SD tag name
    :type tag: string
    :param value: the text for the tag
    :type value: string
    :returns: None
    """
    # Wow. OB's Python interface doesn't make this easy.
    # I take a different route than pybel. I wonder if it's
    # equivalent, or if it's worse.
    tag = _tk._to_openbabel_string(tag)
    value = _tk._to_openbabel_string(value)
    if mol.HasData(tag):
        # If the tag already exists then modify it.
        generic_data = mol.GetData(tag)
        if generic_data.GetDataType() in (_PairData, _CommentData):
            _ob.toPairData(generic_data).SetValue(value)  # Only string values are allowed
            return
        else:
            # This is a special internal name, which I probably can't
            # delete without doing serious harm.
            raise KeyError("Unable to add tag %r: it appears to be a special internal Open Babel name"
                           % (tag,))
        
    # Otherwise, create new entry
    pair_data = _ob.OBPairData()
    pair_data.SetAttribute(tag)
    pair_data.SetValue(value) # Only string values are allowed

    mol.CloneData(pair_data)

@_public
def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    :param mol: the molecule
    :type mol: an Open Babel molecule
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    obj = mol.GetData(_tk._to_openbabel_string(tag))
    if obj is None:
        return None
    return obj.GetValue()


@_public
def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the molecule

    :param mol: the molecule
    :type mol: an Open Babel molecule
    :returns: a list of (string name, string value) pairs
    """
    values = []
    for generic_data in mol.GetData():
        if generic_data.GetDataType() in (_PairData, _CommentData):
            values.append( (generic_data.GetAttribute(),
                            generic_data.GetValue()) )

    # Note: older versions of Open Babel include "OpenBabel Symmetry Classes"
    # as part of the user-data pairs.
    return values

@_public
def get_id(mol):
    """Get the molecule's id using Open Babel's GetTitle()

    :param mol: the molecule
    :type mol: an Open Babel molecule
    :returns: a string
    """
    return mol.GetTitle()

@_public
def set_id(mol, id):
    """Set the molecule's id using Open Babel's SetTitle()

    :param mol: the molecule
    :type mol: an Open Babel molecule
    :param id: the new id
    :type id: string
    :returns: None
    """
    mol.SetTitle(_tk._to_openbabel_string(id))

# accessor aliases for the default fingerprint type object
    
_aliases = {
    "ecfp0": "OpenBabel-ECFP0",
    "ecfp2": "OpenBabel-ECFP2",
    "ecfp4": "OpenBabel-ECFP4",
    "ecfp6": "OpenBabel-ECFP6",
    "ecfp8": "OpenBabel-ECFP8",
    "ecfp10": "OpenBabel-ECFP10",
    "fp2": "OpenBabel-FP2",
    "fp3": "OpenBabel-FP3",
    "fp4": "OpenBabel-FP4",
    "maccs": "OpenBabel-MACCS",
    }
    
def __getattr__(name):
    full_name = _aliases.get(name, None)
    if full_name is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")

    from . import get_fingerprint_type
    return get_fingerprint_type(full_name)

def __dir__():
    return sorted(__all__ + list(_aliases))


#### The following was auto-generated by: utils/mol_io_codegen.py openbabel

from . import io as _io
from ._compat import (
    Any as _Any,
    BinaryIO as _BinaryIO,
    Literal as _Literal,
    Optional as _Optional,
    Union as _Union,
)


_smistring_fmt = get_format("smistring")


@_public
def from_smistring(
    content: str,
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse a SMILES string using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "smistring", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: _Optional[str] = None,
    isomeric: bool = True,
    canonicalization: _Literal[
        True,
        "1",
        "True",
        "true",
        "default",
        "0",
        False,
        "False",
        "false",
        "none",
        "universal",
        "inchified",
        "anticanonical",
    ] = True,
    explicit_hydrogens: bool = False,
    errors: str = "strict",
):
    """Generate a SMILES string from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "smistring", id=id, writer_args={...}, errors=errors)

    The 'canonicalization' value can be one of:
        True, '1', 'True', 'true', or 'default' for standard canonicalization
        False, '0', 'False', 'false' or 'none' for no canonicalization
        'universal' for Universal SMILES (based on the InChI numbering)
        'inchified' for the InChI numbering
        'anticanonical' for a randomly selected initial numbering

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param isomeric: if true, generate isomeric SMILES
    :type isomeric: Boolean (default: True)
    :param canonicalization: canonicalization method
    :type canonicalization: Boolean or string (default: True)
    :param explicit_hydrogens: if true, use explicit hydrogens as [H]
    :type explicit_hydrogens: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "isomeric": isomeric,
            "canonicalization": canonicalization,
            "explicit_hydrogens": explicit_hydrogens,
        },
        error_handler=error_handler,
    )


_smi_fmt = get_format("smi")


@_public
def from_smi(
    content: str,
    *,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    has_header: bool = False,
    errors: str = "strict",
):
    """Parse a SMILES string and id using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "smi", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
            "delimiter": delimiter,
            "has_header": has_header,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: _Optional[str] = None,
    isomeric: bool = True,
    canonicalization: _Literal[
        True,
        "1",
        "True",
        "true",
        "default",
        "0",
        False,
        "False",
        "false",
        "none",
        "universal",
        "inchified",
        "anticanonical",
    ] = "default",
    explicit_hydrogens: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate a SMILES string and id from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "smi", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param isomeric: if true, generate isomeric SMILES
    :type isomeric: Boolean (default: True)
    :param canonicalization: canonicalization method
    :type canonicalization: Boolean or string (default: "default")
    :param explicit_hydrogens: if true, use explicit hydrogens as [H]
    :type explicit_hydrogens: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "isomeric": isomeric,
            "canonicalization": canonicalization,
            "explicit_hydrogens": explicit_hydrogens,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )


@_public
def from_smi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    has_header: bool = False,
    errors: str = "strict",
):
    """Parse a SMILES string and id file using the Open Babel toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "smi", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating Open Babel molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "smi"),
        reader_args={
            "options": options,
            "delimiter": delimiter,
            "has_header": has_header,
        },
        errors=errors,
    )


@_public
def to_smi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    isomeric: bool = True,
    canonicalization: _Literal[
        True,
        "1",
        "True",
        "true",
        "default",
        "0",
        False,
        "False",
        "false",
        "none",
        "universal",
        "inchified",
        "anticanonical",
    ] = "default",
    explicit_hydrogens: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate a SMILES string and id from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "smi", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param options: options string passed to Open Babel
    :type options: string or None
    :param isomeric: if true, generate isomeric SMILES
    :type isomeric: Boolean (default: True)
    :param canonicalization: canonicalization method
    :type canonicalization: Boolean or string (default: "default")
    :param explicit_hydrogens: if true, use explicit hydrogens as [H]
    :type explicit_hydrogens: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "smi"),
        writer_args={
            "options": options,
            "isomeric": isomeric,
            "canonicalization": canonicalization,
            "explicit_hydrogens": explicit_hydrogens,
            "delimiter": delimiter,
        },
        errors=errors,
    )


_sdf_fmt = get_format("sdf")


@_public
def from_sdf(
    content: str,
    *,
    implementation: _Literal[None, "openbabel", "chemfp"] = "openbabel",
    perceive_stereo: bool = False,
    perceive_0d_stereo: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse an SDF record using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "sdf", reader_args={...}, errors=errors)

    :param implementation: SDF record tokenizer implementation
    :type implementation: None or 'openbabel' to use Open Babel, else 'chemfp' (default: "openbabel")
    :param perceive_stereo: not implemented
    :type perceive_stereo: Boolean (default: False)
    :param perceive_0d_stereo: not implemented
    :type perceive_0d_stereo: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sdf_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "implementation": implementation,
            "perceive_stereo": perceive_stereo,
            "perceive_0d_stereo": perceive_0d_stereo,
            "options": options,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sdf(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    always_v3000: bool = False,
    include_atom_class: bool = False,
    include_hcount: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an SDF record from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "sdf", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param always_v3000: if true, always save records in V3000 format
    :type always_v3000: Boolean (default: False)
    :param include_atom_class: if true, include atom classes in the output
    :type include_atom_class: Boolean (default: False)
    :param include_hcount: if true, always include hydrogen counts in the output
    :type include_hcount: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "always_v3000": always_v3000,
            "include_atom_class": include_atom_class,
            "include_hcount": include_hcount,
            "options": options,
        },
        error_handler=error_handler,
    )


@_public
def from_sdf_file(
    source: _Union[None, str, _BinaryIO],
    *,
    implementation: _Literal[None, "openbabel", "chemfp"] = "openbabel",
    perceive_stereo: bool = False,
    perceive_0d_stereo: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse an SDF record file using the Open Babel toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "sdf", reader_args={...}, errors=errors)

    :param implementation: SDF record tokenizer implementation
    :type implementation: None or 'openbabel' to use Open Babel, else 'chemfp' (default: "openbabel")
    :param perceive_stereo: not implemented
    :type perceive_stereo: Boolean (default: False)
    :param perceive_0d_stereo: not implemented
    :type perceive_0d_stereo: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating Open Babel molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "sdf"),
        reader_args={
            "implementation": implementation,
            "perceive_stereo": perceive_stereo,
            "perceive_0d_stereo": perceive_0d_stereo,
            "options": options,
        },
        errors=errors,
    )


@_public
def to_sdf_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    always_v3000: bool = False,
    include_atom_class: bool = False,
    include_hcount: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an SDF record from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param always_v3000: if true, always save records in V3000 format
    :type always_v3000: Boolean (default: False)
    :param include_atom_class: if true, include atom classes in the output
    :type include_atom_class: Boolean (default: False)
    :param include_hcount: if true, always include hydrogen counts in the output
    :type include_hcount: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf"),
        writer_args={
            "always_v3000": always_v3000,
            "include_atom_class": include_atom_class,
            "include_hcount": include_hcount,
            "options": options,
        },
        errors=errors,
    )


_sdf3k_fmt = get_format("sdf3k")


@_public
def to_sdf3k(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    always_v3000: bool = True,
    include_atom_class: bool = False,
    include_hcount: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "sdf3k", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param always_v3000: if true, always save records in V3000 format
    :type always_v3000: Boolean (default: True)
    :param include_atom_class: if true, include atom classes in the output
    :type include_atom_class: Boolean (default: False)
    :param include_hcount: if true, always include hydrogen counts in the output
    :type include_hcount: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf3k_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "always_v3000": always_v3000,
            "include_atom_class": include_atom_class,
            "include_hcount": include_hcount,
            "options": options,
        },
        error_handler=error_handler,
    )


@_public
def to_sdf3k_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    always_v3000: bool = True,
    include_atom_class: bool = False,
    include_hcount: bool = False,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf3k", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param always_v3000: if true, always save records in V3000 format
    :type always_v3000: Boolean (default: True)
    :param include_atom_class: if true, include atom classes in the output
    :type include_atom_class: Boolean (default: False)
    :param include_hcount: if true, always include hydrogen counts in the output
    :type include_hcount: Boolean (default: False)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf3k"),
        writer_args={
            "always_v3000": always_v3000,
            "include_atom_class": include_atom_class,
            "include_hcount": include_hcount,
            "options": options,
        },
        errors=errors,
    )


_fasta_fmt = get_format("fasta")


@_public
def from_fasta(
    content: str,
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse a FASTA record using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "fasta", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _fasta_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_fasta(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a FASTA record from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "fasta", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _fasta_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
        },
        error_handler=error_handler,
    )


@_public
def from_fasta_file(
    source: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse a FASTA record file using the Open Babel toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "fasta", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating Open Babel molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "fasta"),
        reader_args={
            "options": options,
        },
        errors=errors,
    )


@_public
def to_fasta_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a FASTA record from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "fasta", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "fasta"),
        writer_args={
            "options": options,
        },
        errors=errors,
    )


_pdb_fmt = get_format("pdb")


@_public
def from_pdb(
    content: str,
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse a PDB record using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "pdb", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _pdb_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_pdb(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a PDB record from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "pdb", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _pdb_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
        },
        error_handler=error_handler,
    )


@_public
def from_pdb_file(
    source: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse a PDB record file using the Open Babel toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "pdb", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating Open Babel molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "pdb"),
        reader_args={
            "options": options,
        },
        errors=errors,
    )


@_public
def to_pdb_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a PDB record from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "pdb", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "pdb"),
        writer_args={
            "options": options,
        },
        errors=errors,
    )


_inchi_fmt = get_format("inchi")


@_public
def from_inchi(
    content: str,
    *,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchi", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "inchi", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def from_inchi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id file using the Open Babel toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "inchi", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating Open Babel molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "inchi"),
        reader_args={
            "options": options,
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_inchi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchi", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchi"),
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchistring_fmt = get_format("inchistring")


@_public
def from_inchistring(
    content: str,
    *,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Parse an InChI string using the Open Babel toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchistring", reader_args={...}, errors=errors)

    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "options": options,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an InChI string from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "inchistring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
        },
        error_handler=error_handler,
    )


_inchikey_fmt = get_format("inchikey")


@_public
def to_inchikey(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "inchikey", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikey_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def to_inchikey_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an Open Babel molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchikey", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting Open Babel molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchikey"),
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchikeystring_fmt = get_format("inchikeystring")


@_public
def to_inchikeystring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    FixedHLayer: bool = False,
    ReconnectedMetals: bool = False,
    truncspec: bool = None,
    options: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an InChIKey string from an Open Babel molecule

    This is equivalent to calling:
       create_string(mol, "inchikeystring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an Open Babel molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param FixedHLayer: if true, include the fixed hydrogen layer
    :type FixedHLayer: Boolean (default: False)
    :param ReconnectedMetals: if true, reconnect metals
    :type ReconnectedMetals: Boolean (default: False)
    :param truncspec: if true, truncate the InChI
    :type truncspec: Boolean (default: None)
    :param options: options string passed to Open Babel
    :type options: string or None
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an Open Babel molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikeystring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "FixedHLayer": FixedHLayer,
            "ReconnectedMetals": ReconnectedMetals,
            "truncspec": truncspec,
            "options": options,
        },
        error_handler=error_handler,
    )


# Aliases

from_smiles = from_smistring

to_smiles = to_smistring

#### End of auto-generated code.
