# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

# OpenBabel fingerprints are stored as vector<unsigned int>.  On all
# the machines I use, ints have 32 bits.

# OpenBabel bit lengths must be at least sizeof(int)*8 bits long and
# must be a factor of two. I have no idea why this is required.

# OpenBabel supports new fingerprints through a plugin system.  I got
# it working thanks to Noel O'Boyle's excellent work with Cinfony. I
# then found out that the OB API doesn't have any way to get the
# number of bits in the fingerprint. The size is rounded up to the
# next power of two, so FP4 (307 bits) needs 512 bits (16 ints)
# instead of 320 bits (10 ints). That means I can't even get close to
# guessing the bitsize.

# In the end, I hard-coded the supported fingerprints into the system.

import struct

from ._compat import Any, Optional, Union, Literal

from . import types
from . import openbabel_toolkit
from . import _openbabel_toolkit

import openbabel
# If this attribute exists then we're in a Sphinx build environment
_using_sphinx_mock = hasattr(openbabel, "using_sphinx_mock")

if _using_sphinx_mock:
    import openbabel as ob
else:
    if getattr(openbabel, "__version__", "2.0")[:2] == "2.":
        import openbabel as ob
    else:
        import openbabel.openbabel as ob
    del openbabel

    
SOFTWARE = openbabel_toolkit.software
from . import SOFTWARE as CHEMFP_SOFTWARE

# This is a "standard" size according to the struct module
# documentation, so the following is an excess of caution
if struct.calcsize("<I") != 4:
    raise AssertionError("The chemfp.ob module assumes 32 bit integers")

__all__ = []
def _public(obj):
    __all__.append(obj.__name__)
    return obj

############

# I could have written a more general function which created these but
# there's only a few fingerprints lengths to worry about.

# This needs 128 bytes, for 1024 bits
# vectorUnsignedInt will contain 32 32-bit words = 1024 bits

_ob_get_fingerprint = {}
ECFP_AVAILABLE = False # Meaning, "available in Open Babel"
HAS_ECFP = False  # Meaning, supported by chemfp (only for Open Babel 3.0 and later)
def _init():
    names = ("FP2", "FP3", "FP4", "MACCS", "ECFP0", "ECFP2", "ECFP4", "ECFP6", "ECFP8", "ECFP10")
    if _using_sphinx_mock:
        for name in names:
            _ob_get_fingerprint[name] = (None, None)
        return
    # 
    ob.OBConversion()
    for name in names:
        ob_fingerprinter = ob.OBFingerprint.FindFingerprint(name)
        if ob_fingerprinter is None:
            _ob_get_fingerprint[name] = (None, None)
        else:
            _ob_get_fingerprint[name] = (ob_fingerprinter, ob_fingerprinter.GetFingerprint)

    global ECFP_AVAILABLE, HAS_ECFP
    if _ob_get_fingerprint["ECFP2"][0] is not None:
        ECFP_AVAILABLE = True
        if ob.OBReleaseVersion()[:2] != "2.":
            HAS_ECFP = True
            
    if _ob_get_fingerprint["FP2"][0] is None:
        raise ImportError("Unable to load OpenBabel FP2 fingerprinter. Check $BABEL_LIBDIR")
    n = _ob_get_fingerprint["FP2"][0].Getbitsperint()
    if n != 32:
        raise AssertionError(
            "The chemfp.openbabel_types module assumes Open Babel fingerprints have 32 bit integers")
            
_init()

### TODO: Why do I have calc_FP2, calc_FP3, and calc_FP4?
# They were part of the original API. The calc_MACCS() is used for
# fingerprint detection. But they aren't otherwise used.

_pack_1024 = struct.Struct("<" + "I"*32).pack
## def calc_FP2(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP2"][1],
##              _pack_1024 = _pack_1024):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_1024(*fp)

## # This needs 7 bytes, for 56 bits.
## # vectorUnsignedInt will contain 2 32-bit words = 64 bits
_pack_64 = struct.Struct("<II").pack
## def calc_FP3(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP3"][1],
##              _pack_64 = _pack_64):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_64(*fp)[:7]

## # This needs 39 bytes, for 312 bits
## # vectorUnsignedInt will contain 16 32-bit words = 512 bits
_pack_512 = struct.Struct("<" + "I"*16).pack
## def calc_FP4(mol, fp=None,
##              get_fingerprint=_ob_get_fingerprint["FP4"][1],
##              _pack_512 = _pack_512):
##     if fp is None:
##         fp = ob.vectorUnsignedInt()
##     get_fingerprint(mol, fp)
##     return _pack_512(*fp)[:39]

# This needs 21 bytes, for 166 bits
# vectorUnsignedInt will contain 8 32-bit words = 256 bits
# (Remember, although 6 words * 32-bits/word = 192, the OpenBabel
# fingerprint size must be a power of 2, and the closest is 8*32.)
_pack_256 = struct.Struct("<" + "I"*8).pack
def calc_MACCS(mol, fp=None,
               get_fingerprint=_ob_get_fingerprint["MACCS"][1],
               _pack_256 = _pack_256):
    if fp is None:
        fp = ob.vectorUnsignedInt()
    get_fingerprint(mol, fp)
    return _pack_256(*fp)[:21]

# TODO: calc_ECFP<n> versions. If they are really needed.
               


# Open Babel version up to 2.3.0 contained errors in the
# translation of the MACCS patterns from RDKit.
# Post-2.3.0 fixed in version control.
# Chemfp 4.0 dropped support for pre-3.0 versions of Open Babel.
MACCS_VERSION = 2

# MACCS might be missing if BABEL_DATADIR doesn't exist.
HAS_MACCS = False

def _check_for_maccs():
    global HAS_MACCS, MACCS_VERSION
    if _ob_get_fingerprint["MACCS"] == (None, None):
        # MACCS should be here. Report the most likely reason
        import os, warnings
        if "BABEL_DATADIR" not in os.environ:
            warnings.warn("MACCS fingerprint missing; perhaps due to missing BABEL_DATADIR?")
        else:
            warnings.warn("MACCS fingerprint missing; does BABEL_DATADIR point to the correct directory?")
        return

    HAS_MACCS = True

if not _using_sphinx_mock:
    _check_for_maccs()

####

@_public
class OpenBabelBaseFingerprintType(types.FingerprintType):
    #: a description of the Open Babel and chemfp software packages used
    software = types._sphinx_repr_if(
        _using_sphinx_mock, "...", SOFTWARE + " " + CHEMFP_SOFTWARE,
        )
    #: a reference to the :mod:`.openbabel_toolkit` module
    toolkit = types._sphinx_repr_if(
        _using_sphinx_mock, "<module 'chemfp.openbabel_toolkit>'", openbabel_toolkit,
        )
    
    # The Open Babel fingerprinters are not re-entrant so I need
    # to make a new fingerprinter for each make_fingerprinter()

    #### The following was auto-generated by: utils/mol_io_codegen.py --fp openbabel

    def from_smistring(
        self,
        content: str,
        *,
        options: Optional[str] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string

        This is equivalent to calling::

           mol = fptype.toolkit.from_smistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param options: options string passed to Open Babel
        :type options: string or None
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smistring(
            content,
            options=options,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_smi(
        self,
        content: str,
        *,
        options: Optional[str] = None,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = None,
        has_header: bool = False,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_smi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param options: options string passed to Open Babel
        :type options: string or None
        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
        :param has_header: If true, treat the first line of the SMILES file as a header
        :type has_header: Boolean (default: False)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smi(
            content,
            options=options,
            delimiter=delimiter,
            has_header=has_header,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_sdf(
        self,
        content: str,
        *,
        implementation: Literal[None, "openbabel", "chemfp"] = "openbabel",
        perceive_stereo: bool = False,
        perceive_0d_stereo: bool = False,
        options: Optional[str] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an SDF record

        This is equivalent to calling::

           mol = fptype.toolkit.from_sdf(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param implementation: SDF record tokenizer implementation
        :type implementation: None or 'openbabel' to use Open Babel, else 'chemfp' (default: "openbabel")
        :param perceive_stereo: not implemented
        :type perceive_stereo: Boolean (default: False)
        :param perceive_0d_stereo: not implemented
        :type perceive_0d_stereo: Boolean (default: False)
        :param options: options string passed to Open Babel
        :type options: string or None
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_sdf(
            content,
            implementation=implementation,
            perceive_stereo=perceive_stereo,
            perceive_0d_stereo=perceive_0d_stereo,
            options=options,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchi(
        self,
        content: str,
        *,
        options: Optional[str] = None,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param options: options string passed to Open Babel
        :type options: string or None
        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchi(
            content,
            options=options,
            delimiter=delimiter,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchistring(
        self,
        content: str,
        *,
        options: Optional[str] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param options: options string passed to Open Babel
        :type options: string or None
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchistring(
            content,
            options=options,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    # Aliases

    from_smiles = from_smistring

    #### End of auto-generated code.


@_public
@types.adjust_call_doc
class OpenBabelFP2FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP2 fingerprint based on path enumeration

    See https://openbabel.org/wiki/FP2

    This is a Daylight-like path enumeration fingerprint with 1021 bits.

    The OpenBabel-FP2/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP2/1"
    num_bits = 1021
    __call__ = types.FingerprintArguments().get_call_method()

    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP2"][1]
        pack_1024 = _pack_1024
        def openbabel_fp2_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_1024(*fp)
        return openbabel_fp2_fingerprinter

def register_fp2(registry):
    registry.register(OpenBabelFP2FingerprintType_v1)


@_public
@types.adjust_call_doc
class OpenBabelFP3FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP3 fingerprint

    See https://openbabel.org/wiki/FP3

    55 bit fingerprints based on a set of SMARTS patterns defining functional groups.

    The OpenBabel-FP3/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP3/1"
    num_bits = 55
    __call__ = types.FingerprintArguments().get_call_method()
    
    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP3"][1]
        pack_64 = _pack_64
        def openbabel_fp3_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_64(*fp)[:7]
        return openbabel_fp3_fingerprinter

def register_fp3(registry):
    registry.register(OpenBabelFP3FingerprintType_v1)


@_public
@types.adjust_call_doc
class OpenBabelFP4FingerprintType_v1(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    """OpenBabel FP4 fingerprint

    https://openbabel.org/wiki/FP4

    307 bit fingerprints based on a set of SMARTS patterns defining functional groups.

    The OpenBabel-FP4/1 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-FP4/1"
    num_bits = 307
    __call__ = types.FingerprintArguments().get_call_method()
    
    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["FP4"][1]
        pack_512 = _pack_512
        def openbabel_fp4_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_512(*fp)[:39]
        return openbabel_fp4_fingerprinter

def register_fp4(registry):
    registry.register(OpenBabelFP4FingerprintType_v1)

    
@_public
@types.adjust_call_doc
class OpenBabelBaseMACCSFingerprintType(types.NoFingerprintParametersMixin, OpenBabelBaseFingerprintType):
    num_bits = 166
    __call__ = types.FingerprintArguments().get_call_method()

    def make_fingerprinter(self):
        get_fingerprint = _ob_get_fingerprint["MACCS"][1]
        pack_256 = _pack_256
        def openbabel_maccs_fingerprinter(mol):
            fp = ob.vectorUnsignedInt()
            get_fingerprint(mol, fp)
            return pack_256(*fp)[:21]
        return openbabel_maccs_fingerprinter

## class OpenBabelMACCSFingerprintType_v1(OpenBabelBaseMACCSFingerprintType):
##     """Open Babel's implementation of the 166 MACCS keys
## 
##     WARNING: This implementation contains serious bugs! All of
##     the ring sizes are wrong.
## 
##     See https://openbabel.org/wiki/Tutorial:Fingerprints and
##     https://github.com/openbabel/openbabel/blob/master/data/MACCS.txt .
## 
##     The OpenBabel-MACCS/1 :class:`.FingerprintType` has no parameters.
## 
##     Note: this version is only available in older (pre-2012) versions of Open Babel.
##     """
##     name = "OpenBabel-MACCS/1"
##     is_available = (MACCS_VERSION == 1)
##     if not is_available:
##         unavailable_message = "This version of Open Babel does not support OpenBabel-MACCS/1 fingerprints"

@_public
@types.adjust_call_doc
class OpenBabelMACCSFingerprintType_v2(OpenBabelBaseMACCSFingerprintType):
    """Open Babel's implementation of the 166 MACCS keys

    See https://openbabel.org/wiki/Tutorial:Fingerprints and
    https://github.com/openbabel/openbabel/blob/master/data/MACCS.txt .

    Note: Open Babel added support for key 44 on 20 October 2014.  This
    should have been version 3. However, I didn't notice until 1 May
    2017 that there was no chemfp test for it. Since everyone has been
    using it as v2, and very few people used the older version, I
    won't change the version number.

    The OpenBabel-MACCS/2 :class:`.FingerprintType` has no parameters.
    """
    name = "OpenBabel-MACCS/2"
    __call__ = types.FingerprintArguments().get_call_method()

def register_maccs(registry):
    registry.unavailable("1", "This version of Open Babel does not support OpenBabel-MACCS/1 fingerprints")
    registry.register_if(HAS_MACCS, OpenBabelMACCSFingerprintType_v2)


######### ECFP-related code

# Make an ECFP fingerprint function specific to the given fingerprint parameters
ECFP_NUM_BITS = 4096

ecfp_arguments = types.FingerprintArguments()
_add = ecfp_arguments.add_argument

_add("nBits", decoder=types.power_of_2_at_least_32, metavar="INT", default=ECFP_NUM_BITS,
     help = "number of bits in the fingerprint")

# Create the pack function on-demand
class PackTable(dict):
    def __missing__(self, nBits):
        if ((nBits-1) & nBits): # ensure it's a power of 2
            raise ValueError("nBits must be a power of 2: %r" % (nBits,))
        if nBits < 32:
            raise ValueError("nBits is too small: %r" % (nBits,))
        pack = struct.Struct("<" + "I"*(nBits // 32)).pack
        self[nBits] = pack
        return pack

_pack_table = PackTable()
_pack_4096 = _pack_table[4096]

@_public
class OpenBabelECFPBaseFingerprintType(OpenBabelBaseFingerprintType):
    _fingerprinter = _get_fingerprint = None
    
    def __init__(self, fingerprint_kwargs):
        super(OpenBabelECFPBaseFingerprintType, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["nBits"]

    def _make_fingerprinter(self, nBits=ECFP_NUM_BITS):
        get_fingerprint = self._get_fingerprint
        if nBits == 4096:
            pack = _pack_4096
            def openbabel_ecfp_fingerprinter(mol):
                fp = ob.vectorUnsignedInt()
                get_fingerprint(mol, fp)
                return pack(*fp)

        else:
            fingerprinter = self._fingerprinter
            pack = _pack_table[nBits]
            def openbabel_ecfp_fingerprinter(mol):
                fp = ob.vectorUnsignedInt()
                get_fingerprint(mol, fp)
                fingerprinter.Fold(fp, nBits)
                return pack(*fp)
        
        return openbabel_ecfp_fingerprinter

    _arguments = ecfp_arguments
    _format_string = "nBits={nBits}"
        
def _register_efcp(registry, klass):
    if not HAS_ECFP:
        if ECFP_AVAILABLE:
            unavailable_message = "chemfp does not support pre-Open Babel 3.0 ECFP fingerprints"
        else:
            unavailable_message = "This version of Open Babel does not implement the ECFP fingerprints"
        registry.register_if(False, klass, reason = unavailable_message)
    else:
        registry.register(klass)
    
@_public
@types.adjust_call_doc
class OpenBabelECFP0FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP0 fingerprint

    This is a circular fingerprint of diameter 0.
    
    The OpenBabel-ECFP0/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP0/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP0"]
    __call__ = ecfp_arguments.get_call_method()

def register_ecfp0(registry):
    _register_efcp(registry, OpenBabelECFP0FingerprintType_v1)
    

@_public
@types.adjust_call_doc
class OpenBabelECFP2FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP2 fingerprint

    This is a circular fingerprint of diameter 2.
    
    The OpenBabel-ECFP2/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP2/1"
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP2"]
    __call__ = ecfp_arguments.get_call_method()

def register_ecfp2(registry):
    _register_efcp(registry, OpenBabelECFP2FingerprintType_v1)
    
@_public
@types.adjust_call_doc
class OpenBabelECFP4FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP4 fingerprint

    This is a circular fingerprint of diameter 4.
    
    The OpenBabel-ECFP4/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP4/1"
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP4"]
    __call__ = ecfp_arguments.get_call_method()
    
def register_ecfp4(registry):
    _register_efcp(registry, OpenBabelECFP4FingerprintType_v1)
    
@_public
@types.adjust_call_doc
class OpenBabelECFP6FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP6 fingerprint

    This is a circular fingerprint of diameter 6.
    
    The OpenBabel-ECFP6/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP6/1"
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP6"]
    __call__ = ecfp_arguments.get_call_method()

def register_ecfp6(registry):
    _register_efcp(registry, OpenBabelECFP6FingerprintType_v1)
    
@_public
@types.adjust_call_doc
class OpenBabelECFP8FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP8 fingerprint

    This is a circular fingerprint of diameter 8.
    
    The OpenBabel-ECFP8/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP8/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP8"]
    __call__ = ecfp_arguments.get_call_method()

    
def register_ecfp8(registry):
    _register_efcp(registry, OpenBabelECFP8FingerprintType_v1)
    
    
@_public
@types.adjust_call_doc
class OpenBabelECFP10FingerprintType_v1(OpenBabelECFPBaseFingerprintType):
    """Open Babel's implementation of the ECFP10 fingerprint

    This is a circular fingerprint of diameter 10.
    
    The OpenBabel-ECFP10/1 :class:`.FingerprintType` parameter is:

    * nBits - the number of bits in the fingerprint (default: 4096 and
        must be a power of 2)
    """
    name = "OpenBabel-ECFP10/1"
    is_available = HAS_ECFP
    _fingerprinter, _get_fingerprint = _ob_get_fingerprint["ECFP10"]
    __call__ = ecfp_arguments.get_call_method()
    
    
def register_ecfp10(registry):
    _register_efcp(registry, OpenBabelECFP10FingerprintType_v1)
