# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""The chemfp toolkit API wrapper for OpenEye's OEChem toolkit.

This module is also available as ``chemfp.openeye``.
"""

from openeye import oechem as _oechem # OpenEye's OEChem must be installed

if hasattr(_oechem, "sphinx_mock"):
    # This should only happen when we're building the documentation using
    # Sphinx's autodoc with a mock for OEChem.  Sphinx's magic mock isn't good
    # enough so provide a concrete mock which is just barely good enough to let
    # autodoc import the module to extract the docstrings.
    
    class _tk:
        TOOLKIT_VERSION = "OEChem (using sphinx)"
        class registry:
            @staticmethod
            def get_format(format):
                return None
        toolkit = None
else:
    # Keep the wrapper implementation hidden.
    from . import _openeye_toolkit as _tk
    from ._openeye_toolkit import _to_openeye_string


__all__ = []
def _public(f):
    __all__.append(f.__name__)
    return f

name = "openeye"
software = _tk.TOOLKIT_VERSION
is_available = True

@_public
def is_licensed():
    """Return True if the OEChem toolkit license is valid, otherwise False.

    This does not check if the OEGraphSim license is valid. I haven't
    yet figured out how I want to handle that distinction. In the meanwhile
    you'll need to use the OEChem API yourself.

    :returns: True or False
    """
    return _oechem.OEChemIsLicensed()

_registry = _tk.registry
_toolkit = _tk.toolkit

# format API

@_public
def get_formats(include_unavailable=False):
    """Get the list of structure formats that OEChem supports

    If *include_unavailable* is True then also include OEChem formats
    which aren't available to this specific version of OEChem.

    :param include_unavailable: include unavailable formats?
    :type include_unavailable: True or False
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _registry.get_formats(include_unavailable)

@_public
def get_input_formats():
    """Get the list of supported OEChem input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _registry.get_input_formats()

@_public
def get_output_formats():
    """Get the list of supported OEChem output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _registry.get_output_formats()

@_public
def get_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if OEChem does not implement
    the format *format_name* or that format is not available.

    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _registry.get_format(format)

@_public
def get_input_format(format):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError if OEChem does not implement
    the format *format_name* or that format is not an input format.

    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _registry.get_input_format(format)

@_public
def get_output_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if OEChem does not implement
    the format *format_name* or that format is not an output format.
    
    :param format_name: the format name
    :type format_name: a string
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _registry.get_output_format(format)

@_public
def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param source: the structure data source.
    :type source: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _registry.get_input_format_from_source(source, format)

@_public
def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: the structure data source.
    :type destination: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _registry.get_output_format_from_destination(destination, format)


# read/write API

@_public
def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads OEGraphMol molecules from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Note: the reader will clear and reuse the OEGraphMol instance. Make a
    copy if you want to keep the molecule around.

    The *reader_args* dictionary parameters depend on the format. Every
    OEChem format supports:
    
    * aromaticity - one of "default", "openeye", "daylight", "tripos", "mdl", "mmff", or None
    * flavor - a number, string-encoded number, or flavor string

    A "flavor string" is a "|" or "," separated list of format-specific
    flavor terms. It can be a simple as "Default", or a more complex string
    like "Default|-ENDM|DELPHI" which for the PDB reader starts with
    the default settings, removes the ENDM flavor, and adds the CHARGE
    and RADIUS flavors.

    The supported input flavor terms for each format are:

    * SMILES - Canon, Strict, Default
    * sdf - Default
    * skc - Default
    * mol2, mol2h - M2H, Default
    * mmod - FormalCrg, Default
    * pdb - ALL, ALTLOC, BondOrder, CHARGE, Connect, DATA, DELPHI, END, ENDM,
      FORMALCHARGE, FormalCrg, ImplicitH, RADIUS, Rings, SecStruct, TER, TerMask, Default
    * xyz - BondOrder, Connect, FormalCrg, ImplicitH, Rings, Default
    * cdx - SuperAtoms, Default
    * oeb - Default

    You can also pass in a numeric value like 123 or a numeric string like "0".

    In addition, the SMILES record readers have limited support for the
    "delimiter" reader_arg:

    * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None

    Note: the first whitespace after the SMILES string will always be
    treated as a delimiter.

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If None
    then a default Location will be created.

    See :func:`chemfp.openeye_toolkit.read_ids_and_molecules` if you want
    (id, OEGraphMol) pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader parameters passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEGraphMol molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location,
                                   encoding, encoding_errors)

@_public
def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads molecules from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.openeye_toolkit.read_molecules` for details about the other
    parameters.  See :func:`chemfp.openeye_toolkit.read_ids_and_molecules_from_string`
    if you  want to read (id, OEGraphMol) pairs instead of just molecules.

    Note: the reader will clear and reuse the OEGraphMol instance. Make a
    copy if you want to keep the molecule around.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEGraphMol molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

@_public
def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None, encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, OEGraphMol molecule) pairs from a structure file

    See :func:`chemfp.openeye_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, OEGraphMol)
    pairs instead of just the molecules.
    
    Note: the reader will clear and reuse the OEGraphMol instance. Make a
    copy if you want to keep the molecule around.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, OEGraphMol) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args,
                                           errors, location, encoding, encoding_errors)

@_public
def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, OEGraphMol) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.openeye_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.openeye_toolkit.read_molecules_from_string`
    if you just want to read the OEGraphMol molecules instead of (id, OEGraphMol) pairs.
    
    Note: the reader will clear and reuse the OEGraphMol instance. Make a
    copy if you want to keep the molecule around.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, OEGraphMol) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

@_public
def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, OEGraphMol) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. The function
    will reuse the OEGraphMol for successive calls, so make a copy if you want
    to keep it around. However, I haven't really noticed much of a performance
    difference between this and :func:`chemfp.openeye_toolkit.parse_id_and_molecule`
    so I suggest you use that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.openeye_toolkit.read_molecules` for details about the
    other parameters.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, OEGraphMol)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

@_public
def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return an OEGraphMol molecule.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.openeye_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.openeye_toolkit.parse_id_and_molecule`
    if you want the (id, OEGraphMol) pair instead of just the molecule.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an OEGraphMol molecule
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

@_public
def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, OEGraphMol) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.openeye_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.openeye_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.openeye_toolkit.parse_molecule`
    if just want the OEGraphMol molecule and not the the (id, OEGraphMol) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an (id, OEGraphMol molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)


@_public
def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert an OEChem molecule into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an OEChem molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

@_public
def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert an OEChem molecule into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an OEChem molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)

@_public
def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write OEChem molecules to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    an OEChem molecule, an OEChem molecule iterator, or an (id, OEChem molecule) pair
    iterator to a file.

    Molecules are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.

    The *writer_args* dictionary parameters depend on the format. Every
    OEChem format supports:
    
      * aromaticity - one of "default", "openeye", "daylight", "tripos", "mdl", "mmff", or None
      * flavor - a number, string-encoded number, or flavor string

    A "flavor string" is a "|" or "," separated list of format-specific
    flavor terms. It can be as simple as "Default", or a more complex
    string like DEFAULT|-AtomStereo|-BondStero|Canonical to generate a
    canonical SMILES string without stereo information.

    The supported output flavor terms for each format are:

    * SMILES - AtomMaps, AtomStereo, BondStereo, Canonical, ExtBonds, Hydrogens,
      ImpHCount, Isotopes, Kekule, RGroups, SuperAtoms
    * sdf - CurrentParity, MCHG, MDLParity, MISO, MRGP, MV30, NoParity, Default
    * mol2, mol2h - AtomNames, AtomTypeNames, BondTypeNames, Hydrogens, OrderAtoms, Substructure, Default
    * sln - Default
    * pdb - BONDS, BOTH, CHARGE, CurrentResidues, DELPHI, ELEMENT, FORMALCHARGE,
      FormalCrg, HETBONDS, NoResidues, OEResidues, ORDERS, OrderAtoms, RADIUS, TER, Default
    * xyz - Charges, Symbols, Default
    * cdx - Default
    * mopac - CHARGES, XYZ, Default
    * mf - Title, Default
    * oeb - Default
    * inchi, inchikey - Chiral, FixedHLayer, Hydrogens, ReconnectedMetals, Stereo,
      RelativeStereo, RacemicStereo, Default

    You can also pass in a numeric value like 123 or a numeric string like "0".

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer parameters passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats (does not affect OEChem)
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

@_public
def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write OEChem molecule records to a Unicode string.

    See :func:`chemfp.openeye_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output string as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting OEChem molecules
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)

@_public
def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write OEChem molecule records to a byte string.

    See :func:`chemfp.openeye_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output string as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats (does not affect OEChem)
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting OEChem molecules
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location)


##### Molecule API

@_public
def copy_molecule(mol):
    """Return a new OEGraphMol which is a copy of the given OEChem molecule

    :param mol: the molecule to copy
    :type mol: an Open Babel molecule
    :returns: a new OBMol instance
    """
    # I return an OEGraphMol because the read of the API only uses OEGraphMols
    return _oechem.OEGraphMol(mol)

@_public
def add_tag(mol, tag, value):
    """Add an SD tag value to the OEChem molecule

    :param mol: the molecule
    :type mol: an OEChem molecule
    :param tag: the SD tag name
    :type tag: string
    :param value: the text for the tag
    :type value: string
    :returns: None
    """
    _oechem.OEAddSDData(mol, tag, value)

@_public
def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    :param mol: the molecule
    :type mol: an OEChem molecule
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    tag = _to_openeye_string(tag)
    value = _oechem.OEGetSDData(mol, tag)
    if value:
        return value
    if _oechem.OEHasSDData(mol, tag):
        return value
    return None

@_public
def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the molecule

    :param mol: the molecule
    :type mol: an OEChem molecule
    :returns: a list of (string name, string value) pairs
    """
    return [(pair.GetTag(), pair.GetValue()) for pair in _oechem.OEGetSDDataPairs(mol)]

@_public
def get_id(mol):
    """Get the molecule's id using OEChem's GetTitle()
    
    :param mol: the molecule
    :type mol: an OEChem molecule
    :returns: a string
    """
    return mol.GetTitle()

@_public
def set_id(mol, id):
    """Set the molecule's id using OEChem's SetTitle()

    :param mol: the molecule
    :type mol: an OEChem molecule
    :param id: the new id
    :type id: string
    :returns: None
    """
    mol.SetTitle(_to_openeye_string(id))


##### OEChem-specific API

@_public
def is_valid_aromaticity(aromaticity_name):
    """Return True for known aromaticity names, otherwise False"""
    return aromaticity_name in _tk.aromaticity_flavors

# accessor aliases for the default fingerprint type object
    
_aliases = {
    "circular": "OpenEye-Circular",
    "maccs166": "OpenEye-MACCS166",
    "mdl_screen": "OpenEye-MDLScreen",
    "molecule_screen": "OpenEye-MoleculeScreen",
    "path": "OpenEye-Path",
    "smarts_screen": "OpenEye-SMARTSScreen",
    "tree": "OpenEye-Tree",
    }
    
def __getattr__(name):
    full_name = _aliases.get(name, None)
    if full_name is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")

    from . import get_fingerprint_type
    return get_fingerprint_type(full_name)

def __dir__():
    return sorted(__all__ + list(_aliases))

#### The following was auto-generated by: utils/mol_io_codegen.py openeye

from . import io as _io
from ._compat import (
    Any as _Any,
    BinaryIO as _BinaryIO,
    Literal as _Literal,
    Optional as _Optional,
    Union as _Union,
)


_smistring_fmt = get_format("smistring")


@_public
def from_smistring(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse a SMILES string using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "smistring", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canon' = 2
      'Strict' = 1
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "smistring", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'AllBonds' = 2048
      'AtomMaps' = 32  (in default bit flags)
      'AtomStereo' = 8  (in default bit flags)
      'BondStereo' = 16  (in default bit flags)
      'Canonical' = 64  (in default bit flags)
      'ExtBonds' = 512
      'Hydrogens' = 2
      'ImpHCount' = 1024
      'Isotopes' = 1  (in default bit flags)
      'Kekule' = 128
      'RGroups' = 4  (in default bit flags)
      'SuperAtoms' = 256
      'Default' = 125  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


_smi_fmt = get_format("smi")


@_public
def from_smi(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse a SMILES string and id using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "smi", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canon' = 2
      'Strict' = 1
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
            "has_header": has_header,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string and id from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "smi", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'AllBonds' = 2048
      'AtomMaps' = 32  (in default bit flags)
      'AtomStereo' = 8  (in default bit flags)
      'BondStereo' = 16  (in default bit flags)
      'Canonical' = 64  (in default bit flags)
      'ExtBonds' = 512
      'Hydrogens' = 2
      'ImpHCount' = 1024
      'Isotopes' = 1  (in default bit flags)
      'Kekule' = 128
      'RGroups' = 4  (in default bit flags)
      'SuperAtoms' = 256
      'Default' = 125  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_smi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse a SMILES string and id file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "smi", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Canon' = 2
      'Strict' = 1
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "smi"),
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
            "has_header": has_header,
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_smi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a SMILES string and id from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "smi", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'AllBonds' = 2048
      'AtomMaps' = 32  (in default bit flags)
      'AtomStereo' = 8  (in default bit flags)
      'BondStereo' = 16  (in default bit flags)
      'Canonical' = 64  (in default bit flags)
      'ExtBonds' = 512
      'Hydrogens' = 2
      'ImpHCount' = 1024
      'Isotopes' = 1  (in default bit flags)
      'Kekule' = 128
      'RGroups' = 4  (in default bit flags)
      'SuperAtoms' = 256
      'Default' = 125  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "smi"),
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        errors=errors,
    )


_sdf_fmt = get_format("sdf")


@_public
def from_sdf(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse an SDF record using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "sdf", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'FixBondMarks' = 1
      'SuppressEmptyMolSkip' = 4
      'SuppressImp2ExpENHSTE' = 2
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sdf_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sdf(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an SDF record from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "sdf", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Add2D' = 256  (in default bit flags)
      'CurrentParity' = 64
      'MCHG' = 1  (in default bit flags)
      'MDLParity' = 16  (in default bit flags)
      'MISO' = 2  (in default bit flags)
      'MRGP' = 4  (in default bit flags)
      'MV30' = 8  (in default bit flags)
      'NoParity' = 32
      'SuppressImp2ExpENHSTE' = 1024
      'SuppressTimestamps' = 2048
      'UnsetBad2DStereo' = 512
      'Default' = 287  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_sdf_file(
    source: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse an SDF record file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "sdf", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'FixBondMarks' = 1
      'SuppressEmptyMolSkip' = 4
      'SuppressImp2ExpENHSTE' = 2
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "sdf"),
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        errors=errors,
    )


@_public
def to_sdf_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an SDF record from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Add2D' = 256  (in default bit flags)
      'CurrentParity' = 64
      'MCHG' = 1  (in default bit flags)
      'MDLParity' = 16  (in default bit flags)
      'MISO' = 2  (in default bit flags)
      'MRGP' = 4  (in default bit flags)
      'MV30' = 8  (in default bit flags)
      'NoParity' = 32
      'SuppressImp2ExpENHSTE' = 1024
      'SuppressTimestamps' = 2048
      'UnsetBad2DStereo' = 512
      'Default' = 287  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf"),
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        errors=errors,
    )


_sdf3k_fmt = get_format("sdf3k")


@_public
def to_sdf3k(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "sdf3k", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Add2D' = 256  (in default bit flags)
      'CurrentParity' = 64
      'MCHG' = 1  (in default bit flags)
      'MDLParity' = 16  (in default bit flags)
      'MISO' = 2  (in default bit flags)
      'MRGP' = 4  (in default bit flags)
      'MV30' = 8  (in default bit flags)
      'NoParity' = 32
      'SuppressImp2ExpENHSTE' = 1024
      'SuppressTimestamps' = 2048
      'UnsetBad2DStereo' = 512
      'Default' = 287  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf3k_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def to_sdf3k_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf3k", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Add2D' = 256  (in default bit flags)
      'CurrentParity' = 64
      'MCHG' = 1  (in default bit flags)
      'MDLParity' = 16  (in default bit flags)
      'MISO' = 2  (in default bit flags)
      'MRGP' = 4  (in default bit flags)
      'MV30' = 8  (in default bit flags)
      'NoParity' = 32
      'SuppressImp2ExpENHSTE' = 1024
      'SuppressTimestamps' = 2048
      'UnsetBad2DStereo' = 512
      'Default' = 287  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf3k"),
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        errors=errors,
    )


_molfile_fmt = get_format("molfile")


@_public
def from_molfile(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse a molfile using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "molfile", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'FixBondMarks' = 1
      'SuppressEmptyMolSkip' = 4
      'SuppressImp2ExpENHSTE' = 2
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _molfile_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_molfile(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a molfile from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "molfile", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Add2D' = 256  (in default bit flags)
      'CurrentParity' = 64
      'MCHG' = 1  (in default bit flags)
      'MDLParity' = 16  (in default bit flags)
      'MISO' = 2  (in default bit flags)
      'MRGP' = 4  (in default bit flags)
      'MV30' = 8
      'NoParity' = 32
      'SuppressImp2ExpENHSTE' = 1024
      'SuppressTimestamps' = 2048
      'UnsetBad2DStereo' = 512
      'Default' = 279  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _molfile_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


_oez_fmt = get_format("oez")


@_public
def from_oez(
    content: str,
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse an OEZ using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "oez", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _oez_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_oez(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an OEZ from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "oez", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _oez_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_oez_file(
    source: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse an OEZ file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "oez", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "oez"),
        reader_args={
            "flavor": flavor,
        },
        errors=errors,
    )


@_public
def to_oez_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an OEZ from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "oez", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "oez"),
        writer_args={
            "flavor": flavor,
        },
        errors=errors,
    )


_fasta_fmt = get_format("fasta")


@_public
def from_fasta(
    content: str,
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse a FASTA record using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "fasta", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'CustomResidues' = 1
      'EmbeddedSMILES' = 2
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _fasta_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_fasta(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a FASTA record from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "fasta", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _fasta_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_fasta_file(
    source: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse a FASTA record file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "fasta", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'CustomResidues' = 1
      'EmbeddedSMILES' = 2
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "fasta"),
        reader_args={
            "flavor": flavor,
        },
        errors=errors,
    )


@_public
def to_fasta_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a FASTA record from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "fasta", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "fasta"),
        writer_args={
            "flavor": flavor,
        },
        errors=errors,
    )


_sequence_fmt = get_format("sequence")


@_public
def from_sequence(
    content: str,
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse an IUPAC sequence using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "sequence", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'CustomResidues' = 1
      'EmbeddedSMILES' = 2
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sequence_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sequence(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate an IUPAC sequence from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "sequence", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sequence_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


_pdb_fmt = get_format("pdb")


@_public
def from_pdb(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse a PDB record using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "pdb", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'ALL' = 8
      'ALTLOC' = 32526
      'BondOrder' = 1024  (in default bit flags)
      'CHARGE' = 32
      'Connect' = 4096  (in default bit flags)
      'DATA' = 16
      'DELPHI' = 96
      'END' = 2  (in default bit flags)
      'ENDM' = 4  (in default bit flags)
      'FORMALCHARGE' = 128
      'FormalCrg' = 256  (in default bit flags)
      'ImplicitH' = 512  (in default bit flags)
      'RADIUS' = 64
      'Rings' = 2048  (in default bit flags)
      'SecStruct' = 8192  (in default bit flags)
      'TER' = 1
      'Default' = 16134  (in default bit flags)

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _pdb_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_pdb(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a PDB record from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "pdb", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'BONDS' = 1
      'BOTH' = 4  (in default bit flags)
      'CHARGE' = 8
      'CurrentResidues' = 2048
      'DELPHI' = 24
      'ELEMENT' = 64  (in default bit flags)
      'FORMALCHARGE' = 128
      'FormalCrg' = 128
      'HETBONDS' = 256  (in default bit flags)
      'NoResidues' = 1024
      'OEResidues' = 512  (in default bit flags)
      'ORDERS' = 2
      'OrderAtoms' = 4096  (in default bit flags)
      'RADIUS' = 16
      'TER' = 32
      'Default' = 4932  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _pdb_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_pdb_file(
    source: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Parse a PDB record file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "pdb", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'ALL' = 8
      'ALTLOC' = 32526
      'BondOrder' = 1024  (in default bit flags)
      'CHARGE' = 32
      'Connect' = 4096  (in default bit flags)
      'DATA' = 16
      'DELPHI' = 96
      'END' = 2  (in default bit flags)
      'ENDM' = 4  (in default bit flags)
      'FORMALCHARGE' = 128
      'FormalCrg' = 256  (in default bit flags)
      'ImplicitH' = 512  (in default bit flags)
      'RADIUS' = 64
      'Rings' = 2048  (in default bit flags)
      'SecStruct' = 8192  (in default bit flags)
      'TER' = 1
      'Default' = 16134  (in default bit flags)

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "pdb"),
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
        },
        errors=errors,
    )


@_public
def to_pdb_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    errors: str = "strict",
):
    """Generate a PDB record from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "pdb", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'BONDS' = 1
      'BOTH' = 4  (in default bit flags)
      'CHARGE' = 8
      'CurrentResidues' = 2048
      'DELPHI' = 24
      'ELEMENT' = 64  (in default bit flags)
      'FORMALCHARGE' = 128
      'FormalCrg' = 128
      'HETBONDS' = 256  (in default bit flags)
      'NoResidues' = 1024
      'OEResidues' = 512  (in default bit flags)
      'ORDERS' = 2
      'OrderAtoms' = 4096  (in default bit flags)
      'RADIUS' = 16
      'TER' = 32
      'Default' = 4932  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "pdb"),
        writer_args={
            "flavor": flavor,
        },
        errors=errors,
    )


_inchi_fmt = get_format("inchi")


@_public
def from_inchi(
    content: str,
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchi", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
            "has_header": has_header,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "inchi", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'ReconnectedMetals' = 16
      'Stereo' = 2  (in default bit flags)
      'RelativeStereo' = 32
      'RacemicStereo' = 64
      'Default' = 130  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def from_inchi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    aromaticity: _Optional[
        _Literal["none", "openeye", "daylight", "tripos", "mdl", "mmff"]
    ] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id file using the OEChem toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "inchi", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param aromaticity: The OEChem aromaticity model
    :type aromaticity: None, or one of 'none', 'openeye', 'daylight', 'tripos', 'mdl', or 'mmff'
    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating OEChem molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "inchi"),
        reader_args={
            "aromaticity": aromaticity,
            "flavor": flavor,
            "has_header": has_header,
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_inchi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchi", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'ReconnectedMetals' = 16
      'Stereo' = 2  (in default bit flags)
      'RelativeStereo' = 32
      'RacemicStereo' = 64
      'Default' = 130  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchi"),
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchistring_fmt = get_format("inchistring")


@_public
def from_inchistring(
    content: str,
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string using the OEChem toolkit

    This is equivalent to calling:
       parse_molecule(content, "inchistring", reader_args={...}, errors=errors)

    Available bit flag flavors are:
      'Default' = 0

    :param flavor: Input flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "flavor": flavor,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate an InChI string from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "inchistring", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'ReconnectedMetals' = 16
      'Stereo' = 2  (in default bit flags)
      'RelativeStereo' = 32
      'RacemicStereo' = 64
      'Default' = 130  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )


_inchikey_fmt = get_format("inchikey")


@_public
def to_inchikey(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "inchikey", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'RacemicStereo' = 64
      'ReconnectedMetals' = 16
      'RelativeStereo' = 32
      'Stereo' = 2  (in default bit flags)
      'Default' = 130  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikey_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def to_inchikey_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an OEChem molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchikey", writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'RacemicStereo' = 64
      'ReconnectedMetals' = 16
      'RelativeStereo' = 32
      'Stereo' = 2  (in default bit flags)
      'Default' = 130  (in default bit flags)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting OEChem molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchikey"),
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchikeystring_fmt = get_format("inchikeystring")


@_public
def to_inchikeystring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Optional[_Union[int, str]] = "Default",
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate an InChIKey string from an OEChem molecule

    This is equivalent to calling:
       create_string(mol, "inchikeystring", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:
      'Chiral' = 4
      'FixedHLayer' = 8
      'Hydrogens' = 1
      'RacemicStereo' = 64
      'ReconnectedMetals' = 16
      'RelativeStereo' = 32
      'Stereo' = 2  (in default bit flags)
      'Default' = 130  (in default bit flags)

    :param mol: a molecule object
    :type mol: an OEChem molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: None, integer or string with "|"- or ","-separated terms (default: "Default")
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an OEChem molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikeystring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )


# Aliases

from_smiles = from_smistring

to_smiles = to_smistring

#### End of auto-generated code.
