# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

"""The chemfp toolkit API wrapper for the RDKit toolkit.

This module is also available as ``chemfp.rdkit``.
"""

import rdkit # RDKit must be installed to use chemfp.rdkit_toolkit

if hasattr(rdkit, "sphinx_mock"):
    # This should only happen when we're building the documentation using
    # Sphinx's autodoc with a mock for RDKit.  Sphinx's magic mock isn't good
    # enough so provide a concrete mock which is just barely good enough to let
    # autodoc import the module to extract the docstrings.
    
    class _tk:
        SOFTWARE = "RDKit (using sphinx)"
        class format_registry:
            @staticmethod
            def get_format(format):
                return None
        toolkit = None
else:
    # Don't provide a direct link to RDKit.
    del rdkit
    # And keep the wrapper implementation hidden.
    from . import _rdkit_toolkit as _tk

__all__ = []
def _public(f):
    __all__.append(f.__name__)
    return f

name = "rdkit"
software = _tk.SOFTWARE
is_available = True

@_public
def is_licensed():
    """Return True - RDKit is always licensed

    :returns: True
    """
    return True


_format_registry = _tk.format_registry
_toolkit = _tk.toolkit

# format API

@_public
def get_formats(include_unavailable=False):
    """Get the list of structure formats that RDKit supports

    If *include_unavailable* is True then also include RDKit formats
    which aren't available to this specific version of RDKit, such
    as the InChI formats if your RDKit installation wasn't compiled
    with InChI support.

    :param include_unavailable: include unavailable formats?
    :type include_unavailable: True or False
    :returns: a list of Format objects
    """
    return _format_registry.get_formats(include_unavailable)

@_public
def get_input_formats():
    """Get the list of supported RDKit input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_formats()

@_public
def get_output_formats():
    """Get the list of supported RDKit output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_formats()

@_public
def get_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if RDKit does not implement
    the format *format_name* or that format is not available.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_format(format)

@_public
def get_input_format(format):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError if RDKit does not implement
    the format *format_name* or that format is not an input format.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_format(format)

@_public
def get_output_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if RDKit does not implement
    the format *format_name* or that format is not an output format.
    
    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_format(format)

@_public
def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.
    
    :param source: the structure data source.
    :type source: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format_from_source(source, format)

@_public
def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: The structure data source.
    :type destination: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format_from_destination(destination, format)

# read/write API

@_public
def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads RDKit molecules from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Note: the reader returns a new RDKit molecule each time.

    The *reader_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * has_header - True or False
      * sanitize - True or default sanitizes; False for unsanitized processing

    * InChI
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing
      
    * SDF
    
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * strictParsing - True or default for strict parsing; False for lenient parsing

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If None
    then a default Location will be created.

    See :func:`chemfp.rdkit_toolkit.read_ids_and_molecules` if you want (id, molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader parameters passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location,
                                   encoding, encoding_errors)

@_public
def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads RDKit molecules from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.rdkit_toolkit.read_molecules` for details about the other
    parameters.  See :func:`chemfp.rdkit_toolkit.read_ids_and_molecules_from_string`
    if you  want to read (id, RDKit) pairs instead of just molecules.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

@_public
def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None, encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, RDKit molecule) pairs from a structure file

    See :func:`chemfp.rdkit_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, RDKit molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, RDKit molecule) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location,
                                           encoding, encoding_errors)

@_public
def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, RDKit molecule) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.rdkit_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.rdkit_toolkit.read_molecules_from_string`
    if you just want to read the RDKit molecules instead of (id, molecule) pairs.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, RDKit molecule) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

@_public
def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, RDKit molecule) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. However, I haven't
    really noticed much of a performance difference between this and
    :func:`chemfp.rdkit_toolkit.parse_id_and_molecule` so you can probably
    so I suggest you use that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.rdkit_toolkit.read_molecules` for details about the
    other parameters.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, RDKit molecule)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

@_public
def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return an RDKit molecule.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.rdkit_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.rdkit_toolkit.parse_id_and_molecule`
    if you want the (id, RDKit molecule) pair instead of just the molecule.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an RDKit molecule
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

@_public
def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, RDKit molecule) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.rdkit_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.rdkit_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.rdkit_toolkit.parse_molecule`
    if just want the RDKit molecule and not the the (id, RDKit molecule) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an (id, RDKit molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)


@_public
def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert an RDKit molecule into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an RDKit molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a Unicode string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

@_public
def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert an RDKit molecule into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: an RDKit molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a byte string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)


@_public
def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write RDKit molecules to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    an RDKit molecule, an RDKit molecule iterator, or an (id, RDKit molecule) pair
    iterator to a file.
    
    Molecules are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.
    
    The *writer_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * isomericSmiles - True to generate isomeric SMILES
      * kekuleSmiles - True to generate SMILES in Kekule form
      * canonical - True to generate a canonical SMILES
      * allBondsExplicit - True to write explict '-' and ':' bonds, even if they can be inferred; default is False
      * allHsExplicit - True to write explicit hydrogen counts; default is False
      * cxsmiles - True to include CXSMILES annotations; default is False

    InChI and InChIKey
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * include_id - True or default to include the id as the second column; False has no id column
      * options - an options string passed to the underlying InChI library
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing

    SDF
    
      * includeStereo - True include stereo information; False or default does not
      * kekulize - True or default creates the connection table with bonds in Kekeule form
      * v3k - True to alway export in V3000 format

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer parameters passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

@_public
def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a string.

    See :func:`chemfp.rdkit_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting RDKit molecules
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)

@_public
def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a text string.

    See :func:`chemfp.rdkit_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting RDKit molecules
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location, level)


# molecule manipulation API

@_public
def copy_molecule(mol):
    """Return a new RDKit molecule which is a copy of the given molecule

    :param mol: the molecule to copy
    :type mol: an RDKit molecule
    :returns: a new RDKit Mol instance
    """
    # This (and its copy.copy() equivalent) doesn't copy the properties.
    from rdkit import Chem
    new_mol = Chem.Mol(mol.ToBinary())
    # I have to do it myself.
    for name in mol.GetPropNames():
        new_mol.SetProp(name, mol.GetProp(name))
    try:
        new_mol.SetProp("_Name", mol.GetProp("_Name"))
    except KeyError:
        pass
    return new_mol

@_public
def add_tag(mol, tag, value):
    """Add an SD tag value to the RDKit molecule

    :param mol: the molecule
    :type mol: an RDKit molecule
    :param tag: the SD tag name
    :type tag: string
    :param value: the text for the tag
    :type value: string
    :returns: None
    """
    mol.SetProp(tag, value)

@_public
def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    :param mol: the molecule
    :type mol: an RDKit molecule
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    try:
        return mol.GetProp(tag)
    except KeyError:
        return None

@_public
def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the molecule

    :param mol: the molecule
    :type mol: an RDKit molecule
    :returns: a list of (string name, string value) pairs
    """
    return [(name, mol.GetProp(name)) for name in mol.GetPropNames()]

@_public
def get_id(mol):
    """Get the molecule's id from RDKit's _Name property

    :param mol: the molecule
    :type mol: an RDKit molecule
    :returns: a string
    """
    try:
        return mol.GetProp("_Name")
    except KeyError:
        return None

@_public
def set_id(mol, id):
    """Set the molecule's id as RDKit's _Name property

    :param mol: the molecule
    :type mol: an RDKit molecule
    :param id: the new id
    :type id: string
    :returns: None
    """
    mol.SetProp("_Name", id)

# accessor aliases for the default fingerprint type object
    
_aliases = {
    "avalon": "RDKit-Avalon",
    "maccs166": "RDKit-MACCS166",
    "morgan": "RDKit-Morgan",
    "atom_pair": "RDKit-AtomPair",
    "pattern": "RDKit-Pattern",
    "rdk": "RDKit-Fingerprint",
    "secfp": "RDKit-SECFP",
    "torsion": "RDKit-Torsion",
    }
    
def __getattr__(name):
    full_name = _aliases.get(name, None)
    if full_name is None:
        raise AttributeError(f"module {__name__} has no attribute {name}")

    from . import get_fingerprint_type
    return get_fingerprint_type(full_name)

def __dir__():
    return sorted(__all__ + list(_aliases))

#### The following was auto-generated by: utils/mol_io_codegen.py rdkit

from . import io as _io
from ._compat import (
    Any as _Any,
    BinaryIO as _BinaryIO,
    Literal as _Literal,
    Optional as _Optional,
    Union as _Union,
)


_smistring_fmt = get_format("smistring")


@_public
def from_smistring(
    content: str,
    *,
    sanitize: bool = True,
    errors: str = "strict",
):
    """Parse a SMILES string using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "smistring", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    isomericSmiles: bool = True,
    kekuleSmiles: bool = False,
    canonical: bool = True,
    allBondsExplicit: bool = False,
    allHsExplicit: bool = False,
    cxsmiles: bool = False,
    errors: str = "strict",
):
    """Generate a SMILES string from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "smistring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param isomericSmiles: If true, generate an isomeric SMILES
    :type isomericSmiles: Boolean (default: True)
    :param kekuleSmiles: If true, generate Kekule SMILES
    :type kekuleSmiles: Boolean (default: False)
    :param canonical: If true, generate a canonical SMILES
    :type canonical: Boolean (default: True)
    :param allBondsExplicit: If true, include bond symbols even for single and aromatic bond
    :type allBondsExplicit: Boolean (default: False)
    :param allHsExplicit: If true, include hydrogen counts for every atom
    :type allHsExplicit: Boolean (default: False)
    :param cxsmiles: If true, generate CXSmiles
    :type cxsmiles: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "isomericSmiles": isomericSmiles,
            "kekuleSmiles": kekuleSmiles,
            "canonical": canonical,
            "allBondsExplicit": allBondsExplicit,
            "allHsExplicit": allHsExplicit,
            "cxsmiles": cxsmiles,
        },
        error_handler=error_handler,
    )


_smi_fmt = get_format("smi")


@_public
def from_smi(
    content: str,
    *,
    sanitize: bool = True,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse a SMILES string and id using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "smi", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _smi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "has_header": False,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_smi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    isomericSmiles: bool = True,
    kekuleSmiles: bool = False,
    canonical: bool = True,
    allBondsExplicit: bool = False,
    allHsExplicit: bool = False,
    cxsmiles: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate a SMILES string and id from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "smi", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param isomericSmiles: If true, generate an isomeric SMILES
    :type isomericSmiles: Boolean (default: True)
    :param kekuleSmiles: If true, generate Kekule SMILES
    :type kekuleSmiles: Boolean (default: False)
    :param canonical: If true, generate a canonical SMILES
    :type canonical: Boolean (default: True)
    :param allBondsExplicit: If true, include bond symbols even for single and aromatic bond
    :type allBondsExplicit: Boolean (default: False)
    :param allHsExplicit: If true, include hydrogen counts for every atom
    :type allHsExplicit: Boolean (default: False)
    :param cxsmiles: If true, generate CXSmiles
    :type cxsmiles: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _smi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "isomericSmiles": isomericSmiles,
            "kekuleSmiles": kekuleSmiles,
            "canonical": canonical,
            "allBondsExplicit": allBondsExplicit,
            "allHsExplicit": allHsExplicit,
            "cxsmiles": cxsmiles,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )


@_public
def from_smi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    sanitize: bool = True,
    has_header: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse a SMILES string and id file using the RDKit toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "smi", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param has_header: If true, treat the first line of the SMILES file as a header
    :type has_header: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "smi"),
        reader_args={
            "sanitize": sanitize,
            "has_header": has_header,
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_smi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    isomericSmiles: bool = True,
    kekuleSmiles: bool = False,
    canonical: bool = True,
    allBondsExplicit: bool = False,
    allHsExplicit: bool = False,
    cxsmiles: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Generate a SMILES string and id from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "smi", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param isomericSmiles: If true, generate an isomeric SMILES
    :type isomericSmiles: Boolean (default: True)
    :param kekuleSmiles: If true, generate Kekule SMILES
    :type kekuleSmiles: Boolean (default: False)
    :param canonical: If true, generate a canonical SMILES
    :type canonical: Boolean (default: True)
    :param allBondsExplicit: If true, include bond symbols even for single and aromatic bond
    :type allBondsExplicit: Boolean (default: False)
    :param allHsExplicit: If true, include hydrogen counts for every atom
    :type allHsExplicit: Boolean (default: False)
    :param cxsmiles: If true, generate CXSmiles
    :type cxsmiles: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "smi"),
        writer_args={
            "isomericSmiles": isomericSmiles,
            "kekuleSmiles": kekuleSmiles,
            "canonical": canonical,
            "allBondsExplicit": allBondsExplicit,
            "allHsExplicit": allHsExplicit,
            "cxsmiles": cxsmiles,
            "delimiter": delimiter,
        },
        errors=errors,
    )


_sdf_fmt = get_format("sdf")


@_public
def from_sdf(
    content: str,
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    strictParsing: bool = True,
    includeTags: bool = True,
    errors: str = "strict",
):
    """Parse an SDF record using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "sdf", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param strictParsing: If true, require stricter adherence to the SDF specification
    :type strictParsing: Boolean (default: True)
    :param includeTags: if true, extract the struture data tag fields
    :type includeTags: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sdf_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "strictParsing": strictParsing,
            "includeTags": includeTags,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sdf(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    includeStereo: bool = False,
    kekulize: bool = True,
    v3k: bool = False,
    errors: str = "strict",
):
    """Generate an SDF record from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "sdf", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param includeStereo: if true, include stereochemistry information in the record
    :type includeStereo: Boolean (default: False)
    :param kekulize: if true, Kekulize the molecule before creating the record
    :type kekulize: Boolean (default: True)
    :param v3k: if true, always write in V3000 format
    :type v3k: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "includeStereo": includeStereo,
            "kekulize": kekulize,
            "v3k": v3k,
        },
        error_handler=error_handler,
    )


@_public
def from_sdf_file(
    source: _Union[None, str, _BinaryIO],
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    strictParsing: bool = True,
    includeTags: bool = True,
    errors: str = "strict",
):
    """Parse an SDF record file using the RDKit toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "sdf", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param strictParsing: If true, require stricter adherence to the SDF specification
    :type strictParsing: Boolean (default: True)
    :param includeTags: if true, extract the struture data tag fields
    :type includeTags: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "sdf"),
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "strictParsing": strictParsing,
            "includeTags": includeTags,
        },
        errors=errors,
    )


@_public
def to_sdf_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    includeStereo: bool = False,
    kekulize: bool = True,
    v3k: bool = False,
    errors: str = "strict",
):
    """Generate an SDF record from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param includeStereo: if true, include stereochemistry information in the record
    :type includeStereo: Boolean (default: False)
    :param kekulize: if true, Kekulize the molecule before creating the record
    :type kekulize: Boolean (default: True)
    :param v3k: if true, always write in V3000 format
    :type v3k: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf"),
        writer_args={
            "includeStereo": includeStereo,
            "kekulize": kekulize,
            "v3k": v3k,
        },
        errors=errors,
    )


_sdf3k_fmt = get_format("sdf3k")


@_public
def to_sdf3k(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    includeStereo: bool = False,
    kekulize: bool = True,
    v3k: bool = True,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "sdf3k", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param includeStereo: if true, include stereochemistry information in the record
    :type includeStereo: Boolean (default: False)
    :param kekulize: if true, Kekulize the molecule before creating the record
    :type kekulize: Boolean (default: True)
    :param v3k: if true, always write in V3000 format
    :type v3k: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sdf3k_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "includeStereo": includeStereo,
            "kekulize": kekulize,
            "v3k": v3k,
        },
        error_handler=error_handler,
    )


@_public
def to_sdf3k_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    includeStereo: bool = False,
    kekulize: bool = True,
    v3k: bool = True,
    errors: str = "strict",
):
    """Generate an SDF record in V3000 format from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "sdf3k", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param includeStereo: if true, include stereochemistry information in the record
    :type includeStereo: Boolean (default: False)
    :param kekulize: if true, Kekulize the molecule before creating the record
    :type kekulize: Boolean (default: True)
    :param v3k: if true, always write in V3000 format
    :type v3k: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "sdf3k"),
        writer_args={
            "includeStereo": includeStereo,
            "kekulize": kekulize,
            "v3k": v3k,
        },
        errors=errors,
    )


_molfile_fmt = get_format("molfile")


@_public
def from_molfile(
    content: str,
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    strictParsing: bool = True,
    errors: str = "strict",
):
    """Parse a molfile using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "molfile", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param strictParsing: If true, require stricter adherence to the SDF specification
    :type strictParsing: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _molfile_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "strictParsing": strictParsing,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_molfile(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    includeStereo: bool = False,
    kekulize: bool = True,
    v3k: bool = False,
    errors: str = "strict",
):
    """Generate a molfile from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "molfile", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param includeStereo: if true, include stereochemistry information in the record
    :type includeStereo: Boolean (default: False)
    :param kekulize: if true, Kekulize the molecule before creating the record
    :type kekulize: Boolean (default: True)
    :param v3k: if true, always write in V3000 format
    :type v3k: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _molfile_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "includeStereo": includeStereo,
            "kekulize": kekulize,
            "v3k": v3k,
        },
        error_handler=error_handler,
    )


_rdbinmol_fmt = get_format("rdbinmol")


@_public
def from_rdbinmol(
    content: str,
    *,
    errors: str = "strict",
):
    """Parse an RDKit binary molecule byte string using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "rdbinmol", reader_args={...}, errors=errors)

    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _rdbinmol_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={},
        error_handler=error_handler,
    )
    return mol


@_public
def to_rdbinmol(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an RDKit binary molecule byte string from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "rdbinmol", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _rdbinmol_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={},
        error_handler=error_handler,
    )


_fasta_fmt = get_format("fasta")


@_public
def from_fasta(
    content: str,
    *,
    sanitize: bool = True,
    flavor: _Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9] = 0,
    errors: str = "strict",
):
    """Parse a FASTA record using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "fasta", reader_args={...}, errors=errors)

    Possible flavor values are:

     *  0 = L-amino acids
     *  1 = D-amino acids
     *  2 = RNA, no cap
     *  3 = RNA, 5' cap
     *  4 = RNA, 3' cap
     *  5 = RNA, both caps
     *  6 = DNA, no cap
     *  7 = DNA, 5' cap
     *  8 = DNA, 3' cap
     *  9 = DNA, both caps

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param flavor: The sequence type (amino acid, RNA, or DNA), and how to handle caps
    :type flavor: integer from 0-9, inclusive (default: 0)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _fasta_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_fasta(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a FASTA record from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "fasta", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _fasta_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={},
        error_handler=error_handler,
    )


@_public
def from_fasta_file(
    source: _Union[None, str, _BinaryIO],
    *,
    sanitize: bool = True,
    flavor: _Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9] = 0,
    errors: str = "strict",
):
    """Parse a FASTA record file using the RDKit toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "fasta", reader_args={...}, errors=errors)

    Possible flavor values are:

     *  0 = L-amino acids
     *  1 = D-amino acids
     *  2 = RNA, no cap
     *  3 = RNA, 5' cap
     *  4 = RNA, 3' cap
     *  5 = RNA, both caps
     *  6 = DNA, no cap
     *  7 = DNA, 5' cap
     *  8 = DNA, 3' cap
     *  9 = DNA, both caps

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param flavor: The sequence type (amino acid, RNA, or DNA), and how to handle caps
    :type flavor: integer from 0-9, inclusive (default: 0)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "fasta"),
        reader_args={
            "sanitize": sanitize,
            "flavor": flavor,
        },
        errors=errors,
    )


@_public
def to_fasta_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    errors: str = "strict",
):
    """Generate a FASTA record from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "fasta", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "fasta"),
        writer_args={},
        errors=errors,
    )


_sequence_fmt = get_format("sequence")


@_public
def from_sequence(
    content: str,
    *,
    sanitize: bool = True,
    flavor: _Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9] = 0,
    errors: str = "strict",
):
    """Parse an IUPAC sequence using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "sequence", reader_args={...}, errors=errors)

    Possible flavor values are:

     *  0 = L-amino acids
     *  1 = D-amino acids
     *  2 = RNA, no cap
     *  3 = RNA, 5' cap
     *  4 = RNA, 3' cap
     *  5 = RNA, both caps
     *  6 = DNA, no cap
     *  7 = DNA, 5' cap
     *  8 = DNA, 3' cap
     *  9 = DNA, both caps

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param flavor: The sequence type (amino acid, RNA, or DNA), and how to handle caps
    :type flavor: integer from 0-9, inclusive (default: 0)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _sequence_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "flavor": flavor,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_sequence(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate an IUPAC sequence from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "sequence", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _sequence_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={},
        error_handler=error_handler,
    )


_helm_fmt = get_format("helm")


@_public
def from_helm(
    content: str,
    *,
    sanitize: bool = True,
    errors: str = "strict",
):
    """Parse a HELM string using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "helm", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _helm_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_helm(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    errors: str = "strict",
):
    """Generate a HELM string from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "helm", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _helm_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={},
        error_handler=error_handler,
    )


_pdb_fmt = get_format("pdb")


@_public
def from_pdb(
    content: str,
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    flavor: _Literal[0] = 0,
    proximityBonding: bool = True,
    errors: str = "strict",
):
    """Parse a PDB record using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "pdb", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param flavor: The value 0 (may change in the future)
    :type flavor: 0
    :param proximityBonding: If true, connect atoms based on a proximity search
    :type proximityBonding: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _pdb_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "flavor": flavor,
            "proximityBonding": proximityBonding,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_pdb(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    flavor: _Union[int, str] = 0,
    errors: str = "strict",
):
    """Generate a PDB record from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "pdb", id=id, writer_args={...}, errors=errors)

    Available bit flag flavors are:

     *  1 = 'MODEL' = write MODEL/ENDMDL
     *  2 = 'NO_CONECT' = no CONECT records
     *  4 = 'BOTH_CONECT' = CONECT records in both directions
     *  8 = 'NO_BOND_ORDER' = use only one CONECT even for higher bond orders
     * 16 = 'MASTER' = write MASTER record
     * 32 = 'TER' = write TER record

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param flavor: Output flavor bit flags
    :type flavor: An integer or string of '|'- or ','-separated terms
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _pdb_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "flavor": flavor,
        },
        error_handler=error_handler,
    )


@_public
def from_pdb_file(
    source: _Union[None, str, _BinaryIO],
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    flavor: _Literal[0] = 0,
    proximityBonding: bool = True,
    errors: str = "strict",
):
    """Parse a PDB record file using the RDKit toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "pdb", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param flavor: The value 0 (may change in the future)
    :type flavor: 0
    :param proximityBonding: If true, connect atoms based on a proximity search
    :type proximityBonding: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "pdb"),
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "flavor": flavor,
            "proximityBonding": proximityBonding,
        },
        errors=errors,
    )


@_public
def to_pdb_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    flavor: _Union[int, str] = 0,
    errors: str = "strict",
):
    """Generate a PDB record from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "pdb", writer_args={...}, errors=errors)

    Available bit flag flavors are:

     *  1 = 'MODEL' = write MODEL/ENDMDL
     *  2 = 'NO_CONECT' = no CONECT records
     *  4 = 'BOTH_CONECT' = CONECT records in both directions
     *  8 = 'NO_BOND_ORDER' = use only one CONECT even for higher bond orders
     * 16 = 'MASTER' = write MASTER record
     * 32 = 'TER' = write TER record

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param flavor: Output flavor bit flags
    :type flavor: An integer or string of '|'- or ','-separated terms
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "pdb"),
        writer_args={
            "flavor": flavor,
        },
        errors=errors,
    )


_inchi_fmt = get_format("inchi")


@_public
def from_inchi(
    content: str,
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "inchi", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchi_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchi(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "inchi", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchi_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def from_inchi_file(
    source: _Union[None, str, _BinaryIO],
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    errors: str = "strict",
):
    """Parse an InChI string and id file using the RDKit toolkit

    This is mostly equivalent to calling:
       read_molecules(source, "inchi", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating RDKit molecules
    """
    # Using the higher-level/public API
    return read_molecules(
        source,
        _io.get_compressed_input_format(source, "inchi"),
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
        },
        errors=errors,
    )


@_public
def to_inchi_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChI string and id from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchi", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchi"),
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchistring_fmt = get_format("inchistring")


@_public
def from_inchistring(
    content: str,
    *,
    sanitize: bool = True,
    removeHs: bool = True,
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    errors: str = "strict",
):
    """Parse an InChI string using the RDKit toolkit

    This is equivalent to calling:

       parse_molecule(content, "inchistring", reader_args={...}, errors=errors)

    :param sanitize: If true, sanitize the molecule after parsing
    :type sanitize: Boolean (default: True)
    :param removeHs: If true, remove simple hydrogens from the molecular graph
    :type removeHs: Boolean (default: True)
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    id, mol = _inchistring_fmt._format_config.parse_id_and_molecule(
        content,
        id_tag=None,
        reader_args={
            "sanitize": sanitize,
            "removeHs": removeHs,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
        },
        error_handler=error_handler,
    )
    return mol


@_public
def to_inchistring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    errors: str = "strict",
):
    """Generate an InChI string from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "inchistring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchistring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
        },
        error_handler=error_handler,
    )


_inchikey_fmt = get_format("inchikey")


@_public
def to_inchikey(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "inchikey", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikey_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        error_handler=error_handler,
    )


@_public
def to_inchikey_file(
    destination: _Union[None, str, _BinaryIO],
    *,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    delimiter: _Optional[
        _Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
    ] = None,
    include_id: bool = True,
    errors: str = "strict",
):
    """Generate an InChIKey string and id from an RDKit molecule

    This is mostly equivalent to calling:
       open_molecule_writer(destination, "inchikey", writer_args={...}, errors=errors)

    :param destination: where to write the molecules
    :type destination: None, a filename string, or a file-like object
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param delimiter: The separator between the SMILES and the id
    :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
    :param include_id: if true, include the molecule id in the output
    :type include_id: Boolean (default: True)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting RDKit molecules
    """
    # Using the higher-level/public API
    return open_molecule_writer(
        destination,
        _io.get_compressed_output_format(destination, "inchikey"),
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
            "delimiter": delimiter,
            "include_id": include_id,
        },
        errors=errors,
    )


_inchikeystring_fmt = get_format("inchikeystring")


@_public
def to_inchikeystring(
    mol: _Any,
    *,
    id: _Optional[str] = None,
    options: str = "",
    logLevel: _Optional[int] = None,
    treatWarningAsError: bool = False,
    errors: str = "strict",
):
    """Generate an InChIKey string from an RDKit molecule

    This is equivalent to calling:
       create_string(mol, "inchikeystring", id=id, writer_args={...}, errors=errors)

    :param mol: a molecule object
    :type mol: an RDKit molecule
    :param id: an alternate identifier for the output record, if relevant
    :type id: None or a string (default: None)
    :param options: an configuration string to pass to the InChI API
    :type options: a string (default: "")
    :param logLevel: the log level for the InChI API
    :type logLevel: an integer, or None to disable logging completely (default: None)
    :param treatWarningAsError: treat any InChI warnings as an error
    :type treatWarningAsError: Boolean (default: False)
    :param errors: specify how to handle errors
    :type errors: one of "strict", "ignore", or "log"
    :returns: an RDKit molecule object
    """
    # Using the lower-level/internal API
    error_handler = _io.get_parse_error_handler(errors)
    return _inchikeystring_fmt._format_config.create_string(
        mol,
        id=id,
        writer_args={
            "options": options,
            "logLevel": logLevel,
            "treatWarningAsError": treatWarningAsError,
        },
        error_handler=error_handler,
    )


# Aliases

from_smiles = from_smistring

to_smiles = to_smistring

#### End of auto-generated code.
