# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

try:
    from functools import cache
except ImportError:
    # Python 3.7 support
    from functools import lru_cache
    def cache(f):
        return lru_cache()(f)

from rdkit import Chem
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem.MACCSkeys import GenMACCSKeys
from rdkit import DataStructs

# If this attribute exists then we're in a Sphinx build environment
_using_sphinx_mock = hasattr(Chem, "using_sphinx_mock")

try:
    from rdkit.Avalon import pyAvalonTools
    HAS_AVALON = True
except ImportError:
    pyAvalonTools = None
    HAS_AVALON = False

from . import rdkit_toolkit as _tk
from . import types

from . import io
from .encodings import from_binary_lsb as _from_binary_lsb

from ._compat import Any, Optional, Union, Literal


# Used to check for version-dependent fingerprints
@cache
def get_version_probe_mol():
    return Chem.MolFromSmiles(r"CC1=CC(=NN1CC(=O)NNC(=O)\C=C\C2=C(C=CC=C2Cl)F)C")

software = _tk.software
from . import SOFTWARE as CHEMFP_SOFTWARE

__all__ = []
def _public(obj):
    __all__.append(obj.__name__)
    return obj

########


all_arguments = types.FingerprintArguments()
_add = all_arguments.add_argument

if _using_sphinx_mock:
    class _tk:
        software = ""
    CHEMFP_SOFTWARE = ""

@_public
class RDKitBaseFingerprintType(types.ThreadsafeFingerprinterMixin,
                               types.FingerprintType):
    #: a description of the RDKit and chemfp software packages used
    software = types._sphinx_repr_if(
        _using_sphinx_mock, "...", _tk.software + " " + CHEMFP_SOFTWARE,
        )
    
    #: a reference to the :mod:`.rdkit_toolkit` module
    toolkit = types._sphinx_repr_if(
        _using_sphinx_mock, "<module 'chemfp.rdkit_toolkit'>", _tk,
        )

    #### The following was auto-generated by: utils/mol_io_codegen.py --fp rdkit

    def from_smistring(
        self,
        content: str,
        *,
        sanitize: bool = True,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string

        This is equivalent to calling::

           mol = fptype.toolkit.from_smistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smistring(
            content,
            sanitize=sanitize,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_smi(
        self,
        content: str,
        *,
        sanitize: bool = True,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a SMILES string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_smi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_smi(
            content,
            sanitize=sanitize,
            delimiter=delimiter,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_sdf(
        self,
        content: str,
        *,
        sanitize: bool = True,
        removeHs: bool = True,
        strictParsing: bool = True,
        includeTags: bool = True,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an SDF record

        This is equivalent to calling::

           mol = fptype.toolkit.from_sdf(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param removeHs: If true, remove simple hydrogens from the molecular graph
        :type removeHs: Boolean (default: True)
        :param strictParsing: If true, require stricter adherence to the SDF specification
        :type strictParsing: Boolean (default: True)
        :param includeTags: if true, extract the struture data tag fields
        :type includeTags: Boolean (default: True)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_sdf(
            content,
            sanitize=sanitize,
            removeHs=removeHs,
            strictParsing=strictParsing,
            includeTags=includeTags,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_molfile(
        self,
        content: str,
        *,
        sanitize: bool = True,
        removeHs: bool = True,
        strictParsing: bool = True,
        errors: str = "strict",
    ):
        """Generate a fingerprint from a molfile

        This is equivalent to calling::

           mol = fptype.toolkit.from_molfile(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param removeHs: If true, remove simple hydrogens from the molecular graph
        :type removeHs: Boolean (default: True)
        :param strictParsing: If true, require stricter adherence to the SDF specification
        :type strictParsing: Boolean (default: True)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_molfile(
            content,
            sanitize=sanitize,
            removeHs=removeHs,
            strictParsing=strictParsing,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchi(
        self,
        content: str,
        *,
        sanitize: bool = True,
        removeHs: bool = True,
        logLevel: Optional[int] = None,
        treatWarningAsError: bool = False,
        delimiter: Optional[
            Literal["to_eol", "space", "tab", "comma", "whitespace", "native", " ", "\t"]
        ] = None,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string and id

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchi(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param removeHs: If true, remove simple hydrogens from the molecular graph
        :type removeHs: Boolean (default: True)
        :param logLevel: the log level for the InChI API
        :type logLevel: an integer, or None to disable logging completely (default: None)
        :param treatWarningAsError: treat any InChI warnings as an error
        :type treatWarningAsError: Boolean (default: False)
        :param delimiter: The separator between the SMILES and the id
        :type delimiter: One of None, 'to_eol', 'space', 'tab', 'comma', 'whitespace', 'native', or the space or tab characters (default: None)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchi(
            content,
            sanitize=sanitize,
            removeHs=removeHs,
            logLevel=logLevel,
            treatWarningAsError=treatWarningAsError,
            delimiter=delimiter,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    def from_inchistring(
        self,
        content: str,
        *,
        sanitize: bool = True,
        removeHs: bool = True,
        logLevel: Optional[int] = None,
        treatWarningAsError: bool = False,
        errors: str = "strict",
    ):
        """Generate a fingerprint from an InChI string

        This is equivalent to calling::

           mol = fptype.toolkit.from_inchistring(content, ..., errors=errors)
           fp = fptype.from_mol(mol) if (mol is not None) else None

        :param sanitize: If true, sanitize the molecule after parsing
        :type sanitize: Boolean (default: True)
        :param removeHs: If true, remove simple hydrogens from the molecular graph
        :type removeHs: Boolean (default: True)
        :param logLevel: the log level for the InChI API
        :type logLevel: an integer, or None to disable logging completely (default: None)
        :param treatWarningAsError: treat any InChI warnings as an error
        :type treatWarningAsError: Boolean (default: False)
        :param errors: specify how to handle errors
        :type errors: one of "strict", "ignore", or "log"
        :returns: a fingerprint byte string
        """
        mol = self.toolkit.from_inchistring(
            content,
            sanitize=sanitize,
            removeHs=removeHs,
            logLevel=logLevel,
            treatWarningAsError=treatWarningAsError,
            errors=errors,
        )
        return self.from_mol(mol) if (mol is not None) else None

    # Aliases

    from_smiles = from_smistring

    #### End of auto-generated code.


@_public
class VariableSizeFingerprint(RDKitBaseFingerprintType):
    """This is a variable-size fingerprint type, specified by the user"""
    def __init__(self, fingerprint_kwargs):
        super(VariableSizeFingerprint, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["fpSize"]

@_public
class FixedSizeFingerprint(RDKitBaseFingerprintType):
    """This is a fixed-size fingerprint type"""
    pass


###### The topological fingerprinter

# TODO: Catch up to all of the options added since chemfp-1.1
#   (need to add "branchedPaths" and "useBondOrder"
# 
# RDKFingerprint( (Mol) mol, (int) minPath=1, (int) maxPath=7, (int) fpSize=2048,
#                 (int) nBitsPerHash=2, (bool) useHs=True, (float) tgtDensity=0.0,
#                 (int) minSize=128, (bool) branchedPaths=True  (bool) useBondOrder=True,
#                 (AtomPairsParameters) atomInvariants=0, (AtomPairsParameters) fromAtoms=0,
#                 (AtomPairsParameters) atomBits=None)

def _check_fromAtoms(fromAtoms):
    if fromAtoms is None:
        return
    if len(fromAtoms) == 0:
        raise ValueError("fromAtoms must contain at least one atom index")
    for atom_index in fromAtoms:
        if atom_index < 0:
            raise ValueError("fromAtoms must contain non-negative integers")
    return

# Some constants shared by the fingerprinter and the command-line code.

NUM_BITS = 2048
MIN_PATH = 1
MAX_PATH = 7
BITS_PER_HASH = 2
USE_HS = 1
assert USE_HS == 1, "Don't make this 0 unless you know what you are doing"

RDK_FINGERPRINT_branchedPaths=1
RDK_FINGERPRINT_useBondOrder=1


def make_rdk_fingerprinter(minPath=MIN_PATH, maxPath=MAX_PATH, fpSize=NUM_BITS,
                           nBitsPerHash=BITS_PER_HASH, useHs=USE_HS,
                           branchedPaths=RDK_FINGERPRINT_branchedPaths,
                           useBondOrder=RDK_FINGERPRINT_useBondOrder,
                           fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    if not (minPath > 0):
        raise ValueError("minPath must be 1 or greater")
    if not (maxPath >= minPath):
        raise ValueError("maxPath must not be smaller than minPath")
    if not (nBitsPerHash > 0):
        raise ValueError("nBitsPerHash must be 1 or greater")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def rdk_fingerprinter(mol):
            fp = Chem.RDKFingerprint(
                mol, minPath=minPath, maxPath=maxPath, fpSize=fpSize,
                nBitsPerHash=nBitsPerHash, useHs=useHs,
                branchedPaths=branchedPaths, useBondOrder=useBondOrder)
            return DataStructs.BitVectToBinaryText(fp)
    else:
        def rdk_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = Chem.RDKFingerprint(
                mol, minPath=minPath, maxPath=maxPath, fpSize=fpSize,
                nBitsPerHash=nBitsPerHash, useHs=useHs,
                branchedPaths=branchedPaths, useBondOrder=useBondOrder,
                fromAtoms=new_fromAtoms)
            return DataStructs.BitVectToBinaryText(fp)
    
    return rdk_fingerprinter

generic_fpSize = types.FingerprintArgument(
    name="fpSize", default=None, metavar="INT", help=
        "number of bits in the fingerprint. Default of 2048 for RDK, Morgan, "
        "topological torsion, atom pair, pattern and SECFP fingerprints, "
        "and 512 for Avalon fingerprints",
    decoder=types.positive_int)

generic_radius = types.FingerprintArgument(
    name="radius", default=None, metavar="INT", help=
        "radius for the Morgan or SECFP fingerprints. Default of 2 for Morgan, 3 for SECFP",
    decoder=types.nonnegative_int)

generic_nBitsPerEntry = types.FingerprintArgument(
    name = "nBitsPerEntry", default=None, metavar = "INT",
    decoder = types.positive_int, 
    help = "number of bits per entry")

generic_includeChirality = types.FingerprintArgument(
    name = "includeChirality", default=None, metavar="0|1",
    decoder = types.zero_or_one,
    encoder = types.encode_bool_as_zero_or_one,
    help = "include chirality information in the atom invariants")


_add("fpSize", decoder=types.positive_int, metavar="INT", default=NUM_BITS,
     help = "number of bits in the fingerprint (applies to RDK, Morgan, topological torsion, atom pair, and pattern fingerprints")

_add("minPath", decoder=types.positive_int, metavar="INT", default=MIN_PATH,
     help = "minimum number of bonds to include in the subgraph")

_add("maxPath", decoder=types.positive_int, metavar="INT", default=MAX_PATH,
     help = "maximum number of bonds to include in the subgraph")

_add("nBitsPerHash", decoder=types.positive_int, metavar="INT",
     default=BITS_PER_HASH, help = "number of bits to set per path")

_add("useHs",
         decoder = types.zero_or_one, metavar = "0|1",
         encoder = types.encode_bool_as_zero_or_one,
         default = USE_HS,
         help = "include information about the number of hydrogens on each atom")

def encode_fromAtom_list(fromAtoms):
    if not fromAtoms:
        return None
    return ",".join(("%d" % index) for index in fromAtoms)

_add("fromAtoms", decoder=types.fromAtom_list, encoder=encode_fromAtom_list, metavar="INT,INT,...",
     default=None, help = "specify the atom indices to use")


# Not supporting the tgtDensity and minSize options.
# This program generates fixed-length fingerprints.

# Figure out which RDK fingerprint versions
@cache
def RDK_VERSION():
    return {
        b"~\x89\x90\x04\xdb\xe9t\xc5\x87\xfe\xfc\xbf\xf8r\x01\xd6\xe9\xd7\xbbZv\xfeM\xc9\xfb\xf7{\xb7\xbd\xacl\xc9\xff\xb6s\xd5\x84\xf2\x8e\xdb\xcf\xcc\xb7\x02\xfa-\xa6\xee2\x1c\x19\x11\xe8q;\x08\x96NTN\xb2\xd7\x99\xa9\xca\xe7\xc3\x0c\x9bqP!\xd5\xe6\xf5\xff\x92\xf6\xed\xfa\xc4\xb4x\xf5\xd6\xde~\xfb\xcdp\xefv\xfb\xcd\xbb\x7f\xb9\xda\xf8?\xad\xd9\xf6\xe4\x94\xc3\xfb\xbf\xefv9\xff\xfe\x9bd\xf3\x15\x9d\x9bm\xeb\x0f\xaa$\xf5\x01\xef\xb7": "1",
        b"""\xcd\xd7\xb9\xc7\xd8\xa7\x18\n\xff\x90R\xc2\xfbS\xf5\xba\xd1\xf5\xf6F\x0f\x8d\xf1\x07\xbf\t_\x995`\xcaH\xf1>\x045q\'5\xce>\x8f\xf5\x18L\xb6\xae\xca\xde\xd5\xbd\xcd\xff/K"m\xdc\xec\xd3\x02\xb9\xf5\xc0\xb6_\xe2nv\xe36\xf5\xd5j\x1f\x90\xc6\xd1/\xb3]"=[\xbd\xb0F\xb9\xcfd\x96[e\xf7x\xafQ\xc7\xfa\xb3\xbf\xf3\xdfb\xb7]\xd7\xfe\xb7\xd6\xfd\x1f{\xd7\xbd\xf1\'P\xfaU\x89l^f\xde\xb5}\xaf""": "2",
        }[make_rdk_fingerprinter(fpSize=1024)(get_version_probe_mol())]

# Version 3.2 added support for fromAtoms. The type string needs to be
# backwards compatible to strings without 'fromAtoms=', so only add the
# term if fromAtoms is actually used.
## class FormatFromAtoms(object):
##     def __init__(self, template):
##         self.template = template
##     def format(self, **kwargs):
##         fromAtoms_str = kwargs["fromAtoms"]
##         del kwargs["fromAtoms"]
##         output_str = self.template.format(**kwargs)
##         if fromAtoms_str:
##             output_str += " fromAtoms=" + fromAtoms_str
##         return output_str

class FormatOptionalParameters(object):
    def __init__(self, template, optional_defaults):
        self.template = template
        self.optional_defaults = optional_defaults
        
    def format(self, **kwargs):
        output_str = self.template.format(**kwargs)

        for param_name, ignore_val in self.optional_defaults:
            val = kwargs.pop(param_name)
            if val != ignore_val:
                output_str += " %s=%s" % (param_name, val)
        return output_str

@_public
@types.adjust_call_doc
class RDKitFingerprintType_v1(VariableSizeFingerprint):
    """RDKit's Daylight-like fingerprint based on linear path and branched tree enumeration, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.RDKFingerprint

    The RDKit-Fingerprint/1 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * minPath - minimum number of bonds (default: 1)
    * maxPath - maximum number of bonds (default: 7)
    * nBitsPerHash - number of bits to set for each path hash (default: 2)
    * useHs - include information about the number of hydrogens on each atom? (default: True)

    Note: this version is only available in ancient (pre-2014) versions of RDKit 
    """
    name = "RDKit-Fingerprint/1"
    _make_fingerprinter = staticmethod(make_rdk_fingerprinter)
    _arguments = all_arguments.select_names("minPath maxPath fpSize nBitsPerHash useHs")
    _format_string = ("minPath={minPath} maxPath={maxPath} fpSize={fpSize} "
                      "nBitsPerHash={nBitsPerHash} useHs={useHs}")
    __call__ = _arguments.get_call_method()

rdkit_fingerprint_arguments = all_arguments.select_names(
    "minPath maxPath fpSize nBitsPerHash useHs fromAtoms")

rdkit_fingerprint_arguments.add_argument(
    "branchedPaths", decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = RDK_FINGERPRINT_branchedPaths,
    help="if set both branched and unbranched paths will be used in the fingerprint")

rdkit_fingerprint_arguments.add_argument(
    "useBondOrder", decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = RDK_FINGERPRINT_useBondOrder,
    help="if set both bond orders will be used in the path hashes")
    
@_public
@types.adjust_call_doc
class RDKitFingerprintType_v2(VariableSizeFingerprint):
    """RDKit's Daylight-like fingerprint based on linear path and branched tree enumeration, version 2

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.RDKFingerprint

    The RDKit-Fingerprint/2 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * minPath - minimum number of bonds (default: 1)
    * maxPath - maximum number of bonds (default: 7)
    * nBitsPerHash - number of bits to set for each path hash (default: 2)
    * useHs - include information about the number of hydrogens on each atom? (default: True)
    * branchedPaths - include both branched and unbranched paths (default: True)
    * useBondOrder - use both bond orders in the path hashes (default: True)
    * fromAtoms - a comma-separated list of atom indices which must be part of the path enumeration
    """
    name = "RDKit-Fingerprint/2"
    _make_fingerprinter = staticmethod(make_rdk_fingerprinter)
    _arguments = rdkit_fingerprint_arguments
    _format_string = FormatOptionalParameters(
        "minPath={minPath} maxPath={maxPath} fpSize={fpSize} "
        "nBitsPerHash={nBitsPerHash} useHs={useHs}",
        (("branchedPaths", "1"), ("useBondOrder", "1"), ("fromAtoms", None)))
    __call__ = _arguments.get_call_method()

def register_rdk(registry):
    registry.register_if(RDK_VERSION() == "1", RDKitFingerprintType_v1)
    registry.register_if(RDK_VERSION() == "2", RDKitFingerprintType_v2)


########### The MACCS fingerprinter


def maccs166_fingerprinter(mol):
    try:
        fp = GenMACCSKeys(mol)
    except RuntimeError:
        # RDKit doesn't like to process the empty molecule. (tested Dec. 2013)
        # Return the empty fingerprint for that case.
        if mol.GetNumAtoms() == 0:
            return "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        raise  # pragma: no cover - don't think it's possible to get here
        
    # In RDKit the first bit is always bit 1 .. bit 0 is empty (?!?!)
    bitstring_with_167_bits = fp.ToBitString()
    # I want the bits to start at 0, so I do a manual left shift
    # XXX optimize? This takes about 1-2% of the overall run time,
    # which doesn't seem like enough to worry about.
    return _from_binary_lsb(bitstring_with_167_bits[1:])[1]

def make_maccs166_fingerprinter():
    return maccs166_fingerprinter


@cache
def MACCS_VERSION():
    probe_mol = Chem.MolFromSmiles("[He]")
    has_key_44 = GenMACCSKeys(probe_mol)[44] != 0
    return {False: "1", True: "2"}[has_key_44]

@_public
@types.adjust_call_doc
class RDKitMACCSFingerprintType_v1(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """RDKit's implementation of the 166 MACCS keys, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html#rdkit.Chem.rdMolDescriptors.GetMACCSKeysFingerprint

    The RDKit-MACCS166/1 fingerprints have no parameters.

    This comes from an ancient version of RDKit which does not support MACCS key 44 ("OTHER").
    """
    num_bits = 166
    name = "RDKit-MACCS166/1"
    _make_fingerprinter = staticmethod(make_maccs166_fingerprinter)
    __call__ = types.FingerprintArguments().get_call_method()


@_public
@types.adjust_call_doc
class RDKitMACCSFingerprintType_v2(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """RDKit's implementation of the 166 MACCS keys, version 2

    See https://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html#rdkit.Chem.rdMolDescriptors.GetMACCSKeysFingerprint

    The RDKit-MACCS166/2 fingerprints have no parameters. RDKit version
    added this version in late 2014 to support MACCS key 44 ("OTHER").
    """
    num_bits = 166
    name = "RDKit-MACCS166/2"
    _make_fingerprinter = staticmethod(make_maccs166_fingerprinter)
    __call__ = types.FingerprintArguments().get_call_method()

def register_maccs166(registry):
    registry.register_if(MACCS_VERSION() == "1", RDKitMACCSFingerprintType_v1)
    registry.register_if(MACCS_VERSION() == "2", RDKitMACCSFingerprintType_v2)


########### The Morgan fingerprinter

# GetMorganFingerprintAsBitVect(
#     (Mol)mol, (int)radius , (int)nBits=2048,
#     (AtomPairsParameters)invariants=[], (AtomPairsParameters) fromAtoms=[],
#     (bool)useChirality=False, (bool)useBondTypes=True, (bool)useFeatures=False,
#     (AtomPairsParameters)bitInfo=None,
#     (bool)includeRedundantEnvironments=False,  # Added 2020-03
# )
        

# Some constants shared by the fingerprinter and the command-line code.

MORGAN_radius = 2
MORGAN_useFeatures = 0
MORGAN_useChirality = 0
MORGAN_useBondTypes = 1
MORGAN_includeRedundantEnvironments = 0

_GetMorganFingerprintAsBitVect = rdMolDescriptors.GetMorganFingerprintAsBitVect
if _using_sphinx_mock or ("includeRedundantEnvironments" in _GetMorganFingerprintAsBitVect.__doc__):
    GetMorganFingerprintAsBitVect = _GetMorganFingerprintAsBitVect
else:
    _issued_includeRedundantEnvironments_warning = False
    def GetMorganFingerprintAsBitVect(*args, **kwargs):
        global _issued_includeRedundantEnvironments_warning
        val = kwargs.pop("includeRedundantEnvironments")
        if val and not _issued_includeRedundantEnvironments_warning:
            _issued_includeRedundantEnvironments_warning = True
            import sys
            sys.stderr.write("WARNING: Ignoring RDKit-Morgan includeRedundantEnvironments=1, "
                                 "which this version of RDKit does not support.\n")

        return _GetMorganFingerprintAsBitVect(*args, **kwargs)

def make_morgan_fingerprinter(fpSize=NUM_BITS,
                              radius=MORGAN_radius,
                              useFeatures=MORGAN_useFeatures,
                              useChirality=MORGAN_useChirality,
                              useBondTypes=MORGAN_useBondTypes,
                              includeRedundantEnvironments=MORGAN_includeRedundantEnvironments,
                              fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    if not (radius >= 0):
        raise ValueError("radius must be 0 or greater")
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        def morgan_fingerprinter(mol):
            fp = GetMorganFingerprintAsBitVect(
                mol, radius, nBits=fpSize, useChirality=useChirality,
                useBondTypes=useBondTypes, useFeatures=useFeatures,
                includeRedundantEnvironments=includeRedundantEnvironments)
            return DataStructs.BitVectToBinaryText(fp)
    else:
        def morgan_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = GetMorganFingerprintAsBitVect(
                mol, radius, nBits=fpSize, useChirality=useChirality,
                useBondTypes=useBondTypes, useFeatures=useFeatures,
                includeRedundantEnvironments=includeRedundantEnvironments,
                fromAtoms=new_fromAtoms)
            return DataStructs.BitVectToBinaryText(fp)

    return morgan_fingerprinter


# Morgan
morgan_arguments = types.FingerprintArguments()

morgan_arguments = all_arguments.select_names("fpSize fromAtoms")

morgan_arguments.add_argument(
    "radius", decoder=types.nonnegative_int, metavar="INT",
    default = MORGAN_radius,
    help = "radius for the Morgan algorithm")

morgan_arguments.add_argument(
    "useFeatures", decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = MORGAN_useFeatures,
    help = "use chemical-feature invariants")

morgan_arguments.add_argument(
    "useChirality", decoder = types.zero_or_one, metavar="0|1",
    encoder=types.encode_bool_as_zero_or_one,
    default = MORGAN_useChirality,
    help = "include chirality information")

morgan_arguments.add_argument(
    "useBondTypes", decoder = types.zero_or_one, metavar="0|1",
    encoder=types.encode_bool_as_zero_or_one,
    default = MORGAN_useBondTypes,
    help = "include bond type information")

morgan_arguments.add_argument(
    "includeRedundantEnvironments", decoder = types.zero_or_one, metavar = "0|1",
    encoder=types.encode_bool_as_zero_or_one,
    default = MORGAN_includeRedundantEnvironments,
    help = "if set, the check for redundant atom environments will not be done")


@_public
@types.adjust_call_doc
class RDKitMorganFingerprintType_v1(VariableSizeFingerprint):
    """RDKit Morgan (ECFP-like) fingerprints, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html#rdkit.Chem.rdMolDescriptors.GetMorganFingerprintAsBitVect

    The RDKit-Morgan/1 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * radius - radius for the Morgan algorithm (default: 2)
    * useFeatures - use chemical-feature invariants (default: 0)
    * useChirality - use chirality information (default: 0)
    * useBondTypes - include bond type information (default: 1)
    * includeRedundantEnvironments - if set, the check for redundant atom
      environments will not be done (added in RDKit 2020-3) (default: 0)
    * fromAtoms - a comma-separated list of atom indices to use as centers
    """
    name = "RDKit-Morgan/1"
    _arguments = morgan_arguments
    _format_string = FormatOptionalParameters(
        "radius={radius} fpSize={fpSize} useFeatures={useFeatures} "
        "useChirality={useChirality} useBondTypes={useBondTypes}",
        (("includeRedundantEnvironments", "0"), ("fromAtoms", None)))
    _make_fingerprinter = staticmethod(make_morgan_fingerprinter)
    __call__ = _arguments.get_call_method()

def register_morgan(registry):
    registry.register(RDKitMorganFingerprintType_v1)

########### Atom Pair fingerprinter

# GetHashedAtomPairFingerprintAsBitVect(
#   (Mol) mol, (int) nBits=2048, (int) minLength=1, (int) maxLength=30,
#   (AtomPairsParameters) fromAtoms=0, (AtomPairsParameters) ignoreAtoms=0,
#   (AtomPairsParameters) atomInvariants=0, (int) nBitsPerEntry=4,
#   (bool)includeChirality=False, (bool)use2D=True

MIN_LENGTH = 1
MAX_LENGTH = 30
ATOM_PAIR_nBitsPerEntry = 4
ATOM_PAIR_includeChirality = 0
ATOM_PAIR_use2D = 1

def make_atom_pair_fingerprinter(fpSize=NUM_BITS,
                                 minLength=MIN_LENGTH,
                                 maxLength=MAX_LENGTH,
                                 nBitsPerEntry=ATOM_PAIR_nBitsPerEntry,
                                 includeChirality=ATOM_PAIR_includeChirality,
                                 use2D=ATOM_PAIR_use2D,
                                 fromAtoms=None):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    if not (minLength >= 0):
        raise ValueError("minLength must be 0 or greater")
    if not (maxLength >= minLength):
        raise ValueError("minLength must not be larger than maxLength")
    if not (nBitsPerEntry >= 1):
        raise ValueError("nBitsPerEntry must be 1 or greater")
    
    _check_fromAtoms(fromAtoms)

    if fromAtoms is None:
        if use2D:
            def pair_fingerprinter(mol):
                fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                    mol, nBits=fpSize, minLength=minLength, maxLength=maxLength,
                    nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality,
                    use2D=use2D)
                return DataStructs.BitVectToBinaryText(fp)
        else:
            def pair_fingerprinter(mol):
                # return an empty fingerprint if there are no conformers
                if not mol.GetNumConformers():
                    return b"\0" * ((fpSize+7)//8)
                fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                    mol, nBits=fpSize, minLength=minLength, maxLength=maxLength,
                    nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality,
                    use2D=False)
                return DataStructs.BitVectToBinaryText(fp)
    else:
        if use2D:
            def pair_fingerprinter(mol):
                n = mol.GetNumAtoms()
                new_fromAtoms = [i for i in fromAtoms if i < n]
                if not new_fromAtoms:
                    return b"\0" * ((fpSize+7)//8)
                fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                    mol, nBits=fpSize, minLength=minLength, maxLength=maxLength,
                    fromAtoms=new_fromAtoms,
                    nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality,
                    use2D=use2D)

                return DataStructs.BitVectToBinaryText(fp)
        else:
            def pair_fingerprinter(mol):
                # return an empty fingerprint if there are no conformers
                if not mol.GetNumConformers():
                    return b"\0" * ((fpSize+7)//8)
                n = mol.GetNumAtoms()
                new_fromAtoms = [i for i in fromAtoms if i < n]
                if not new_fromAtoms:
                    return b"\0" * ((fpSize+7)//8)
                fp = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(
                    mol, nBits=fpSize, minLength=minLength, maxLength=maxLength,
                    fromAtoms=new_fromAtoms,
                    nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality,
                    use2D=use2D)

                return DataStructs.BitVectToBinaryText(fp)
    return pair_fingerprinter

@cache
def ATOM_PAIR_VERSION():
    return {
        b"\xfdB\xfe\xbd\xfa\xdd\xff\xf5\xff\x05\xdf?\xe3\xc3\xff\xfb": "1",
        b"w\xf7\xff\xf7\xff\x17\x01\x7f\x7f\xff\xff\x7f\xff?\xff\xff": "2",
        }[make_atom_pair_fingerprinter(128)(get_version_probe_mol())]

atom_pair_arguments = all_arguments.select_names("fpSize fromAtoms")

# pair
atom_pair_arguments.add_argument(
    "minLength",
    decoder = types.nonnegative_int, metavar = "INT",
    default = MIN_LENGTH,
    help = "minimum bond count for a pair")

atom_pair_arguments.add_argument(
    "maxLength",
    decoder = types.nonnegative_int, metavar = "INT",
    default = MAX_LENGTH,
    help = "maximum bond count for a pair")

atom_pair_arguments.add_argument(
    "nBitsPerEntry",
    decoder = types.positive_int, metavar = "INT",
    default = ATOM_PAIR_nBitsPerEntry,
    help = "number of bits per entry")

atom_pair_arguments.add_argument(
    "includeChirality",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = ATOM_PAIR_includeChirality,
    help = "include chirality information in the atom invariants")

atom_pair_arguments.add_argument(
    "use2D",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = ATOM_PAIR_use2D,
    help = "use 2D instead of 3D distance matrix")

@_public
class RDKitBaseAtomPairFingerprintType(VariableSizeFingerprint):
    """Base class for the RDKitAtomPair fingerprint types"""
    _make_fingerprinter = staticmethod(make_atom_pair_fingerprinter)

@_public
@types.adjust_call_doc
class RDKitAtomPairFingerprint_v1(RDKitBaseAtomPairFingerprintType):
    """RDKit atom pair fingerprints, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html#rdkit.Chem.rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect

    The RDKit-AtomPair/1 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * minLength - minimum bond count for a pair (default: 1)
    * maxLength - maximum bond count for a pair (default: 30)
    
    Note: this version was only available in ancient (pre-2012) versions of RDKit.
    Chemfp no longer supports those versions of RDKit.
    """
    name = "RDKit-AtomPair/1"
    _arguments = atom_pair_arguments.select_names("fpSize minLength maxLength")
    _format_string = "fpSize={fpSize} minLength={minLength} maxLength={maxLength}"
    __call__ = _arguments.get_call_method()

@_public
@types.adjust_call_doc
class RDKitAtomPairFingerprint_v2(RDKitBaseAtomPairFingerprintType):
    """RDKit atom pair fingerprints, version 2

    See https://rdkit.org/docs/source/rdkit.Chem.rdMolDescriptors.html#rdkit.Chem.rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect

    The RDKit-AtomPair/2 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * minLength - minimum bond count for a pair (default: 1 bond)
    * maxLength - maximum bond count for a pair (default: 30, max: 63)
    * nBitsPerEntry - number of bits to use in simulating counts (default: 4)
    * includeChirality - if set, chirality will be used in the atom invariants (default: 0)
    * use2D - if 1, use a 2D distance matrix, if 0 use the 3D matrix from the first
        set of conformers, or return an empty fingerprint if no conformers (default: 1)
    * fromAtoms - a comma-separated list of atom indices which must be in the pair
    """
    name = "RDKit-AtomPair/2"
    _arguments = atom_pair_arguments
    _format_string = FormatOptionalParameters(
        "fpSize={fpSize} minLength={minLength} maxLength={maxLength}",
        (("nBitsPerEntry", "4"), ("includeChirality", "0"), ("use2D", "1"), ("fromAtoms", None))
        )
    __call__ = _arguments.get_call_method()
         
def register_atom_pair(registry):
    registry.register_if(ATOM_PAIR_VERSION() == "1", RDKitAtomPairFingerprint_v1)
    registry.register_if(ATOM_PAIR_VERSION() == "2", RDKitAtomPairFingerprint_v2)
        
 
########### Torsion fingerprinter

# GetHashedTopologicalTorsionFingerprintAsBitVect( 
#   (Mol) mol, (int) nBits=2048, (int) targetSize=4, (AtomPairsParameters) fromAtoms=0,
#   (AtomPairsParameters) ignoreAtoms=0, (AtomPairsParameters) atomInvariants=0,
#   (int) nBitsPerEntry=4, (bool) includeChirality=False)

# A bit complicated because the torsion type changed over time

TARGET_SIZE = 4
TORSION_nBitsPerEntry = 4
TORSION_includeChirality = 0

def make_torsion_fingerprinter(fpSize=NUM_BITS,
                               targetSize=TARGET_SIZE,
                               fromAtoms=None,
                               nBitsPerEntry=TORSION_nBitsPerEntry,
                               includeChirality=TORSION_includeChirality):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    if not (targetSize >= 0):
        raise ValueError("targetSize must be 0 or greater")
    _check_fromAtoms(fromAtoms)
    if not (nBitsPerEntry > 0):
        raise ValueError("nBitsPerEntry must be 1 or greater")

    if fromAtoms is None:
        def torsion_fingerprinter(mol):
            fp = rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(
                mol, nBits=fpSize, targetSize=targetSize,
                nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality)
            return DataStructs.BitVectToBinaryText(fp)
    else:
        def torsion_fingerprinter(mol):
            n = mol.GetNumAtoms()
            new_fromAtoms = [i for i in fromAtoms if i < n]
            if not new_fromAtoms:
                return b"\0" * ((fpSize+7)//8)
            fp = rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect(
                mol, nBits=fpSize, targetSize=targetSize,
                fromAtoms=new_fromAtoms,
                nBitsPerEntry=nBitsPerEntry, includeChirality=includeChirality)
            return DataStructs.BitVectToBinaryText(fp)
    return torsion_fingerprinter

@cache
def TORSION_VERSION():
    return {
        b"\xc2\x10@\x83\x010\x18\xa4,\x00\x80B\xc0\x00\x08\x00": "1",
        b"\x13\x11\x103\x00\x007\x00\x00p\x01\x111\x0107": "2",
        }[make_torsion_fingerprinter(128)(get_version_probe_mol())]

# torsion
_add("targetSize", decoder=types.positive_int, metavar="INT",
     default=TARGET_SIZE, help = "number of bonds per torsion")


@_public
@types.adjust_call_doc
class RDKitBaseTorsionFingerprintType(VariableSizeFingerprint):
    _arguments = all_arguments.select_names("fpSize targetSize")
    _format_string = "fpSize={fpSize} targetSize={targetSize}"
    _make_fingerprinter = staticmethod(make_torsion_fingerprinter)
    __call__ = _arguments.get_call_method()


@_public
class RDKitTorsionFingerprintType_v1(RDKitBaseTorsionFingerprintType):
    """RDKit torsion fingerprints, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.AtomPairs.Torsions.html

    An implementation of Topological-torsion fingerprints, as
    described in: R. Nilakantan, N. Bauman, J. S. Dixon,
    R. Venkataraghavan; "Topological Torsion: A New Molecular
    Descriptor for SAR Applications.  Comparison with Other
    Descriptors" JCICS 27, 82-85 (1987).

    The RDKit-Torsion/1 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * targetSize - number of bonds per torsion (default: 4)

    Note: this version is only available in older (pre-2014) versions of RDKit 
    Chemfp no longer supports those versions of RDKit.
    """
    name = "RDKit-Torsion/1"


torsion_fingerprint_parameters = all_arguments.select_names("fpSize targetSize fromAtoms")
torsion_fingerprint_parameters.add_argument(
    "nBitsPerEntry",
    decoder = types.positive_int, metavar = "INT",
    default = TORSION_nBitsPerEntry,
    help = "number of bits per entry")
torsion_fingerprint_parameters.add_argument(
    "includeChirality",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = TORSION_includeChirality,
    help = "include chirality information")
    
@_public
@types.adjust_call_doc
class RDKitTorsionFingerprintType_v2(RDKitBaseTorsionFingerprintType):
    """RDKit torsion fingerprints, version 2

    See https://rdkit.org/docs/source/rdkit.Chem.AtomPairs.Torsions.html

    An implementation of Topological-torsion fingerprints, as
    described in: R. Nilakantan, N. Bauman, J. S. Dixon,
    R. Venkataraghavan; "Topological Torsion: A New Molecular
    Descriptor for SAR Applications.  Comparison with Other
    Descriptors" JCICS 27, 82-85 (1987).

    The RDKit-Torsion/2 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * targetSize - number of bonds per torsion (default: 4)
    * nBitsPerEntry - number of bits to set per entry (default: 4)
    * includeChirality - include chirality information (default: 0)
    * fromAtoms - a comma-separated list of atom indices which must be part of the torsion
    """
    _arguments = torsion_fingerprint_parameters
    _format_string = FormatOptionalParameters(
        "fpSize={fpSize} targetSize={targetSize}",
        (("nBitsPerEntry", "4"), ("includeChirality", "0"), ("fromAtoms", None)),
        )
    __call__ = _arguments.get_call_method()
    name = "RDKit-Torsion/2"

def register_torsion(registry):
    registry.register_if(TORSION_VERSION()=="1", RDKitTorsionFingerprintType_v1)
    registry.register_if(TORSION_VERSION()=="2", RDKitTorsionFingerprintType_v2)

 
########### Pattern fingerprinter

# PatternFingerprint((Mol) mol, (int) fpSize=2048, (list) atomCounts=[], (ExplicitBitVect) setOnlyBits=None)
# "NOTE: This function is experimental. The API or results may change from release to release."

def make_pattern_fingerprinter(fpSize=NUM_BITS):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    def pattern_fingerprinter(mol):
        fp = Chem.PatternFingerprint(mol, fpSize=fpSize)
        return DataStructs.BitVectToBinaryText(fp)
    return pattern_fingerprinter

@_public
@types.adjust_call_doc
class RDKitBasePatternFingerprint(VariableSizeFingerprint):
    _arguments = all_arguments.select_names("fpSize")
    _format_string = "fpSize={fpSize}"
    _make_fingerprinter = staticmethod(make_pattern_fingerprinter)
    __call__ = _arguments.get_call_method()

@_public
class RDKitPatternFingerprint_v1(RDKitBasePatternFingerprint):
    """RDKit's experimental substructure screen fingerprint, version 1

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.PatternFingerprint

    The RDKit-Pattern/1 fingerprint has no parameters.

    Note: this version is only available in ancient versions of RDKit.
    Chemfp no longer supports those versions of RDKit.
    """
    name = "RDKit-Pattern/1"

@_public
class RDKitPatternFingerprint_v2(RDKitBasePatternFingerprint):
    """RDKit's experimental substructure screen fingerprint, version 2

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.PatternFingerprint

    The RDKit-Pattern/2 fingerprint has no parameters.

    Note: this version is only available in ancient versions of RDKit.
    Chemfp no longer supports those versions of RDKit.
    """
    name = "RDKit-Pattern/2"
    
@_public
class RDKitPatternFingerprint_v3(RDKitBasePatternFingerprint):
    """RDKit's experimental substructure screen fingerprint, version 3

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.PatternFingerprint

    The RDKit-Pattern/3 fingerprint has no parameters. This version was
    released 2017.03.1.

    Note: Chemfp no longer supports those versions of RDKit.
    """
    name = "RDKit-Pattern/3"

@_public
class RDKitPatternFingerprint_v4(RDKitBasePatternFingerprint):
    """RDKit's experimental substructure screen fingerprint, version 4

    See https://rdkit.org/docs/source/rdkit.Chem.rdmolops.html#rdkit.Chem.rdmolops.PatternFingerprint

    The RDKit-Pattern/4 fingerprint has no parameters. This version was
    introduced in August 2017 for the 2017.09.1 release.
    """
    name = "RDKit-Pattern/4"

def _check_pattern_version():
    _pattern = make_pattern_fingerprinter(fpSize=256)(get_version_probe_mol())
    if _pattern == b'\xfb\xce\xfe\xfd\xf8\xdb{l\x97\xfb\xfa\xf3\xfb\xff\xf4Y\xfc\x7f\xb9m\xff\xf7\xff\xff\xfe\x1fw\xef\xd3\xcfRy':
        return "1"

    _pattern2 = make_pattern_fingerprinter(fpSize=256)(Chem.MolFromSmiles("O"))
    
    if _pattern == b'{\xde\xfe\xfd\xfa\xdbsl\x97i\xfa\xb2\xfb\xff\xf49\xfd\x7f\xb9m\xff\xf3\xff\xff\xfe\x7fwo\xd3\x8f\xd2y':
        # The probe molecule isn't able to distinguish between v2 and v3
        if _pattern2 == b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "2"
        if _pattern2 == b"\x00\x02\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "3"

    if _pattern == b'{\xdf\xfe\xfd\xfa\xdfs|\x97i\xfa\xb2\xfb\xff\xf49\xfd\x7f\xb9o\xff\xf3\xff\xff\xfe\x7fwo\xdb\x8f\xd2y':
        # Might as well double-check _pattern2
        if _pattern2 == b"\x00\x02\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00":
            return "4"

    return "unknown"
    

@cache
def PATTERN_VERSION():
    if hasattr(Chem, "PatternFingerprint"):
        PATTERN_VERSION = _check_pattern_version()
    else:
        PATTERN_VERSION = "unknown"
    return PATTERN_VERSION

def register_pattern(registry):
    registry.register_if(PATTERN_VERSION()=="1", RDKitPatternFingerprint_v1)
    registry.register_if(PATTERN_VERSION()=="2", RDKitPatternFingerprint_v2)
    registry.register_if(PATTERN_VERSION()=="3", RDKitPatternFingerprint_v3)
    registry.register_if(PATTERN_VERSION()=="4", RDKitPatternFingerprint_v4)


########### Avalon fingerprinter

# GetAvalonFP( (object)mol [, (int)nBits=512 [, (bool)isQuery=False [, (bool)resetVect=False [, (int)bitFlags=15761407]]]])

# I don't know what the bitFlags do.
# avalonSSSBits        =    32767 = 0x007fff
# avalonSimilarityBits = 15761407 = 0xf07fff

AVALON_NBITS = 512
IS_QUERY = 0
BIT_FLAGS = 15761407

avalon_arguments = types.FingerprintArguments()
avalon_arguments.add_argument("fpSize", decoder=types.positive_int, metavar="INT", default=AVALON_NBITS,
     help = "number of bits in the Avalon fingerprint")
avalon_arguments.add_argument(
    "isQuery",
    decoder = types.zero_or_one, metavar = "0_or_1",
    encoder = types.encode_bool_as_zero_or_one,
    default = IS_QUERY,
    help = "is the fingerprint for a query structure? (1 if yes, 0 if no)")

avalon_arguments.add_argument(
    "bitFlags",
    decoder = types.positive_int, metavar = "INT",
    default = BIT_FLAGS,
    help = "bit flags, SSSBits are 32767 and similarity bits are 15761407")

def make_avalon_fingerprinter(fpSize=AVALON_NBITS, isQuery=IS_QUERY, bitFlags=BIT_FLAGS):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    isQuery = 1 if isQuery else 0
    def avalon_fingerprinter(mol):
        fp = pyAvalonTools.GetAvalonFP(mol, nBits=fpSize, isQuery=isQuery, bitFlags=bitFlags)
        return DataStructs.BitVectToBinaryText(fp)
    return avalon_fingerprinter

@_public
@types.adjust_call_doc
class RDKitAvalonFingerprintType_v1(VariableSizeFingerprint):
    """Avalon fingerprints

    The Avalon Cheminformatics toolkit is available from
    https://sourceforge.net/projects/avalontoolkit/ . It is not part of
    the core RDKit distribution. Instead, RDKit has a compile-time
    option to download and include it as part of the build process.

    The Avalon fingerprint are described in the supplemental information
    for "QSAR - How Good Is It in Practice? Comparison of Descriptor
    Sets on an Unbiased Cross Section of Corporate Data Sets", Peter
    Gedeck, Bernhard Rohde, and Christian Bartels, J. Chem. Inf. Model.,
    2006, 46 (5), pp 1924-1936, DOI: 10.1021/ci050413p. The supplemental
    information is available from
    https://pubs.acs.org/doi/suppl/10.1021/ci050413p

    It uses a set of feature classes which "have been fine-tuned to
    provide good screen-out for the set of substructure queries
    encounted at Novartis while limiting redundancy." The classes are
    ATOM_COUNT, ATOM_SYMBOL_PATH, AUGMENTED_ATOM, AUGMENTED_BOND,
    HCOUNT_PAIR, HCOUNT_PATH, RING_PATH, BOND_PATH, HCOUNT_CLASS_PATH,
    ATOM_CLASS_PATH, RING_PATTERN, RING_SIZE_COUNTS, DEGREE_PATHS,
    CLASS_SPIDERS, FEATURE_PAIRS and ALL_PATTERNS.
    """

    name = "RDKit-Avalon/1"
    _make_fingerprinter = staticmethod(make_avalon_fingerprinter)
    _arguments = avalon_arguments
    _format_string = ("fpSize={fpSize} "
                      "isQuery={isQuery} "
                      "bitFlags={bitFlags}")
    __call__ = _arguments.get_call_method()

def register_avalon(registry):
    registry.register_if(HAS_AVALON, RDKitAvalonFingerprintType_v1)

########### SECFP fingerprinter

  ## EncodeSECFPMol(std::vector<ROMol>& mols,
  ##      int radius = 3,
  ##      bool rings = true,
  ##      bool isomeric = false,
  ##      bool kekulize = false,  # was 'true' before RDKit-2021-03-1 release
  ##      int min_radius = 1,
  ##      int length = 2048);

try:
    from rdkit.Chem import rdMHFPFingerprint
except ImportError:
    HAS_MHFP = False
else:
    HAS_MHFP = True
    # SECFP fingerprints don't use the MHFP parameters so use (0, 0)
    #  instead of (n_permutations=2048, seed=42)
    _secfp_encoder = rdMHFPFingerprint.MHFPEncoder(0, 0)

SECFP_FPSIZE = 2048
SECFP_RADIUS = 3
SECFP_RINGS = 1
SECFP_ISOMERIC = 0

SECFP_KEKULIZE = 0
SECFP_MIN_RADIUS = 1

secfp_arguments = types.FingerprintArguments()
secfp_arguments.add_argument(
    "fpSize",
    decoder = types.positive_int, metavar = "INT",
    default = SECFP_FPSIZE,
    help = "number of bits in the SECFP fingerprint")
secfp_arguments.add_argument(
    "radius",
    decoder = types.nonnegative_int, metavar = "INT",
    default = SECFP_RADIUS,
    help = "maximum radius of the substructure generated at each atom")
secfp_arguments.add_argument(
    "rings",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default=SECFP_RINGS,
    help = "if 1, add SSSR ring to the fingerprint")
secfp_arguments.add_argument(
    "isomeric",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = SECFP_ISOMERIC,
    help = "if 1, use isomeric SMILES instead of non-isomeric SMILES")
secfp_arguments.add_argument(
    "kekulize",
    decoder = types.zero_or_one, metavar = "0|1",
    encoder = types.encode_bool_as_zero_or_one,
    default = SECFP_KEKULIZE,
    help = "if 1, use Kekule SMILES instead of aromatic SMILES")
secfp_arguments.add_argument(
    "min_radius",
    decoder = types.nonnegative_int, metavar = "INT",
    default = SECFP_MIN_RADIUS,
    help = "minimum radius used to extract n-grams")

def make_secfp_fingerprinter(
        fpSize=SECFP_FPSIZE,
        radius=SECFP_RADIUS,
        rings=SECFP_RINGS,
        isomeric=SECFP_ISOMERIC,
        kekulize=SECFP_KEKULIZE,
        min_radius=SECFP_MIN_RADIUS):
    if not (fpSize > 0):
        raise ValueError("fpSize must be 1 or greater")
    if not (radius >= 0):
        raise ValueError("radius must be 0 or greater")
    if min_radius > radius:
        raise ValueError("min_radius must not be larger than radius")
    EncodeSECFPMol = _secfp_encoder.EncodeSECFPMol
    BitVectToBinaryText = DataStructs.BitVectToBinaryText
    def secfp_fingerprinter(mol):
        fp = EncodeSECFPMol(mol, radius, rings, isomeric, kekulize, min_radius, fpSize)
        return BitVectToBinaryText(fp)
    return secfp_fingerprinter

@_public
@types.adjust_call_doc
class RDKitSECFPFingerprintType_v1(VariableSizeFingerprint):
    """SECFP fingerprints

    The SMILES Extended Connectivity Fingerprint, as described in:
      Probst, D., Reymond, J. A probabilistic molecular fingerprint for
      big data settings. J Cheminform 10, 66
      (2018). https://doi.org/10.1186/s13321-018-0321-8
      https://jcheminf.biomedcentral.com/articles/10.1186/s13321-018-0321-8
      
    These are circular fingerprints which encode the circular region as
    a fragment SMILES, which is then hashed to produce the fingerprint bits.

    The RDKit-SECFP/1 :class:`.FingerprintType` parameters are:

    * fpSize - number of bits in the fingerprint (default: 2048)
    * radius - analogous to the radius for the Morgan algorithm (default: 3)
    * rings - include ring membership (default: 1)
    * isomeric - use isomeric SMILES (default: 0)
    * kekulize - Kekulize the molecule and use Kekule SMILES (default: 0)
    * min_radius - minimum radius for the Morgan algorithm (default: 1)

    """
    name = "RDKit-SECFP/1"
    _make_fingerprinter = staticmethod(make_secfp_fingerprinter)
    _arguments = secfp_arguments
    _format_string = ("fpSize={fpSize} "
                      "radius={radius} "
                      "rings={rings} "
                      "isomeric={isomeric} "
                      "kekulize={kekulize} "
                      "min_radius={min_radius}")
    __call__ = _arguments.get_call_method()
        
def register_secfp(registry):
    registry.register_if(HAS_MHFP, RDKitSECFPFingerprintType_v1)

