"""Record types used by the text toolkit.

This is an internal chemfp module. It should not be imported by programs
which use the public API. (Let me know if anything else should be part
of the public API.)

This module contains class definitions for the record objects returned
by the :mod:`.text_toolkit`, which are part of the public API.

"""
# Don't import the actual methods.
# That would introduce a circular import dependency.
from . import _text_toolkit as _tk

__all__ = []
def _public(obj):
    __all__.append(obj.__name__)
    return obj

def _fancy_repr(record):
    if record is not None:
        if len(record) > 60:
            record_s = repr(record[:57])
            record_s = record_s[:-1] + " ..." + record_s[-1]
        else:
            record_s = repr(record)
    else:
        record_s = "None"
    
    return record_s


@_public
class TextRecord(object):
    """Base class for the text_toolkit 'molecules', which work with the records as text.

    The :mod:`chemfp.text_toolkit` implements the toolkit API, but
    it doesn't know chemistry. Instead of returning real molecule objects, with
    atoms and bonds, it returns TextRecord subclass instances that hold the
    record as a text string.

    As an implementation detail (which means its subject to change) there is
    a subclass for each of the support formats.
    
    * :class:`SDFRecord` - holds "sdf" records
    * :class:`SmiRecord` - holds "smi" records (the full line from a "smi" SMILES file)
    * :class:`CanRecord` - holds "can" records (the full line from a "can" SMILES file)
    * :class:`UsmRecord` - holds "usm" records (the full line from a "usm" SMILES file)
    * :class:`SmiStringRecord` - holds "smistring" records (only the "smistring" SMILES string; no id)
    * :class:`CanStringRecord` - holds "canstring" records (only the "canstring" SMILES string; no id)
    * :class:`UsmStringRecord` - holds "usmstring" records (only the "usmstring" SMILES string; no id)

    All of the classes have the following attributes:
    .. py:attribute:: id

       The record identifier as a Unicode string, or None if there is no identifier

    .. py:attribute:: id_bytes

       The record identifier as a byte string, or None if there is no identifier

    .. py:attribute:: record

       The record, as a string. For the smistring, canstring, and usmstring
       formats, this is only the SMILES string.

    .. py:attribute:: record_format

       One of "sdf", "smi", "can", "usm", "smistring", "canstring", or "usmstring".

    The SMILES classes have an attribute:

    .. py:attribute:: smiles

       The SMILES string component of the record.

    """

    __init__ = NotImplemented  # Must call a child class
    
    def add_tag(self, tag, value):
        """Add an SD tag value to the TextRecord

        This methods does nothing if the record is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: string
        :param value: the text for the tag
        :type value: string
        :returns: None
        """
        pass
    
    def get_tag(self, tag):
        """Get the named SD tag value, or None if it doesn't exist or is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: byte or Unicode string
        :returns: a Unicode string, or None
        """
        return None
    
    def get_tag_as_bytes(self, tag):
        """Get the named SD tag value, or None if it doesn't exist or is not an "sdf" record.

        :param tag: the SD tag name
        :type tag: byte string
        :returns: a byte string, or None
        """
        return None
    
    def get_tag_pairs(self):
        """Get a list of all SD tag (name, value) pairs for the TextRecord using Unicode strings

        This function returns an empty list if the record is not an "sdf" record.

        :returns: a list of (Unicode string name, Unicode string value) pairs
        """
        return []
    
    def get_tag_pairs_as_bytes(self):
        """Get a list of all SD tag (name, value) pairs for the TextRecord using byte strings

        This function returns an empty list if the record is not an "sdf" record.

        :returns: a list of (byte string name, byte string value) pairs
        """
        return []

    def copy(self):
        """Return a new record which is a copy of the given record"""
        raise NotImplementedError


@_public
class SDFRecord(TextRecord):
    "Holds an SDF record. See :class:`.TextRecord` for API details"
    record_format = "sdf"
    __slots__ = ("_id", "_id_bytes", "record", "encoding", "encoding_errors")
    def __init__(self, id_bytes, record, encoding, encoding_errors):
        self._id_bytes = id_bytes
        try:
            self._id = id_bytes.decode(encoding, encoding_errors)
        except UnicodeDecodeError:
            self._id = None
        self.record = record
        self.encoding = encoding
        self.encoding_errors = encoding_errors


    def copy(self):
        return SDFRecord(self.id_bytes, self.record, self.encoding, self.encoding_errors)

    def __repr__(self):
        return "SDFRecord(id_bytes=%r(id=%r), record=%s, encoding=%r, encoding_errors=%r)" % (
            self.id_bytes, self.id, _fancy_repr(self.record), self.encoding, self.encoding_errors)

    @property
    def id(self):
        return self._id
    @id.setter
    def id(self, id):
        self._id_bytes = id.encode(self.encoding, self.encoding_errors)
        self._id = id
        
    @property
    def id_bytes(self):
        return self._id_bytes
    @id_bytes.setter
    def id_bytes(self, id_bytes):
        self._id = id_bytes.decode(self.encoding, self.encoding_errors)
        self._id_bytes = id_bytes
    
    def add_tag(self, tag, value):
        if isinstance(tag, str):
            tag = tag.encode(self.encoding, self.encoding_errors)
        if isinstance(value, str):
            value = value.encode(self.encoding, self.encoding_errors)
                             
        self.record = _tk.add_sdf_record_tag(self.record, tag, value)
    
    def get_tag(self, tag):
        return _tk.get_sdf_record_tag(self.record, tag, self.encoding, self.encoding_errors)
    
    def get_tag_as_bytes(self, tag):
        return _tk.get_sdf_record_tag_as_bytes(self.record, tag)
        
    def get_tag_pairs(self, encoding=None, encoding_errors=None):
        if encoding is None:
            encoding = self.encoding
        if encoding_errors is None:
            encoding_errors = self.encoding_errors
        return [(k.decode(encoding, encoding_errors), v.decode(encoding, encoding_errors))
                     for (k, v) in _tk.get_sdf_tag_pairs(self.record)]
    
    def get_tag_pairs_as_bytes(self):
        return _tk.get_sdf_tag_pairs(self.record)
    

    copy.__doc__ = TextRecord.copy.__doc__
    add_tag.__doc__ = TextRecord.add_tag.__doc__
    get_tag.__doc__ = TextRecord.get_tag.__doc__
    get_tag_pairs.__doc__ = TextRecord.get_tag_pairs.__doc__

@_public
class BaseSmiRecord(TextRecord):
    """The base record type for SMILES files"""
    __slots__ = ("id", "record", "smiles", "encoding", "encoding_errors")
    def __init__(self, id, record, smiles, encoding, encoding_errors):
        self.id = id
        self.record = record
        self.smiles = smiles
        self.encoding = encoding
        self.encoding_errors = encoding_errors

    def copy(self):
        return self.__class__(self.id, self.record, self.smiles, self.encoding, self.encoding_errors)

    def __repr__(self):
        smiles_repr = _fancy_repr(self.smiles)
        record_repr = _fancy_repr(self.record)
        return "%s(id=%r, record=%s, smiles=%s, encoding=%r, encoding_errors=%r)" % (
            self.__class__.__name__, self.id, record_repr, smiles_repr, self.encoding, self.encoding_errors)
    
@_public
class SmiRecord(BaseSmiRecord):
    """Holds a "smi" record. See :class:`.TextRecord` for API details"""
    record_format = "smi"

@_public
class CanRecord(BaseSmiRecord):
    """Holds a "can" record. See :class:`.TextRecord` for API details"""
    record_format = "can"

@_public
class UsmRecord(BaseSmiRecord):
    """Holds a "usm" record. See :class:`.TextRecord` for API details"""
    record_format = "usm"

class _SmiStringRecord(TextRecord):
    __slots__ = ("id", "record", "smiles")
    encoding = "utf8"
    encoding_errors = "strict"
    def __init__(self, id, record, smiles):
        self.id = id
        self.record = record
        self.smiles = smiles

    def copy(self):
        return self.__class__(self.id, self.record, self.smiles)

    def __repr__(self):
        record_repr = _fancy_repr(self.record)
        return "%s(id=%r, record=%s, smiles=%r)" % (
            self.__class__.__name__, self.id, record_repr, self.smiles)

@_public
class SmiStringRecord(_SmiStringRecord):
    """Holds a "smistring" record. See :class:`.TextRecord` for API details"""
    record_format = "smistring"

@_public
class CanStringRecord(_SmiStringRecord):
    """Holds a "canstring" record. See :class:`.TextRecord` for API details"""
    record_format = "canstring"

@_public
class UsmStringRecord(_SmiStringRecord):
    """Holds a "usmstring" record. See :class:`.TextRecord` for API details"""
    record_format = "usmstring"
    

