# Copyright (c) 2010-2022 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See https://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

import sys
import re
import pkg_resources
import traceback

from . import _compat

__all__ = []
def _public(f):
    __all__.append(f.__name__)
    return f

# Break a cyclical loop:
def _has_toolkit(toolkit_name):
    from . import has_toolkit
    globals()["_has_toolkit"] = has_toolkit
    return has_toolkit(toolkit_name)
    
##

# This isn't yet a public class.
# It should also derive from ChemFPError
# However, there's an import loop to resolve.
@_public
class FingerprintTypeValueError(ValueError):
    pass

@_public
class FingerprintUnavailable(ValueError):
    def __init__(self, base_name, toolkit_name, version, reason):
        self.base_name = base_name
        self.toolkit_name = toolkit_name
        self.version = version
        self.reason = reason
        if version is None:
            self.name = base_name
        else:
            self.name = f"{base_name}/{version}"
        msg = f"Unable to use {self.name}: {reason}"
        super().__init__(msg)

    def copy(self):
        return FingerprintUnavailable(self.base_name, self.toolkit_name, self.version, self.reason)

@_public
class FingerprintTypeUnknownError(FingerprintTypeValueError):
    def __init__(self, family_name, registry, toolkit_name=None):
        msg = "Unknown fingerprint type %s" % (_compat.myrepr(family_name),)
        super().__init__(msg)
        self._errmsg = msg
        self.family_name = family_name
        self._registry = registry
        self._aliases = {}
        self._toolkit_name = toolkit_name
        # TODO: Migrate the *2fps programs so they include this information
        # rather than always list all (relevant) fingerprint types.
        self._include_help = False

    def __str__(self):
        if self._include_help:
            return self._errmsg + ". " + self.get_help()
        else:
            return self._errmsg

    def include_help(self):
        self._include_help = True
        
    def add_aliases(self, aliases):
        "Include aliases {alias name -> chemfp family name} in the help"
        self._aliases.update(aliases)

    def set_toolkit_name(self, toolkit_name):
        "Set the toolkit name used to get the list of available fingerprint types"
        self._toolkit_name = toolkit_name

    def get_help(self):
        "Get a help message about likely or possible names"
        suggestions = self.get_suggestions()
        if suggestions:
            if len(suggestions) == 1:
                msg = "Closest match is:"
            else:
                msg = "Closest matches are:"
            entries = suggestions
        else:
            msg = "Known families are:"
            entries = self.get_all_names()

        rows = []
        for (name, alias) in entries:
            if alias is None:
                rows.append(f"\n  {name}")
            else:
                rows.append(f"\n  {name} (alias for {alias})")
        row_s = "".join(rows)
        return f"{msg}{row_s}"
        
    def get_all_names(self):
        """Return a list of available fingerprint names as 2-element tuples

        "This is part of the internal API"
        """

        if self._toolkit_name is None:
            # can't yet mix include_unavailable=True and toolkit_name!=None
            family_names = self._registry.get_family_names(
                    include_unavailable = True,
                    )
        else:
            family_names = self._registry.get_family_names(
                    include_unavailable = False,
                    toolkit_name = self._toolkit_name
                    )
            
        names = dict.fromkeys(family_names, None)
        names.update(self._aliases)
        return sorted(names.items(), key = lambda pair: pair[0].lower())

    def get_suggestions(self, cutoff=0.6):
        """Return a list of likely fingerprint names as 2-element tuples

        "This is part of the internal API"
        """
        import difflib
        from collections import defaultdict

        search_name = self.family_name.lower()
        
        # Map from lower-case name to original names.
        # (In principle, chemfp type names are case sensitive,
        # so it's possible to have names which only differ by case.)
        name_map = defaultdict(list)
        for name, alias in self.get_all_names():
            name_map[name.lower()].append((name, alias))

        if not name_map:
            return []
        
        suggestions = []
        for lc_name in difflib.get_close_matches(
                search_name, name_map, n = len(name_map), cutoff=cutoff):
            suggestions.extend(name_map[lc_name])

        # Now add substring search results.
        # (I want "maccs" or "babel" to work.)
        if len(search_name) >= 3:
            seen = set(entry[0] for entry in suggestions)
            for lc_name, entries in name_map.items():
                if search_name in lc_name:
                    for entry in entries:
                        entry_name = entry[0]
                        if entry_name not in seen:
                            suggestions.append(entry)
                            seen.add(entry_name)
        
        return suggestions

        

######## Parse a type string

# Type strings are of the form
#
#    <name>  (<WS> <argument-name> '=' <argument-value>)*
#
#  where <name> is of the form:
#
#    <base name> ('/' <version>)?
#
# The base name name and version match [!-.0-~]+  (printable ASCII except for "/")
# The keys match [a-zA-Z_][a-zA-Z0-9_]* . The values match [!-~]+

_name_pat = re.compile(r"([!-.0-~]+)(/([!-.0-~]+))?$")

_not_allowed_char_pat = re.compile(r"[^!-.0-~]")
def _contains_special_characters(name):
    return _not_allowed_char_pat.search(name) is not None


def _verify_name(name):
    "Verify that the type string name is correct"
    # Quick check to see if it's in the right form.
    m = _name_pat.match(name)
    if m is not None:
        return m.group(1), m.group(3)

    # Something went wrong. Let's see if I can give a better report.
    if not name:
        raise FingerprintTypeValueError("Unable to use an empty fingerprint type name")
    if not name.strip():
        raise FingerprintTypeValueError("Unable to use a blank fingerprint type name: %r" % (name,))

    n = name.count("/")
    if n == 0:
        if _contains_special_characters(name):
            raise FingerprintTypeValueError("Fingerprint type name contains an unexpected character: %r" % (name,))
        return name, None
    if n == 1:
        base_name, version = name.split("/")
        
        if not base_name:
            raise FingerprintTypeValueError("Missing fingerprint type name before the first '/': %r" % (name,))
        if _contains_special_characters(base_name):
            raise FingerprintTypeValueError("Fingerprint type name contains an unexpected character: %r" % (name,))
        
        if not version:
            raise FingerprintTypeValueError("Missing version after the first '/' of the fingerprint type: %r" % (name,))
        if _contains_special_characters(version):
            raise FingerprintTypeValueError("Fingerprint type version contains an unexpected character: %r" % (name,))

        return base_name, version

    raise FingerprintTypeValueError("Too many '/'s in the fingerprint type name: %r" % (name,))


    
class TokenizedType:
    """A tokenized type string

    `name` is the type name, which may or may not have a version
    string. `base_name` is the type name without the version,
    and `version` is the version component, or None if there is
    no version.

    `settings` is a list of argument pairs as the 2-element
    tuple (argument name, argument value), where both the name
    and value are strings.
    """
    
    def __init__(self, base_name, version, settings):
        self.base_name = base_name
        self.version = version
        if version is None:
            self.name = base_name
        else:
            self.name = base_name + "/" + version
        self.settings = settings

def tokenize_type_string(type):
    """Convert a type string into a TokenizedType instance

    This is most easily understood as an example.
    
    >>> tokens = tokenize_type_string("ABC/123 a=1 b=xyz")
    >>> tokens.name, tokens.base_name, tokens.version
    ('ABC/123', 'ABC', '123')
    >>> tokens.settings
    [('a', '1'), ('b', 'xyz')]
    """
    base_name, version, settings = _split_type_string(type)
    
    # It starts off correct. Now parse the name/value settings.

    seen = set()
    tokens = []
    
    for term in settings:
        try:
            left, right = term.split("=")
        except ValueError:
            raise FingerprintTypeValueError("Term %r in type %r must have one and only one '='" %
                                                (term, type))
        if left in seen:
            raise FingerprintTypeValueError("Duplicate parameter %r in type %r" % (left, type))

        seen.add(left)
        tokens.append( (left, right) )
    
    return TokenizedType(base_name, version, tokens)

def _split_type_string(type):
    """Break up a type string into its (base name, version, settings)"""
    if not type:
        if type is None:
            raise FingerprintTypeValueError("Must specify a fingerprint type string")
        raise FingerprintTypeValueError("Unable to use an empty fingerprint type")
    if type.isspace():
        raise FingerprintTypeValueError("Empty fingerprint type (%r)" % (type,))
    if type[:1].isspace():
        raise FingerprintTypeValueError("Fingerprint type must not begin with whitespace: %r" % (type,))

    terms = type.split()
    if not terms:
        raise AssertionError("Unexpected empty fingerprint type (%r)" % (type,))

    name = terms[0]
    base_name, version = _verify_name(name)
    return base_name, version, terms[1:]


######## Manage information about how to convert string parameters into actual arguments

# A FingerprintArguments instance contains information about the
# arguments for a type, including a help message and the ways to
# convert argument values to and from a string, plus some extra
# information for argparse.ArgumentParser.add_argument().

class FingerprintArgument:
    """Information about a fingerprint type argument and converters to/from a string
    """
    def __init__(self, name, default, help, decoder, encoder=str,
                 action=None, metavar=None):
        #: the argument name
        self.name = name
        #: default value for the argument
        self.default = default
        #: help message describing the name (used for command-line arguments)
        self.help = help
        #: function to convert a string to its actual value
        self.decoder = decoder
        #: function to convert a value to a string
        self.encoder = encoder
        #: value used for argparse's add_argument
        self.action = action
        #: value used for argparse's add_argument
        self.metavar = metavar

class _unchanged: # Helper for the created __call__ function
    def __repr__(self):
        return "<current value>"
_unchanged = _unchanged()
        
class FingerprintArguments:
    """Information about how to convert fingerprint type arguments

    `arguments` is a dictionary mapping an argument name to its
    FingerprintArgument.

    """
    def __init__(self, arguments={}):
        self.arguments = arguments.copy()

    # TODO: add more of the dictionary API?
    def __getitem__(self, name):
        return self.arguments[name]

    def __contains__(self, name):
        return name in self.arguments
        

    def select_names(self, names):
        """Make a new FingerprintArguments() using the selected argument names

        `names` is either list of argument names or a string containing
        whitespace separated names.

        :param name: the argument names for the new FingerprintArguments
        :type name: list of strings, or a whitespace-separated string with names
        :returns: a new FingerprintArguments using only the selected names
        """
        if isinstance(names, _compat.filename_types):
            names = names.split()
        new_arguments = {}
        for name in names:
            new_arguments[name] = self.arguments[name]
        return FingerprintArguments(new_arguments)

    def add_argument(self, name, default, help, decoder, 
                     encoder=str, action=None, metavar=None,
                     add_default_to_help=True):
        """Add a new argument

        Add a new argument named `name` with default value `default`.
        If `add_default_to_help` is True then the `help` message is
        combined with information about the default value to make
        the complete help message. Otherwise the help string is used
        as-is.
        
        Argument values are externally represented as strings.
        The `decoder` and `encoder` callables convert the string
        to/from the actual value. The `decoder` may raise a
        ValueError if there's an error. (For example, Python's
        `float` builtin is a valid decoder.)

        The `action` and `metavar` values are passed unchanged
        to argparse.
        
        :param name: the name of the argument
        :type name: string
        :param decoder: a callable which converts a string to a value
                   or raises a ValueError if there's an error
        :param default: the default value for the argument
        :param help: text passed to argparse
        :param encoder: a callable which converts a valid value to
                   a string
        :param action: action value passed to argparse
        :param metavar: metavar value passed to argparse
        :param add_default_to_help: if True, combine help and encoder(default)
                   to make the actual help message
        """
        if name in self.arguments:
            raise FingerprintTypeValueError("Argument %r already specified" % (name,))

        if add_default_to_help:
            help = "%s (default=%s)" % (help, encoder(default))
            
        self.arguments[name] = FingerprintArgument(name, default, help,
                                                   decoder, encoder, action, metavar)


    def decode_type_parameters(self, parameters):
        """Convert name/encoded_value pairs into a fingerprint_args dict

        `parameters` is an iterable of (name, string encoded value) pairs.
        Unknown names and duplicate names raise a ValueError. Default
        values are used for missing arguments.

        :param parameters: an iterable of (name, string encoded value) pairs
        :returns: a dictionary of names and decoded values
        """
        arguments = self.arguments

        # Set up the defaults
        kwargs = dict( (name, arg.default) for (name, arg) in arguments.items() )

        # Go through the parameters
        seen = set()
        for name, value in parameters:
            # Check for unknown and duplicate names
            if name not in arguments:
                raise FingerprintTypeValueError("Unsupported fingerprint parameter name %r" % (name,))
            if name in seen:
                raise FingerprintTypeValueError("Duplicate fingerprint parameter name %r" % (name,))
            seen.add(name)
            arg = arguments[name]
            try:
                kwargs[name] = arg.decoder(value)
            except ValueError as err:
                raise FingerprintTypeValueError("Unable to parse fingerprint parameter %r value %r: %s" % (name, value, err))

        return kwargs

    def merge_with_defaults(self, fingerprint_args):
        """Fill in missing parameters in a dictionary of fingerprint arguments.

        `fingerprint_args` is a dictionary containing parameter names and
        (decoded) values. The output is a new dictionary containing the names
        and values of the input dictionary, plus default values for the missing
        parameter names.

        Raises a ValueError if one of the `fingerprint_args` dictionary keys
        is not a known parameter.
        """
        arguments = self.arguments
        
        # Set up the defaults
        kwargs = dict( (name, arg.default) for (name, arg) in arguments.items())

        for name, value in fingerprint_args.items():
            if name not in arguments:
                raise FingerprintTypeValueError("Unsupported fingerprint parameter %r (=%r)" % (name, value))
            # Allow people to specify 'None' to mean "use the default value"
            if value is not None:
                kwargs[name] = value
        
        return kwargs

    def encode_fingerprint_args(self, fingerprint_args):
        """Convert parameter names and values into a dict of name and string-encoded values

        `fingerprint_args` is a dictionary of parameter names and (decoded) values.
        The result is a new dictionary with the same names and string encoded values.

        Raises a ValueError if one of the `fingerprint_args` dictionary keys
        is not a known parameter, or if one of the values could not be encoded.
        """
        arguments = self.arguments
        d = {}
        for (name, value) in fingerprint_args.items():
            try:
                encoder = arguments[name].encoder
            except KeyError:
                raise FingerprintTypeValueError("Unsupported fingerprint argument %r (=%r)" % (name, value))
            try:
                d[name] = encoder(value)
            except ValueError as err:
                raise FingerprintTypeValueError(str(err) + (" in argument %s" % (name,)))
        return d

    def get_call_method(self):
        """Create a __call__ method for a fingerprint type subclass with the correct kwargs and defaults

        This creates something like:

            def __call__(self, *, fpBits=_unchanged):
                d = self.fingerprint_kwargs.copy()
                if fpBits is not _unchanged:
                    d["fpBits"] = fpBits
                return self.get_fingerprint_family().from_kwargs(d)
        """
        return _get_call_method(self.arguments)

# BaseFingerprintType.__call__.__doc__ but for no parameters
def _no_parameters():
    """Create a new instance of this fingerprint type.

    The new fingerprint type is a clone of the current fingerprint type.

    :returns: a new :class:`chemfp.types.FingerprintType`
    """
    
# BaseFingerprintType.__call__.__doc__ but for one parameters
def _one_parameter():
    """Create a new instance of this fingerprint type.

    The new fingerprint type is a clone of the current fingerprint type,
    with the same fingerprint family and same parameter value.

    Call with a keyword argument to change that value.

    :returns: a new :class:`chemfp.types.FingerprintType`
    """
    
# BaseFingerprintType.__call__.__doc__ but for many keyword parameters
def _many_parameters():
    """Create a new instance of this fingerprint type.

    The new fingerprint type is a clone of the current fingerprint type,
    with the same fingerprint family and same parameter value.

    If a keyword argument is specified, its value will replace the
    current value. Unspecified keyword arguments will keep the
    current value.

    :returns: a new :class:`chemfp.types.FingerprintType`
    """



def _get_call_method(fingerprint_arguments):
    import ast
    import types
    from ast import (
        arg,
        arguments,
        Assign,
        Attribute,
        Call,
        Compare,
        Constant,
        FunctionDef,
        If,
        Index,
        IsNot,
        Load,
        Module,
        Name,
        Return,
        Store,
        Str,
        Subscript,
        )
    kwonlyargs = []
    kw_defaults = []
    if_tests = []

    if sys.version_info[:2] <= (3, 8):
        def SLICE(name):
            return Index(value=Str(s=name))
    else:
        def SLICE(name):
            return Constant(value=name)
        
    lineno = 2
    for name in fingerprint_arguments:
        kwonlyargs.append(ast.arg(arg=name))
        kw_defaults.append(Name(id='_unchanged', ctx=Load()))

        if_tests.append(
            If(
             lineno=lineno,
             col_offset = 0,
             test=Compare(
              left=Name(id=name, ctx=Load()),
              ops=[
               IsNot()],
              comparators=[
               Name(id='_unchanged', ctx=Load())]),
             body=[
              Assign(
               targets=[
                Subscript(
                 value=Name(id='d', ctx=Load()),
                 slice=SLICE(name),
                 ctx=Store())],
               value=Name(id=name, ctx=Load()))],
             orelse=[])
             )
        lineno += 1
                 
    module_ast = Module(
        lineno = 1,
        col_offset = 0,
            
        body=[
         FunctionDef(
          name='__call__',
          args=arguments(
           posonlyargs=[],
           args=[
            arg(arg='self')],
           kwonlyargs=kwonlyargs,
           kw_defaults=kw_defaults,
           defaults=[]),
          body=[
           Assign(
            targets=[
             Name(id='d', ctx=Store())],
            value=Call(
             func=Attribute(
              value=Attribute(
               value=Name(id='self', ctx=Load()),
               attr='fingerprint_kwargs',
               ctx=Load()),
              attr='copy',
              ctx=Load()),
             args=[],
             keywords=[]))
            ] + if_tests + [
           Return(
            lineno = lineno,
            col_offset = 0,
            value=Call(
             func=Attribute(
              value=Call(
               func=Attribute(
                value=Name(id='self', ctx=Load()),
                attr='get_fingerprint_family',
                ctx=Load()),
               args=[],
               keywords=[]),
              attr='from_kwargs',
              ctx=Load()),
             args=[
              Name(id='d', ctx=Load())],
             keywords=[]))],
          decorator_list=[])],
        type_ignores=[])

    ast.fix_missing_locations(module_ast)
    module_code = compile(module_ast, "<fptype-call>", "exec")

    for const in module_code.co_consts:
        if isinstance(const, types.CodeType):
            break
    else:
        raise AssertionError("Could not find created `__call__` function")

    func = types.FunctionType(const, {"_unchanged": _unchanged}, name="__call__")
    func.__kwdefaults__ = dict.fromkeys(fingerprint_arguments, _unchanged)
    # func.__doc__ = BaseFingerprintType.__call__.__doc__
    if not fingerprint_arguments:
        func.__doc__ = _no_parameters.__doc__
    elif len(fingerprint_arguments) == 1:
        func.__doc__ = _one_parameter.__doc__
    else:
        func.__doc__ = _many_parameters.__doc__
    return func

# Tweak the docstring for the autogenerated __call__ in the derived FingerprintType.
def adjust_call_doc(klass):
    full_name = f"{klass.__module__}.{klass.__name__}"
    klass.__call__.__doc__ = klass.__call__.__doc__.replace(
        "chemfp.types.FingerprintType", full_name,
        )
    return klass


# Helper functions to validate fingerprint type arguments

def positive_int(s):
    # Don't do int(s) because that allows "+3" and " 3 ", which I don't want
    if not s.isdigit():
        raise FingerprintTypeValueError("must be 1 or greater")
    i = int(s)
    if i == 0:
        raise FingerprintTypeValueError("must be 1 or greater")
    return i

def nonnegative_int(s):
    if not s.isdigit():
        raise FingerprintTypeValueError("must be 0 or greater")
    return int(s)

def zero_or_one(s):
    if s == "0":
        return 0
    if s == "1":
        return 1
    raise FingerprintTypeValueError("must be 0 or 1")

def encode_bool_as_zero_or_one(val):
    if val:
        return "1"
    return "0"

def power_of_2_at_least_32(s):
    if not s.isdigit():
        raise FingerprintTypeValueError("must a power of 2 and at least 32")
    i = int(s)
    if i < 32 or bin(i).count("1") != 1:
        raise FingerprintTypeValueError("must a power of 2 and at least 32")
    return i
    
def fromAtom_list(s):
    t = s.strip()
    if not t:
        # Should I allow the empty string?
        # If so, does that mean no atoms or all atoms?
        raise FingerprintTypeValueError(
            "must contain a comma-separated list of atom indices")

    terms = t.split(",")
    seen = set()
    for term in terms:
        if not term.isdigit():
            raise FingerprintTypeValueError(
                "term %r must be a non-negative integer"
                % (term,))
        seen.add(int(term))
    return sorted(seen)
    

######## The Fingerprint Family classes

class BaseFingerprintFamily:
    name = None
    toolkit = None
    def __call__(self, fingerprint_kwargs=None):
        raise NotImplementedError("must be implemented in the subclass")

    def from_kwargs(self, fingerprint_kwargs=None):
        raise NotImplementedError("must be implemented in the subclass")

    def from_text_settings(self, settings=None):
        raise NotImplementedError("must be implemented in the subclass")

    def get_kwargs_from_text_settings(self, settings=None):
        raise NotImplementedError("must be implemented in the subclass")


# The BaseFingerprintFamily is the "real" base class.
# I don't know why I have a BaseFingerprintFamily. Remove it? XXX

@_public
class FingerprintFamily(BaseFingerprintFamily):
    """A FingerprintFamily is used to create a FingerprintType or get information about its parameters

    Two reasons to use a FingerprintFamily (instead of using
    :func:`chemfp.get_fingerprint_type` or :func:`chemfp.get_fingerprint_type_from_text_settings`) are:

    * figure out the default arguments;
    * given a text settings or parameter dictionary, use the keys from the default
      argument keys to remove other parameters before creating a FingerprintType
      (otherwise the creation function will raise an exception)

    All fingerprint families have the following attributes:

    * name - the type name, including version
    * toolkit - the toolkit API for the underlying chemistry toolkit, or None
    
    """
    def __init__(self, fingerprint_class):
        """The constructor is not part of the public API."""
        self._fingerprint_class = fingerprint_class
        self._arguments = getattr(self._fingerprint_class, "_arguments", None)

    def __repr__(self):
        """Return a string like 'FingerprintFamily(<RDKit-Fingerprint/2>)'"""
        return "FingerprintFamily(<%s>)" % (self._fingerprint_class.name,)

    @property
    def name(self):
        """The full fingerprint name, including the version"""
        return self._fingerprint_class.name

    @property
    def base_name(self):
        """The base fingerprint name, without the version"""
        return self.name.split("/")[0]
        
    @property
    def version(self):
        """The fingerprint version (the part after the '/' in the full name)"""
        return self.name.split("/")[1]
    
    @property
    def toolkit(self):
        """The chemfp toolkit wrapper for the chemistry toolkit used to generate this fingerprint, or None"""
        return self._fingerprint_class.toolkit

    def __call__(self, **fingerprint_kwargs):
        """Create a fingerprint type; keyword arguments can override the defaults

        The argument values are native Python values, not string-encoded values::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family(fpSize=1024)
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param fingerprint_kwargs: the fingerprint parameters
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        return self.from_kwargs(fingerprint_kwargs)

    def from_kwargs(self, fingerprint_kwargs=None):
        """Create a fingerprint type; items in the *fingerprint_kwargs* dictionary can override the defaults

        The dictionary values are native Python values, not string-encoded values::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family.from_kwargs({"fpSize": 1024})
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param fingerprint_kwargs: the fingerprint parameters
        :type fingerprint_kwargs: a dictionary where the values are Python objects
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        if fingerprint_kwargs is None:
            fingerprint_kwargs = {}

        if self._arguments is not None:
            fingerprint_kwargs = self._arguments.merge_with_defaults(fingerprint_kwargs)

        return self._fingerprint_class(fingerprint_kwargs)
        ## except TypeError as err:
        ##     raise ValueError("Unable to create a %r fingerprint type: %s"
        ##                      % (self._fingerprint_class.name, err))

    def from_text_settings(self, settings=None):
        """Create a fingerprint type; *settings* is a dictionary with string-encoded value that can override the defaults

        The dictionary values are string-encoded values, not native Python values.
        This function exists to help handle command-line arguments and setting files.::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> fptype = family.from_text_settings()
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=2048 nBitsPerHash=2 useHs=1'
          >>> fptype = family.from_text_settings({"fpSize": "1024"})
          >>> fptype.get_type()
          'RDKit-Fingerprint/2 minPath=1 maxPath=7 fpSize=1024 nBitsPerHash=2 useHs=1'

        The function will raise an exception for unknown arguments.

        :param settings: the fingerprint text settings
        :type settings: a dictionary where the values are string-encoded
        :returns: an object implementing the :class:`chemfp.types.FingerprintType` API
        """
        if settings is None:
            return self._fingerprint_class(self._arguments.merge_with_defaults({}))
        fingerprint_kwargs = self.get_kwargs_from_text_settings(settings)
        return self._fingerprint_class(fingerprint_kwargs)

    def get_kwargs_from_text_settings(self, settings=None):
        """Convert a dictionary of string-encoded fingerprint parameters into native Python values

        String-encoded values ("text settings") can come from the command-line,
        a configuration file, a web reqest, or other text sources. The fingerprint
        types need actual Python values. This method converts the first to the second::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> family.get_kwargs_from_text_settings()
          {'maxPath': 7, 'fpSize': 2048, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}
          >>> family.get_kwargs_from_text_settings({"fpSize": "128", "maxPath": "5"})
          {'maxPath': 5, 'fpSize': 128, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}

        :param settings: the fingerprint text settings
        :type settings: a dictionary where the values are string-encoded
        :returns: an dictionary of (decoded) fingerprint parameters
        """
        if self._arguments is None:
            if settings:
                raise FingerprintTypeValueError("This fingerprint does not take settings")
            return {}
        if settings is None:
           settings = {}
        return self._arguments.decode_type_parameters(settings.items())

    def get_defaults(self):
        """Return the default parameters as a dictionary

        The dictionary values are native Python objects::

          >>> import chemfp
          >>> family = chemfp.get_fingerprint_family("RDKit-Fingerprint")
          >>> family.get_defaults()
          {'maxPath': 7, 'fpSize': 2048, 'nBitsPerHash': 2, 'minPath': 1, 'useHs': 1}

        :returns: an dictionary of fingerprint parameters
        """
        if self._arguments is None:
            return {}
        return self._arguments.merge_with_defaults({})


########

@_public
class NoFingerprintParametersMixin:
    """This fingerprint type does not support parameters"""
    def __init__(self, fingerprint_kwargs):
        if fingerprint_kwargs:
            raise FingerprintTypeValueError("This fingerprint does not support parameters")
        super(NoFingerprintParametersMixin, self).__init__(fingerprint_kwargs)

@_public
class BaseFingerprintType:
    #: the fingerprint name
    name = None
    #: a reference to the underlying toolkit wrapper
    toolkit = None
    #: a description of the software package(s) used
    software = None
    
    #: an internal flag indicating if the fingerprinter can
    #: raise an exception when processing a molecule
    fingerprinter_can_fail = False

    def __init__(self, fingerprint_kwargs):
        self.fingerprint_kwargs = fingerprint_kwargs
        base_name, mid, version = self.name.partition("/")
        #: the part of the name before the '/'
        self.base_name = base_name
        if not mid:
            version = None
        #: the part of the name after the '/'
        self.version = version

    def __call__(self, **kwargs):
        """Create a new instance of this fingerprint type.
        
        The new fingerprint type is a clone of the current fingerprint type,
        with the same fingerprint family and same parameters.

        Use `**kwargs` to specify new values. Any remaining parameters
        will be left unchanged.
        
        :returns: a new :class:`chemfp.types.FingerprintType`
        """
        d = self.fingerprint_kwargs.copy()
        d.update(kwargs)
        return self.get_fingerprint_family().from_kwargs(d)
    
    def get_type(self):
        """Get the full type string (name and parameters) for this fingerprint type

        :returns: a canonical fingerprint type string, including its parameters
        """
        raise NotImplementedError("Must be implemented in a derived class")

    def make_fingerprinter(self):
        """Make a 'fingerprinter'; a callable which takes a molecule and returns a fingerprint

        :returns: a function object which takes a molecule and return a fingerprint
        """
        raise NotImplementedError("Must be implemented in the derived class")


    def read_molecule_fingerprints(self, source, format=None, id_tag=None,
                                    reader_args=None, errors="strict", location=None):
        """Read fingerprints from a structure source as a FingerprintIterator

        Iterate through the *format* structure records in *source*. If *format*
        is None then auto-detect the format based on the *source*. Use the
        fingerprint type to compute the fingerprint. For SD files, use *id_tag*
        to get the record id from the given SD tag instead of the title line.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and goes to the next
        record, and "ignore" goes to the next record.

        The *location* parameter takes a Location instance. If None then a default
        Location will be created.
        

        :param source: the structure source
        :type source: a filename, file object, or None to read from stdin
        :param format: the input structure format
        :type format: a format name string, or Format object, or None to auto-detect
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :param location: object used to track parser state information
        :type location: a Location object, or None
        :returns: a :class:`chemfp.FingerprintIterator` which iterates over the (id, fingerprint) pair
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def read_molecule_fingerprints_from_string(self, content, format=None, id_tag=None,
                                                reader_args=None, errors="strict", location=None):
        """Read fingerprints from structure records in a string, as a FingerprintIterator

        Iterate through the *format* structure records in *content*. Use the
        fingerprint type to compute the fingerprint. For SD files, use *id_tag*
        to get the record id from the given SD tag instead of the title line.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and goes to the next
        record, and "ignore" goes to the next record.

        The *location* parameter takes a Location instance. If None then a default
        Location will be created.
        
        :param content: the string containing structure records
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :param location: object used to track parser state information
        :type location: a Location object, or None
        :returns: a :class:`chemfp.FingerprintIterator` which iterates over the (id, fingerprint) pair
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def parse_molecule_fingerprint(self, content, format, reader_args=None, errors="strict"):
        """Parse the first molecule record of the content then compute and return the fingerprint

        Read the first molecule from *content*, which contains records
        in the given *format*. Compute and return its fingerprint.

        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        the fingerprint, and "ignore" returns None for the fingerprint without
        any extra message.
        
        :param content: the string containing at least one structure record
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: the fingerprint as a byte string
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def parse_id_and_molecule_fingerprint(self, content, format, id_tag=None,
                                          reader_args=None, errors="strict"):
        """Parse the first molecule record of the content then compute and return the id and fingerprint

        Read the first molecule from *content*, which contains records in
        the given *format*. Compute its fingerprint and get the molecule id.
        For an SD record use *id_tag* to get the record id from the given
        SD tag instead of from the title line.

        Return the id and fingerprint as the (id, fingerprint) pair.
        
        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        values it cannot compute, and "ignore" is like "report" but without
        the error message. For "report" and "ignore", if the molecule cannot
        be parsed then the result will be (None, None). If the fingerprint
        cannot be computed then the result will be (id, None).

        :param content: the string containing at least one structure record
        :type source: a string
        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: a pair of (id string, fingerprint byte string)
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def make_id_and_molecule_fingerprint_parser(self, format, id_tag=None, reader_args=None, errors="strict"):
        """Make a function which parses molecule from a record and returns the id and computed fingerprint

        This is a very specialized function, designed for performance, but it
        doesn't appear to give any advantage. You likely don't need it.

        Return a function which parses a content string containing structure
        records in the given *format* to get a molecule. Use the molecule to
        compute the fingerprint and get its id. For an SD record use *id_tag*
        to get the record id from the given SD tag instead of from the
        title line.

        The new function will return the (id, fingerprint) pair.
        
        The *reader_args* dictionary parameters depend on the toolkit and format.
        For details see the docstring for ``self.toolkit.read_molecules``.

        The *errors* parameter specifies how to handle errors. "strict" raises
        an exception, "report" sends a message to stderr and return None for
        values it cannot compute, and "ignore" is like "report" but without
        the error message. For "report" and "ignore", if the molecule cannot
        be parsed then the result will be (None, None). If the fingerprint
        cannot be computed then the result will be (id, None).

        :param format: the input structure format
        :type format: a format name string, or Format object
        :param id_tag: SD tag containing the record id
        :type id_tag: string, or None to use the record title
        :param reader_args: reader parameters passed to the underlying toolkit
        :type reader_args: a dictionary
        :param errors: specify how to handle errors
        :type errors: one of "strict", "report", or "ignore"
        :returns: a function which takes a content string and returns an (id, fingerprint) pair
        """

    def compute_fingerprint(self, mol):
        """Compute and return the fingerprint byte string for the toolkit molecule

        :param mol: a toolkit molecule
        :returns: the fingerprint as a byte string
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def compute_fingerprints(self, mols):
        """Compute and return the fingerprint for each toolkit molecule in an iterator

        This function is a slightly optimized version of::

          for mol in mols:
            yield self.compute_fingerprint(mol)

        :param mols: an iterable of toolkit molecules
        :returns: a generator of fingerprints, one per molecule
        """
        raise NotImplementedError("Must be implemented in the derived class")

    def get_fingerprint_family(self):
        """Return the fingerprint family for this fingerprint type

        :returns: a :class:`FingerprintFamily`
        """
        return registry.get_family(self.name)

def _compute_molecule_fingerprints(reader, fingerprinter, error_handler):
    if error_handler is None:
        for id, mol in reader:
            if mol is None:
                yield id, None
            else:
                yield id, fingerprinter(mol)
    else:
        # Some CDK fingerprinters may fail
        location = reader.location
        for id, mol in reader:
            if mol is None:
                yield id, None
            else:
                try:
                    yield id, fingerprinter(mol)
                except Exception as err:
                    error_handler.error(
                        "Cannot generate fingerprint: %s" % (err,),
                        location)
                    continue

@_public
class ThreadsafeFingerprinterMixin:
    """This fingerprint type is thread-safe.

    The make_fingerprint() method always returns the same fingerprinter.
    """
    def __init__(self, fingerprint_kwargs):
        super(ThreadsafeFingerprinterMixin, self).__init__(fingerprint_kwargs)
        self._fingerprinter = self._make_fingerprinter(**fingerprint_kwargs)

    def make_fingerprinter(self):
        return self._fingerprinter

    make_fingerprinter.__doc__ = BaseFingerprintType.__doc__

@_public
class FingerprintType(BaseFingerprintType):
    """The base to all fingerprint types

    A fingerprint type has the following public attributes:

    * :attr:`.FingerprintType.name` - the fingerprint name, including the version
    * :attr:`.FingerprintType.base_name` - the fingerprint name, without the version
    * :attr:`.FingerprintType.version` - the fingerprint version part of the name
    * :attr:`.FingerprintType.toolkit` - the chemfp wrapper module to the appropriate chemistry toolkit
    * :attr:`.FingerprintType.software` - a description of the software package(s) involved
    * :attr:`.FingerprintType.num_bits` - the number of bits in this fingerprint type
    * :attr:`.FingerprintType.fingerprint_kwargs` - a dictionary containing any fingerprint arguments

    The built-in fingerprint types are:

    * :class:`chemfp.openbabel_types.OpenBabelFP2FingerprintType_v1` - ``OpenBabel-FP2/1`` -
      Open Babel FP2
    * :class:`chemfp.openbabel_types.OpenBabelFP3FingerprintType_v1` - ``OpenBabel-FP3/1`` -
      Open Babel FP3
    * :class:`chemfp.openbabel_types.OpenBabelFP4FingerprintType_v1` - ``OpenBabel-FP4/1`` -
      Open Babel FP4
    * :class:`chemfp.openbabel_types.OpenBabelMACCSFingerprintType_v1` - ``OpenBabel-MACCS/1`` -
      Open Babel 166 MACCS keys
    * :class:`chemfp.openbabel_types.OpenBabelMACCSFingerprintType_v2` - ``OpenBabel-MACCS/2`` -
      Open Babel 166 MACCS keys
    * :class:`chemfp.openbabel_types.OpenBabelECFP0FingerprintType_v1` - ``OpenBabel-ECFP0`` -
      Open Babel ECFP0
    * :class:`chemfp.openbabel_types.OpenBabelECFP2FingerprintType_v1` - ``OpenBabel-ECFP2`` -
      Open Babel ECFP2
    * :class:`chemfp.openbabel_types.OpenBabelECFP4FingerprintType_v1` - ``OpenBabel-ECFP4`` -
      Open Babel ECFP4
    * :class:`chemfp.openbabel_types.OpenBabelECFP6FingerprintType_v1` - ``OpenBabel-ECFP6`` -
      Open Babel ECFP6
    * :class:`chemfp.openbabel_types.OpenBabelECFP8FingerprintType_v1` - ``OpenBabel-ECFP8`` -
      Open Babel ECFP8
    * :class:`chemfp.openbabel_types.OpenBabelECFP10FingerprintType_v1` - ``OpenBabel-ECFP10`` -
      Open Babel ECFP10
      
    * :class:`chemfp.openbabel_patterns.SubstructOpenBabelFingerprinter_v1` - ``ChemFP-Substruct-OpenBabel/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with Open Babel
    * :class:`chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v1` - ``RDMACCS-OpenBabel/1`` -
      chemfp's own 166 MACCS keys implemented with Open Babel (does not include key 44)
    * :class:`chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v2` - ``RDMACCS-OpenBabel/1`` -
      chemfp's own 166 MACCS keys implemented with Open Babel

    * :class:`chemfp.openeye_types.OpenEyeCircularFingerprintType_v2` - ``OpenEye-Circular/2`` -
      OEGraphSim circular fingerprints
    * :class:`chemfp.openeye_types.OpenEyeMACCSFingerprintType_v2` - ``OpenEye-MACCS166/2`` -
      OEGraphSim 166 MACCS keys
    * :class:`chemfp.openeye_types.OpenEyePathFingerprintType_v2` - ``OpenEye-Path/2`` -
      OEGraphSim path fingerprints
    * :class:`chemfp.openeye_types.OpenEyeTreeFingerprintType_v2` - ``OpenEye-Tree/2`` -
      OEGraphSim tree fingerprints
    * :class:`chemfp.openeye_types.OpenEyeMoleculeScreenFingerprintType_v1` - ``OpenEye-MoleculeScreen/1`` -
      OEGraphSim molecule screens
    * :class:`chemfp.openeye_types.OpenEyeSMARTSScreenFingerprintType_v1` - ``OpenEye-SMARTSScreen/1`` -
      OEGraphSim SMARTS screens
    * :class:`chemfp.openeye_types.OpenEyeMDLScreenFingerprintType_v1` - ``OpenEye-MDLScreen/1`` -
      OEGraphSim MDL screens

    * :class:`chemfp.openeye_patterns.SubstructOpenEyeFingerprinter_v1` - ``ChemFP-Substruct-OpenEye/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with OEChem
    * :class:`chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v1` - ``RDMACCS-OpenEye/1`` -
      chemfp's own 166 MACCS keys implemented with OEChem (does not include key 44)
    * :class:`chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v2` - ``RDMACCS-OpenEye/2`` -
      chemfp's own 166 MACCS keys implemented with OEChem

    * :class:`chemfp.rdkit_types.RDKitFingerprintType_v2` - RDKit-Fingerprint/2 - RDKit path and tree fingerprint
    * :class:`chemfp.rdkit_types.RDKitMACCSFingerprintType_v2` - ``RDKit-MACCS/2`` -
      RDKit 166 MACCS keys
    * :class:`chemfp.rdkit_types.RDKitMorganFingerprintType_v1` - ``RDKit-Morgan/1`` -
      RDKit circular fingerprints
    * :class:`chemfp.rdkit_types.RDKitAtomPairFingerprint_v2` - ``RDKit-AtomPair/2`` -
      RDKit atom pair fingerprints
    * :class:`chemfp.rdkit_types.RDKitTorsionFingerprintType_v2` - ``RDKit-Torsion/2`` -
      RDKit torsion fingerprints
    * :class:`chemfp.rdkit_types.RDKitAvalonFingerprintType_v2` - ``RDKit-Avalon/2`` -
      RDKit's interface to the Avalon fingerprints
    * :class:`chemfp.rdkit_types.RDKitPatternFingerprint_v4` - ``RDKit-Pattern/4`` -
      RDKit's substructure pattern fingerprints
    * :class:`chemfp.rdkit_types.RDKitSECFPFingerprintType_v1` - ``RDKit-SECFP/1`` -
      SMILES Extended Connectivity Fingerprint from Probst et al.

    * :class:`chemfp.rdkit_patterns.SubstructRDKitFingerprintType_v1` - ``ChemFP-Substruct-RDKit/1`` -
      chemfp's 881 CACTVS/PubChem-like keys implemented with RDKit
    * :class:`chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v1` - ``RDMACCS-RDKit/1`` -
      chemfp's own 166 MACCS keys implemented with OEChem (does not include key 44)
    * :class:`chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v2` - ``RDMACCS-RDKit/2`` -
      chemfp's own 166 MACCS keys implemented with OEChem


    """
    toolkit = None
    software = None
    _arguments = None
    _format_string = None
    _make_fingerprinter = None

    def __repr__(self):
        return f"{self.__class__.__name__}(<{self.get_type()}>)"
    
    # Fingerprint methods
    def get_type(self):
        if self.fingerprint_kwargs:
            encoded_args = self._arguments.encode_fingerprint_args(self.fingerprint_kwargs)
            return self.name + " " + self._format_string.format(**encoded_args)
        else:
            return self.name

    def make_fingerprinter(self):
        # I assume the fingerprinter contains state, so always make a new one.
        # (Use the Threadsafe mixin if it's thread-safe.)
        return self._make_fingerprinter(**self.fingerprint_kwargs)


    def get_metadata(self, sources=None):
        """Return a Metadata appropriate for the given fingerprint type.

        This is most commonly used to make a :class:`chemfp.Metadata`
        that can be passed into a :class:`chemfp.FingerprintWriter`.
        
        If *sources* is a string or a list of strings then it will passed
        to the newly created Metadata instance. It should contain filenames
        or other description of the fingerprint sources.

        :param sources: fingerprint source filenames or other description
        :type sources: None, a string, or list of strings
        :returns: a :class:`chemfp.Metadata`
        """
        from chemfp import Metadata, io
        return Metadata(num_bits = self.num_bits,
                        sources = sources,
                        software = self.software,
                        type = self.get_type(),
                        date = io.utcnow())

        
    def read_molecule_fingerprints(self, source, format=None, id_tag=None,
                                    reader_args=None, errors="strict", location=None):
        """Read fingerprints from a structure source. Returns a stream of (id, fp) pairs"""
        from chemfp import FingerprintIterator
        if self.fingerprinter_can_fail:
            from . import io
            errors = error_handler = io.get_parse_error_handler(errors)
        else:
            error_handler = None
        reader = self.toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location)
        fingerprinter = self.make_fingerprinter()
        return FingerprintIterator(self.get_metadata(source),
                                   _compute_molecule_fingerprints(reader, fingerprinter, error_handler),
                                   reader.location, reader.close)

    def read_molecule_fingerprints_from_string(self, content, format=None, id_tag=None,
                                                reader_args=None, errors="strict", location=None):
        """Read fingerprints from a string. Returns a stream of (id, fp) pairs"""
        from chemfp import FingerprintIterator
        if self.fingerprinter_can_fail:
            from . import io
            errors = error_handler = io.get_parse_error_handler(errors)
        else:
            error_handler = None
        reader = self.toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                                 errors, location)
        fingerprinter = self.make_fingerprinter()
        return FingerprintIterator(self.get_metadata(),
                                   _compute_molecule_fingerprints(reader, fingerprinter, error_handler),
                                   reader.location, reader.close)
                                

    def parse_molecule_fingerprint(self, content, format, reader_args=None, errors="strict"):
        """Return the corresponding fingerprint byte string for the structure record"""
        mol = self.toolkit.parse_molecule(content, format, reader_args=reader_args, errors=errors)
        if mol is None:
            return None
        fingerprinter = self.make_fingerprinter()
        return fingerprinter(mol)

    def parse_id_and_molecule_fingerprint(self, content, format, id_tag=None,
                                          reader_args=None, errors="strict"):
        """Return the corresponding id and fingerprint byte string for the structure record"""
        id, mol = self.toolkit.parse_id_and_molecule(content, format, id_tag=id_tag,
                                                     reader_args=reader_args, errors=errors)
        if mol is None:
            return id, None
        fingerprinter = self.make_fingerprinter()
        return id, fingerprinter(mol)

    def make_id_and_molecule_fingerprint_parser(self, format, id_tag=None, reader_args=None, errors="strict"):
        """Make a function which parses molecule from a record and returns the id and computed fingerprint"""
        parse_id_and_molecule = self.toolkit.make_id_and_molecule_parser(
            format=format, id_tag=id_tag, reader_args=reader_args, errors=errors)
        fingerprinter = self.make_fingerprinter()

        def parse_id_and_molecule_fingerprint(content):
            id, mol = parse_id_and_molecule(content)
            if mol is not None:
                return id, fingerprinter(mol)
            else:
                return id, None

        return parse_id_and_molecule_fingerprint

    def from_mol(self, mol):
        """Return the corresponding fingerprint byte string for the molecule"""
        fingerprinter = self.make_fingerprinter()
        return fingerprinter(mol)

    def from_mols(self, mols):
        """Return the corresponding fingerprint byte strings for a stream of molecules"""
        fingerprinter = self.make_fingerprinter()
        for mol in mols:
            yield fingerprinter(mol)
    
    def compute_fingerprint(self, mol):
        """Return the corresponding fingerprint byte string for the molecule

        .. deprecated:: 4.0
        
            Use `from_mol` instead.
        """
        fingerprinter = self.make_fingerprinter()
        return fingerprinter(mol)

    def compute_fingerprints(self, mols):
        """Return the corresponding fingerprint byte strings for a stream of molecules

        .. deprecated:: 4.0
        
            Use `from_mols` instead.
        """
        fingerprinter = self.make_fingerprinter()
        for mol in mols:
            yield fingerprinter(mol)

    # XXX Hack because I don't want to spend the time to remove the BaseFingerprint
    
    get_type.__doc__ = BaseFingerprintType.get_type.__doc__
    make_fingerprinter.__doc__ = BaseFingerprintType.make_fingerprinter.__doc__
    read_molecule_fingerprints.__doc__ = BaseFingerprintType.read_molecule_fingerprints.__doc__
    read_molecule_fingerprints_from_string.__doc__ = BaseFingerprintType.read_molecule_fingerprints_from_string.__doc__
    parse_molecule_fingerprint.__doc__ = BaseFingerprintType.parse_molecule_fingerprint.__doc__
    parse_id_and_molecule_fingerprint.__doc__ = BaseFingerprintType.parse_id_and_molecule_fingerprint.__doc__
    make_id_and_molecule_fingerprint_parser.__doc__ = BaseFingerprintType.make_id_and_molecule_fingerprint_parser.__doc__
    compute_fingerprint.__doc__ = BaseFingerprintType.compute_fingerprint.__doc__
    compute_fingerprints.__doc__ = BaseFingerprintType.compute_fingerprints.__doc__

########
def _get_module(module_path):
    if isinstance(module_path, str):
        module = __import__(module_path)
        for name in module_path.split(".")[1:]:
            module = getattr(module, name)
    else:
        module = module_path
    return module


# Part of the code uses a a list of (key, value) tuples. Other parts
# use a dict API and expect to call .items() to get those
# tuples. FakeSortedDict is an adapter of the first to the second.

class FakeSortedDict:
    def __init__(self, items):
        self._items = items
    def __bool__(self):
        return bool(self._items)
    def items(self):
        return self._items
    def iteritems(self):
        return self._items
    def __repr__(self):
        return "FakeSortedDict(%r)" % (self._items,)


# Only the "register_*" methods are part of the public API.

class FingerprintFamilyRegistry:
    def __init__(self, family_resolvers):
        self._family_resolvers = family_resolvers
        
    def register_alias(self, toolkit_name, alias, name):
        2/0 # TODO

    def register_fingerprint_type(self, toolkit_name, fingerprint_type):
        2/0 # TODO
            
    def _resolve_name(self, family_name, toolkit_name=None):
        if "/" in family_name:
            base_name, _, version = family_name.partition("/")
        else:
            base_name = family_name
            version = None
            
        if base_name in self._family_resolvers:
            return self._family_resolvers[base_name].resolve(version)
        
        raise FingerprintTypeUnknownError(family_name, self, toolkit_name)

    def _resolve_all(self, toolkit_name=None):
        for name, resolver in self._family_resolvers.items():
            resolver.resolve_all()
    
    def get_all_fingerprint_names(self, toolkit_name=None):
        if toolkit_name is None:
            names = set(self._family_resolvers)
            # TODO: aliases
        else:
            names = set(name for (name, resolver) in self._family_resolvers.items()
                        if resolver.toolkit_name == toolkit_name)
        return names
        

    def get_family(self, family_name):
        return self._resolve_name(family_name)

    def get_families(self, toolkit_name=None):
        # Return only versioned families
        families = []
        for base_name, resolver in self._family_resolvers.items():
            if toolkit_name is not None and toolkit_name != resolver.toolkit_name:
                continue
            families.extend(resolver.get_families())
        return families

    def get_family_names(self, include_unavailable, toolkit_name=None):
        # The set contains both the versioned and unversioned family names
        names = set()
        for resolver in self._family_resolvers.values():
            if toolkit_name is not None and toolkit_name != resolver.toolkit_name:
                continue
            names.update(resolver.get_family_names(include_unavailable))
        return names
 
    def has_family(self, typename):
        try:
            self._resolve_name(typename)
        except ValueError:
            return False
        else:
            return True
    
    def get_fingerprint_type(self, type, fingerprint_kwargs=None):
        parsed = tokenize_type_string(type)
        name = parsed.name
        family = self._resolve_name(name)

        try:
            final_kwargs = family.get_kwargs_from_text_settings(FakeSortedDict(parsed.settings))
        except ValueError as err:
            raise FingerprintTypeValueError("Cannot get fingerprint type %r: %s" % (type, err)) from None
        
        if fingerprint_kwargs:
            final_kwargs.update(fingerprint_kwargs)
        try:
            return family.from_kwargs(final_kwargs)
        except ValueError as err:
            raise FingerprintTypeValueError("Cannot get fingerprint type %r: %s" % (type, err)) from None
            

    def get_fingerprint_type_from_text_settings(self, type, settings):
        parsed = tokenize_type_string(type)
        name = parsed.name
        self._resolve_name(name)
        family = self._resolve_name(name)
        
        try:
            type_kwargs = family.get_kwargs_from_text_settings(FakeSortedDict(parsed.settings))
        except ValueError as err:
            raise FingerprintTypeValueError("Error with type %r: %s" % (type, err))

        if settings:
            # This is a bit complicated. The "get_kwargs_from_text_settings" returns a complete
            # dictionary, with defaults filled in, when I only want the new terms.
            try:
                settings_kwargs = family.get_kwargs_from_text_settings(settings)
            except ValueError as err:
                raise FingerprintTypeValueError("Error with type %r: %s" % (type, err)) from None

            # Override the type string values using only the new values in settings
            # and not the defaults.
            for name in settings:
                type_kwargs[name] = settings_kwargs[name]

        return family(**type_kwargs)

_pretty_toolkit_name = {
    "rdkit": "RDKit",
    "openbabel": "Open Babel",
    "openeye": "OEGraphSim",
    "cdk": "CDK",
    }
    
class FingerprintTypeRegistry:
    api_version = 1
    def __init__(self, base_name, toolkit_name):
        self.base_name = base_name
        self.toolkit_name = toolkit_name
        self._unavailable = []
        self._available = []
        
    def unavailable(self, version, reason, priority=0):
        item = (
            priority,
            version,
            FingerprintUnavailable(self.base_name, self.toolkit_name, version, reason),
            )
        self._unavailable.append(item)

    def _check_prefix(self, fingerprint_type):
        name = fingerprint_type.name
        expected = self.base_name + "/"
        if not name.startswith(expected):
            raise ValueError(f"Fingerprint type name {name!r} must start with {expected!r}")
        return name.partition("/")[2]
        
    def register(self, fingerprint_type, priority=0):
        version = self._check_prefix(fingerprint_type)
        item = (
            priority,
            version,
            FingerprintFamily(fingerprint_type),
            )
        self._available.append(item)

    def register_if(self, is_available, fingerprint_type, reason=None, priority=0):
        if is_available:
            return self.register(fingerprint_type, priority)
        else:
            version = self._check_prefix(fingerprint_type)
            if reason is None:
                pretty_name = _pretty_toolkit_name.get(self.toolkit_name, self.toolkit_name)
                reason = (
                    f"This version of {pretty_name} does not support the {fingerprint_type.name} fingerprint"
                    )
            return self.unavailable(version, reason, priority)
    
class FamilyResolver:
    def __init__(self, base_name, toolkit_name):
        self.base_name = base_name
        self.toolkit_name = toolkit_name
        self.entry_points = []
        
        self._lookup = None # (version | None) -> FingerprintFamily | FingerprintUnavailable
        
    def _load(self):
        if self._lookup is not None:
            return
        if not _has_toolkit(self.toolkit_name):
            self._lookup = {
                None: FingerprintUnavailable(
                    self.base_name, self.toolkit_name, None,
                    f"toolkit {self.toolkit_name!r} is not available"),
                }
            return
        
        unavailable = []
        available = []
        
        for entry_point in self.entry_points:
            try:
                func = entry_point.resolve()
            except Exception as err:
                sys.stderr.write(f"Cannot load entry point {entry_point}: {err} - Skipping.\n")
                traceback.print_exc()
                continue
            # Always use a new registry so failures are all-or-none.
            fptype_registry = FingerprintTypeRegistry(self.base_name, self.toolkit_name)
            try:
                func(fptype_registry)
            except Exception as err:
                sys.stderr.write(f"Cannot call entry point {entry_point}: {err} - Skipping.\n")
                traceback.print_exc()
                continue
            
            unavailable.extend(fptype_registry._unavailable)
            available.extend(fptype_registry._available)


        lookup = {}
        unavailable.sort()
        for _, _, obj in unavailable:
            lookup[obj.version] = obj
        if unavailable:
            lookup[None] = unavailable[-1][-1]

        available.sort()
        for _, _, obj in available:
            lookup[obj.version] = obj

        if available:
            lookup[None] = available[-1][-1]

        if None not in lookup:
            lookup[None] = FingerprintUnavailable(
                self.base_name, self.toolkit_name, None,
                "Unable to load any entry points for this fingerprint type",
                )
        
        self._lookup = lookup
            
    def resolve(self, version):
        self._load()
        try:
            obj = self._lookup[version]
        except KeyError:
            available_versions = sorted(
                repr(name) for name, obj in self._lookup.items()
                if name is not None and not isinstance(obj, ValueError)
                )
            if len(available_versions) > 1:
                *left, right = available_versions
                left_str = ", ".join(left)
                msg = f"available versions: {left_str} or {right}"
            elif len(available_versions) == 1:
                msg = f"available version: {available_versions[0]}"
            else:
                raise self._lookup[None].copy() from None # No implementations available
                
            raise FingerprintUnavailable( # This version isn't available.
                self.base_name, self.toolkit_name, version,
                f"Unknown fingerprint type version ({msg})"
                ) from None
        
        if isinstance(obj, FingerprintUnavailable):
            # Always raise a new copy. (Don't know if that's important, but it feels right.)
            raise obj.copy()  # known fingerprint type/version but not available
        return obj

    def get_available(self):
        self._load()
        return [obj for key, obj in self._lookup.items() if key is not None and not isinstance(obj, ValueError)]

    def get_family_names(self, include_unavailable=False):
        # Includes versioned and unversioned names
        self._load()
        def V(base_name, version):
            if version is None:
                return base_name
            else:
                return f"{base_name}/{version}"
            
        if include_unavailable:
            return set(V(obj.base_name, version) for (version, obj) in self._lookup.items())
        else:
            return set(V(obj.base_name, version) for (version, obj) in self._lookup.items()
                           if not isinstance(obj, ValueError))

    def get_families(self):
        # Only return versioned families
        self._load()
        return [obj for obj in self._lookup.values() if not isinstance(obj, ValueError)]
        
    
def load_family_resolvers():
    resolvers = {}
    for toolkit_name in ("rdkit", "openeye", "openbabel", "cdk"):
        for ep in pkg_resources.iter_entry_points(f"chemfp.{toolkit_name}.types"):
            name = ep.name
            
            resolver = resolvers.get(name, None)
            if resolver is None:
                resolver = resolvers[name] = FamilyResolver(name, toolkit_name)
            elif resolver.toolkit_name != toolkit_name:
                sys.stderr.write(
                    f"Cannot register fingerprint type {name!r} from entry point {ep} because "
                    f"it was previously registered for a different toolkit in {resolver.entry_points[0]}\n")
                continue

            resolver.entry_points.append(ep)
    
    return resolvers
            
registry = FingerprintFamilyRegistry(load_family_resolvers())

def parse_type(name):
    import warnings
    warnings.warn("Please use chemfp.get_fingerprint_type() instead of chemfp.types.parse_type()")
    return registry.get_fingerprint_type(name)

####

# Helper function to display the appropriate string during Sphinx documentation
def _sphinx_repr_if(using_sphinx_mock, doc, real):
    if using_sphinx_mock:
        return _SPHINX_REPR(doc)
    else:
        return real
        
class _SPHINX_REPR:
    def __init__(self, for_sphinx):
        self.for_sphinx = for_sphinx
    def __repr__(self):
        return self.for_sphinx

